/*
  * Khoros: $Id: utilities.c,v 1.8 1996/08/10 03:14:21 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: utilities.c,v 1.8 1996/08/10 03:14:21 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Various Private Utility Routines For the kexpr Library
   >>>>		
   >>>>  Private:
   >>>>                 kexpr_get_idlist()
   >>>>                 kexpr_delete_idlist()
   >>>>                 kexpr_get_symbol()
   >>>>                 kexpr_reget_symbol()
   >>>>                 kexpr_add_symbol()
   >>>>                 kexpr_delete_symbol()
   >>>>                 kexpr_free_symbol()
   >>>>                 kexpr_free_symlist()
   >>>>                 kexpr_eval_expression()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"
#include "y.tab.h"

static IdList  *var_head = NULL;

long	current_id;
jmp_buf	jump_dest;
int    	error_found = FALSE;
int    	jump_dest_enabled = FALSE;
Symbol  FinalValue, *CompiledSymbol;
char	*sptr, *string_array, *kexpr_error_string;


/*-----------------------------------------------------------
|
|  Routine Name: kexpr_get_idlist - get the idlist
|
|       Purpose: kexpr_get_idlist() gets the idlist associated
|		 for a particular id.  If no idlist exists for the
|		 id then a list is created and added to the "var_head".
|
|         Input: id - the variable identifier
|
|        Output: RETURNS: returns the idlist on success, NULL otherwise
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
IdList *kexpr_get_idlist(
   long id)
{
	IdList *idlist;


	idlist = var_head;
	while (idlist != NULL)
	{
	   if (idlist->id == id)
	      break;

	   idlist = idlist->next;
	}

	if (idlist == NULL)
	{
	   if ((idlist = (IdList *) kcalloc((size_t) 1, sizeof(IdList))) == NULL)
	      return(NULL);

	   idlist->id	    = id;
	   idlist->varlist = NULL;
	   idlist->parent  = NULL;
	   idlist->next    = var_head;
	   var_head         = idlist;
	}
	return(idlist);
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_delete_idlist - delete the idlist
|
|       Purpose: kexpr_delete_idlist() deletes the idlist associated
|		 with a particular id.
|
|         Input: id - the variable identifier
|
|        Output: RETURNS: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
int kexpr_delete_idlist(
   long id)
{
	IdList *idlist, *list;


	list = idlist = var_head;
	while (list != NULL)
	{
	   if (list->id == id)
	      break;

	   idlist = list;
	   list = list->next;
	}

	/*
	 *  If we don't find the id in the current list then just return
	 */
	if (list == NULL)
	   return(FALSE);

	/*
	 *  If the variable list is at the head then advance 'var_head' to the
	 *  next variable list, otherwise we just link around the variable list.
	 */
	if (idlist == list)
	   var_head = list->next;
	else
	   idlist->next = list->next;

	/*
	 *  Make sure that none of the other variable list use the deleted
	 *  idlist as a parent.
	 */
	idlist = var_head;
	while (idlist != NULL)
	{
	   if (idlist->parent == list)
	      idlist->parent = NULL;

	   idlist = idlist->next;
	}
	kexpr_free_symlist(list->varlist);
	kfree_and_NULL(list);

	return(TRUE);
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_get_symbol - get a symbol
|
|       Purpose: kexpr_get_symbol() gets the symbol associated with
|		 a particular idlist and name identifier.  If this
|		 list does not contain the particular symbol and
|		 the idlist is a child then the kexpr_get_symbol()
|		 is recursed with the parent's idlist.
|
|         Input: idlist - the list of variable list from which
|			   we search the symbol
|		 name    - the name of the symbol we are looking
|			   for.
|		 parent  - whether we should be allowed to search
|			   the parent's idlist if one exists.
|
|        Output: RETURNS: returns the symbol on success, NULL otherwise
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
Symbol *kexpr_get_symbol(
   IdList *idlist,
   char    *name,
   int     parent)
{
	Symbol	   *symbol;
	SymbolList *symlist;


	/*
	 *  Race through the symbol list looking to see if we can find the
	 *  symbol.
	 */
	symlist = idlist->varlist;
	while (symlist != NULL)
	{
	   symbol = symlist->symbol;
	   if (kstrcmp(name, symbol->name) == 0)
	      return(symbol);

	   symlist = symlist->next;
	}

	/*
	 *  If the list doesn't have the symbol then try list's parent
	 *  if one exists.
	 */
	if (idlist->parent != NULL && parent == TRUE)
	   symbol = kexpr_get_symbol(idlist->parent, name, parent);
	else
	   symbol = NULL;

	return(symbol);
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_reget_symbol - reget a symbol
|
|       Purpose: When a symbol is undefined, we may need to search
|		 the parent or global list to see if the symbol is
|		 defined elsewhere.  This mechanism allows a lazy
|		 creation and evaluation of expressions.
|
|         Input: idlist - the list of variable list from which
|			   we search the symbol
|		 oldsymbol  - the old symbol we are looking to replace.
|        Output:
|	Returns: returns the new symbol on success, NULL otherwise
|
|    Written By: Mark Young
|          Date: Jul 27, 1995
|
------------------------------------------------------------*/
Symbol *kexpr_reget_symbol(
   IdList *idlist,
   Symbol  *oldsymbol)
{
	Symbol	   *symbol;
	SymbolList *symlist, *tmplist;

	/*
	 *  See if the symbol is defined elsewhere
	 */
	if (idlist->parent != NULL)
	   symbol = kexpr_get_symbol(idlist->parent, oldsymbol->name, TRUE);
	else
	   symbol = NULL;

	/* if no other symbol exists then return NULL */
	if (symbol == NULL)
	   return(NULL);

	/*
	 *  Race through all the idlist's (not just the one in which symbol
	 *  is defined) and replace the old symbol with the new one.
	 */
	idlist = var_head;
	while (idlist != NULL)
	{
	   symlist = idlist->varlist;
	   while (symlist != NULL)
	   {
	      if (symlist->symbol == oldsymbol)
	         symlist->symbol = symbol;

	      if (symlist->symbol != NULL)
	      {
	         tmplist = symlist->symbol->deflist;
	         while (tmplist != NULL)
	         {
	            if (tmplist->symbol == oldsymbol)
	               tmplist->symbol = symbol;

	            tmplist = tmplist->next;
	         }
	      }
	      symlist = symlist->next;
	   }
	   idlist = idlist->next;
	}
	kfree_and_NULL(oldsymbol);
	return(symbol);
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_add_symbol - add a symbol to a idlist
|
|       Purpose: kexpr_add_symbol() adds the symbol associated
|		 with a particular idlist and name identifier.
|
|         Input: idlist - the list of variable list to which
|			   we add the symbol
|		 name    - the name of the symbol we are adding
|                type    - the type of the symbol we are adding
|                value   - the current value of that symbol
|
|        Output: RETURNS: returns the symbol on success, NULL otherwise
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
Symbol *kexpr_add_symbol(
   IdList *idlist,
   char    *name,
   int     type,
   double  value)
{
	Symbol	    *symbol;
	SymbolList  *symlist;


	if ((symlist = (SymbolList *) kcalloc((size_t) 1, sizeof(SymbolList))) == NULL)
	   return(NULL);

	if ((symbol = (Symbol *) kcalloc((size_t)1, sizeof(Symbol))) == NULL)
	   return(NULL);

	symbol->name =  kstrdup(name);
	symbol->type =  type;
	symbol->eval =  FALSE;
	symbol->Value   = value;
	symbol->idlist = idlist;

	symlist->symbol  = symbol;
	symlist->next    = idlist->varlist;
	idlist->varlist = symlist;
	return(symbol);
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_delete_symbol - delete a symbol from a idlist
|
|       Purpose: kexpr_delete_symbol() deletes the symbol associated
|		 with a particular idlist and name identifier.  If
|		 this list does not contain the particular symbol then
|		 we return false, otherwise we delete it from the idlist
|		 and then the kexpr_get_symbol() is recursed with the
|		 parent's idlist.
|
|         Input: id     - the id from which the symbol came from
|		 symbol - the name of the symbol we are adding
|
|        Output: RETURNS: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
int kexpr_delete_symbol(
   long   id,
   Symbol *symbol)
{
	IdList	   *idlist;
	SymbolList *symlist, *list;

	/*
	 * Check to see if it is in the id idlist list.
	 */
	if ((idlist = kexpr_get_idlist(id)) == NULL)
	   return(FALSE);

	list = symlist = idlist->varlist;
	while (list != NULL)
	{
	   if (list->symbol == symbol)
	      break;

	   symlist = list;
	   list = list->next;
	}

	if (list == NULL)
	   return(FALSE);

	if (symlist == list)
	   idlist->varlist = list->next;
	else
	   symlist->next = list->next;

	list->next = NULL;
	kexpr_free_symlist(list);
	return(TRUE);
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_free_symbol - free a symbol
|
|       Purpose: kexpr_free_symbol() is called when we want to free
|		 the contents of a symbol, but not the symbol itself.
|		 We do this before re-using the symbol.
|
|         Input: symbol - the current symbol to be freed
|
|        Output: None
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
void kexpr_free_symbol(
   Symbol *symbol)
{
	Symbol     temp;
	SymbolList list;


	/*
	 *  If the current symbol is NULL then just return
	 */
	if (symbol == NULL)
	   return;

	/*
	 *  Free all the resources that are allocated with the symbol,
	 *  ie). the symlist and the string (if the symbol is of type
	 *  string).
	 */
	if (symbol->type == STRING && symbol->String != NULL)
	   kfree_and_NULL(symbol->String);
	else if (symbol->type != UNDEFINED && symbol->deflist != NULL &&
	         (symbol->type == VARIABLE || symbol->type == EXPRESSION))
	{
	   
	   list.symbol = symbol;
	   list.next   = NULL;
	   kexpr_eval_symlist(&list, FALSE, &temp);
	   symbol->Value = temp.Value;
	}

	if (symbol->deflist != NULL)
	   kexpr_free_symlist(symbol->deflist);

	symbol->deflist = NULL;
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_free_symlist - free a symbol list
|
|       Purpose: kexpr_free_symlist() is called when we want to free
|		 the contents of a symbol list.
|
|         Input: symbol - the current symbol to be freed
|
|        Output: None
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
void kexpr_free_symlist(
   SymbolList *symlist)
{
	Symbol	   *symbol;
	SymbolList *temp;


	/*
	 *  Free all the resources that are allocated with the symlist,
	 *  ie). the symbol and the symlist itself
	 */
	while (symlist != NULL)
	{
	   symbol = symlist->symbol;
	   if (symbol->type == NUMBER || symbol->type == STRING ||
	       symbol->type == OPERATOR)
	   {
	      kexpr_free_symbol(symbol);
	      kfree_and_NULL(symbol);
	   }
	   temp = symlist->next;
	   kfree_and_NULL(symlist);
	   symlist = temp;
	}
}

/*-----------------------------------------------------------
|
|  Routine Name: kexpr_eval_expression - evaluate an expression
|
|       Purpose: kexpr_eval_expression() evaluates the string
|		 and returns whether we succesfully evaluated
|		 that expression.
|
|         Input: id     - the variable identifier
|		 string - the string to be evaluated
|
|        Output: error  - if an error occurred the message is
|			  stored in the string array and False
|			  returned.  If this field is NULL then
|			  the error message is printed to kstderr.
|		 RETURNS: returns the idlist on success, NULL otherwise
|
|    Written By: Mark Young  
|          Date: Thu Jun 25 1992
|
------------------------------------------------------------*/
int kexpr_eval_expression(
   long id,
   char *mystr,
   char *error)
{
	error_found = FALSE;
	kexpr_error_string = error;
	if (setjmp(jump_dest) == 0)
	{
	   current_id = id;
	   jump_dest_enabled = TRUE;
	   mystr = kstring_cleanup(mystr, mystr);
	   string_array = sptr = mystr;
	   (void) kexpr_parse();
	}
	jump_dest_enabled = FALSE;
	return(error_found == FALSE);
}
