/*
 * Khoros: $Id: File.c,v 1.15 1996/06/30 22:36:36 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: File.c,v 1.15 1996/06/30 22:36:36 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		UNIX File Transport Routines
   >>>>		
   >>>>   Static:
   >>>>                 file_tempnam()
   >>>>                 file_open()
   >>>>                 file_close()
   >>>>                 file_read()
   >>>>                 file_write()
   >>>>                 file_lseek()
   >>>>                 file_access()
   >>>>                 file_unlink()
   >>>>                 file_lock()
   >>>>                 file_descriptors()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"	


static int file_tempnam	    PROTO((const char *, const char *, char *));
static int     file_open    PROTO((const char *, int, mode_t, kfile *));
static int     file_close   PROTO((kfile *));
static ssize_t file_read    PROTO((kfile *, kaddr, size_t));
static ssize_t file_write   PROTO((kfile *, kaddr, size_t));
static off_t   file_lseek   PROTO((kfile *, off_t, int));
static int     file_access  PROTO((const char *, const char *, int));
static int     file_unlink  PROTO((const char *, const char *));
static int     file_lock    PROTO((kfile *, int));
static int file_descriptors PROTO((kfile *, int *, int *));

TransportInformation file_transport[] =
{
    {
	"Standard Unix Files",
	"file",
	TRUE,
	file_tempnam,
	file_open,
	file_close,
	file_read,
	file_write,
	file_lseek,
	file_access,
	file_unlink,
	file_lock,
	file_descriptors,
    }
};

/*
 *  Internal resource structures for "standard file" transport
 */
typedef struct
{
	int fid;
} ResourceStruct;


/*-------------------------------------------------------------
|  
|  Routine Name: file_tempnam
|  
|       Purpose: This function initializes a tempnam for a "file"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"filename"
|		  	"/path/filename"
|		  	"./path/filename"
|		  	"file=XXXXXXXXXX"
|		  	     ...etc....
|
|         Input:  identifier - the transport identifier
|		  template   - the filename template to be modeled
|        Output:  result     - the newly created transport name
|		  returns -1 or 0 depending whether we
|		  sucessfully created the template
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int file_tempnam(
   const char *identifier,
   const char *template,
   char *result)
{
	char *temp, path[KLENGTH];

	/*
	 *  If the template is "-" or "#" this means that it is kstdin,
	 *  stdout, or kstderr, which means don't make a tempnam simply
	 *  copy this as the output file
	 */
	if (kstrcmp(template, "-") == 0 || kstrcmp(template, "#") == 0)
	{
	   kstrcpy(result, template);
	   return(0);
	}

	/*
	 *  Make a tempnam from the template
	 */
	(void) kstrcpy(path, template);
	if ((temp = ktmpnam(path)) == NULL)
	   return(-1);

	if (identifier == NULL)
	   (void) kstrcpy(result, temp);
	else
	   (void) sprintf(result,"%s=%s", identifier, temp);

	kfree(temp);
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_open
|  
|       Purpose: This function opens a "file" connection.  It is the
|	         internal driver to open a file, which is called by
|	         the user when they call kopen().  kopen() calls when
|	         it gets an open request with a filename of the following
|	         syntax:
|  
|		  	"filename"
|		  	"/path/filename"
|		  	"./path/filename"
|		  	"file=XXXXXXXXXX"
|		  	     ...etc....
|  
|         Input: 
|        Output:  returns 0 or -1 if an error occurs
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int file_open(
   const char  *path,
   int   flags,
   mode_t mode,
   kfile *file)
{
	int fid;
	ResourceStruct *resources;
	const char *check = path;
	char *tmp, *cmd, temp[KLENGTH];

	if ((tmp = kstrrchr(path, '.')) != NULL &&
		 (kstrcmp(tmp, ".Z") == 0 || kstrcmp(tmp, ".gz") == 0))
	{
	   if (kfile_iswrite(file))
	   {
	      cmd = kstrcmp(tmp, ".gz") == 0 ? "gzip"   : "compress";
	      ksprintf(temp,"pipe=%s -c > %s", cmd, path);
	   }
	   else
	   {
	      cmd = kstrcmp(tmp, ".gz") == 0 ? "gunzip" : "uncompress";
	      ksprintf(temp,"pipe=%s -c %s", cmd, path);
	   }
	   return(kfile_rename(file, temp));
	}
	else if (kstrcmp(path, "-") == 0 || kstrcmp(path, "#") == 0)
	   return(kfile_rename(file, "pipe="));

	if ((fid = open(path, flags, mode)) < 0)
	{
	   char name[KLENGTH];

	   while (*check != '\0')
	   {
	      if (!isalnum(*check) && *check != '/' && *check != '.' &&
		  *check != '_' && *check != '~' && *check != '#' &&
		  *check != '-' && *check != '+')
	      {
	         if (*check != '\\')
	            return(-1);
		 check++;
	      }
	      check++;
	   }
	   kdirname(path, name);

	   if (!(flags & KOPEN_CREAT) || errno != ENOENT ||
	       !kmake_dir(name, 0777) ||
	       (fid = open(path, flags, mode)) < 0)
	   {
	      kinfo(KSYSLIB, "file_open:  Unable to open file '%s'", path);
	      return(-1);
	   }
	}

	if ((resources = (ResourceStruct *) kcalloc((size_t) 1,
				sizeof(ResourceStruct))) == NULL)
	{
	   kinfo(KSYSLIB, "file_open: Not enough memory to kcalloc (%d) bytes"
             " for the khoros file structure.\n", sizeof(ResourceStruct));
	   return(-1);
	}
	resources->fid = fid;
	file->resources = (kaddr) resources;
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_close
|  
|       Purpose: This function closes a "file".  It is the internal driver
|	         to close a data file, which is called by the user
|	         when they call kclose().  kclose() calls the "file" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the initial kopen().
|  
|         Input:  file - the kfile structure.
|        Output:  returns whether we were able to close the file
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int file_close(
   kfile *file)
{
	int status = 0;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call close() since "kfile" is just a regular file
	 *  descriptor.
	 */
	status = close(resources->fid);
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_read
|  
|       Purpose: This function reads a "file".  It is the internal driver
|	         to read data from a file, which is called by the user
|	         when they call kread().  kread() calls the "file" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr  - the pointer to store the data into.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes read from the file
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t file_read(
   kfile *file,
   kaddr ptr,
   size_t nbytes)
{
	ssize_t numread;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call read() since "kfile" is just a regular file
	 *  descriptor.
	 */
        numread = read(resources->fid, ptr, nbytes);
	return(numread);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_write
|  
|       Purpose: This function writes to a "file".  It is the internal driver
|	         to write data from the supplied data array to a file, which
|	         is called by the user when they call kwrite().  kwrite() calls
|	         the "file" internal drivers by the nature of the transport,
|	         which is dictated by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to write the data from.
|		  nbytes - the number of bytes to write.
|        Output:  returns the number of bytes written to the kfile
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t file_write(
   kfile *file,
   kaddr ptr,
   size_t nbytes)
{
	ssize_t numwrite;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call write() since "kfile" is just a regular file
	 *  descriptor.
	 */
        numwrite = write(resources->fid, ptr, nbytes);
	return(numwrite);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_lseek
|  
|       Purpose: This function is used to do a "lseek".  It is the internal
|	         driver to rewind to a specific point so that the data can be
|	         skipped or re-read.  This is called when the user calls
|	         klseek().  klseek() calls the "file" internal drivers by the
|	         nature of the transport, which is dictated by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  offset - the offset in which to seek
|		  whence - the control of how the offset will be applied
|        Output:  returns the number of bytes written to the kfile
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static off_t file_lseek(
   kfile *file,
   off_t  offset,
   int    whence)
{
	off_t	   pos;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call lseek() since "kfile" is just a regular file
	 *  descriptor.
	 */
	pos = lseek(resources->fid, offset, whence);
	return(pos);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_access
|  
|       Purpose: This function is used to do see if a file is accessable.
|	         This is called when the user calls kaccess().  kaccess()
|	         calls the "file" internal drivers by the nature of the
|	         transport, which is dictated by the ktransport_identifier().
|  
|         Input:  name - the original path name passed to kaccess
|		  path - the filename that they wish to access
|		  mode - the access mode information
|        Output:  returns the result
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int file_access(
   const char *name,
   const char *path,
   int  mode)
{
	int	   status;

	/*
	 *  Simply call access() since "path" is just a regular filename
	 */
	if (kstrcmp(path, "-") == 0)
	   return(0);

	status = access(path, mode);
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_unlink
|  
|       Purpose: This function initializes a tempnam for a "file"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"filename"
|		  	"/path/filename"
|		  	"./path/filename"
|		  	"file=XXXXXXXXXX"
|		  	     ...etc....
|  
|         Input:  name - the original path name passed to kunlink
|		  path - the filename that they wish to unlink
|        Output:  returns 0 or -1 depending whether we
|		  sucessfully created the template
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int file_unlink(
   const char *name,
   const char *path)
{
	int	status;

	status = unlink(path);
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_lock
|  
|       Purpose: This function locks a "file" transport.  Given
|	         a request with a filename of the following syntax:
|  
|		  	"filename"
|		  	"/path/filename"
|		  	"./path/filename"
|		  	"file=XXXXXXXXXX"
|		  	     ...etc....
|  
|         Input: 
|        Output:  returns 0 or -1 depending whether we
|		  sucessfully created the template
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int file_lock(
   kfile *file,
   int   operation)
{
        struct  flock lock;
        int     status = 0, cmd;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

        /*
         *  Set the lock to be the whole file...
         */
        lock.l_len    = 0;
        lock.l_start  = 0;
        lock.l_whence = KSEEK_SET;

        /*
         *  Translate the lock operation in fcntl() lock type.
         */
        if (operation & KLOCK_SH)
           lock.l_type = F_RDLCK;
        else if (operation & KLOCK_EX)
           lock.l_type = F_WRLCK;
        else if (operation & KLOCK_UN)
           lock.l_type = F_UNLCK;
        else
           return(-1);

        /*
         *  Blocking is translated into the type of set lock command for
         *  fcntl.  The F_SETLK means set file lock, but don't block, and
         *  F_SETLKW means set file lock, but go ahead and wait.
         */
        if (operation & KLOCK_NB)
           cmd = F_SETLK;
        else
           cmd = F_SETLKW;

        status = fcntl(resources->fid, cmd, &lock) == -1 ? -1 : 0;
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: file_descriptors
|  
|       Purpose: This function shouldn't exist, but there is occasion
|	         when the user wants to know the actual file descriptor
|	         associated with the file descriptor.
|  
|         Input: file structure
|        Output:  inum - the input file descriptor
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int file_descriptors(
   kfile *file,
   int   *inum,
   int   *onum)
{
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	if (inum != NULL) *inum = resources->fid;
	if (onum != NULL) *onum = resources->fid;
	return(0);
}
