/*
 * Khoros: $Id: env.c,v 1.11 1996/05/30 20:26:48 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: env.c,v 1.11 1996/05/30 20:26:48 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Environment Utilities
   >>>>		
   >>>>   Static:
   >>>>                 _init_envp()
   >>>>   Public:
   >>>>                 kgetenv()
   >>>>                 kputenv()
   >>>>                 kremenv()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"	

/*-----------------------------------------------------------
|
|  Routine Name: _init_envp - initialize environment variables
|
|       Purpose: This routine converts the system standard environ
|		 variable into a Khoros managed array so that
|		 the karray_* routines can be used to add and remove
|		 fields to the environment.
|    Written By: Mark Young and Steven Jorgensen
|          Date: Jan 16, 1996
| Modifications:
|
------------------------------------------------------------*/
static void _init_envp(void)
{
   char **envp = kprog_get_envp(), **temp = NULL, *string;
   static int envp_init = FALSE;
   size_t num, size;

   /*=================================================================
    *  If we are changing the environ for the first time then we need
    *  copy the current environ into an array of strings kallocated by
    *  kmalloc.
    *================================================================*/
   if (envp_init == FALSE && *envp != NULL)
   {
      envp_init = TRUE;

      for (num = size = 0; envp[num] != NULL; num++)
      {
         if ((string = kstrdup(envp[num])) != NULL)
	    temp = karray_insert(temp, KSTRING, string, size++,KLIST_TAIL,TRUE);
      }
      envp = karray_insert(temp, KSTRING, NULL, size, KLIST_TAIL, TRUE);
      kprog_set_envp(envp);
   }
}

/************************************************************
*
*  Routine Name: kgetenv - get an environment variable from the environ list.
*       Purpose: This module is used to get an environment variable
*		 from the environ list.  For example, if the user has
*                their TMPDIR environment variable set to /usr/var/tmp, 
*                and this routine is called with (name = "TMPDIR"), then
*                "/usr/var/tmp" will be returned.
*         Input: name - the name of the environament variable to look for
*        Output: none
*       Returns: the environment string upon success, NULL on failure
*  Restrictions: The string returned is simply an address into the environ
*		 list, so it should \fBNOT\fP be freed, or changed in any way
*		 by the calling routine.
*    Written By: Mark Young  
*          Date: Sep 27, 1992 11:55
*************************************************************/

char *kgetenv(
   const char *name)
{
	char *string, **envp;
	size_t num = 0, len = kstrlen(name);


	_init_envp();
	envp = kprog_get_envp();

	while (envp[num] != NULL)
	{
	   string = envp[num];
	   if (kstrncmp(name, string, len) == 0 && (int) string[len] == '=')
	   {
	      string = &string[len+1];
	      return(string);
	   }
	   num++;
	}
	return(NULL);
}

/************************************************************
*
*  Routine Name: kputenv - put an environment variable into the environ list.
*       Purpose: This module is used to put an environment variable
*		 in the environ list.  If "PUTENV" is defined then
*		 we call putenv() otherwise we do our own putenv.  The
*		 only difference is that kputenv() will malloc space
*		 for the incoming environment variable.
*         Input: name - string to be added into putenv
*       Returns: (0) on success, (-1) otherwise
*    Written By: Mark Young  
*          Date: Jul 03, 1992 13:58
*  Side Effects: the input "name" is copied into the environment list
* Modifications: Converted from vputenv in Khoros 1.0 (SJ)
*************************************************************/

int kputenv(
   const char *name)
{
	size_t len, num;
	char   *tmp, *string, **envp;


	_init_envp();

	if ((tmp = kstrdup(name)) == NULL)
	   return(-1);

        envp = kprog_get_envp();
	if ((string = kstrchr(tmp, '=')) != NULL)
	   len = string - tmp;
	else
	   len = kstrlen(tmp);

	/*
	 *  Search to see if this is an existing environment variable.  If
	 *  so then replace the existing one and return.  Otherwise add it
	 *  to the environment variable list.
	 */
	num = 0;
	while (envp[num] != NULL)
	{
	   string = envp[num];
	   if (kstrncmp(tmp, string, len) == 0 && (int) string[len] == '=')
	   {
	      kfree_and_NULL(string);
	      envp[num] = tmp;
	      return(0);
	   }
	   num++;
	}
	envp[num++] = tmp;
	envp = (char **) karray_insert(envp,KSTRING,NULL,num, KLIST_TAIL, TRUE);
	kprog_set_envp(envp);
	return(0);
}


/************************************************************
*
*  Routine Name: kremenv - remove an environment variable from the environment
*       Purpose: This module is used to remove an environment variable
*		 from the environ list.  Since there does not seem
*		 to be a standard routine for this on many machines, we
*		 wrote our own.
*         Input: name - a string containing the environment variable to remove.
*       Returns: (0) on success, (-1) otherwise
*  Restrictions: This routine will not work if the environ list is
*		 not stored as an array of strings.
*    Written By: Mark Young
*          Date: Jul 03, 1992 14:04
*  Side Effects: Does not kfree_and_NULL string associated with environment variable
*		 deleted, since we cannot know how it was created.
* Modifications: Converted from vremenv in Khoros 1.0 (SJ)
*************************************************************/

int kremenv(
   const char *name)
{
	size_t  len, num = (size_t) 0;
	char   *string, **temp, **envp;


	if (name == NULL)
	   return(-1);

	_init_envp();

	envp = kprog_get_envp();
	len = kstrlen(name);
	while (envp[num] != NULL)
	{
	   string = envp[num];
	   if (kstrncmp(name, string, len) == 0 && (int) string[len] == '=')
	   {
	      while (envp[num] != NULL) num++;
	      temp = (char **) karray_delete(envp, KSTRING, string, num);
	      temp[num-1] = NULL;
	      kprog_set_envp(temp);
	      kfree_and_NULL(string);
	      return(0);
	   }
	   num++;
	}
	return(-1);
}
