/*
 * Khoros: $Id: kstruct.c,v 1.15 1996/08/10 03:15:03 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: kstruct.c,v 1.15 1996/08/10 03:15:03 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Structure Service Utilities
   >>>>		
   >>>>   Public:
   >>>>                 kstruct_define()
   >>>>                 kstruct_undefine()
   >>>>                 kstruct_check()
   >>>>                 kstruct_free()
   >>>>                 kstruct_compare()
   >>>>                 kstruct_duplicate()
   >>>>                 kstruct_flatten()
   >>>>                 kstruct_unflatten()
   >>>>                 kstruct_setinfo()
   >>>>                 kstruct_getinfo()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"	

typedef struct {
    char  *name;
    size_t size;
    int  (*read_struct)    PROTO((int, kaddr, int));
    int  (*write_struct)   PROTO((int, kaddr, int));
    int  (*compare_struct) PROTO((kaddr, kaddr, int));
    void (*free_struct)    PROTO((kaddr, int, int));
} StructEntry;

klist *struct_list;

/************************************************************
*  
*    Routine Name: kstruct_define - define a structure entry
*         Purpose: This module is used to define a structure entry, which
*		   can then be used by the kread_generic()/kwrite_generic()
*		   routines to read and write arbitary structures.
*           Input: name - the structure name
*		   size - the size of the structure 'sizeof(struct)'
*		   read_struct  - the read routine to call for reading
*		   write_struct - the write routine to call for writing
*		   compare_struct  - the compare routine to call for comparing
*		   free_struct  - the free routine to call for freeing memory
*         Returns: returns the structure data type on success, KUNDEFINED
*		   otherwise
*      Written By: Mark Young
*            Date: Aug 15, 1995
*************************************************************/
int kstruct_define(
   const char *name,
   size_t  size,
   int   (*read_struct) PROTO((
      int,
      kaddr,
      int)),
   int   (*write_struct) PROTO((
      int,
      kaddr,
      int)),
   int   (*compare_struct) PROTO((
      kaddr,
      kaddr,
      int)),
   void  (*free_struct) PROTO((
      kaddr,
      int,
      int)))
{
	int          token;
	StructEntry *entry;

	token = 0 - ((int) kstring_to_token(name));
	if (klist_locate(struct_list, (kaddr) token) != NULL)
	   return(token);

	if ((entry = (StructEntry *) kcalloc((size_t)1, sizeof(StructEntry))) == NULL)
	   return(KUNDEFINED);

	entry->size = size;
	entry->name = kstrdup(name);
	entry->read_struct    = read_struct;
	entry->write_struct   = write_struct;
	entry->compare_struct = compare_struct;
	entry->free_struct    = free_struct;

	struct_list = klist_add(struct_list, (kaddr) token, entry);
	return(token);
}

/************************************************************
*  
*    Routine Name: kstruct_undefine - undefine a structure entry
*         Purpose: This module is used to undefine a structure entry, which
*		   is used by the kread_generic()/kwrite_generic()
*		   routines to read and write arbitary structures.
*           Input: type - the structure type in which to undefine
*	   Output: 
*         Returns: TRUE (1) on success, FALSE (0) otherwise
*      Written By: Mark Young
*            Date: Aug 15, 1995
*************************************************************/
int kstruct_undefine(
   int  type)
{
	klist *list;
	StructEntry *entry;

	if ((list = klist_locate(struct_list, (kaddr) type)) == NULL)
	   return(FALSE);

	entry = (StructEntry *) klist_clientdata(list);
	kfree_and_NULL(entry->name); kfree_and_NULL(entry);
	struct_list = klist_delete(struct_list, (kaddr) type);
	return(TRUE);
}

/************************************************************
*  
*    Routine Name: kstruct_check - check to see if a datatype is defined
*         Purpose: This module is used to check to see if a datatype structure 
*		   is defined.  It takes the name of the structure originally
*		   used to define the entry.  If the structure exists
*		   kstruct_check() returns the structure data type define,
*		   which can then be used by the kread_generic()/
*		   kwrite_generic() routines to read and write arbitary
*		   structures.
*           Input: name - the structure name
*	   Output: 
*         Returns: returns the structure data type on success, KUNDEFINED
*		   otherwise
*      Written By: Mark Young
*            Date: Aug 15, 1995
*************************************************************/
int kstruct_check(
   char *name)
{
	int    token;

	token = 0 - ((int) kstring_to_token(name));
	if (klist_locate(struct_list, (kaddr) token) != NULL)
	   return(token);

	return(KUNDEFINED);
}

/************************************************************
*  
*    Routine Name: kstruct_free - frees a structure and any associated memory
*         Purpose: This module is used to call the free handler associated with
*		   a structure data definition.  The structure free handler is
*		   called to free the assocaited data with the structure.
*           Input: data      - the structure to be freed
*		   type      - the structure type in which to free
*		   free_self - a flag indicating whether or not to
*			       free itself.  This parameter is passed
*			       to the free routine associated with the
*			       type.
*	   Output: 
*         Returns: returns TRUE on success, FALSE otherwise
*		   otherwise
*      Written By: Mark Young
*            Date: Sept 4, 1995
*     Declaration: void kstruct_free(
*     ! kaddr data,
*     ! int   type,
*     ! int   freeself)
*************************************************************/
static void freeit(
   kaddr clientdata)
{
	khash *hash = (khash *) clientdata;
	khash_free(hash, NULL);
}

void kstruct_free(
   kaddr data,
   int   type,
   int   freeself)
{
	khash *hdata;
	int   i, record = FALSE;
	void (*free_struct) PROTO((kaddr, int, int));

	static size_t numptrs = 0;
	static int saveptrs = FALSE;
	static khash *hash  = NULL;
	static kaddr *ptrs  = NULL;
	static int   *types = NULL;
	static int   *fself = NULL;

	/* if the data is NULL then nothing needs to be done */
	if (data == NULL)
	   return;

	if (saveptrs == FALSE)
	{
	   hash = khash_init(KINT, 100);
	   record = saveptrs = TRUE;
	}

	if (khash_clientdata(hash, type, &hdata) == FALSE)
	{
	   hdata = khash_init(KSTRUCT, KLENGTH);
	   khash_add(hash, type, hdata);
	}
	else if (khash_check(hdata, data) == TRUE)
	   return;

	/* keep fast hash list of pointers we've seen */
	khash_add(hdata, data, NULL);

	ptrs  = karray_insert(ptrs, KSTRUCT, data, numptrs, KLIST_TAIL, TRUE);
	fself = karray_insert(fself, KINT, freeself, numptrs, KLIST_TAIL, TRUE);
	types = karray_insert(types, KINT, type, numptrs++, KLIST_TAIL, TRUE);
	if (record == FALSE)
	   return;

	/* retrieve the free structure specified by the type */
	kstruct_getinfo(type, NULL, NULL, NULL, NULL, NULL, &free_struct);
	for (i = 0; i < numptrs; i++)
        {
#if 0
	   /*-- steve [jk] debug sanity check --*/
	   kprintf(": 0x%x : freeself %s : freeing %s \n", 
		   (int) ptrs[i], (fs[i]==TRUE) ? "TRUE " : "FALSE",
		   kdefine_to_datatype(types[i]));
#endif
	   if (type != types[i])
	   {
	      type = types[i];
	      kstruct_getinfo(type, NULL,NULL,NULL,NULL,NULL, &free_struct);
	   }
	   if (free_struct) free_struct(ptrs[i], types[i], fself[i]);
	}

	for (i = 0; i < numptrs; i++)
	   if (types[i] == 0) kfree(ptrs[i]);

	karray_free(ptrs, KSTRUCT, numptrs, NULL), ptrs  = NULL;
	karray_free(types, KINT, numptrs, NULL), types = NULL;
	karray_free(fself, KINT, numptrs, NULL), fself = NULL;
	numptrs  = 0;

	/* free all the pointers now that we're through screwing around */
	khash_free(hash, freeit), hash = NULL;
	saveptrs = FALSE;
}

/************************************************************
*  
*    Routine Name: kstruct_compare - compares two structures 
*         Purpose: This module is used to compare two structures to see if they
*		   are the same.  This is done by calling the associated compare
*		   handler associated with a structure data definition.
*		   The structure compare handler is called to compare the
*		   assocaited data with the two structures.
*           Input: data1 - the first structure to be compared
*		   data2 - the second structure to be compared
*		   type - the structure type in which to be compared
*	   Output: 
*         Returns: returns TRUE on success, FALSE otherwise
*		   otherwise
*      Written By: Mark Young
*            Date: Sept 4, 1995
*************************************************************/
int kstruct_compare(
   kaddr data1,
   kaddr data2,
   int   type)
{
	int (*compare_struct) PROTO((kaddr, kaddr, int));
	int i, status = TRUE, record = FALSE;

	static size_t numptrs = 0;
	static kaddr *ptrs = NULL;
	static int *types = NULL;
	static int saveptrs = FALSE;

	/*
	 *  If either data is NULL then we better not call the compare
	 *  routines.  Besides if both are NULL then they are equal.
	 */
	if (data1 == NULL || data2 == NULL)
	   return((!data1 && !data2) ? TRUE : FALSE);

	if (saveptrs == FALSE)
	   record = saveptrs = TRUE;

	/* looks strange, but we need to know if we have already done this */
	for (i = 0; i < numptrs; i += 2)
	   if (((ptrs[i] == data1 && ptrs[i+1] == data2) ||
	        (ptrs[i] == data2 && ptrs[i+1] == data1)) &&
		types[i] == type && types[i+1] == type)
	      return(status);

	ptrs = karray_insert(ptrs, KSTRUCT, data1, numptrs, KLIST_TAIL, TRUE);
	types = karray_insert(types, KINT, type, numptrs++, KLIST_TAIL, TRUE);
	ptrs = karray_insert(ptrs, KSTRUCT, data2, numptrs, KLIST_TAIL, TRUE);
	types = karray_insert(types, KINT, type, numptrs++, KLIST_TAIL, TRUE);
	if (record == FALSE)
	   return(status);

	/* retrieve the free structure specified by the type */
	kstruct_getinfo(type, NULL, NULL, NULL, NULL, &compare_struct, NULL);
        for (i = 0; i < numptrs; i += 2)
        {
           if (type != types[i])
           {
              type = types[i];
              kstruct_getinfo(type, NULL,NULL,NULL,NULL, &compare_struct, NULL);
           }

           if (compare_struct &&
	       (status = compare_struct(ptrs[i], ptrs[i+1], type)) == FALSE)
	      break;
        }

	if (record == TRUE)
	{
	   karray_free(ptrs,  KSTRUCT, numptrs, NULL);
	   karray_free(types, KINT, numptrs, NULL);
	   ptrs = NULL;
	   types = NULL;
	   numptrs  = 0;
	   saveptrs = FALSE;
	}
	return(status);
}

/************************************************************
*
*    Routine Name: kstruct_duplicate - duplicates a structure
*         Purpose: This module is used to duplicate structure according to
*		   the supplied data type define.
*           Input: data - the structure to be duplicated
*		   type - the structure type in which to duplicate
*	   Output: 
*         Returns: returns the duplicated structure on success, NULL otherwise
*		   otherwise
*      Written By: Mark Young
*            Date: Sept 4, 1995
*************************************************************/
kaddr kstruct_duplicate(
   kaddr data,
   int   type,
   kaddr duplicate)
{
	int   id;

	if ((id = ktmpfid()) == -1)
	   return(NULL);

	if (duplicate != NULL)
	{
	   kwrite_generic(id, data, (size_t)1, type);
	   (void) klseek(id, (off_t) 0, KSEEK_SET);
	   kread_generic(id, duplicate, (size_t)1, type);
	}
	else
	{
	   kwrite_pointer(id, &data, type);
	   (void) klseek(id, (off_t) 0, KSEEK_SET);
	   kread_pointer(id, &duplicate, type);
	}
	kclose(id);
	return(duplicate);
}

/************************************************************
*  
*    Routine Name: kstruct_flatten - flattens a structure
*         Purpose: This module is used to flatten a structure according to
*		   the supplied data type define.
*           Input: data - the structure to be flattened
*		   type - the structure type in which to flatten
*	   Output: size - if not NULL, then returns the size of the flattened
*			  array
*         Returns: returns the flattened structure on success, NULL otherwise
*		   otherwise
*      Written By: Mark Young
*            Date: Sept 4, 1995
*************************************************************/
kaddr kstruct_flatten(
   kaddr   data,
   int     type,
   size_t *size)
{
	int   fid;
	kaddr flatten;

	if ((fid = ktmpfid()) == -1)
	   return(NULL);

	kwrite_pointer(fid, &data, type);
	(void) klseek(fid, (off_t) 0, KSEEK_SET);
	flatten = kfile_readdata(fid, NULL, size);
	kclose(fid);
	return(flatten);
}

/************************************************************
*  
*    Routine Name: kstruct_unflatten - unflattens data into a structure
*         Purpose: This module is used to unflatten data into a structure
*		   according to the supplied data type define.
*           Input: data - the flattened data to be unflattened into a structure
*		   type - the structure type in which to unflatten
*		   size - the size of the flatten data
*	   Output:
*         Returns: returns the unflattened structure on success, NULL otherwise
*		   otherwise
*      Written By: Mark Young
*            Date: Sept 4, 1995
*************************************************************/
kaddr kstruct_unflatten(
   kaddr  data,
   int    type,
   size_t size)
{
	int   fid;
	kaddr unflatten;

	if ((fid = ktmpfid()) == -1)
	   return(NULL);

	kfile_writedata(fid, data, size);
	(void) klseek(fid, (off_t) 0, KSEEK_SET);
	kread_pointer(fid, &unflatten, type);
	kclose(fid);
	return(unflatten);
}

/************************************************************
*  
*    Routine Name: kstruct_setinfo - override the info in a structure entry
*         Purpose: This module is used to override the information currently
*		   stored within a structure entry.
*
*		   The name, size, read and write structure can be overridden.
*		   If a non-NULL name is passed in then the current structure
*		   name is replaced.  If the size is not -1 then the structure
*		   size is replaced.  If the read or write structure routines
*		   are not NULL then the respective routines are replaced.
*           Input: type - the structure type in which to replace certain info
*		   name - the structure name (if not NULL)
*		   size - the size of the structure (if not -1)
*		   read_struct  - the read routine (if not NULL)
*		   write_struct - the write routine (if not NULL)
*		   compare_struct - the compare routine (if not NULL)
*		   free_struct  - the free routine (if not NULL)
*	   Output: 
*         Returns: TRUE (1) on success, FALSE (0) otherwise
*      Written By: Mark Young
*            Date: Aug 15, 1995
*************************************************************/
int kstruct_setinfo(
   int    type,
   const char *name,
   ssize_t  size,
   int   (*read_struct) PROTO((
      int,
      kaddr,
      int)),
   int   (*write_struct) PROTO((
      int,
      kaddr,
      int)),
   int   (*compare_struct) PROTO((
      kaddr,
      kaddr,
      int)),
   void  (*free_struct) PROTO((
      kaddr,
      int, 
      int)))
{
        klist *list;
        StructEntry *entry;
 
        if ((list = klist_locate(struct_list, (kaddr) type)) == NULL)
           return(FALSE);
 
        entry = (StructEntry *) klist_clientdata(list);
	if (size != -1)   entry->size = (size_t) size;
	if (name != NULL) entry->name = kstrdup(name);
	if (read_struct    != NULL) entry->read_struct  = read_struct;
	if (write_struct   != NULL) entry->write_struct = write_struct;
	if (compare_struct != NULL) entry->compare_struct = compare_struct;
	if (free_struct    != NULL) entry->free_struct = free_struct;
	return(TRUE);
}

/************************************************************
*  
*    Routine Name: kstruct_getinfo - retrieve the info in a structure entry
*         Purpose: This module is used to retrieve the information currently
*		   stored within a structure entry.
*
*		   The name, size, read and write structure can be retrieved.
*		   Only non-NULL entries will be returned.  So only pointers
*		   to the desired information need to be passed in.
*           Input: type - the structure type in which to replace certain info
*	   Output: name - the structure name (if not NULL)
*		   size - the size of the structure (if not NULL)
*		   read_struct  - the read routine (if not NULL)
*		   write_struct - the write routine (if not NULL)
*		   compare_struct - the compare routine (if not NULL)
*		   free_struct  - the free routine (if not NULL)
*         Returns: TRUE (1) on success, FALSE (0) otherwise
*      Written By: Mark Young
*            Date: Aug 15, 1995
*************************************************************/
int kstruct_getinfo(
   int     type,
   char   **name,
   size_t  *size,
   int    (**read_struct) PROTO((
      int,
      kaddr,
      int)),
   int    (**write_struct) PROTO((
      int,
      kaddr,
      int)),
   int    (**compare_struct) PROTO((
      kaddr,
      kaddr,
      int)),
   void   (**free_struct) PROTO((
      kaddr,
      int,
      int)))
{
        klist *list;
        StructEntry *entry;
 
	if (size != NULL) *size = 0;
	if (name != NULL) *name = NULL;
	if (read_struct    != NULL) *read_struct    = NULL;
	if (write_struct   != NULL) *write_struct   = NULL;
	if (compare_struct != NULL) *compare_struct = NULL;
	if (free_struct    != NULL) *free_struct    = NULL;
        if ((list = klist_locate(struct_list, (kaddr) type)) == NULL)
           return(FALSE);
 
        entry = (StructEntry *) klist_clientdata(list);
	if (size != NULL) *size = entry->size;
	if (name != NULL) *name = entry->name;
	if (read_struct    != NULL) *read_struct    = entry->read_struct;
	if (write_struct   != NULL) *write_struct   = entry->write_struct;
	if (compare_struct != NULL) *compare_struct = entry->compare_struct;
	if (free_struct    != NULL) *free_struct    = entry->free_struct;
	return(TRUE);
}
