#
# tools-test.pl - library of support routines for the tools testsuites
#

#-----------------------------------------------------------------------
# ktest_initialize() - initialize environment and working directory
#	$testID - a string which identifies the test case
#
# This function initializes the testing environment for the tools.
# We determine a unique working directory, and set this up with a
# custom Toolboxes file, set KHOROS_TOOLBOX
#-----------------------------------------------------------------------
sub ktest_initialize
{
   local($testID) = @_;
   local($TMPDIR);
   local($hostname);
   local(*TOOLBOX, *OLDTB);
   local($ToolboxFile);
   local($tbfile);


   $TMPDIR = $ENV{'TMPDIR'} || '/usr/tmp';

   chop($hostname = `uname -n`);

   #--------------------------------------------------------------------
   #-- create a temporary working directory just for this test
   #--------------------------------------------------------------------
   $WORKDIR = "$TMPDIR/$testID-$hostname";
   mkdir($WORKDIR, 0755) unless -d $WORKDIR;

   #--------------------------------------------------------------------
   #-- create a Toolboxes file in work directory, and copy over the
   #-- contents of whatever is referenced by KHOROS_TOOLBOX
   #--------------------------------------------------------------------
   $ToolboxFile = "$WORKDIR/Toolboxes";
   open(TOOLBOX, "> $ToolboxFile") || do
   {
      die "Can't create Toolboxes file $ToolboxFile: $!\n";
   };

   foreach $tbfile (split(/:/, $ENV{'KHOROS_TOOLBOX'}))
   {
      open(TBFILE, "< $tbfile") || die "Can't read toolbox file $tbfile: $!\n";
      while (<TBFILE>)
      {
	 print TOOLBOX;
      }
      close TBFILE;
   }
   close TOOLBOX;

   #--------------------------------------------------------------------
   #-- now KHOROS_TOOLBOX points to our own little Toolboxes file
   #--------------------------------------------------------------------
   $ENV{'KHOROS_TOOLBOX'} = $ToolboxFile;

   #--------------------------------------------------------------------
   #-- Return the path to the working directory
   #-- All toolboxes and such should be created in here.
   #--------------------------------------------------------------------
   return $WORKDIR;
}

#-----------------------------------------------------------------------
# ktest_end() - mark the end of a test-suite test case
#	$WORKDIR - the path for working directory returned by ktest_initialize
#
# This function is called as the last thing in a test-suite.
# It cleans up behind us, removing the working directory.
#-----------------------------------------------------------------------
sub ktest_end
{
   local($WORKDIR) = @_;


   system("/bin/rm -rf $WORKDIR");
}

#-----------------------------------------------------------------------
# ktest_get_toolboxes() - return a list of the available toolboxes
#
# This function looks at the KHOROS_TOOLBOX environment variable
# and goes through all toolbox files, building a list of toolboxes seen.
# The list of toolboxes is returned.
#-----------------------------------------------------------------------
sub ktest_get_toolboxes
{
   local(@toolbox_files);
   local(@toolboxes);


   @toolbox_files = split(/:/, $ENV{'KHOROS_TOOLBOX'});

   foreach $tbfile (@toolbox_files)
   {
      open(TBFILE, "< $tbfile") || do
      {
	 &ktest_fatal("Unable to read toolbox file $tbfile: $!\n");
	 exit
      };

      while (<TBFILE>)
      {
	 next if /^\s*\#/;
	 push(@toolboxes, $1) if /^([^:]+):/;
      }
      close TBFILE;
   }

   return @toolboxes;
}

#-----------------------------------------------------------------------
# ktest_new_tbname() - return an unused toolbox name
#
# This function is used to pick a name for a new toolbox where
# you don't care what the new toolbox is called.
#-----------------------------------------------------------------------
sub ktest_new_tbname
{
   local(@cur_toolboxes);
   local($tbname);
   local($index);


   @cur_toolboxes = &ktest_get_toolboxes();
   $index = 0;
   do
   {
      $tbname = sprintf("kcmstest%d", $index);
      @foundList = grep(/^$tbname$/i, @cur_toolboxes);
      ++$index;
   } while (@foundList > 0);
   return $tbname;
}

#-----------------------------------------------------------------------
# ktest_fatal() - print an error message and exit with error status
#	$message - the error message to print out
#
# This function prints out the passed error message, then exits with
# an errror status of 1.
#-----------------------------------------------------------------------
sub ktest_fatal
{
   local($message) = @_;

   print $message;
   exit(1);
}

1;
