 /*
  * Khoros: $Id: lkflip.c,v 1.16 1996/07/05 02:10:51 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lkflip.c,v 1.16 1996/07/05 02:10:51 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for kflip
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lkflip
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lkflip - reflect data in object along specified axes
* 
*       Purpose: Library Routine for kflip.  This routine performs
*                data reflection about one, all, or any combination of
*                axes.  The axes supported are: KWIDTH, KHEIGHT, KDEPTH, 
*		 KTIME, and KELEMENTS.
*
*		 Note that kflip only reflects data stored in the
*		 value and mask segments.  If the input or source
*		 object contains a location, mask, or map segment,
*		 data is not reflected.  If it is desired that the
*		 output or destination object contain these segments,
*		 then the calling routine should copy the desired
*		 segments.
*
*		 If the source object has the attribute KPDS_KERNEL_ORIGIN
*		 set to specify a "hot spot" in the data set, then the
*		 kernel origin is flipped along the same specified axes.
*
*         Input: i (src)        - the source data object 
*                num_dim	- No. of dimensions or axes to operate on 
*                dim_list       - list of dimensions or axes on which to operate
*
*        Output: o (dest)       - the destination data object 
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: 
*    Written By: Ramiro Jordan, Becky Bishop
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: January 18, 1996
****************************************************************/
/* -library_def */
int lkflip( kobject src,
            int num_dim, 
            long *dim_list, 
            kobject dest )
/* -library_def_end */

/* -library_code */
{
   kobject src_ref, dest_ref;   /* ref. objs for source and destination */
   char *lib = "kdatamanip";    /* library name (for error reporting) */
   char *rtn = "lkflip";        /* routine name (for error reporting) */

   int val_exists;              /* TRUE if object has VALUE data */
   int mask_exists;             /* TRUE if object has MASK data */

   int kernel_origin_exists;	/* TRUE if object has KPDS_KERNEL_ORIGIN set */
   int w_ko;			/* width coordinate of kernel origin */ 
   int h_ko;			/* height coordinate of kernel origin */
   int d_ko;			/* depth coordinate of kernel origin */
   int t_ko;			/* time coordinate of kernel origin */
   int e_ko;			/* elements coordinate of kernel origin */

   int data_type;               /* source object data type */
   int pdata_type;              /* data type for processing */
   int w, h, d, t, e;           /* number of data elements in each direction */
   int data_length;             /* numer of data points to process */
   int num_regions;             /* number of regions to process */
   int i, loop_count, count;    /* loop counters */
   int region_size[5];

   klist *objlist = NULL;	/* Allocated object list used by KCALL macro */
   
   unsigned char *bdata = NULL, *bval = NULL;
   long *idata = NULL, *ival = NULL;
   double *ddata = NULL, *dval = NULL;
   kdcomplex *cmplx_data = NULL, *cmplx_val = NULL;

  /* Check which data is available for lkflip. */
   val_exists  = kpds_query_value(src);
   mask_exists = kpds_query_mask(src);

  /*------------------------------------------------------------------------
  | Check to see if the source and destination objects passed into
  | the lroutine (lkflip) are valid data objects.  With these checks
  | pushed into the lroutine, the driver routine need not necessarily check
  | the status on the kpds_input and kpds_output calls, though it is a good
  | idea to do it.
  | These checks are provided here to make the lroutine complete in its
  | implementation so that it can be called by other programs.
  | (Reusable code..).
  --------------------------------------------------------------------------*/
   if (src == KOBJECT_INVALID) 
   {
      kerror(lib, rtn, "Source object is not a valid data object.");
      return(FALSE);
   }
   if (dest == KOBJECT_INVALID) 
   {
      kerror(lib, rtn, "Destination object is not a valid data object.");
      return (FALSE);
   }

  /*
   *  Set up a reference to the source object so that the library does
   *  not have any side effects on the source object's attributes.
   */
   KCALL((src_ref = kpds_reference_object(src)) != KOBJECT_INVALID);
   objlist = klist_add(objlist, src_ref, "KOBJECT");

  /*
   * Set up a reference to the destination object to facilitate processing.
   * Do not want to introduce side effects on the destination attributes.
   */
   KCALL((dest_ref = kpds_reference_object(dest)) != KOBJECT_INVALID);
   objlist = klist_add(objlist, dest_ref, "KOBJECT");

  /*
   * If source object has location, map and/or time data the program
   * does nothing.  Reflection is performed in objects that have VALUE
   * and/or MASK data.
   */ 

  /*
   * Process VALUE data.
   */
   if (val_exists) 
   {
     /*-----------------------------------------------------------------------
     | Now get the presentation attributes on the source reference object.
     | These will not be applied to the destination object.
     |
     | The source object may contain a collection of data, thus we get
     | the KPDS_VALUE_SIZE attribute which provides the following information:
     |          - w = number of parameters in the width direction
     |          - h = number of parameters in the height direction 
     |          - d = number of parameters in the depth direction
     |          - t = number of parameters in the time direction 
     |          - e = number of parameters in the elements direction 
     ------------------------------------------------------------------------*/
      KCALL(kpds_get_attribute(src_ref, KPDS_VALUE_SIZE, &w, &h, &d, &t, &e));

      /* 
       * Query to see if the attribute KPDS_KERNEL_ORIGIN has been
       * set and if so, set a boolean variable so that the kernel
       * origin is flipped as well.
       */
      if (kpds_get_attribute(src_ref, 
		KPDS_KERNEL_ORIGIN, &w_ko, &h_ko, &d_ko, &t_ko, &e_ko) == TRUE)
         kernel_origin_exists = TRUE;
      else
         kernel_origin_exists = FALSE;

     /*
      * Obtain data type of source object to be processed.
      *     - data_type = source object data type
      */
      KCALL(kpds_get_attribute(src_ref, KPDS_VALUE_DATA_TYPE, &data_type));

     /*-----------------------------------------------------------------------
     | Select the appropriate common data type for processing the data.
     | Want to minimize any corruption on the data integrity.
     | Recommended data types are:
     |      - KUBYTE = unsigned character 
     |      - KLONG = signed long integer
     |      - KULONG = unsigned long integer
     |      - KDOUBLE = double precision floating point
     |      - KDCOMPLEX = double precision complex
     ------------------------------------------------------------------------*/

      pdata_type = kdatatype_cast_process(data_type,data_type,
                   (KUBYTE | KLONG | KULONG | KDOUBLE | KDCOMPLEX));

     /*-----------------------------------------------------------------------
     | If data is byte representable then process it as unsigned char
     | (KUBYTE).  If data is integer representable then process it as
     | long int (KLONG).  If data is float representable then process
     | it as double float (KDOUBLE).  If data is COMPLEX representable
     | then process it as double complex (KDCOMPLEX).  In the destination
     | object the data type will be automatically casted to the appropriate
     | data type.
     ------------------------------------------------------------------------*/
      if ( (pdata_type == KUBYTE) || (pdata_type == KLONG) ||
           (pdata_type == KULONG) || (pdata_type == KDOUBLE) ||
           (pdata_type == KDCOMPLEX) ) 
      {
         KCALL(kpds_set_attribute(src_ref, KPDS_VALUE_DATA_TYPE, pdata_type));

         KCALL(kpds_set_attribute(dest_ref, KPDS_VALUE_DATA_TYPE, pdata_type));

         kinfo(KVERBOSE,"Input   data type is: %d\n",data_type);
         kinfo(KVERBOSE,"Process data type is: %d\n",pdata_type);
      }

      for (loop_count = 0; loop_count < num_dim; loop_count++) 
      {
          region_size[0] = region_size[1] = region_size[2] = 1;
          region_size[3] = region_size[4] = 1;
          if (loop_count == 1) 
          {
             KCALL((src_ref = kpds_reference_object(dest_ref)) 
			!= KOBJECT_INVALID);
             objlist = klist_add(objlist, src_ref, "KOBJECT");
          }
          switch (dim_list[loop_count])
          {
                 case KWIDTH:
                                region_size[0] = w;
                                break;
                 case KHEIGHT:
                                region_size[1] = h;
                                break;
                 case KDEPTH:
                                region_size[2] = d;
                                break;
                 case KTIME:
                                region_size[3] = t;
                                break;
                 case KELEMENTS:
                                region_size[4] = e;
                                break;
          }
          KCALL(kpds_set_attributes(src_ref,KPDS_VALUE_REGION_SIZE,
                             region_size[0],region_size[1],
                             region_size[2],region_size[3],
                             region_size[4],KPDS_VALUE_POSITION,
                             0,0,0,0,0,NULL));
          KCALL(kpds_set_attributes(dest_ref,KPDS_VALUE_REGION_SIZE,
                             region_size[0],region_size[1],
                             region_size[2],region_size[3],
                             region_size[4],KPDS_VALUE_POSITION,
                             0,0,0,0,0,NULL));
          data_length = region_size[0]*region_size[1]*region_size[2]*
                        region_size[3]*region_size[4];
          num_regions = (w*h*d*t*e)/data_length;
          for (count=0; count < num_regions; count++) 
          {
            /*-----------------------------------------------------------------
            | Process KUBYTE data
            -----------------------------------------------------------------*/
             if ( (pdata_type == KUBYTE) ) 
             {
                KCALL(!((bdata = (unsigned char *) (kpds_get_data(src_ref,
                                 KPDS_VALUE_REGION, bdata))) == NULL));
                if (bval == NULL) 
                {
                   KCALL(!((bval = (unsigned char *)
                  kmalloc((size_t)data_length*sizeof(unsigned char))) == NULL));
	           objlist = klist_add(objlist, bval, "KMALLOC");
                }

                for (i=0; i<data_length; i++) 
		{
                  bval[i] = bdata[data_length-1-i];
                }

                KCALL(kpds_put_data(dest_ref, KPDS_VALUE_REGION, bval));
             }
            /*-----------------------------------------------------------------
            | Process KLONG and KULONG data.
            -----------------------------------------------------------------*/
             if ( (pdata_type == KLONG)  || (pdata_type == KULONG) ) 
	     {
                KCALL(!((idata = (long *) (kpds_get_data(src_ref,
                                 KPDS_VALUE_REGION, idata))) == NULL));

                if (ival == NULL) 
		{
                   KCALL(!((ival = (long *)
                         kmalloc((size_t)data_length*sizeof(long))) == NULL));
		   objlist = klist_add(objlist, ival, "KMALLOC");
                }

                for (i=0; i<data_length; i++) 
		{
                  ival[i] = idata[data_length-1-i];
                }

                KCALL(kpds_put_data(dest_ref, KPDS_VALUE_REGION, ival));
             }
            /*-----------------------------------------------------------------
            | Process KDOUBLE data.
            -----------------------------------------------------------------*/
             if ( (pdata_type == KDOUBLE) ) 
	     {
                KCALL(!((ddata = (double *) (kpds_get_data(src_ref, 
			           KPDS_VALUE_REGION, ddata))) == NULL));

		if (dval == NULL) 
		{
		   KCALL(!((dval = (double *)
		         kmalloc((size_t)data_length*sizeof(double))) == NULL));
		   objlist = klist_add(objlist, dval, "KMALLOC");
		}

                for (i=0; i<data_length; i++) 
                  dval[i] = ddata[data_length-1-i];

                KCALL(kpds_put_data(dest_ref, KPDS_VALUE_REGION, dval));
             }
            /*-----------------------------------------------------------------
            | Process KDCOMPLEX data.
            -----------------------------------------------------------------*/
             if ( (pdata_type == KDCOMPLEX) ) 
	     {
                KCALL(!((cmplx_data = (kdcomplex *) (kpds_get_data(src_ref,
                                    KPDS_VALUE_REGION, cmplx_data))) == NULL));

		if (cmplx_val == NULL) 
		{
		   KCALL(!((cmplx_val = (kdcomplex *)
                     kmalloc((size_t)data_length*sizeof(kdcomplex))) == NULL));
		   objlist = klist_add(objlist, cmplx_val, "KMALLOC");
                }

                for (i=0; i<data_length; i++) 
                  cmplx_val[i] = cmplx_data[data_length-1-i];

                KCALL(kpds_put_data(dest_ref, KPDS_VALUE_REGION, cmplx_val));
             }

          }
          if (bdata) kfree_and_NULL(bdata);
          if (bval) kfree_and_NULL(bval);
          if (idata) kfree_and_NULL(idata);
          if (ival) kfree_and_NULL(ival);
          if (ddata) kfree_and_NULL(ddata);
          if (dval) kfree_and_NULL(dval);
          if (cmplx_data) kfree_and_NULL(cmplx_data);
          if (cmplx_val) kfree_and_NULL(cmplx_val);
      }

      if (kernel_origin_exists)
      {
         for (loop_count = 0; loop_count < num_dim; loop_count++) 
         {
	    switch(dim_list[loop_count])
            {
               case KWIDTH:     w_ko = w_ko;
	   		        h_ko = w - (h_ko + 1);
				d_ko = d_ko;
				t_ko = t_ko;
				e_ko = e_ko;
                                break;

               case KHEIGHT:    w_ko = h - (w_ko + 1);
				h_ko = h_ko;
				d_ko = d_ko;
				t_ko = t_ko;
				e_ko = e_ko;
                                break;

               case KDEPTH:     w_ko = w_ko;
				h_ko = h_ko;
				d_ko = d - (d_ko + 1);
				t_ko = t_ko;
				e_ko = e_ko;
                                break;

               case KTIME:      w_ko = w_ko;
				h_ko = h_ko;
				d_ko = d_ko;
				t_ko = t - (t_ko + 1);
				e_ko = e_ko;
                                break;

               case KELEMENTS:  w_ko = w_ko;
                                h_ko = h_ko;
                                d_ko = d_ko;
                                t_ko = t_ko;
                                e_ko = e - (e_ko + 1);
                                break;
	
	       default:	        kerror(lib, rtn, 
			                    "Invalid flip axis specified");
			        return(FALSE);
            }
         }
	 KCALL(kpds_set_attribute(dest_ref, KPDS_KERNEL_ORIGIN,
				  w_ko, h_ko, d_ko, t_ko, e_ko));
      }

   } /* end processing for VALUE data */

  /*--------------------------------------------------------------------------
  | Process MASK data.
  --------------------------------------------------------------------------*/
   if (mask_exists) 
   {
     /*-----------------------------------------------------------------------
     | Now set the presentation attributes on the source reference object.
     | These will not be applied to the destination object.
     |
     | The source object may contain a collection of images, thus we get
     | the KPDS_MASK_SIZE attribute which provides the following information:
     |          - w = number of parameters in the width direction
     |          - h = number of parameters in the height direction 
     |          - d = number of parameters in the depth direction
     |          - t = number of parameters in the time direction 
     |          - e = number of parameters in the elements direction 
     ----------------------------------------------------------------------*/
      KCALL(kpds_get_attribute(src_ref, KPDS_MASK_SIZE, &w, &h, &d, &t, &e));

     /*-----------------------------------------------------------------------
     | Obtain data type of source object to be processed.
     |     - data_type = source object data type
     ----------------------------------------------------------------------*/
      KCALL(kpds_get_attribute(src_ref, KPDS_MASK_DATA_TYPE, &data_type));

     /*-----------------------------------------------------------------------
     | Select the appropriate common data type for processing the data.
     | Want to minimize any corruption on the data integrity.
     | Recommended data types are:
     |      - KUBYTE = unsigned character 
     |      - KLONG = signed long integer
     |      - KULONG = unsigned long integer
     |      - KDOUBLE = double precision floating point
     |      - KDCOMPLEX = double precision complex
     ----------------------------------------------------------------------*/

      pdata_type = kdatatype_cast_process(data_type,data_type,
                   (KUBYTE | KLONG | KULONG | KDOUBLE | KDCOMPLEX) );

     /*-----------------------------------------------------------------------
     | If data is byte representable then process it as unsigned char
     | (KUBYTE).  If data is integer representable then process it as
     | long int (KLONG).  If data is float representable then process
     | it as double float (KDOUBLE).  If data is COMPLEX representable
     | then process it as double complex (KDCOMPLEX).  In the destination
     | object the data type will be automatically casted to the appropriate
     | data type.
     ----------------------------------------------------------------------*/
      if ( (pdata_type == KUBYTE) || (pdata_type == KLONG) ||
           (pdata_type == KULONG) || (pdata_type == KDOUBLE) ||
           (pdata_type == KDCOMPLEX) ) 
      {
         KCALL(kpds_set_attribute(src_ref, KPDS_MASK_DATA_TYPE, pdata_type));

         KCALL(kpds_set_attribute(dest_ref, KPDS_MASK_DATA_TYPE, pdata_type));
      }

      for (loop_count = 0; loop_count < num_dim; loop_count++) 
      {
          region_size[0] = region_size[1] = region_size[2] = 1;
          region_size[3] = region_size[4] = 1;
          if (loop_count == 1) 
	  {
             KCALL((src_ref = kpds_reference_object(dest_ref)) 
		        != KOBJECT_INVALID) 
             objlist = klist_add(objlist, src_ref, "KOBJECT");
          }

          switch (dim_list[loop_count])
          {
                 case KWIDTH:
                                region_size[0] = w;
                                break;
                 case KHEIGHT:
                                region_size[1] = h;
                                break;
                 case KDEPTH:
                                region_size[2] = d;
                                break;
                 case KTIME:
                                region_size[3] = t;
                                break;
                 case KELEMENTS:
                                region_size[4] = e;
                                break;
          }

          KCALL(kpds_set_attributes(src_ref,KPDS_MASK_REGION_SIZE,
                             region_size[0],region_size[1],
                             region_size[2],region_size[3],
                             region_size[4],KPDS_MASK_POSITION,
                             0,0,0,0,0,NULL));
          KCALL(kpds_set_attributes(dest_ref,KPDS_MASK_REGION_SIZE,
                             region_size[0],region_size[1],
                             region_size[2],region_size[3],
                             region_size[4],KPDS_MASK_POSITION,
                             0,0,0,0,0,NULL));
          data_length = region_size[0]*region_size[1]*region_size[2]*
                        region_size[3]*region_size[4];
          num_regions = (w*h*d*t*e)/data_length;
          for (count=0; count < num_regions; count++) {
            /*-----------------------------------------------------------------
            | Process KUBYTE data.
            -----------------------------------------------------------------*/
             if ( (pdata_type == KUBYTE) ) 
	     {
                KCALL(!((bdata = (unsigned char *) (kpds_get_data(src_ref,
                                 KPDS_MASK_REGION, bdata))) == NULL));
                if (bval == NULL) 
		{
                   KCALL(!((bval = (unsigned char *)
                        kmalloc((size_t)data_length*sizeof(unsigned char))) == NULL));
		   objlist = klist_add(objlist, bval, "KMALLOC");
                }

                for (i=0; i<data_length; i++) 
                  bval[i] = bdata[data_length-1-i];

                KCALL(kpds_put_data(dest_ref, KPDS_MASK_REGION, bval));
             }
            /*-----------------------------------------------------------------
            | Process KLONG and KULONG data.
            -----------------------------------------------------------------*/
             if ( (pdata_type == KLONG)  || (pdata_type == KULONG) ) 
	     {
                KCALL(!((idata = (long *) (kpds_get_data(src_ref, 
			         KPDS_MASK_REGION, idata))) == NULL));

                if (ival == NULL) 
		{
                   KCALL(!((ival = (long *)
			 kmalloc((size_t)data_length*sizeof(long))) == NULL));
		   objlist = klist_add(objlist, ival, "KMALLOC");
                }
                for (i=0; i<data_length; i++) 
                  ival[i] = idata[data_length-1-i];

                KCALL(kpds_put_data(dest_ref, KPDS_MASK_REGION, ival));
             }
            /*-----------------------------------------------------------------
            | Process KDOUBLE data.
            -----------------------------------------------------------------*/
             if ( (pdata_type == KDOUBLE) ) 
	     {
                KCALL(!((ddata = (double *) (kpds_get_data(src_ref, 
		                 KPDS_MASK_REGION, ddata))) == NULL));

                if (dval == NULL) 
		{
                   KCALL(!((dval = (double *)
			 kmalloc((size_t)data_length*sizeof(double))) == NULL));
		   objlist = klist_add(objlist, dval, "KMALLOC");
		}

                for (i=0; i<data_length; i++) 
                  dval[i] = ddata[data_length-1-i];

                KCALL(kpds_put_data(dest_ref, KPDS_MASK_REGION, dval));
             }
            /*-----------------------------------------------------------------
            | Process KDCOMPLEX data.
            -----------------------------------------------------------------*/
             if ( (pdata_type == KDCOMPLEX) ) 
	     {
                KCALL(!((cmplx_data = (kdcomplex *) (kpds_get_data(src_ref,
                                     KPDS_MASK_REGION, cmplx_data))) == NULL));

                if (cmplx_val == NULL) 
		{
                   KCALL(!((cmplx_val = (kdcomplex *)
                         kmalloc((size_t)data_length*sizeof(kdcomplex))) == NULL));
		   objlist = klist_add(objlist, cmplx_val, "KMALLOC");
                }

                for (i=0; i<data_length; i++) 
                  cmplx_val[i] = cmplx_data[data_length-1-i];

                KCALL(kpds_put_data(dest_ref, KPDS_MASK_REGION, cmplx_val));
             }

          }
          if (bdata) kfree_and_NULL(bdata);
          if (bval) kfree_and_NULL(bval);
          if (idata) kfree_and_NULL(idata);
          if (ival) kfree_and_NULL(ival);
          if (ddata) kfree_and_NULL(ddata);
          if (dval) kfree_and_NULL(dval);
          if (cmplx_data) kfree_and_NULL(cmplx_data);
          if (cmplx_val) kfree_and_NULL(cmplx_val);
      }

   } /* end processing for MASK data */
  
   (void) kpds_close_object(src_ref); 
   (void) kpds_close_object(dest_ref); 

   return (TRUE);
}
/* -library_code_end */
