 /*
  * Khoros: $Id: lkhisto.c,v 1.15 1996/07/05 02:11:01 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lkhisto.c,v 1.15 1996/07/05 02:11:01 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for khisto
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lkhisto
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lkhisto - compute histograms for data objects
* 
*       Purpose: lkhisto computes histogram information for data objects.
*                The unit of data to be used to compute each histogram can be
*                selected to lie along any combination of the axes 5D data 
*                model.
*
*                For example, it is possible to compute a histogram for each
*                HxD plane in a data set comprised of a sequence of multiband
*                volumes.
*
*		 If the input data object has a map, the input value data
*		 is optionally pulled through the input map before computing the
*		 histogram.
*
*                Output histograms are stored as columns in the output object,
*                one column per unit of data.
*
*                The histogram data type is UNSIGNED LONG.
*
*                If a gating object is supplied, then only input data that has
*                a corresponding non-zero value in the gating object is
*                histogrammed.
*
*                The out-of-bounds flags can be used to catch any counts that
*                lie outside of the histogram array.
*
*         Input: in_obj - object to be histogrammed
*                gate_obj - optional gate object (ROI)
*		 w,h,d,t,e - flags that, if set, indicate
*			that the processing unit for independent
*			histograms should include the given axis
*	         min - minimum side of most negative bin
*	         binwidth - width of each bin
*	         bins - number of bins
*	         oob - out-of-bounds flag; append two extra bins
*			   to the histogram to count those values that
*			   fall out of the histogram bounds
*		 map - if the input object has a map, then pull the data through the map (1) or leave the data unmapped (0)
*	         normalize - normalize the integrated histogram to [0..1] (1) or just leave it as computed (0)
*
*        Output: kobject out_obj - contains histograms for the
*                       processing units in the data
*                kobject integ_obj - contains integrated histograms (NULL if integrated histograms are not desired)
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: 
*    Written By: Scott Wilson
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lkhisto(
  kobject in_obj,
  kobject gate_obj,
  int w, int h, int d, int t, int e,
  double min,
  double binwidth,
  int bins,
  int oob,
  int map,
  int normalize,
  kobject out_obj,
  kobject integ_obj)
/* -library_def_end */

/* -library_code */
{
        char *lib = "kdatamanip", *rtn = "lkhisto";
	klist *objlist=NULL;
	unsigned int i;
	unsigned long *hist; /* Histogram array pointer */
	double *integ=NULL;  /* Integrated histogram array pointer */
        double accum;        /* Integration accumulator */
	unsigned wc,hc,dc,tc,ec; /* Dimensions of data set */
	unsigned int suw,suh,sud,sut,sue,junk; /* Dimensions of subunit */
	unsigned int uw,uh,ud,ut,ue;           /* Dimensions of unit */
        unsigned int nuw,nuh,nud,nut,nue; /* Number of units in 
                                             each direction */
        unsigned int nsw,nsh,nsd,nst,nse; /* Number of subunits per unit in 
                                             each direction*/
	double *b=NULL,*bmask=NULL,*bgate=NULL; /* Pointers for data */
	int n;
        unsigned int type,hnum;
	unsigned int numptsunit,numptssub;
	unsigned int numu_w,numu_h,numu_d,numu_t,numu_e; /* Unit counters */
	unsigned int nums_w,nums_h,nums_d,nums_t,nums_e; /* Subunit counters */

	unsigned int upos_w,upos_h,upos_d,upos_t,upos_e; /* Unit pos'n */
	unsigned int spos_w,spos_h,spos_d,spos_t,spos_e; /* Subunit pos'n */

        unsigned int map_exists,mask_exists;

	unsigned int wcon,hcon,dcon,tcon,econ; /* Optimal region constraints */
	unsigned int wflag,hflag,dflag,tflag,eflag; /* Optimal region flags */

        /* Make sure we have valid objects */
        if (in_obj == KOBJECT_INVALID || out_obj == KOBJECT_INVALID)
          {
            kerror(lib, rtn, "Bogus input or output");
            return(FALSE);
          }

        /* Check out the histogram params to make sure they mean something. */
	if (bins <= 1 || binwidth <= 0)
          {
            kerror(lib, rtn, "Bogus histogram parameters");
            return(FALSE);
          }

        /* Reference the input object to avoid side effects, then add the
           reference to the list of goodies to be autmatically free'd on
           error. */
        KCALL((in_obj = kpds_reference_object(in_obj)) != KOBJECT_INVALID);
	objlist = klist_add(objlist,in_obj,"KOBJECT");

        mask_exists=kpds_query_mask(in_obj);

        map_exists =kpds_query_map(in_obj);
        /* If map exists, optionally pull all the data thru the map. */
        if (map_exists)
          {
            if (map) KCALL(kpds_set_attribute(in_obj,KPDS_MAPPING_MODE,
                                                     KMAPPED));
          }

        /* Constrain the input data to double real for reading */
	KCALL(kpds_set_attributes(in_obj,
                                 KPDS_VALUE_DATA_TYPE, KDOUBLE,
                                 KPDS_VALUE_COMPLEX_CONVERT, KREAL, NULL));
        if (mask_exists) KCALL(kpds_set_attribute(in_obj,
                                 KPDS_MASK_DATA_TYPE,KDOUBLE));
 
	/* If gating object is supplied, reference it and SET it for KMAPPED 
           and KDOUBLE presentation. Also add the reference to the autofree 
           list. */
        if (gate_obj)
          {
            KCALL((gate_obj = kpds_reference_object(gate_obj)) 
		!= KOBJECT_INVALID); 
	    objlist = klist_add(objlist,gate_obj,"KOBJECT");
            if (kpds_query_mask(gate_obj))
              {
                kerror(lib, rtn, "Gating object cannot contain mask data.");
                (void) kpds_close_object(in_obj);
                (void) kpds_close_object(gate_obj);
                return(FALSE);
              }
            if (kpds_query_map(gate_obj)) 
                KCALL(kpds_set_attribute(gate_obj,KPDS_MAPPING_MODE,KMAPPED));
            KCALL(kpds_set_attributes(gate_obj,
                                 KPDS_VALUE_DATA_TYPE, KDOUBLE,
                                 KPDS_VALUE_COMPLEX_CONVERT, KREAL, NULL));
          }

        /* See how big the data is */
	KCALL(kpds_get_attributes(in_obj,
                                 KPDS_VALUE_SIZE,&wc,&hc,&dc,&tc,&ec,
                                 KPDS_VALUE_DATA_TYPE,&type,NULL));

        /* If NO flags are set, set to WxH for default histogramming unit */
        if (w+h+d+t+e == 0) w=h=1;

        /* What we need to do is first figure out how many histogram units 
           there are. This is determined only by the size of the data and the 
           unit directions specified by the user. While we're at it, get
           the dimensions of each unit and the number of such units in
           each direction of the dataset. */
        numptsunit = 1;
        numptsunit *= (uw = w ? wc : 1);
        numptsunit *= (uh = h ? hc : 1);
        numptsunit *= (ud = d ? dc : 1);
        numptsunit *= (ue = t ? tc : 1);
        numptsunit *= (ut = e ? ec : 1);
        nuw = w ? 1 : wc;
        nuh = h ? 1 : hc;
        nud = d ? 1 : dc;
        nut = t ? 1 : tc;
        nue = e ? 1 : ec;

        /* Now we need to figure out the size and number of "optimal regions"
           that form each histogram unit. We'll use the constrained
           optimal region size attribute to do this.

           First, set up constraints such that the optimal region will be of 
           size 1 on each axis. */
        wcon  = hcon  = dcon  = tcon  = econ  = 1;
        wflag = hflag = dflag = tflag = eflag = KOPTIMAL_FIXED_VALUE;
        /* Now remove the constraint for each axis that forms a 
           histogram unit. */
        if (w) wflag = KOPTIMAL_FACTOR;
        if (h) hflag = KOPTIMAL_FACTOR;
        if (d) dflag = KOPTIMAL_FACTOR;
        if (t) tflag = KOPTIMAL_FACTOR;
        if (e) eflag = KOPTIMAL_FACTOR;

        /* Use the constraint values and flags to pull back the optimal region 
           size for a subunit. */
	KCALL(kpds_set_attribute(in_obj,KPDS_VALUE_OPTIMAL_REGION_FLAGS,
             wflag,hflag,dflag,tflag,eflag));
	KCALL(kpds_set_attribute(in_obj,KPDS_VALUE_OPTIMAL_REGION_CONSTRAINT,
              wcon,hcon,dcon,tcon,econ));
        KCALL(kpds_get_attribute(in_obj,
              KPDS_VALUE_OPTIMAL_REGION_SIZE,&suw,&suh,&sud,&sut,&sue,&junk));

        /* Compute number of points per subunit, number of subunits per
           unit, and the number of subunits per unit. */
        nsw = uw/suw;
        nsh = uh/suh;
        nsd = ud/sud;
        nst = ut/sut;
        nse = ue/sue;
        numptssub = suw*suh*sud*sut*sue;

        /* Allocate the histogramming array and add it to the autofree list */
        hnum = bins;			/* Just the histogram */
        if (oob) hnum += 2;		/* Add out-of-bounds bins */
        KCALL(!((hist = (unsigned long *)kmalloc((size_t)hnum*sizeof(unsigned long))) == NULL));
	objlist = klist_add(objlist,hist,"KMALLOC");
        if (integ_obj != KOBJECT_INVALID)
          {
            KCALL(!((integ = (double *)kmalloc((size_t)hnum*sizeof(double))) == NULL));
	    objlist = klist_add(objlist,integ,"KMALLOC");
          }

        /* Allocate space for the various data buffers */
        KCALL(!((b     = (double *)kmalloc((size_t)(suw*suh*sud*sut*sue*sizeof(double)))) == NULL));
        KCALL(!((bmask = (double *)kmalloc((size_t)(suw*suh*sud*sut*sue*sizeof(double)))) == NULL));
        KCALL(!((bgate = (double *)kmalloc((size_t)(suw*suh*sud*sut*sue*sizeof(double)))) == NULL));
	objlist = klist_add(objlist,b,"KMALLOC");
	objlist = klist_add(objlist,bmask,"KMALLOC");
	objlist = klist_add(objlist,bgate,"KMALLOC");

        /* Create the output value segment if it's not there already and
           set it's size and data type attributes.  */
        if (!kpds_query_value(out_obj)) KCALL(kpds_create_value(out_obj));
        KCALL(kpds_set_attributes(out_obj,
                                 KPDS_VALUE_SIZE,nuw*nuh*nud*nut*nue,hnum,1,1,1,
                                 KPDS_VALUE_DATA_TYPE,KULONG,NULL));

        /* Kill the output object mask,map,location,time segs if they exist */
        if (kpds_query_mask(out_obj))     KCALL(kpds_destroy_mask(out_obj));
        if (kpds_query_map(out_obj))      KCALL(kpds_destroy_map(out_obj));
        if (kpds_query_location(out_obj)) KCALL(kpds_destroy_location(out_obj));
        if (kpds_query_time(out_obj))     KCALL(kpds_destroy_time(out_obj));

        /* Do the same for the optional integrated histo output object */
        if (integ_obj != KOBJECT_INVALID)
          {
            if (!kpds_query_value(integ_obj)) KCALL(kpds_create_value(integ_obj));
            KCALL(kpds_set_attributes(integ_obj,
                                 KPDS_VALUE_SIZE,nuw*nuh*nud*nut*nue,hnum,1,1,1,
                                 KPDS_VALUE_DATA_TYPE,KDOUBLE,NULL));
            if (kpds_query_mask(integ_obj))     KCALL(kpds_destroy_mask(integ_obj));
            if (kpds_query_map(integ_obj))      KCALL(kpds_destroy_map(integ_obj));
            if (kpds_query_location(integ_obj)) KCALL(kpds_destroy_location(integ_obj));
            if (kpds_query_time(integ_obj))     KCALL(kpds_destroy_time(integ_obj));
          }

        /* Ref the output objects to avoid side effects and add them to the
           autofree list */
        KCALL((out_obj = kpds_reference_object(out_obj)) != KOBJECT_INVALID);
	objlist = klist_add(objlist,out_obj,"KOBJECT");
        if (integ_obj != KOBJECT_INVALID)
          {
            KCALL((integ_obj = kpds_reference_object(integ_obj)) != KOBJECT_INVALID);
	    objlist = klist_add(objlist,integ_obj,"KOBJECT");
          }

        /* Set the subunit sizes on the input, gate, and output objects */
        KCALL(kpds_set_attribute(in_obj,
                                  KPDS_VALUE_REGION_SIZE,suw,suh,sud,sut,sue));
        if (gate_obj) KCALL(kpds_set_attribute(gate_obj,
                                  KPDS_VALUE_REGION_SIZE,suw,suh,sud,sut,sue));
        KCALL(kpds_set_attribute(out_obj,
                                  KPDS_VALUE_REGION_SIZE,1,hnum,1,1,1));
        if (integ_obj != KOBJECT_INVALID)
          {
            KCALL(kpds_set_attribute(integ_obj,
                                  KPDS_VALUE_REGION_SIZE,1,hnum,1,1,1));
          }

        /* If a mask is in place set the region size attributes appropriately */
	if (mask_exists) KCALL(kpds_set_attribute(in_obj,
                                   KPDS_MASK_REGION_SIZE,suw,suh,sud,sut,sue));

        /* Handle each unit in WHDTE block order */
	for (numu_e=0; numu_e<nue; numu_e++)
          {
	    for (numu_t=0; numu_t<nut; numu_t++)
              {
	        for (numu_d=0; numu_d<nud; numu_d++)
                  {
	            for (numu_h=0; numu_h<nuh; numu_h++)
                      {
	                for (numu_w=0; numu_w<nuw; numu_w++)
                          {
                            /* Zero the histo array */
	                    (void)kbzero((kaddr)hist,(size_t)hnum*sizeof(unsigned long));

                            /* Compute the offset into the data of the origin of each unit */
                            upos_w = numu_w*uw;
                            upos_h = numu_h*uh;
                            upos_d = numu_d*ud;
                            upos_t = numu_t*ut;
                            upos_e = numu_e*ue;
 
                            /* Step through each of the subunits in WHDTE block order */
	                    for (nums_e=0; nums_e<nse; nums_e++)
                              {
	                        for (nums_t=0; nums_t<nst; nums_t++)
                                  {
	                            for (nums_d=0; nums_d<nsd; nums_d++)
                                      {
	                                for (nums_h=0; nums_h<nsh; nums_h++)
                                          {
	                                    for (nums_w=0; nums_w<nsw; nums_w++)
                                              {
                                                /* Compute the offset into the data of the origin of each subunit,
                                                   a relative computation that used the unit position as a reference. */
                                                spos_w = upos_w + nums_w*suw;
                                                spos_h = upos_h + nums_h*suh;
                                                spos_d = upos_d + nums_d*sud;
                                                spos_t = upos_t + nums_t*sut;
                                                spos_e = upos_e + nums_e*sue;

                                                /* Set the position on the input object to the location of the subunit
                                                   origin. This guarantees that we get the *right* subunit, not simply
                                                   the next one that the autoincrement code thinks we should get (which
                                                   is normally the next one along the lowest index order. */ 
                                                KCALL(kpds_set_attribute(in_obj,KPDS_VALUE_POSITION,spos_w,spos_h,spos_d,spos_t,spos_e));
                                                if (gate_obj)
                                                  KCALL(kpds_set_attribute(gate_obj,KPDS_VALUE_POSITION,spos_w,spos_h,spos_d,spos_t,spos_e));
                                                if (mask_exists)
                                                  KCALL(kpds_set_attribute(in_obj,KPDS_MASK_POSITION,spos_w,spos_h,spos_d,spos_t,spos_e));

                                                /* Grab a region of goodies from the input object */
	                                        KCALL(kpds_get_data(in_obj,KPDS_VALUE_REGION, (kaddr)b));
	                                        if (gate_obj) KCALL(kpds_get_data(gate_obj,KPDS_VALUE_REGION, (kaddr)bgate));
                                                if (mask_exists) KCALL(kpds_get_data(in_obj,KPDS_MASK_REGION,(kaddr)bmask));

                                                /* Form the histogram */
                                                if (!mask_exists && gate_obj==NULL)
                                                  {
                                                    for (i=0; i<numptssub; i++)
                                                      {
                                                        n = (b[i]-min)/binwidth;
                                                        if (b[i] == min+binwidth*bins) n = bins-1;
                                                        if (n>=0 && n<bins) hist[n]++;
		                                        if (oob)
                                                          { /* If out-of-bounds bins requested */
                                                            if (n < 0) hist[hnum-2]++;
                                                            if (n >= bins) hist[hnum-1]++;
                                                          }
                                                      }
                                                  }
                                                else if (mask_exists && gate_obj==NULL)
                                                  {
                                                    for (i=0; i<numptssub; i++)
                                                      {
                                                        if (bmask[i] != 0)
                                                          {
                                                            n = (b[i]-min)/binwidth;
                                                            if (b[i] == min+binwidth*bins) n = bins-1;
                                                            if (n >= 0 && n <bins) hist[n]++;
		                                            if (oob)
                                                              { /* If out-of-bounds bins requested */
                                                                if (n < 0) hist[hnum-2]++;
                                                                if (n >= bins) hist[hnum-1]++;
                                                              }
                                                          }
                                                      }
                                                  }
                                                else if (!mask_exists && gate_obj)
                                                  {
                                                    for (i=0; i<numptssub; i++)
                                                      {
                                                        if (bgate[i] != 0)
                                                          {
                                                            n = (b[i]-min)/binwidth;
                                                            if (b[i] == min+binwidth*bins) n = bins-1;
                                                            if (n >= 0 && n <bins) hist[n]++;
                                                            if (oob)
                                                              { /* If out-of-bounds bins requested */
                                                                if (n < 0) hist[hnum-2]++;
                                                                if (n >= bins) hist[hnum-1]++;
                                                              }
                                                          }
                                                      }
                                                  }
                                                else if (mask_exists && gate_obj)
                                                  {
                                                    for (i=0; i<numptssub; i++)
                                                      {
                                                        if (bmask[i] != 0 && bgate[i] != 0)
                                                          {
                                                            n = (b[i]-min)/binwidth;
                                                            if (b[i] == min+binwidth*bins) n = bins-1;
                                                            if (n >= 0 && n <bins) hist[n]++;
                                                            if (oob)
                                                              { /* If out-of-bounds bins requested */
                                                                if (n < 0) hist[hnum-2]++;
                                                                if (n >= bins) hist[hnum-1]++;
                                                              }
                                                          }
                                                      }
                                                  } /* End if compound if statement */
                                              }
                                          }
                                      }
                                  }
                              } /* End of subunit loops */

                            /* Store the resulting histogram */
                            KCALL(kpds_put_data(out_obj, KPDS_VALUE_REGION, (kaddr)hist));

                            /* If the integrated histogram object is requested, integrate the
                               histogram (and optionally normalize it to [0...1]) and store it. */
                            if (integ_obj != KOBJECT_INVALID)
                              {
	                        (void)kbzero((kaddr)integ,(size_t)hnum*sizeof(double));
                                accum = 0.0;
                                for (i=0; i<bins; i++)
                                  {
                                    accum += hist[i];
                                    integ[i] = accum;
                                  }
                                if (oob)
                                  {
                                    accum += hist[hnum-2];
                                    integ[hnum-2] = accum;
                                    accum += hist[hnum-1];
                                    integ[hnum-1] = accum;
                                  }
                                if (normalize)
                                  {
                                    for (i=0; i<bins; i++) integ[i] /= (double)accum;
                                    if (oob)
                                      {
                                        integ[hnum-2] /= accum;
                                        integ[hnum-1] /= accum;
                                      }
                                  }
                                KCALL(kpds_put_data(integ_obj, KPDS_VALUE_REGION, (kaddr)integ));
                              }
                          }
                      }
                  }
              }
          } /* End of unit loops */

	(void)klist_free(objlist, (kfunc_void)lkcall_free);

	return TRUE;
}
/* -library_code_end */
