#! /bin/sh

echo "running test1.sh for kformats ... process ($$)"

# ------------------------------------------------------------------------
# report test_num test_obj cntrl_obj test_descr
#
# The report function uses ksegcmp to compare the test output data object
# ($2) against the correct result ($3), and reports success or failure of
# the test.  If the result of ksegcmp is false, status is set to 1.
# ------------------------------------------------------------------------
report()
{
test_num=$1
test_obj="$2"
cntrl_obj="$3"
test_descr="$4"

ksegcmp -i1 "$test_obj" -i2 "$cntrl_obj" -tol 1e-2 -sat 1 -s -rt 1 -all
# ksegcmp -i1 "$test_obj" -i2 "$cntrl_obj" -tol 1e-5 -sat 1 -s -rt 1 -all
if [ $? = 0 ]
then
   echo "Passed:  Test  $test_num --> "$test_descr""
else
   echo "FAILED:  Test  $test_num --> "$test_descr""
   status=1
fi
}

#
#  Initialize status to be success (ie. 0).  Upon failure, status is
#  initialized to 1.
#
status=0

# ------------------------------------------------------------------------
#  set path for test data, and path for temporary directory
# ------------------------------------------------------------------------
tbpath=`kecho -tb DATAMANIP -echo path`
dpath="$tbpath/data/testdata"
tmpdir=${TMPDIR:-/tmp}
BINARY_PATH=${BINARY_PATH:-`kecho -tb DATAMANIP -echo binpath`}

# ------------------------------------------------------------------------
# Test  1: convert kdcomplex KDF file to ascii
# ------------------------------------------------------------------------
test_num=1
descr="convert kdcomplex KDF file to ascii"
${BINARY_PATH}/kformats -i $dpath/kdcomplex.kdf -o $tmpdir/out.$$ -ascii
cat <<EOF > $tmpdir/ascii.$$
(5831.208259,0) (0,0) (0, 0)  
(0,0)  (2976.344864,0)  (0,0)  
(0,0)  (0,0)  (449.4468775,0)
EOF
report $test_num "$tmpdir/out.$$" "$tmpdir/ascii.$$" "$descr"
/bin/rm $tmpdir/ascii.$$ $tmpdir/out.$$

# ------------------------------------------------------------------------
# Test  2: convert complex ascii file to kdcomplex KDF
# ------------------------------------------------------------------------
test_num=2
descr="convert complex ascii file to kdcomplex KDF"
cat <<EOF > $tmpdir/ascii.$$
(5831.208259,0) (0,0) (0, 0)
(0,0)  (2976.344864,0)  (0,0)
(0,0)  (0,0)  (449.4468775,0)
EOF
${BINARY_PATH}/kformats -i $tmpdir/ascii.$$ -o $tmpdir/out.$$ -kdf
report $test_num "$tmpdir/out.$$" "$dpath/kdcomplex.kdf" "$descr"
/bin/rm $tmpdir/out.$$ $tmpdir/ascii.$$

# ------------------------------------------------------------------------
# Test  3: convert bit VIFF file to bit KDF
# ------------------------------------------------------------------------
test_num=3
descr="convert bit VIFF file to bit KDF"
cat <<EOF > $tmpdir/ascii.$$
0 1
1 1
0 0
0 1
0 1
EOF
kasc2val -i1 $tmpdir/ascii.$$ -wsize 2 -hsize 5 -type 1 -o $tmpdir/input.$$
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -viff
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  4: convert ubyte multielement KDF file to VIFF
# ------------------------------------------------------------------------
test_num=4
descr="convert ubyte multielement KDF file to VIFF (large data set)"

kgconst -wsize 28 -hsize 38 -esize 24 -real 1 -type 3 -o $tmpdir/input.$$

set KHOROS_KDMS_BUFSIZE=8
export KHOROS_KDMS_BUFSIZE
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -viff
unset KHOROS_KDMS_BUFSIZE

report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  5: convert bit KDF file to bit SUN raster
# ------------------------------------------------------------------------
test_num=5
descr="convert bit KDF file to bit SUN raster"
cat <<EOF > $tmpdir/ascii.$$
0 1 0 1
1 1 0 1
0 0 0 0
0 1 1 0
0 1 1 0
EOF
kasc2val -i1 $tmpdir/ascii.$$ -wsize 4 -hsize 5 -type 1 -o $tmpdir/input.$$
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -rast
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  6: convert ubyte KDF file to SUN raster (single plane)
# ------------------------------------------------------------------------
test_num=6
descr="convert ubyte KDF file to SUN raster (single plane)"
cat <<EOF > $tmpdir/ascii.$$
0 1 0 5
1 1 0 111
0 0 0 0
0 1 1 0
0 1 211 255
EOF
kasc2val -i1 $tmpdir/ascii.$$ -wsize 4 -hsize 5 -type 3 -o $tmpdir/input.$$
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -rast
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  7: convert ubyte KDF file to SUN raster (3 plane)
# ------------------------------------------------------------------------
test_num=7
descr="convert ubyte KDF file to SUN raster (3 plane)"
cat <<EOF > $tmpdir/ascii.$$
0 1 0 5
1 1 0 111
0 0 0 0
0 1 1 0
0 1 211 255
0 0 0 0
0 0 0 0
0 0 0 3
0 0 0 0
0 0 0 0
21 21 21 21
22 21 21 21
23 21 21 21
24 21 21 21
25 61 71 81
EOF
kasc2val -i1 $tmpdir/ascii.$$ -wsize 4 -hsize 5 -esize 3 -type 3 \
         -o $tmpdir/input.$$
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -rast
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  8: convert ubyte KDF file to SUN raster (single plane with 3 col ubyte map)
# ------------------------------------------------------------------------
test_num=8
descr="convert ubyte KDF file to SUN raster (single plane with 3 col ubyte map)"
cat <<EOF > $tmpdir/ascii.$$
0 1 2 3
1 2 3 4
4 3 2 1
3 2 1 0
0 0 0 0
EOF
kasc2val -i1 $tmpdir/ascii.$$ -wsize 4 -hsize 5 -type 3 -o $tmpdir/input.$$
cat <<EOF > $tmpdir/ascii.$$
0 0 0
1 2 3
4 5 6
7 8 9
253 254 255
EOF
kasc2map -i1 $tmpdir/ascii.$$ -i2 $tmpdir/input.$$ -wsize 3 -hsize 5 -type 3 \
         -o $tmpdir/input.$$
kcolorspace -i $tmpdir/input.$$ -o $tmpdir/input.$$ -type 2
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -rast
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  9: convert ubyte KDF file to raw
# ------------------------------------------------------------------------
test_num=9
descr="convert ubyte KDF file to raw"
cat <<EOF > $tmpdir/ascii.$$
0 1 2 3
1 2 3 4
4 3 2 1
3 2 1 0
0 0 0 0
EOF
kasc2val -i1 $tmpdir/ascii.$$ -o $tmpdir/input.$$ -type 3 -wsize 4 -hsize 5
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -raw
kimportraw -i1 $tmpdir/output.$$ -o $tmpdir/raw1.$$ -type 3 -wsize 4 -hsize 5
kimportraw -i1 test1.results/raw.out -o $tmpdir/raw2.$$ -type 3 -wsize 4 -hsize 5
report $test_num "$tmpdir/raw1.$$" "$tmpdir/raw2.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$  $tmpdir/raw1.$$ $tmpdir/raw2.$$

# ------------------------------------------------------------------------
# Test  10: convert PGM-raw file to KDF
#
# This test data set was generated by the following program, which is
# based on the description found in the pgm man page. The data is fixed
# up to look like whitespace (ASCII 32 is a space), to make sure that a
# bug does not exist in reading the data. 
#
# #include <stdio.h>
# #include <sys/types.h>
# #include <sys/stat.h>
# #include <fcntl.h>
#
# main()
#  {
#    unsigned char hdr[] = {'P','5',' ','4',' ','7',' ','3','2','\n'};
#    unsigned char a[20]={
#                         32,32,2,3,
#                         1,2,3,4,
#                         4,3,2,1,
#                         3,2,1,0,
#                         0,0,0,0
#                        };
#
#   int f;
#
#   f = open("data.pgm.raw",O_CREAT|O_WRONLY|O_TRUNC,0666);
#   write(f,hdr,10);
#   write(f,a,20);
#   close(f);
# }
# ------------------------------------------------------------------------
test_num=10
descr="convert PGM-raw file to KDF"
kformats -i test1.data/data.pgm.raw -o $tmpdir/output.$$ -kdf
cat <<EOF > $tmpdir/ascii.$$
32 32 2 3
1 2 3 4
4 3 2 1
3 2 1 0
0 0 0 0
EOF
kasc2val -i1 $tmpdir/ascii.$$ -o $tmpdir/input.$$ -type 6 -wsize 4 -hsize 5
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  11: convert PGM-ascii file to KDF
# ------------------------------------------------------------------------
test_num=11
descr="convert PGM-ascii file to KDF"
kformats -i test1.data/data.pgm.ascii -o $tmpdir/output.$$ -kdf
cat <<EOF > $tmpdir/ascii.$$
    0 1 2 3
    4 5 6 7
    8 9 10 11
    1 1 1 1
    0 1 2 3
    4 5 6 7
    8 9 10 11
EOF
kasc2val -i1 $tmpdir/ascii.$$ -o $tmpdir/input.$$ -type 6 -wsize 4 -hsize 7
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  12a: convert KDF to PGM-raw, check using ksegcmp
# ------------------------------------------------------------------------
test_num=12a
descr="convert KDF to PGM-raw, check using ksegcmp"
cat <<EOF > $tmpdir/ascii.$$
    0 1 2 3
    4 5 6 7
    8 9 10 11
    1 1 1 1
    0 1 2 3
    4 5 6 7
    8 9 10 11
EOF
kasc2val -i1 $tmpdir/ascii.$$ -o $tmpdir/input.$$ -type 3 -wsize 4 -hsize 7
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -pnm
report $test_num "$tmpdir/output.$$" "$tmpdir/input.$$" "$descr"
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

# ------------------------------------------------------------------------
# Test  12b: convert KDF to PGM-raw, check using diff. This test is done
# specifically to make sure that the maxval in the PGM file is set
# correctly.
#
# The code that generated the test file is:
#
# #include <stdio.h>
# #include <sys/types.h>
# #include <sys/stat.h>
# #include <fcntl.h>
# 
# main()
#   {
#     unsigned char hdr[] = {'P','5','\n','4',' ','7','\n','1','1','\n'};
#     unsigned char a[28]={
#     0,1,2,3,
#     4,5,6,7,
#     8,9,10,11,
#     1,1,1,1,
#     0,1,2,3,
#     4,5,6,7,
#     8,9,10,11
#                         };
# 
#    int f;
# 
#    f = open("data2.pgm.raw",O_CREAT|O_WRONLY|O_TRUNC,0666);
#    write(f,hdr,10);
#    write(f,a,28);
#    close(f);
# }
# ------------------------------------------------------------------------
test_num=12b
descr="convert KDF to PGM-raw, check using diff, check maximum gray value"
cat <<EOF > $tmpdir/ascii.$$
    0 1 2 3
    4 5 6 7
    8 9 10 11
    1 1 1 1
    0 1 2 3
    4 5 6 7
    8 9 10 11
EOF
kasc2val -i1 $tmpdir/ascii.$$ -o $tmpdir/input.$$ -type 3 -wsize 4 -hsize 7
kformats -i $tmpdir/input.$$ -o $tmpdir/output.$$ -pnm
diff $tmpdir/output.$$ test1.data/data2.pgm.raw
if [ $? = 0 ]
then
   echo "Passed:  Test  $test_num --> "$test_descr""
else
   echo "FAILED:  Test  $test_num --> "$test_descr""
   status=1
fi
/bin/rm $tmpdir/output.$$ $tmpdir/ascii.$$ $tmpdir/input.$$

exit $status

