 /*
  * Khoros: $Id: memory.c,v 1.15 1996/05/30 18:10:01 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: memory.c,v 1.15 1996/05/30 18:10:01 mirage Exp $";
#endif


/*
 * Copyright (C) 1993, 1994, 1995, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Memory Code
   >>>>		
   >>>>  Private:
   >>>>                 kdas_memory_construct()
   >>>>                 kdas_memory_recast()
   >>>>                 kdas_memory_resize()
   >>>>                 kdas_memory_reorder()
   >>>>                 kdas_memory_clip()
   >>>>                 kdas_memory_interp()
   >>>>   Static:
   >>>>                 __memory_clip()
   >>>>                 __memory_bit_clip()
   >>>>                 __memory_interp()
   >>>>                 __memory_bit_interp()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"
#include <float.h>

static void __memory_clip       PROTO((kubyte *, kubyte *, int, int *, int *, 
				       int *, int *, int *, int, int));

static void __memory_bit_clip   PROTO((kubyte *, kubyte *, int, int *, 
				       int *, int *, int *, int *, int, 
				       int, int, int));

static void __memory_interp     PROTO((kubyte *, kubyte *, int, int, int *, 
				       int *, int *, int *, int *, int *, 
				       int *, int *, int *, int *, int *, 
				       int *, double *, int, int, int));

static void __memory_bit_interp PROTO((kubyte *, kubyte *, int, int, int *, 
				       int *, int *, int *, int *, int *, 
				       int *, int *, int *, int *, int *, 
				       int *, double *, int, int, int, int,
				       int));

/*-- useful debugging macro for printing bytes in binary --*/
#define BYTE_PRINT(a)	                                      \
   {                                                          \
      kubyte  m[] ={0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01}; \
      int i;                                                  \
      for (i = 0; i < 8; i++)                                 \
      {                                                       \
	 if (a & m[i])                                        \
	    kprintf("*");     				      \
	 else                 				      \
	    kprintf("-");     				      \
      }                       				      \
      kprintf(" ");           				      \
   }

/****    =============================================================
 ****    Memory Methods
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_memory_construct - construct some memory
|
|       Purpose: This routine will construct a block of memory
|	 	 of the given size, dimension, and datatype.
|	
|         Input: dimension - the dimensionality of the segment
|		 size      - the size array of the segment
|		 datatype  - the datatype of the segment
|		 rval      - real value to initialize to
|		 ival      - imaginary value to initialize to
|
|        Output: none
|
|       Returns: pointer to the constructed memory
|
|    Written By: Steve Kubica
|          Date: May 23, 1994
| Modifications:
|
------------------------------------------------------------*/
kaddr
kdas_memory_construct(
   int 	  dimension,
   int 	 *size,
   int 	  datatype,
   double rval,
   double ival)
{
   kaddr *ndata;
   int    overall_size = 1;
   int    i;
   

   /*-- compute the overall size as a product of size and datatype --*/
   for (i = 0; i < dimension; i++)
      overall_size *= size[i];

   /*-- avoid preseting if we don't have to --*/
   if (rval != 0.0 || ival != 0.0)
   {
      ndata = kmalloc((size_t)overall_size * kdata_size(datatype));
      kdas_preset_data(ndata, dimension, size, datatype, rval, ival);
   }
   else
      ndata = kcalloc((size_t)overall_size, (size_t)kdata_size(datatype));
   
   return ndata;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_memory_recast - recast data in a memory
|
|       Purpose: This routine will recast the data in a given
|		 block of memory and place it into a new block
|		 of memory.
|
|		 The new memory block will be returned and the old
|		 one freed.
|
|         Input: data      - data to recast
|                dimension - the dimensionality of the segment
|		 size      - the size array of the segment
|		 odatatype - the original datatype of the segment
|		 ndatatype - the new datatype to cast to 
|		 
|        Output: none
|
|       Returns: A new data block with the recast data.
|
|    Written By: Steve Kubica
|          Date: May 23, 1994
| Modifications:
|
------------------------------------------------------------*/
kaddr
kdas_memory_recast(
   kaddr   data,
   int     dimension,
   int    *size,
   int     odatatype,
   int     ndatatype)
{
   kaddr ndata;
   int   num;
   int   i;


   /*-- create a new block of data in which to return the recast data --*/
   ndata = kdas_memory_construct(dimension, size, ndatatype, 0.0, 0.0);

   /*-- compute the number of elements in the the block --*/
   for (i = 0, num = 1; i < dimension; i++)
      num *= size[i];

   /*-- recast the data --*/
   kdas_cast_data(data, odatatype, ndatatype, dimension, size, KREAL, ndata);

   kfree(data);
   
   return ndata;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_memory_resize - resize data in memory
|
|       Purpose: This routine will resize the data in a given
|		 block of memory and place it into a new block
|		 of memory.
|
|		 The new memory block will be returned and the old
|		 one freed.
|
|         Input: data      - data to resize
|                dimension - the dimensionality of the segment
|		 datatype  - the datatype of the segment
|		 osize     - the original size array of the segment
|		 nsize     - the new size array of the segment
|		 rval      - real value to use when padding to 
|			     larger size
|		 ival      - imaginary value to use when padding 
|			     to larger size
|        Output: none
|
|       Returns: A new data block with the resized data.
|
|    Written By: Steve Kubica
|          Date: May 23, 1994
| Modifications:
|
------------------------------------------------------------*/
kaddr
kdas_memory_resize(
   kaddr   data,
   int     dimension,
   int     datatype,
   int    *osize,
   int    *nsize,
   double  rval,
   double  ival)
{
   kaddr ndata;
   kaddr block;
   int   position[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;
   int   msize[KDAS_MAX_DIM];
   int   num;
   int   i;


   /*-- create a new memory block in which to return the resized data --*/
   ndata = kdas_memory_construct(dimension, nsize, datatype, rval, ival);

   /*-- compute the minimumest size --*/
   for (i = 0, num = 1; i < dimension; i++)
   {
      msize[i] = kmin(nsize[i], osize[i]);
      num *= msize[i];
   }

   /*-- allocate data block for resizing --*/
   block = kmalloc((size_t)(num*kdata_size(datatype)));

   /*-- read all of the old memory --*/
   kdas_memory_clip(data, block, dimension, datatype, osize, msize, NULL,
		    position, position, msize, KDAS_READ);

   /*-- write into the new memory --*/
   kdas_memory_clip(ndata, block, dimension, datatype, nsize, msize, NULL,
		    position, position, msize, KDAS_WRITE);

   /*-- cleanup --*/
   kfree(block);
   kfree(data);

   return ndata;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_memory_reorder - reorder data in memory
|
|       Purpose: This routine will reorder the data in a given
|		 block of memory and place it into a new block
|		 of memory.
|
|		 The new memory block will be returned and the old
|		 one freed.
|
|         Input: data      - data to reorder
|                dimension - the dimensionality of the segment
|		 datatype  - the datatype of the segment
|                osize     - size array in the original order
|                nsize     - size array in the new order
|		 oorder    - the original order array of the segment
|		 norder    - the order size array of the segment
|		 
|        Output: none
|
|       Returns: A new data block with the reordered data.
|
|    Written By: Steve Kubica
|          Date: February 25, 1996
| Modifications:
|
------------------------------------------------------------*/
kaddr
kdas_memory_reorder(
   kaddr   data,
   int     dimension,
   int     datatype,
   int    *osize,
   int    *nsize,
   int    *oorder,
   int    *norder)
{
   kaddr ndata;

   /*-- reorder the data into the new block --*/
   ndata = kdas_transpose_data(data, dimension, datatype, osize, nsize, 
			       oorder, norder, NULL);
   /*-- cleanup --*/
   if (ndata != data)
      kfree(data);

   return ndata;
}

/****    =============================================================
 ****    Memory Clipper Routines
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_memory_clip
|
|       Purpose: This routine is an n-dimensional clipping routine
|		 which will clip an arbitrary block of memory either
|		 into or out of data space stored in memory.
|
|	        DATA "SPACE" IN MEMORY       DESTINATION "BLOCK" IN MEMORY
|
|                                                     ___dposition
|		 +-+-+-+-+-+-+-+-+-+             +-+-/-+-+-+-+-+       
|		 | | | | | | | | | | <- write    | |/| | | | | |
|	         |-O-+-+-+-+-+-+-+-+             +-O-+-+-+-+-+-+
|	         |/|*|*|*|*|*| | | |   /|___|\   | |*|*|*|*|*| |
|	      	 /-+-+-+-+-+-+-+-+-+  /       \  +-+-+-+-+-+-+-+
|    spositon___/| |*|*|*|*|*| | | |  \  ___  /  | |*|*|*|*|*| |
|		 +-+-+-+-+-+-+-+-+-+   \|   |/   +-+-+-+-+-+-+-+
|		 | |*|*|*|*|*| | | |             | |*|*|*|*|*| |
|		 +-+-+-+-+-+-+-+-+-+     read -> +-+-+-+-+-+-+-+
|	
|	       [size indicated by ssize]       [size indicated by bsize]
|                                  
|		       [dimension and datatype apply to both]
|
|                        +-+-+-+-+-+
|		         |*|*|*|*|*|    
|                        +-+-+-+-+-+    the size of the actual data 
|		         |*|*|*|*|*|    unit being transfered is 
|                        +-+-+-+-+-+    indicated by unit size
|		         |*|*|*|*|*|
|                        +-+-+-+-+-+
|
|		 It is assumed that space is pointing to the
|		 beginning of the dataspace and block is pointing
|		 to the beginning of the blockspace when this routine 
|		 is called.
|
|         Input: space     - data space to clip into and out of
|		 block     - the block of data
|                dimension - dimensionality of the data
|                datatype  - datatype
|		 ssize     - the total size of the data space
|		 bsize     - the total size of the block space
|	         asize     - the apparent size of the data space (unused)
|                sposition - beginning corner of the unit in the data space
|                bposition - beginning corner of the unit in the block space
|                unit      - size of the block 
|		 action    - KDAS_READ for reading, KDAS_WRITE for writing
|
|        Output: block of memory is clipped either to or from the data space
|
|       Returns: nothing
|
|    Written By: Steve Kubica and Jeremy Worley
|          Date: May 22, 1995
| Modifications:
|
------------------------------------------------------------*/
void
kdas_memory_clip(
   kaddr  space,
   kaddr  block,
   int    dimension,
   int    datatype,
   int   *ssize,
   int   *bsize,
   int   *asize,
   int   *sposition,
   int   *bposition,
   int   *unit,
   int    action)
{
   int soffset;
   int boffset;
   int dim = dimension;
   int datasize = kdata_size(datatype);

   int sskip[KDAS_MAX_DIM];
   int bskip[KDAS_MAX_DIM];
   int sstride[KDAS_MAX_DIM];
   int bstride[KDAS_MAX_DIM];

   int _ssize[KDAS_MAX_DIM];
   int _bsize[KDAS_MAX_DIM];
   int _unit[KDAS_MAX_DIM];
   int _sposition[KDAS_MAX_DIM];
   int _bposition[KDAS_MAX_DIM];


   /*-- make copies of these, as they can change if we collapse them --*/
   kmemcpy(_ssize,     ssize,      sizeof(int) * dim);
   kmemcpy(_bsize,     bsize,      sizeof(int) * dim);
   kmemcpy(_unit,      unit,       sizeof(int) * dim);
   kmemcpy(_sposition, sposition,  sizeof(int) * dim);
   kmemcpy(_bposition, bposition,  sizeof(int) * dim);

   /*-- collapse the space down if possible --*/
   if (datatype != KBIT)
      kdas_collapse_space(&dim,_ssize,_bsize,_unit,_sposition,_bposition);

   /*-- compute the initial offsets into the dataspace --*/
   kdas_compute_offsets(dim, datatype, datasize, datasize, _ssize, _bsize, 
			_sposition, _bposition, &soffset, &boffset);
   
   /*-- compute the skips needed to traverse the data space --*/
   kdas_compute_strides(dim, datatype, datasize, datasize, _ssize, _bsize, 
			sstride, bstride);
   kdas_compute_skips(dim, sstride, bstride, _unit, _unit, sskip, bskip);

   /*-- clip the data --*/
   if (datatype == KBIT)
   {
      int sshift   = _sposition[0] % 8;
      int bshift   = _bposition[0] % 8;
      int linesize = ((_unit[0] + 7) >> 3); /*-- byte aligned line --*/

      __memory_bit_clip((kubyte *) space, (kubyte *) block, dim-1, 
			_unit, &soffset, &boffset, sskip, bskip,
			linesize, sshift, bshift, action);
   }
   else
      __memory_clip((kubyte *) space, (kubyte *) block, dim-1, 
		    _unit, &soffset, &boffset, sskip, bskip, 
		    (int)(_unit[0]*kdata_size(datatype)), action);

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) __memory_clip
|
|       Purpose: This routine is a recursive orthogonal clipper which
|		 clips to and from memory.
|
|		 On each recursion, if we have recursed down to the 0th
|		 dimension, we copy a line of data.  Otherwise, we
|		 skip a unit in our current dimension and recurse 
|		 down by one dimension.
|
|         Input: space    - data space to clip into and out of
|		 block    - the block of data
|		 rank     - current dimension being considered
|		 unit     - size of the block to clip
|		 soffset  - current offset into the data space 
|		 boffset  - current offset into the data block
|		 sskip    - skips to use when traversing data space
|		 bskip    - skips to use when traversing data block
|		 linesize - the base size of the 0th dimension * datatype
|		 action   - KDAS_READ for reading, KDAS_WRITE for writing
|	
|        Output: block of memory is clipped either to or from the data space
|
|       Returns: nothing
|
|    Written By: Steve Kubica and Jeremy Worley
|          Date: May 22, 1994
| Modifications:
|
------------------------------------------------------------*/
static void 
__memory_clip(
   kubyte *space,
   kubyte *block,
   int     rank,
   int    *unit,
   int    *soffset,
   int    *boffset,
   int    *sskip,
   int    *bskip,
   int     linesize,
   int     action)
{
   register int i;
   register int len  = unit[rank];
   register int sskp = sskip[rank];
   register int bskp = bskip[rank];


   if (rank > 0)
   {
      for (i = 0; i < len; i++, *soffset += sskp, *boffset += bskp)
	 __memory_clip(space, block, rank-1, unit, soffset, boffset, 
		       sskip, bskip, linesize, action);
   }
   else
   {
      if (action == KDAS_READ)
	 kmemcpy(block+*boffset, space+*soffset, (unsigned long) linesize);
      else
	 kmemcpy(space+*soffset, block+*boffset, (unsigned long) linesize);

      *soffset += linesize;
      *boffset += linesize;
   }
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) __memory_bit_clip
|
|       Purpose: This routine is a recursive orthogonal bit clipper
|		 which clips from one block of memory to another.
|
|		 On each recursion, if we have recursed down to the 0th
|		 dimension, we copy a line of data.  Otherwise, we
|		 move skip a unit in our current dimension and recurse 
|		 down by one dimension.   
|
|		 This routine deals with headache of non-byte alligned
|		 bit data.  The shifts relative to the unit boundary
|		 byte in both the data space and data block are 
|		 accounted for in this routine.
|
|         Input: space    - data space to clip into and out of
|		 block    - the block of data
|		 rank     - current dimension being considered
|		 unit     - size of the block to clip
|		 soffset  - current offset into the data space 
|		 boffset  - current offset into the data block
|		 sskip    - skips to use when traversing data space
|		 bskip    - skips to use when traversing data block
|		 linesize - the base size of the 0th dimension * datatype
|		 sshift   - # of bits in data space that data is shifted by
|		 bshift   - # of bits in block that data is shifted by
|		 action   - KDAS_READ for reading, KDAS_WRITE for writing
|	
|        Output: block of memory is clipped either to or from the data space
|
|       Returns: nothing
|	
|    Written By: Steve Kubica and Jeremy Worley
|          Date: April 20, 1994
| Modifications:
|
------------------------------------------------------------*/
static void 
__memory_bit_clip(
   kubyte *space,
   kubyte *block,
   int     rank,
   int    *unit,
   int    *soffset,
   int    *boffset,
   int    *sskip,
   int    *bskip,
   int     linesize,
   int     sshift,
   int     bshift,
   int     action)
{
   register int i;
   register int len  = unit[rank];
   register int sskp = sskip[rank];
   register int bskp = bskip[rank];
   register int do_read = (action == KDAS_READ);

   if (rank > 0)
   {
      for (i = 0; i < len; i++, *soffset += sskp, *boffset += bskp)
	 __memory_bit_clip(space, block, rank-1, unit, soffset, boffset, 
			   sskip, bskip, linesize, sshift, bshift, action);
   }
   else
   {      
      kubyte  himask[] = {0xff,0x7f,0x3f,0x1f,0x0f,0x07,0x03,0x01};
      kubyte  lomask[] = {0x80,0xc0,0xe0,0xf0,0xf8,0xfc,0xfe,0xff};
      kubyte *dst, *src;
      int     ds, ss, c;

      if (do_read)
	 dst = block+*boffset, src = space+*soffset, ds = bshift, ss = sshift;
      else
	 dst = space+*soffset, src = block+*boffset, ds = sshift, ss = bshift;

      if (ss == ds)
      {
	 for (i = 0; i < linesize - 1; i++)
	    *dst++ = *src++;

	 if ((c = (unit[0] % 8)) != 0) /*-- last byte : only get left bits --*/
	    *dst &= himask[c], *dst |= *src & ~himask[c];
	 else
	    *dst = *src;
      }
      else if (ss > ds)
      {
	 for (i = 0; i < linesize; i++)
	 {
	    *dst    = 0x00;
	    *dst   |= (*src++ << (ss - ds))     & lomask[8 - ss + ds];
	    *dst++ |= (*src   >> (8 - ss + ds)) & himask[8 - ss + ds];
	 }
      }
      else /* ss < ds */
      {
	 for (i = 0; i < linesize - 1; i++)
	 {
	    *dst   &= ~himask[ds - ss]; 
	    *dst++ |= (*src   >> (ds - ss))     & himask[ds - ss];
	    *dst   &= ~lomask[ds - ss]; 
	    *dst   |= (*src++ << (8 - ds + ss)) & lomask[ds - ss];
	 }

	 /* *dst &= ~himask[ds - ss]; */
	 /* kprintf(" :"); BYTE_PRINT(*dst); */
	 *dst |= (*src >> (ds - ss)) & himask[ds - ss];
	 /* kprintf(" :"); BYTE_PRINT(*dst); */
	 /* kprintf("\n"); */

      }
      /*-- don't increment soffset and boffset, skips will do this for us --*/
   }
      
   return;
}

/****    =============================================================
 ****    Memory Interpolater Routines
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_memory_interp
|
|       Purpose: This routine is an n-dimensional clipping routine
|		 which will clip an arbitrary block of memory either
|		 into or out of data space stored in memory.  The 
|		 source space may "appear" to be a size which
|		 is larger or smaller than its actual size, in which
|		 case this routine will zero order interpolate the
|		 data line-by-line as it is being clipped to or from
|		 the destination to make it appear as if the the data
|		 was clipped from a space which was the apparent size.
|
|	   DATA "SPACE" IN MEMORY          DATA "BLOCK" IN MEMORY
|                                         shown at an apparent size
|
|						  ___bposition
|	     +-+-+-+-+-+-+-+-+-+             +-+-/-+-+-+-+-+-+
|	     | | | | | | | | | | <- write    | |/| | | | | | |
|	     |-O-+-+-+-+-+-+-+-+             +-O-+-+-+-+-+-+-+
|	     |/|1|2|3| | | | | |   /|___|\   | |1|1|2|2|3|3| |
|	     /-+-+-+-+-+-+-+-+-+  /       \  +-+-+-+-+-+-+-+-+
| spositon__/| |4|5|6| | | | | |  \  ___  /  | |1|1|2|2|3|3| |
|            +-+-+-+-+-+-+-+-+-+   \|   |/   +-+-+-+-+-+-+-+-+
|...appears  | |7|8|9| | | | | |             | |4|4|5|5|6|6| |
|  to be at  +-+-+-+-+-+-+-+-+-+     read -> +-+-+-+-+-+-+-+-+
| aposition  | | | | | | | | | |             | |4|4|5|5|6|6| |
|	     +-+-+-+-+-+-+-+-+-+             +-+-+-+-+-+-+-+-+
|	     | | | | | | | | | |             | |7|7|8|8|9|9| |
|	     +-+-+-+-+-+-+-+-+-+             +-+-+-+-+-+-+-+-+
|	     | | | | | | | | | |             | |7|7|8|8|9|9| |
|	     +-+-+-+-+-+-+-+-+-+             +-+-+-+-+-+-+-+-+
|	  
|	 [size indicated by ssize]       [actual size indicated by bsize]
|    [apparent size indicated by asize]
|				    
|	        [dimension and datatype apply to both]
|
|                        +-+-+-+-+-+-+
|		         |1|1|2|2|3|3|    
|                        +-+-+-+-+-+-+    the size of the actual data 
|		         |1|1|2|2|3|3|    unit being transfered is 
|                        +-+-+-+-+-+-+    indicated by unit size in 
|		         |4|4|5|5|6|6|    apparent space
|                        +-+-+-+-+-+-+
|		         |4|4|5|5|6|6|    
|                        +-+-+-+-+-+-+
|		         |7|7|8|8|9|9|    
|                        +-+-+-+-+-+-+
|		         |7|7|8|8|9|9|    
|                        +-+-+-+-+-+-+
|
|		 It is assumed that space is pointing to the
|		 beginning of the dataspace and block is pointing
|		 to the beginning of the blockspace when this routine 
|		 is called.
|
|         Input: space     - data space to interpolate into and out of
|		 block     - the block of data
|                dimension - dimensionality of the data
|                datatype  - datatype 
|		 ssize     - the total size of the data space
|		 bsize     - the total size of the block space
|	 	 asize     - the total apparent size of the data space
|                aposition - beginning corner of the unit in the apparent space
|                bposition - beginning corner of the unit in the block space
|                aunit     - size of the block in apparent space
|		 action    - KDAS_READ for reading, KDAS_WRITE for writing
|
|        Output: block of memory is clipped either to or from the data space
|
|       Returns: nothing
|
|    Written By: Steve Kubica and Jeremy Worley
|          Date: May 22, 1995
| Modifications:
|
------------------------------------------------------------*/
void
kdas_memory_interp(
   kaddr  space,
   kaddr  block,
   int    dimension,
   int    datatype,
   int   *ssize,
   int   *bsize,
   int   *asize,
   int   *aposition,
   int   *bposition,
   int   *aunit,
   int    action)
{
   int soffset;
   int boffset;
   int dim = dimension;
   int ds  = kdata_size(datatype);

   int sskip[KDAS_MAX_DIM];
   int bskip[KDAS_MAX_DIM];
   int sstride[KDAS_MAX_DIM];
   int bstride[KDAS_MAX_DIM];
   int sposition[KDAS_MAX_DIM];
   int sunit[KDAS_MAX_DIM];

   double fskip[KDAS_MAX_DIM];

   int i;

   /*----------------------------------------------------------------
   |   determine the fractional skips needed to traverse the source :
   |   we will also determine our actual block size and position in 
   |   our actual data space from our apparent block size and position
   |   in the apparent space.    mathematically, I am not sure why
   |   this is such a mess, but it makes the testsuite pass ;-)
   \-----------------------------------------------------------------*/
   for (i = 0; i < dimension; i++)
   {
      fskip[i]     = (double)ssize[i] / (double)asize[i];
      sposition[i] = (int) kfloor((double) aposition[i] * fskip[i]);
      sunit[i]     = (int) kceil((double) aunit[i] * fskip[i]);

      /*-- backend clipping if we've rounded up too far for the unit --*/
      if (sposition[i] + sunit[i] > ssize[i])
	 sunit[i]--;
   }

   /*-- for bit data, we need to track line position outside of soffset --*/
   if (datatype == KBIT)
      sposition[0] = 0;
   
   /*-- compute the initial offsets into the dataspace --*/
   kdas_compute_offsets(dim, datatype, ds, ds, ssize, bsize, 
			sposition, bposition, &soffset, &boffset);

   /*-- compute the skips needed to traverse the data space --*/
   kdas_compute_strides(dim, datatype, ds, ds, ssize, bsize, sstride, bstride);
   kdas_compute_skips(dim, sstride, bstride, sunit, aunit, sskip, bskip);

   /*-- clip the data from the source to the destination --*/
   if (datatype == KBIT)
   {
      int ashift = aposition[0] % 8;
      int bshift = bposition[0] % 8;

      __memory_bit_interp((kubyte *) space, (kubyte *) block, dim-1, dim,
			  aunit, sunit, ssize, bsize, aposition, bposition, 
			  &soffset, &boffset, sstride, bstride, 
			  sskip, bskip, fskip, datatype, 
			  ds, ashift, bshift, action);
   }
   else
      __memory_interp((kubyte *) space, (kubyte *) block, dim-1, dim, 
		      aunit, sunit, ssize, bsize, aposition, bposition, 
		      &soffset, &boffset,  sstride, bstride, 
		      sskip, bskip, fskip, datatype, ds, action);
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) __memory_interp
|
|       Purpose: This routine is a recursive orthogonal clipper which
|		 clips to and from memory interpolating as it clips.
|
|		 On each recursion, if we have recursed down to the 0th
|		 dimension, we interpolate a line of data.  After each
|		 line we may either back up a line, or skip some lines
|		 depending on the interpolation factor in that dimension.
|
|         Input: space     - data space to clip into and out of
|		 block     - the block of data
|		 rank      - current dimension being considered
|		 dim       - the dimensionality of the data
|		 unit      - size of the block to clip in the apparent space
|		 sunit     - size of the block to clip in the actual space
|		 ssize     - actual size of the data source
|		 bsize     - actual size of the data block
|		 aposition - current position in the apparent data space
|		 bposition - current position in the data block
|		 soffset   - current offset into the data space 
|		 boffset   - current offset into the data block
|		 sstride   - strides to use when traversing the data space
|		 bstride   - strides to use when traversing the data block
|		 sskip     - skips to use when traversing data space
|		 bskip     - skips to use when traversing data block
|		 fskip     - floating point scale factors to transform
|		 	     between the apparent size and actual size of
|			     the data space
|		 datatype  - the datatype of the data
|		 datasize  - the base size the datatype
|		 action    - KDAS_READ for reading, KDAS_WRITE for writing
|	
|        Output: block of memory is clipped either to or from the data space
|
|       Returns: nothing
|
|    Written By: Steve Kubica and Jeremy Worley
|          Date: May 22, 1994
| Modifications:
|
------------------------------------------------------------*/
static void 
__memory_interp(
   kubyte *space,
   kubyte *block,
   int     rank,
   int     dim,
   int    *unit,
   int    *sunit,
   int    *ssize,
   int    *bsize,
   int    *aposition,
   int    *bposition,
   int    *soffset,
   int    *boffset,
   int    *sstride,
   int    *bstride,
   int    *sskip,
   int    *bskip,
   double *fskip,
   int     datatype,
   int     datasize,
   int     action)
{
   register int i;
   register int len;
   double   f, b, skp;
   int      pos;
   

   if (rank > 0)
   {
      /*----------------------------------------------------------------
      |   set up our backtracking variables : we will backtrack in the
      |   dataspace if reading, and in the blockspace if writing.
      \-----------------------------------------------------------------*/
      if (action == KDAS_READ)
         len = unit[rank],  pos = aposition[rank], skp = fskip[rank];
      else
         len = sunit[rank], pos = bposition[rank], skp = 1.0/fskip[rank];

      f = kabs(pos + 1) * skp;
      b = f - skp + FLT_EPSILON;

      for (i = 0; i < len; i++, b = f, f += skp + FLT_EPSILON)
      {
	 __memory_interp(space, block, rank-1, dim, unit, sunit, 
			 ssize, bsize, aposition, bposition, 
			 soffset, boffset, sstride, bstride, sskip, 
			 bskip, fskip, datatype, datasize, action);

	 /*------------------------------------------------------------
         |   the rank-1 case has moved us to the next rank unit.
	 |   we may now want to back up by one unit, depending on how
	 |   far we are along this dimension.
	 \-------------------------------------------------------------*/
	 if (action == KDAS_READ)
	    *soffset += ((int)f - (int)b - 1) * sstride[rank];
	 else
	    *boffset += ((int)f - (int)b - 1) * bstride[rank];
      }

      if (rank < dim - 1)
	 *soffset += sskip[rank+1], *boffset += bskip[rank+1];
   }
   else
   {
      if (action == KDAS_READ)
	 kdata_resample(space+*soffset, sunit[0], unit[0], datatype, 
			KZERO_ORDER, KREAL, fskip[0], 0.0, 0.0, 
			kfraction(aposition[0] * fskip[0]), 0.0,
			block+*boffset);
      else
	 kdata_resample(block+*boffset, unit[0], sunit[0], datatype,
			KZERO_ORDER, KREAL, (1.0/fskip[0]), 0.0, 0.0,
			0.0, 0.0, space+*soffset);

      *soffset += ssize[0] * datasize;
      *boffset += bsize[0] * datasize;
   }
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) __memory_bit_interp
|
|       Purpose: a recursive orthogonal clipinterp routine.
|         Input:
|        Output:
|       Returns:
|
|    Written By: Jeremy Worley
|          Date: June 4, 1994
| Modifications:
|
------------------------------------------------------------*/
static void 
__memory_bit_interp(
   kubyte *space,
   kubyte *block,
   int     rank,
   int     dim,
   int    *unit,
   int    *sunit,
   int    *ssize,
   int    *bsize,
   int    *aposition,
   int    *bposition,
   int    *soffset,
   int    *boffset,
   int    *sstride,
   int    *bstride,
   int    *sskip,
   int    *bskip,
   double *fskip,
   int     datatype,
   int     datasize,
   int     sshift,
   int     bshift,
   int     action)
{
   register int i;
   register int len;
   double   f, b, skp;
   int      pos;
   

   if (rank > 0)
   {
      /*----------------------------------------------------------------
      |   set up our backtracking variables : we will backtrack in the
      |   dataspace if reading, and in the blockspace if writing.
      \-----------------------------------------------------------------*/
      if (action == KDAS_READ)
         len = unit[rank],  pos = aposition[rank], skp = fskip[rank];
      else
         len = sunit[rank], pos = bposition[rank], skp = 1.0/fskip[rank];

      f = kabs(pos + 1) * skp;
      b = f - skp + FLT_EPSILON;

      for (i = 0; i < len; i++, b = f, f += skp + FLT_EPSILON)
      {
	 __memory_bit_interp(space, block, rank-1, dim, unit, sunit, 
			     ssize, bsize, aposition, bposition, 
			     soffset, boffset, sstride, bstride, sskip, 
			     bskip, fskip, datatype, datasize, 
			     sshift, bshift, action);

	 /*------------------------------------------------------------
         |   the rank-1 case has moved us to the next rank unit.
	 |   we may now want to back up by one unit, depending on how
	 |   far we are along this dimension.
	 \-------------------------------------------------------------*/
	 if (action == KDAS_READ)
	    *soffset += ((int)f - (int)b - 1) * sstride[rank];
	 else
	    *boffset += ((int)f - (int)b - 1) * bstride[rank];
      }

      if (rank < dim - 1)
	 *soffset += sskip[rank+1], *boffset += bskip[rank+1];
   }
   else
   {
      if (action == KDAS_READ)
	 kdas_bit_resample(space+*soffset, sunit[0], unit[0], bshift, 
			   (double)aposition[0] * fskip[0], block+*boffset);
      else
	 kdas_bit_resample(block+*boffset, unit[0], sunit[0], sshift,
			   (double)aposition[0] / fskip[0], space+*soffset);
      
      *soffset += (ssize[0] + 7) >> 3;
      *boffset += (bsize[0] + 7) >> 3;
   }
   return;
}
