 /*
  * Khoros: $Id: pipeline.c,v 1.9 1996/05/29 01:38:34 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: pipeline.c,v 1.9 1996/05/29 01:38:34 mirage Exp $";
#endif


/*
 * Copyright (C) 1993, 1994, 1995, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Pipeline Code
   >>>>		
   >>>>  Private:
   >>>>                 kdas_pipeline_data()
   >>>>                 kdas_transpose_data()
   >>>>   Static:
   >>>>                 _convert()
   >>>>                 _scale()
   >>>>                 _normalize()
   >>>>                 _transpose()
   >>>>                 _avoid_transpose()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

static int  _convert   	     PROTO((kaddr, kaddr, int, int *, int, int, int, 
		       		    double, double));
static int  _scale     	     PROTO((kaddr, int, int *, int, double, double, 
				    double));
static int  _normalize 	     PROTO((kaddr, int, int *, int, double, double));
static void _transpose 	     PROTO((unsigned char *, unsigned char *, int, int,
				    int, int *, int *, int *, int *, int *));
static int  _avoid_transpose PROTO((int, int *, int *, int *, int *));

/****    =============================================================
 ****    Temporary Buffers
 ****    =============================================================
 ****/

static kdas_buffer *buffer1 = NULL;
static kdas_buffer *buffer2 = NULL;
static kdas_buffer *buffer3 = NULL;

/****    =============================================================
 ****    Pipeline Code
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_pipeline_data
|
|       Purpose: This routine will pass a block of data through
|		 a presentation pipeline.  This pipeline will 
|		 perform the following stages :
|	
|   	  	   o  complex conversion
|   	  	   o  upcasting
|   	  	   o  transposition
|   	  	   o  scaling
|   	  	   o  normalization
|   	  	   o  downcasting
|
|         Input: presentation - the presentation which dictates how to
|		                invoke each of the pipeline stages
|	
|		 upcast       - invoke the upcast stage
|		 downcast     - invoke the downcast stage
|		 convert      - invoke the complex convert stage
|	
|		 size - the size of the data region being processed in 
|			the pipeline
|
|		 reading - TRUE if the data is being read out of the segment
|			   FALSE if the data is being written to the segment
|
|		 data - the data being passed through the pipeline
|
|        Output: none
|
|       Returns: the pipelined data
|
|    Written By: Steve Kubica
|          Date: October 12, 1994
| Modifications:
|
------------------------------------------------------------*/
kaddr
kdas_pipeline_data(
   kdas_presentation *pres, 
   int                upcast,
   int                downcast,
   int                convert,
   int               *size,
   int                reading,
   kaddr              data)
{
   int dimension = pres->dimension;

   int sdatatype;  /*-- keep this set to our most current working datatype --*/
   int ddatatype;

   kaddr sdata = data;  /*-- keep pointing to our most up-to-date source --*/
   kaddr ddata = data;  /*-- we'll return this --*/

   /*-- determine source and destination datatypes : depends on direction --*/
   sdatatype = (reading) ? pres->segment->datatype : pres->datatype;
   ddatatype = (reading) ? pres->datatype : pres->segment->datatype;

   /*------------------------------------------------------------------
   |   handle complex conversion : we want to do this first while
   |                               we are still in a complex datatype.
   \-------------------------------------------------------------------*/
   if (convert)
   {
      int ccdatatype;  /*-- complex convert from source to this datatype --*/
      
      switch (sdatatype)
      {
	 case KDCOMPLEX: 
	    ccdatatype = KDOUBLE; 
	    break;
	 case KCOMPLEX:
	    ccdatatype = KFLOAT;
	    break;
	 case KDOUBLE:
	    ccdatatype = KDCOMPLEX;
	    break;
	 case KFLOAT:
	    ccdatatype = KCOMPLEX;
	    break;
	 default:
	    ccdatatype = KCOMPLEX;
	    break;
      }

      buffer1 = kdas_adjust_buffer_size(buffer1, dimension, size, ccdatatype);
      ddata = buffer1->data;

      if (!_convert(sdata, ddata, dimension, size, pres->complex_convert,
		    sdatatype, ccdatatype, pres->rpad, pres->ipad))
	 return NULL;

      sdatatype = ccdatatype;

      sdata = ddata;
   }

   /*------------------------------------------------------------------
   |   upcasting : this will occur if the destination datatype has 
   |               a smaller range or less precision than the source 
   |		   datatype.  we do this first because scaling and 
   |		   normalization are less prone to loss of precision
   |               and overflow if they operate on the "higher order" 
   |               data type. 
   \-------------------------------------------------------------------*/
   if (upcast)
   {
      buffer2 = kdas_adjust_buffer_size(buffer2, dimension, size, ddatatype);
      ddata = buffer2->data;

      if (!kdas_cast_data(sdata, sdatatype, ddatatype, dimension, size, 
			  KREAL, ddata))
         return FALSE;

      sdata = ddata;
      sdatatype = ddatatype;
   }
   
   /*------------------------------------------------------------------
   |   scale : we can scale in place. 
   \-------------------------------------------------------------------*/
   if (pres->scaling == KSCALE)
      if (!_scale(sdata, dimension, size, sdatatype, pres->scale_factor,
		  pres->scale_roffset, pres->scale_ioffset))
	 return FALSE;

   /*------------------------------------------------------------------
   |   normalize : we can normalize in place. 
   \-------------------------------------------------------------------*/
   if (pres->scaling == KNORMALIZE)
      if (!_normalize(sdata, dimension, size, sdatatype, pres->norm_min,
		      pres->norm_max))
	 return FALSE;

   /*------------------------------------------------------------------
   |   downcast : if we need to cast down, we can safely do it now.
   \-------------------------------------------------------------------*/
   if (downcast)
   {
      buffer3 = kdas_adjust_buffer_size(buffer3, dimension, size, ddatatype);
      ddata = buffer3->data;

      if (!kdas_cast_data(sdata, sdatatype, ddatatype, dimension, size,
			  KREAL, ddata))
         return FALSE;
      
      sdata = ddata;
   }

   return ddata;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _convert
|
|       Purpose: 
|
|         Input:
|        Output:
|       Returns:
|    Written By: Jeremy Worley
|          Date: April 21, 1994
| Modifications:
|
------------------------------------------------------------*/
static int
_convert(
   kaddr  sdata,
   kaddr  ddata,
   int    dimension,
   int   *size,
   int    convert,
   int    sdatatype,
   int    ddatatype,
   double rval,
   double ival)
{
   register int i;
   register int num = 1;
   int status;
   
   for (i = 0; i < dimension; i++)
      num *= size[i];

   status = kdas_cast_data(sdata, sdatatype, ddatatype, dimension, size,
			   convert, ddata);

   /*-- experiment for donna --*/
   if (ddatatype == KCOMPLEX)
   {
      kcomplex *cdata = ddata;
      
      switch (convert)
      {
	 case KREAL:
	    for (i = 0; i < num; i++)
	       cdata[i] = kccomp(kcreal(cdata[i]),(float)ival);
	    break;
	 case KIMAGINARY:
	    for (i = 0; i < num; i++)
	       cdata[i] = kccomp((float)rval, kcimag(cdata[i]));
	    break;
	 default:
	    break;
      }
   }
   else if (ddatatype == KDCOMPLEX)
   {
      kdcomplex *cdata = ddata;
      
      switch (convert)
      {
	 case KREAL:
	    for (i = 0; i < num; i++)
	       cdata[i] = kdccomp(kdcreal(cdata[i]),(double)ival);
	    break;
	 case KIMAGINARY:
	    for (i = 0; i < num; i++)
	       cdata[i] = kdccomp((double)rval, kdcimag(cdata[i]));
	    break;
	 default:
	    break;
      }
   }

   return status;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _scale
|
|       Purpose:
|
|         Input:
|        Output:
|       Returns:
|    Written By: Jeremy Worley
|          Date: April 21, 1994
| Modifications:
|
------------------------------------------------------------*/
static int
_scale(
   kaddr  data,
   int    dimension,
   int   *size,
   int    datatype,
   double scale_factor,
   double roffset,
   double ioffset)
{
   register int i;
   register int num = 1;

   for (i = 0; i < dimension; i++)
      num *= size[i];

   return kdata_scale(data, num, datatype, scale_factor, 0.0,
		      roffset, ioffset, data);
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _normalize
|
|       Purpose:
|
|         Input:
|        Output:
|       Returns:
|    Written By: Jeremy Worley
|          Date: April 21, 1994
| Modifications:
|
------------------------------------------------------------*/
static int
_normalize(
   kaddr  data,
   int 	  dimension,
   int 	 *size,
   int    datatype,
   double norm_min,
   double norm_max)
{
   register int i;
   register int num = 1;

   double factor;
   double offset;

   for (i = 0; i < dimension; i++)
      num *= size[i];

   switch (datatype)
   {
      case KBIT:
	 /*-- easy case...normalization doesn't make sense --*/
	 return TRUE;
      case KBYTE:
	 {
	    signed char *tmp = (signed char *)data;
	    signed char  rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -rmin * factor + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (signed char)(tmp[i] * factor + offset);
	 }
	 break;
      case KUBYTE:
	 {
	    unsigned char *tmp = (unsigned char *)data;
	    unsigned char  rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -((double)rmin * factor) + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (unsigned char)(tmp[i] * factor + offset);
	 }
	 break;
      case KSHORT:
	 {
	    short *tmp = (short *)data;
	    short  rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -rmin * factor + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (short)(tmp[i] * factor + offset);
	 }
	 break;
      case KUSHORT:
	 {
	    unsigned short *tmp = (unsigned short *)data;
	    unsigned short  rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -((double)rmin * factor) + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (unsigned short)(tmp[i] * factor + offset);
	 }
	 break;
      case KINT:
	 {
	    int *tmp = (int *)data;
	    int rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -rmin * factor + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (int)(tmp[i] * factor + offset);
	 }
	 break;
      case KUINT:
	 {
	    unsigned int *tmp = (unsigned int *)data;
	    unsigned int rmin, rmax; 

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -((double)rmin * factor) + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (unsigned int)(tmp[i] * factor + offset);
	 }
	 break;
      case KLONG:
	 {
	    long *tmp = (long *)data;
	    long  rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -rmin * factor + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (long)(tmp[i] * factor + offset);
	 }
	 break;
      case KULONG:
	 {
	    unsigned long *tmp = (unsigned long *)data;
	    unsigned long rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -((double)rmin * factor) + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (unsigned long)(tmp[i] * factor + offset);
	 }
	 break;
      case KFLOAT:
	 {
	    float *tmp = (float *)data;
	    float rmin, rmax;

	    rmin = rmax = tmp[0];

	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0.0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -rmin * factor + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (float)(tmp[i] * factor + offset);
	 }
	 break;
      case KDOUBLE:
	 {
	    double *tmp = (double *)data;
	    double rmin, rmax;

	    rmin = rmax = tmp[0];
	    for (i = 1; i < num; i++)
	    {
	       rmin = kmin(tmp[i], rmin);
	       rmax = kmax(tmp[i], rmax);
	    }

	    if (rmax - rmin != 0.0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -rmin * factor + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = (double)(tmp[i] * factor + offset);
	 }
	 break;
      case KCOMPLEX:
	 {
	    kcomplex *tmp = (kcomplex *) data;
	    double rmin, rmax, imin, imax, mmin, mmax;

	    mmin = mmax = (double)kcmag(tmp[0]);
	    rmin = rmax = (double)kcreal(tmp[0]);	
	    imin = imax = (double)kcimag(tmp[0]);	
	    for (i = 1; i < num; i++)
	    {
	       if (kcmag(tmp[i]) < mmin) {
		  rmin = kcreal(tmp[i]);
		  imin = kcimag(tmp[i]);
		  mmin = kcmag(tmp[i]);	
	       }
	       if (kcmag(tmp[i]) > mmax) {
		  rmax = kcreal(tmp[i]);
		  imax = kcimag(tmp[i]);
		  mmax = kcmag(tmp[i]);	
	       }
	    }
	    if (rmax - rmin != 0.0)
	       factor = (norm_max - norm_min) / ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    for (i = 0; i < num; i++)
	       tmp[i] = kcadd(kcmult(tmp[i],
				     kccomp((double)factor, 0.0)),
			      kccomp((float)rmin, (float)imin));
	 }
	 break;
      case KDCOMPLEX:
	 {
	    kdcomplex *tmp = (kdcomplex *) data;
	    double     rmin, rmax;

	    rmin = rmax = (double)kdcmag(tmp[0]);
	    for (i = 1; i < num; i++)
	    {
	       rmin = (double)kmin(kdcmag(tmp[i]), rmin);
	       rmax = (double)kmax(kdcmag(tmp[i]), rmax);
	    }

	    if (rmax - rmin != 0.0)
	       factor = (norm_max - norm_min) / 
		  ((double)rmax - (double)rmin);
	    else
	       factor = (norm_max - norm_min);

	    offset = -rmin * factor + norm_min;
	    for (i = 0; i < num; i++)
	       tmp[i] = kdcadd(kdcmult(tmp[i], kdccomp((double)factor, 0.0)),
			       kdccomp((double)offset, 0.0));
	 }
	 break;
      default:
	 {
	    kdas_set_error(KDAS_EINVALID_NORMALIZATION);
	    return FALSE;
	 }
   }
   return TRUE;
}

/****    =============================================================
 ****    Transposing Code
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_transpose_data
|
|       Purpose: This routine will transpose a block of data. 
|	
|         Input: 
|        Output: none
|       Returns: the pipelined data
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: October 12, 1994
| Modifications:
|
------------------------------------------------------------*/
kaddr
kdas_transpose_data(
   kaddr  sdata,
   int    dimension,
   int    datatype,
   int   *ssize,
   int   *dsize,
   int   *sorder,
   int   *dorder,
   kaddr  ddata)
{
   int sstride [KDAS_MAX_DIM];
   int dstride [KDAS_MAX_DIM];
   int sskip   [KDAS_MAX_DIM];
   int dskip   [KDAS_MAX_DIM];
   int soffset 	= 0;
   int doffset 	= 0;
   int really_is_bit = FALSE;
   int datasize;

   kaddr sd = sdata;
   kaddr dd = NULL;

   /*---------------------------------------------------------------------
   |   bit data is very painful to reorder, so painful in fact, that we
   |   are going to cast it to byte, reorder it, and then cast it back to
   |   bit.  this isn't optimally fast, but it will generate 
   |   correct results which is more important than speed.
   \---------------------------------------------------------------------*/
   if (datatype == KBIT)
   {
      datatype      = KBYTE;
      really_is_bit = TRUE;

      sd = kdas_memory_construct(dimension, ssize, KBYTE, 0.0, 0.0);
      dd = kdas_memory_construct(dimension, dsize, KBYTE, 0.0, 0.0);

      kdas_cast_data(sdata, KBIT, KBYTE, dimension, ssize, KREAL, sd);
   }

   /*---------------------------------------------------------------------
   |   we may not _really_ have to transpose : in which case we can just
   |   return the data we have.   we still will have to go through the
   |   motions if we have bit data because the transpose may affect
   |   byte alignment of the lines of bit data..
   \---------------------------------------------------------------------*/
   if (!really_is_bit && 
       _avoid_transpose(dimension, sorder, dorder, ssize, dsize))
      return sdata;

   /*-- allocate destination if we need to --*/
   if (ddata == NULL)
      ddata = kdas_memory_construct(dimension, dsize, datatype, 0.0, 0.0);

   if (!really_is_bit)
      dd = ddata;

   /*-- compute our data size for the datatype we are working in --*/
   datasize = kdata_size(datatype);

   /*-- compute the strides through the source and destination spaces --*/
   kdas_compute_strides(dimension, datatype, datasize, datasize, 
			ssize, dsize, sstride, dstride);

   /*-- rotate the source strides according to the index order --*/
   kdas_array_reorder(sstride, dimension, sorder, dorder);

   /*-- compute the skip factors all according to the destination  --*/
   kdas_compute_skips(dimension, sstride, dstride, dsize, dsize, sskip, dskip);

   /*-- finally tranpose the data --*/
   _transpose((kubyte *)sd, (kubyte *)dd, datasize, dimension - 1, 
	      dimension, dsize, &soffset, &doffset, sskip, dskip);

   /*-- if we had bit, then we gotta cast it back --*/
   if (really_is_bit)
   {
      kdas_cast_data(dd, KBYTE, KBIT, dimension, dsize, KREAL, ddata);
      kfree_and_NULL(sd);  kfree_and_NULL(dd);
   }

   return ddata;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _transpose
|
|       Purpose: a recursive orthogonal transpose routine.
|
|         Input:
|        Output:
|       Returns:
|    Written By: Jeremy Worley
|          Date: April 20, 1994
| Modifications:
|
------------------------------------------------------------*/
static void 
_transpose(
   unsigned char *sdata,
   unsigned char *ddata,
   int  datasize,
   int  rank,
   int  dimension,
   int *unit,
   int *soffset,
   int *doffset,
   int *sskip,
   int *dskip)
{
   register int i;

   if (rank >= 0)
   {
      register int len  = unit[rank];
      register int sskp = sskip[rank];
      register int dskp = dskip[rank];

      for (i = 0; i < len; i++, *soffset += sskp, *doffset += dskp)
	 _transpose(sdata, ddata, datasize, rank - 1, dimension, unit,
		    soffset, doffset, sskip, dskip);
   }
   else
      kmemcpy(ddata + *doffset, sdata + *soffset, (size_t) datasize);

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _avoid_transpose
|
|       Purpose: This routine will check two order and size arrays 
|		 to see if they are sufficiently "equivalent" such
|                that a transpose can be avoided.   For instance a
|		 1x10x9 array and a 10x9x1 array are really the
|		 same.   
|
|		 It checks for both the source and destination arrays
|		 for any occurance of the size "1".  The "1" sizes
|		 are all trickled down to the end of the list.  This
|		 should effectively move all the non-1 sizes to the
|		 front of the list, while still preserving their 
|		 relative order.
|
|		 In the end, the arrays are compared and any order
|		 differences where the corresponding sizes are not "1"
|		 imply that we really do need to transpose.
|	
|         Input: dim    - dimension of order and size arrays
|		 sorder - source order array
|		 dorder - destination order array
|		 ssize  - source size array
|		 dsize  - destination size array
|        Output: none
|       Returns: TRUE if we don't really need to transpose, 
|		 FALSE if we really should transpose
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: May 18, 1996
| Modifications:
|
------------------------------------------------------------*/
static int
_avoid_transpose(
   int  dim,
   int *sorder,
   int *dorder,
   int *ssize,
   int *dsize)
{
   int sord[KDAS_MAX_DIM];
   int dord[KDAS_MAX_DIM];
   int ssiz[KDAS_MAX_DIM];
   int dsiz[KDAS_MAX_DIM];
   
   int i;
   int j;
   int tmp;

   kmemcpy(sord, sorder, sizeof(int) * dim);
   kmemcpy(dord, dorder, sizeof(int) * dim);
   kmemcpy(ssiz, ssize,  sizeof(int) * dim);
   kmemcpy(dsiz, dsize,  sizeof(int) * dim);

   /*-- two passes to trickle down all "1" dimensions to end of list --*/
   for (j = 0; j < 2; j++)
      for (i = 0; i < dim-1; i++)
	 if (ssiz[i] == 1) 
	 {
	    /*-- trickle down --*/
	    tmp = sord[i];  sord[i] = sord[i+1];  sord[i+1] = tmp;
	    tmp = ssiz[i];  ssiz[i] = ssiz[i+1];  ssiz[i+1] = tmp;
	 }

   /*-- two passes to trickle down all "1" dimensions to end of list --*/
   for (j = 0; j < 2; j++)
      for (i = 0; i < dim-1; i++)
	 if (dsiz[i] == 1) 
	 {
	    /*-- trickle down --*/
	    tmp = dord[i];  dord[i] = dord[i+1];  dord[i+1] = tmp;
	    tmp = dsiz[i];  dsiz[i] = dsiz[i+1];  dsiz[i+1] = tmp;
	 }

   /*---------------------------------------------------------------------
   |  if the orders are not the same for a position, and the sizes in that
   |  position aren't "1", then we have to transpose 
   \---------------------------------------------------------------------*/
   for (i = 0; i < dim; i++)
      if ((sord[i] != dord[i]) && ((ssiz[i] != 1) || (dsiz[i] != 1)))
	 return FALSE; /*-- gotta transpose --*/ 

   /*-- no transposing necessary --*/
   return TRUE;
}


