 /*
  * Khoros: $Id: utilities.c,v 1.12 1996/05/20 01:03:13 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: utilities.c,v 1.12 1996/05/20 01:03:13 mirage Exp $";
#endif


/*
 * Copyright (C) 1993, 1994, 1995, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Utility Routines
   >>>>		
   >>>>  Private:
   >>>>                 kdas_advance_position()
   >>>>                 kdas_cast_data()
   >>>>                 kdas_preset_data()
   >>>>                 kdas_preset_file()
   >>>>                 kdas_compute_offsets()
   >>>>                 kdas_compute_strides()
   >>>>                 kdas_compute_skips()
   >>>>                 kdas_collapse_space()
   >>>>                 kdas_optimal_region_size()
   >>>>                 kdas_clip_corners()
   >>>>                 kdas_is_big()
   >>>>                 kdas_set_threshold()
   >>>>                 kdas_get_order()
   >>>>                 kdas_array_reorder()
   >>>>                 kdas_array_max()
   >>>>                 kdas_initialized()
   >>>>                 kdas_size_initialized()
   >>>> 		kdas_bit_resample()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/****    =============================================================
 ****    Threshold Size
 ****    =============================================================
 ****/

static int khoros_threshold_size = KHOROS_THRESHOLD_SIZE_DEFAULT;
static int khoros_threshold_size_initialized = FALSE;

/****    =============================================================
 ****    Utility Routines
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_advance_position
|
|       Purpose: This routine advances a given begin point for
|		 a given increment unit.  It will optionally 
|		 either stop at the end of the data set or
|		 wrap around to the beginning of the data set.
|
|         Input: dimension - the dimensionality of the segment
|		 size      - the size of the segment
|		 unit      - the unit to advance by
|		 position  - the position to advance
|	
|        Output: position is advanced in place
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley
|          Date: October 7, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdas_advance_position(
   int  dimension,
   int *size,
   int *unit,
   int *position)
{
   int i;

   /*-- increment to the beginning of the next unit of data --*/
   for (i = 0; i < dimension - 1; i++)
   {
      position[i] += unit[i];
      
      if (position[i] < size[i])
	 return TRUE;
      else
	 position[i] = 0;
   }
   position[dimension - 1] += unit[dimension - 1];

   if (position[dimension - 1] >= size[dimension - 1])
      return FALSE;

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_cast_data
|
|       Purpose: This routine will cast a block of data.
|		 It is just like the routine kdata_cast, 
|		 only it is aware of the bit data type.
|		 Bit data must be cast line-by-line as bit
|		 data is byte-aligned. 
|
|         Input: sdata     - the source data block to be cast
|		 sdatatype - the datatype of the source data block
|		 ddatatype - the datatype to cast the source into
|		 dimension - the dimensionality of the data
|		 size      - the size of the data 
|		 convert   - complex conversion mode.  can be one of
|                            KREAL, KIMAGINARY, KMAGNITUDE, or KPHASE.
|
|        Output: ddata     - the destination data block to hold the
|			     cast data
|
|       Returns: nothing
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: May 17, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdas_cast_data(
   kaddr  sdata, 
   int    sdatatype,
   int    ddatatype,
   int    dimension, 
   int   *size,
   int    convert, 
   kaddr  ddata)
{
   int status = TRUE;
   int i, num = 1;


   if (sdatatype == KBIT)
   {
      kubyte *src = (kubyte *) sdata;
      kubyte *dst = (kubyte *) ddata;
      
      for (i = 1; i < dimension; i++)
	 num *= size[i];

      for (i = 0; i < num; i++) 
      {
	 status |= kdata_cast((kaddr)src, sdatatype, ddatatype, size[0],
			      convert, 1.0, 1.0, (kaddr) dst);
	 src += (size[0] + 7) >> 3;
	 dst += size[0] * kdata_size(ddatatype);
      }
   }
   else if (ddatatype == KBIT)
   {
      kubyte *src = (kubyte *)sdata;
      kubyte *dst = (kubyte *)ddata;
      
      for (i = 1; i < dimension; i++)
	 num *= size[i];

      for (i = 0; i < num; i++) 
      {
	 status |= kdata_cast((kaddr)src, sdatatype, ddatatype, size[0],
			      convert, 1.0, 1.0, (kaddr) dst);
	 src += size[0] * kdata_size(sdatatype);
	 dst += (size[0] + 7) >> 3;
      }
   }
   else
   {
      for (i = 1, num = size[0]; i < dimension; i++)
	 num *= size[i];

      status = kdata_cast(sdata, sdatatype, ddatatype, num, convert, 
			  1.0, 1.0, ddata);
   }
   
   return status;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_preset_data
|
|       Purpose: This routine will initialize a block
|		 of data of an arbitrary dimensionality,
|		 size, and datatype to a given constant
|		 value.
|
|         Input: data - the data block to preset
|		 dimension - the dimensionality of the data
|		 size      - the size of the data 
|		 datatype  - the datatype of the data
|		 rval      - real value to use for initialization
|		 ival      - imaginary value to use for initialization
|
|        Output: the data is initialized in place
|
|       Returns: nothing
|
|    Written By: Jeremy Worley
|          Date: May 17, 1993
| Modifications:
|
------------------------------------------------------------*/
void
kdas_preset_data(
   kaddr  data, 
   int    dimension, 
   int   *size,
   int    datatype, 
   double rval, 
   double ival)
{
   int i;
   int number;


   /*-- bit must be done one line at a time --*/
   if (datatype == KBIT)
   {
      kubyte *d = data;
      int     linesize = (size[0] + 7) >> 3;       /*-- byte aligned lines --*/
      
      for (i = 1, number = 1; i < dimension; i++)  /*-- number of lines --*/
	 number *= size[i];

      for (i = 0; i < number; i++, d += linesize)
	 kdata_fill(d, size[0], KBIT, rval, ival);
   }
   else
   {
      for (i = 1, number = size[0]; i < dimension; i++)
	 number *= size[i];

      kdata_fill(data, number, datatype, rval, ival);
   }

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_preset_file
|
|       Purpose: This routine will initialize a transport
|		 of an arbitrary dimensionality, size, 
|		 and datatype to a given constant value.
|
|         Input: id        - the transport to preset
|		 dimension - the dimensionality of the data
|		 size      - the size of the data 
|		 datatype  - the datatype of the data
|		 rval      - real value to use for initialization
|		 ival      - imaginary value to use for initialization
|
|        Output: the data is initialized in place
|
|       Returns: nothing
|
|    Written By: Jeremy Worley
|          Date: May 17, 1993
| Modifications:
|
------------------------------------------------------------*/
void
kdas_preset_file(
   int    id, 
   int    dimension, 
   int   *size,
   int    datatype, 
   double rval, 
   double ival)
{
   int   position[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;
   int   unit[KDAS_MAX_DIM];
   kaddr data;
   int   i, number;
   long int ip;

   /*-- retain the initial position --*/
   ip = ktell(id);

   kdas_optimal_region_size(dimension, size, datatype, unit);

   /*-- construct a block of optimally-sized data --*/
   data = kdas_memory_construct(dimension, unit, datatype, rval, ival);

   /*-- determine number of points in optimal region --*/
   for (i = 1, number = unit[0]; i < dimension; i++)
      number *= unit[i];
   
   /*-- write out optimal regions to fill out file --*/
   do kwrite_generic(id, data, (size_t) number, datatype);
   while (kdas_advance_position(dimension, size, unit, position));

   kfree(data);

   /*-- set back the file position where we started --*/
   klseek(id, (off_t) ip, SEEK_SET);
   
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_compute_offsets
|
|       Purpose: This routine computes the linear offsets into a 
|		 dataspace for a given position.
|
|		 Consider the following block of memory, indicated 
|		 by '*' :
|
|  position (1,1)__ 0 1 2 3 4
|		   \-+-+-+-+-+     
|		 0 |\| | | | |   
|                  +-\-+-+-+-+   The offset for this example 
|		 1 | |X|*|*| |   will be 6 * size of the datatype.
|                  +-+-+-+-+-+     
|		 2 | |*|*|*| |
|                  +-+-+-+-+-+
|
|		 This gets more complicated for bit data, since
|		 the sizes need to be rounded up to the nearest
|		 byte.
|
|		 This is done for two dataspaces at once, since we typically
|		 need to calculate offsets for both a source and destination.
|
|         Input: dimension - the dimensionality of the data
|		 adatasize - the size of the data type in the first data space
|		 bdatasize - the size of the data type in the second data space
|		 asize     - the size of the first data space
|		 bsize     - the size of the second data space
|		 aposition - the position of the block in the first data space
|		 bposition - the position of the block in the second data space
|		 
|        Output: aoffset   - the resulting offset for the first data space
|        	 boffset   - the resulting offset for the second data space
|	
|       Returns: nothing
|
|    Written By: Jeremy Worley
|          Date: April 21, 1994
| Modifications:
|
------------------------------------------------------------*/
void
kdas_compute_offsets(
   int  dimension,
   int  datatype,
   int  adatasize,
   int  bdatasize,
   int *asize,
   int *bsize,
   int *aposition,
   int *bposition,
   int *aoffset,
   int *boffset)
{
   register int i;
   register int afactor;
   register int bfactor;

   if (datatype == KBIT)
   {
      *aoffset = aposition[0] >> 3;
      *boffset = bposition[0] >> 3;
      
      for (i = 1, afactor = ((asize[0]+7)>>3), bfactor = ((bsize[0]+7)>>3);
	   i < dimension;
	   i++, afactor *= asize[i-1], bfactor *= bsize[i-1])
      {
	 *aoffset += aposition[i] * afactor;
	 *boffset += bposition[i] * bfactor;
      }
   }
   else
   {
      *aoffset = 0;
      *boffset = 0;  

      /*-- compute the offsets into the data spaces --*/
      for (i = 0, afactor = adatasize, bfactor = bdatasize; 
	   i < dimension; 
	   i++, afactor *= asize[i-1], bfactor *= bsize[i-1])
      {
	 *aoffset += aposition[i] * afactor;
	 *boffset += bposition[i] * bfactor;
      }
   }

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_compute_strides
|
|       Purpose: This routine computes the strides required to
|		 traverse through a linear dataspace.  The stride
|		 for each dimension corresponds to the linear 
|		 stride needed to move to the next unit of data 
|		 in that dimension.
|
|		    0 1 2 3 4
|		   +-+-+-+-+-+     For this two-dimensional space,
|		 0 | | | | | |     the strides will be 
|                  +-+-+-+-+-+ 
|		 1 | | | | | |     stride[0] = 1 * datasize
|                  +-+-+-+-+-+     stride[1] = 5 * datasize
|		 2 | | | | | |
|                  +-+-+-+-+-+
|
|		 For each increment by one unit along width, we 
|		 stride linearly by a single datasize unit.  For 
|		 each increment by one unit along height, we stride
|	 	 linearly by five data units.  This is extended out
|		 into n-dimensions.
|
|		 This gets more complicated for bit data, since
|		 the sizes need to be rounded up to the nearest
|		 byte.
|
|		 This is performed on two data spaces at the same
|		 time as we typically need to do this for both a 
|		 source and destination space.
|
|         Input: dimension  - the dimensionality of the data
|		 datatype   - the datatype of the data
|		 adatasize  - the size of the datatype in the first dataspace
|		 bdatasize  - the size of the datatype in the second dataspace
|		 asize      - the size of the first data space
|		 bsize      - the size of the second data space
|
|        Output: astride    - the resulting stride for the first dataspace
|		 bstride    - the resulting stride for the second dataspace
|
|       Returns: nothing
|
|    Written By: Jeremy Worley
|          Date: April 21, 1994
| Modifications:
|
------------------------------------------------------------*/
void
kdas_compute_strides(
   int  dimension,
   int  datatype,
   int  adatasize,
   int  bdatasize,
   int *asize,
   int *bsize,
   int *astride,
   int *bstride)
{
   register int i;

   /*-- compute the striding into the data spaces --*/
   if (datatype == KBIT) 
   {
      astride[0] = bstride[0] = 0;
      astride[1] = (asize[0]+7) >> 3;
      bstride[1] = (bsize[0]+7) >> 3;
   
      for (i = 2; i < dimension; i++) 
      {
	 astride[i] = astride[i - 1] * asize[i - 1];
	 bstride[i] = bstride[i - 1] * bsize[i - 1];
      }
   }
   else 
   {
      for (i = 1, astride[0] = adatasize, bstride[0] = bdatasize; 
	   i < dimension; i++)
      {
	 astride[i] = astride[i - 1] * asize[i - 1];
	 bstride[i] = bstride[i - 1] * bsize[i - 1];
      }
   }
   
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_compute_skips
|
|       Purpose: This routine computes the skips needed to move
|		 from the end of a block of data to the beginning
|		 in a given dimension.
|
|		 Consider the following block of memory, indicated 
|		 by '*' :
|
|		    0 1 2 3 4
|		   +-+-+-+-+-+     For this two-dimensional space,
|		 0 | |*|*|*| |     the skips will be 
|                  +-+-+-+-+-+ 
|		 1 | |*|*|*| |     skip[0] = 0
|                  +-+-+-+-+-+     skip[1] = 2 * datasize
|		 2 | |*|*|*| |
|                  +-+-+-+-+-+
|
|		 After reading a unit along width, there is no skip
|		 to be done.  After reading a series of units along 
|		 height, we need to skip by 2 units to get to 
|		 beginning of the next row.  This is extended out
|		 to n-dimensions.
|
|		 This is performed on two data spaces at the same
|		 time as we typically need to do this for both a 
|		 source and destination space.
|
|         Input: dimension - the dimensionality of the data
|		 astride   - the strides through the first data space
|		 bstride   - the strides through the second data space
|		 aunit     - the unit size relative to the first space
|		 bunit     - the unit size relative to the first space
|		 
|        Output: askip     - the resulting skips for the data space
|	         bskip     - the resulting skips for the second data space
|
|       Returns: nothing
|
|    Written By: Jeremy Worley
|          Date: April 21, 1994
| Modifications:
|
------------------------------------------------------------*/
void
kdas_compute_skips(
   int  dimension,
   int *astride,
   int *bstride,
   int *aunit,
   int *bunit,
   int *askip, 
   int *bskip)
{
   register int i;

   /*-- compute the skip factors from the stride and the unit size --*/
   askip[0] = astride[0];
   bskip[0] = bstride[0];

   for (i = 1; i < dimension; i++)
   {
      askip[i] = astride[i] - astride[i - 1] * aunit[i - 1];
      bskip[i] = bstride[i] - bstride[i - 1] * bunit[i - 1];
   }
   
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_collapse_space - collapses a data space down
|
|       Purpose: This routine will collapse a block size down
|		 in dimensionality in cases where memory units
|		 running along a given dimension are physically 
|		 contiguous in memory.
|
|		 For example, consider the following block of
|		 memory, indicated by '*' :
|
|		    0 1 2 3 4 5 
|		   +-+-+-+-+-+-+
|		 0 | | | | | | |    Memory is stored width by height.
|		   +-+-+-+-+-+-+
|		 1 |*|*|*|*|*|*|
|		   +-+-+-+-+-+-+
|		 2 |*|*|*|*|*|*|
|		   +-+-+-+-+-+-+
|		 3 | | | | | | |
|		   +-+-+-+-+-+-+
|
|		 The data is in a space [6][4] with a beginning point 
|		 at [1][0] and a unit size of [6][2].
|
|		 Notice however that since the data completely spans
|		 the width, the lines of width are contiguous in memory.
|		 Therefore, instead of a two-dimensional block of data,
|		 we could instead consider this as a one-dimensional
|		 block of data.
|		
|		                        1 1 1 1 1 1 1 1 1 2 2 2 2 2
|		    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 8 9 0 1 2 3 4
|		   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|		   | | | | | | |*|*|*|*|*|*|*|*|*|*|*|*| | | | | | |
|		   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  old width index  0 1 2 3 4 5 0 1 2 3 4 5 0 1 2 3 4 5 0 1 2 3 4 5 
|
|		 Once collapsed down, the space is [24] with the
|		 beginning point at [6] and a unit size of [12].
|
|		 This optimizes the clipping we need to do, since
|		 we will be able to do fewer kmemcpy's.
|
|		 Both the source and destination spaces have to 
|		 be "collapsable" for this to work.
|
|         Input: dimension - the dimensionality of the data
|		 asize     - the size of the first data space
|		 bsize     - the size of the second data space
|		 unit      - the unit size of the block
|		 aposition - position in the first block
|		 bposition - position in the second block
|
|        Output: asize, bsize, unit, dimension, and the positions are 
|		 collapsed in place.
|
|       Returns: nothing
|
|    Written By: Steve Kubica and Jeremy Worley
|          Date: March 21, 1994
| Modifications:
|
------------------------------------------------------------*/
void 
kdas_collapse_space(
   int *dimension,
   int *asize,
   int *bsize,
   int *unit,
   int *aposition,
   int *bposition)
{
   int i;
   int j;

   /*-- no sanity checks -- assume all parameters are good --*/

   /*-- collapse any dimension where span of unit equals the size --*/
   for (i = 0; i < *dimension - 1; i++)
   {
      /*-- span must start at the beginning of the space and go to the end --*/
      if ((unit[i]   == asize[i])   && (unit[i]   == bsize[i])   &&
	  (unit[i+1] == asize[i+1]) && (unit[i+1] == bsize[i+1]))
      {
	 /*-- collapse this dimension with the next --*/
	 unit[i]  *= unit[i+1];
	 asize[i] *= asize[i+1];
	 bsize[i] *= bsize[i+1];

	 aposition[i] *= aposition[i+1];
	 bposition[i] *= bposition[i+1];

	 /*-- shift subsequent dimensions down --*/
	 for (j = i+1; j < *dimension; j++)
	 {
	    unit[j]  = unit[j+1];
	    asize[j] = asize[j+1];
	    bsize[j] = bsize[j+1];

	    aposition[j] = aposition[j+1];
	    bposition[j] = bposition[j+1];
	 }
	 (*dimension) -= 1; /*-- reduce dimensionality --*/
	 i--;		    /*-- stay where we are --*/
      }
   }

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_optimal_region_size
|
|       Purpose: This routine determines an optimal region size
|		 for accessing data in such a way that the region
|		 size will be below the khoros_threshold_size.
|
|         Input: dimension - the dimensionality of the segment
|		 size      - the size array of the segment
|		 datatype  - the datatype of the segment
|
|        Output: unit      - the optimal unit size for access
|
|       Returns: nothing
|
|    Written By: Steve Kubica
|          Date: May 23, 1995
| Modifications:
|
------------------------------------------------------------*/
void
kdas_optimal_region_size(
   int  dimension,
   int *size,
   int  datatype,
   int *unit)
{
   int i;
   int d = dimension - 1;


   /*-- start by looking at the whole enchilada --*/
   for (i = 0; i < dimension; i++)
      unit[i] = size[i];

   /*-- cut down the unit size until it isn't big anymore --*/
   while (kdas_is_big(dimension, unit, datatype) && d >= 0)
   {
      if (unit[d] % 2 == 0)       /*-- if possible divide by 2 --*/
	 unit[d] /= 2;
      else if (unit[d] % 3 == 0)  /*-- if possible divide by 3 --*/
	 unit[d] /= 3;
      else if (unit[d] % 5 == 0)  /*-- if possible divide by 5 --*/
	 unit[d] /= 5;
      else if (unit[d] % 7 == 0)  /*-- if possible divide by 7 --*/
	 unit[d] /= 7;
      else                        /*-- okay, I give... bag this dimension --*/
	 unit[d--] = 1;
   }

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_clip_corners
|
|       Purpose: This routine clips a given begin and end point
|	 	 so that it fits within the apparent size of 
|		 a dataspace.
|
|   		 The clipped data region will be entirely contained 
|		 in the data space.  Generally, the region will be 
|		 exactly the same size as the overall block, except 
|		 when the begin or end points of the block being 
|		 extracted go outside of the dataspace.  In this case, 
|		 we need to extract a clipped region which is entirely 
|		 inside the dataspace and insert it appropriately in 
|		 the block of data which we will return to the user.   
|
|		 This is illustrated below, with the clipped region 
|		 indicated by "///"'s.
|
| 				       +------+ 
| 			block we want  |  +---+----+
| 				   \____  |///|    |   data
| 				       +--+---+  ______/
| 					  |        |
| 				  	  +--------+
|                                       
|                                   +-----+
|                                  /|*|*|*| DATA CLIP
|                                 / |-+-+-|   size : clipped_size
|                                /  |*|*|*|   position in data space :
|	          DATA BLOCK    /   +-----+      clipped_begin, clipped_end
|	               +-------/-+       /    position in data block :
|	               |b| | |/| |      /        block_begin
|	               |-+-+-/-+-|     /
|	               | | |/| | |    /         
|	               |-+-+-----|---/---------+
|	               | | |*|*|*| |/| | | | | | DATA SPACE
|	               |-+-+-+-+-|-/-+-+-+-+-+-|   size 
|                      | | |*|*|e|/| | | | | | |
|                      +---------+-+-+-+-+-+-+-|
|                          | | | | | | | | | | |
|         b = begin        |-+-+-+-+/\-+-+-/\+-|
|         e = end          | /\| | /  \/\|/  \ |
|                          |/  \/\/  :   V    \|
|                                    :
|                                    :
|
|		The following information about the clipped region
|		will be returned :
|
|		  [1] the position of the clipped region in the data 
|	  	  [2] the position of the clipped region in the block
|		  [3] the unit size of the clipped region 
|
|		 This information is sufficient in order to clip
|		 from the data space into a region whose size could
|		 be determined by the original begin and end points.
|
|         Input: dimension     - the dimensionality of the data space
|		 size          - the size to clip on
|		 begin         - the beginning point of the data access
|		 		 this is relative to the data space.
|		 end           - the end point of the data access
|		 		 this is relative to the data space.
|
|        Output: clip              - returns TRUE if clipping occured
|		 outside           - returns TRUE if region is out of
|			  	     the bounds of the data space.
|		 clipped_begin     - the clipped begin point, relative 
|		                     to the apparent size 
|		 clipped_end       - the clipped end point, relative
|		 		     to the apparent size
|        	 block_begin       - the position of the clipped region
|                                    in the block 
|        	 clip_unit         - the size of the clipped region
|	
|       Returns: TRUE if all is well, FALSE, if something is amiss
|
|    Written By: Steve Kubica
|          Date: May 25, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_clip_corners(
   int  dimension,			 
   int *size, 
   int *begin, 
   int *end,
   int *clip,
   int *outside,
   int *clipped_begin,
   int *clipped_end,
   int *block_begin,
   int *clip_unit)
{
   int  i;

   /*-- no error checking on input or output arrays - assume they're there --*/

   /*-- assume these are false unless we determine otherwise --*/
   *clip = FALSE;
   *outside = FALSE;

   /*-- determine the clipped size from the begin and end points --*/
   for (i = 0; i < dimension; i++)
   {
      /*-- begin point before data space? --*/
      if (begin[i] < 0)
      {
	 if (end[i] < 0)  /*-- ... is whole thing before? --*/
	    *outside = TRUE;

	 clipped_begin[i] = 0;
	 block_begin[i]   = -begin[i];

	 *clip |= TRUE;
      }
      else
      {
	 clipped_begin[i] = begin[i];
	 block_begin[i] = 0;
      }

      /*-- end point after data space? --*/
      if (end[i] >= size[i])
      {
	 if (begin[i] >= size[i])  /*-- ... is whole thing after? --*/
	    *outside = TRUE;
	 
	 clipped_end[i] = size[i] - 1;

	 *clip |= TRUE;
      }
      else 
	 clipped_end[i] = end[i];

      clip_unit[i] = clipped_end[i] - clipped_begin[i] + 1;
   }

   return TRUE;
}


/*-----------------------------------------------------------
|
|  Routine Name: kdas_is_big - checks if a segment is big
|
|       Purpose: This routine checks the provided segment 
|		 characteristics to see if it is big.  For our 
|		 purposes, the term big means :
|
|  size[0]*size[1]* ... size[dim-1] * sizeof(datatype) < khoros_threshold_size
|	
|		 Where khoros_threshold_size is a variable global
|		 to this file.  The size is in bytes.
|
|		 The size is determined by the environment variable 
|		 KHOROS_DATA_SIZE.  If the environment variable is 
|		 not set, then a default size determined by the
|		 #define of KHOROS_THRESHOLD_SIZE_DEFAULT as it is
|		 defined in "internals.h" is used.  
|
|		 The environment KHOROS_KDMS_BUF_SIZE is also
|		 supported for backwards compatiblity.
|
|		 The routine kdas_set_threshold() can be used
|		 to explicitly set the threshold size.
|
|         Input: dimension - the dimensionality of the segment
|		 size      - the size array of the segment
|		 datatype  - the datatype of the segment
|
|        Output: none
|
|       Returns: TRUE if the segement is BIG, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: March 19, 1994
| Modifications:
|
------------------------------------------------------------*/
int kdas_is_big(
   int  dimension,
   int *size,
   int  datatype)
{
   /*--------------------------------------------------------
   |  
   |   There needs to be some consideration of the segment
   |   based attribute KDAS_BUFFER_THRESHOLD here.
   |
   \--------------------------------------------------------*/
   int overall_size;
   int i;


   /*-- initialize using environment variables, if they are set --*/
   if (!khoros_threshold_size_initialized)
   {
      char *env;

      /*-- try the new cool name first --*/
      env = kgetenv("KHOROS_DATA_SIZE");

      /*-- if that didn't get it, try the old environment variable --*/
      if (env == NULL)
	 env = kgetenv("KHOROS_KDMS_BUFSIZE");
      
      /*-- if one of those got it, then overwrite the default --*/
      if (env != NULL)
	 khoros_threshold_size = atol(env);

      /*-- don't do this again --*/
      khoros_threshold_size_initialized = TRUE;
   }
   
   /*-- at this point, we have a valid size to compare against --*/
   
   /*-- compute the overall size as a product of size and datatype --*/
   for (i = 0, overall_size = 1; i < dimension; i++)
      overall_size *= size[i];

   overall_size *= kdata_size(datatype);
   
   /*-- big means that the overall size is more than our threshold size --*/
   return (overall_size > khoros_threshold_size);
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_set_threshold - sets the global threshold size
|
|       Purpose: This routine is provided as a way to set the
|		 global variable khoros_threshold_size. The 
|		 size is in bytes.
|
|		 This is typically determined by the environment variable 
|		 KHOROS_DATA_SIZE.  If the environment variable is 
|		 not set and this routine is not called, then a default 
|		 size determined by the #define of 
|		 KHOROS_THRESHOLD_SIZE_DEFAULT as it is defined 
|		 in "internals.h" is used.  
|
|		 The environment KHOROS_KDMS_BUF_SIZE is also
|		 supported for backwards compatiblity.
|
|         Input: threshold - threshold size to set
|
|        Output: none
|
|       Returns: none
|
|    Written By: Steve Kubica
|          Date: November 21, 1994
| Modifications:
|
------------------------------------------------------------*/
void kdas_set_threshold(int threshold)
{
   khoros_threshold_size = threshold;
   khoros_threshold_size_initialized = TRUE;
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_get_order
|
|       Purpose: This routine extracts the value of an element
|		 in an array which corresponds to the specified
|		 direction with respect to the index order
|		 indicated by the order array.
|
|         Input: array     - the array to be extracted from
|		 order     - the order of the array
|		 dimension - the dimensionality of the array
|		 direction - the direction to extract
|		 def       - default to use if "direction" does
|		             not exist in the order array.
|
|        Output: returns the value of the array element in direction.
|
|       Returns: nothing
|
|    Written By: Jeremy Worley
|          Date: February 8, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdas_get_order(
   int *array,
   int *order,
   int  dimension,
   int  direction,
   int  def)
{
   int i;

   for (i = 0; i < dimension; i++)
      if (direction == order[i])
	 return array[i];
   
   return def;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_array_reorder
|
|       Purpose: This routine reorders an array in a source index 
|		 order into the order specified in a destination
|		 index order.
|
|         Input: array     - the array to be reordered
|		 dimension - the dimensionality of the array
|		 sorder    - the source order
|		 dorder    - the destination order
|
|        Output: the array is reordered in place
|
|       Returns: nothing
|
|    Written By: Steve Kubica
|          Date: May 23, 1995
| Modifications:
|
------------------------------------------------------------*/
void
kdas_array_reorder(
   int *array,
   int  dimension,
   int *sorder,
   int *dorder)
{
   int i;
   int narray[KDAS_MAX_DIM];

   for (i = 0; i < dimension; i++)
      narray[i] = kdas_get_order(array, sorder, dimension, dorder[i], -1);

   kmemcpy(array, narray, sizeof(int) * dimension);

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_array_max
|
|       Purpose: This is a simple utility routine which 
|		 finds the maximum value in an integer array.
|		 The array values must be 0 or greater.
|
|         Input: array - the order to search
|		 dimension - the dimensionality of the array
|
|        Output: none
|
|       Returns: the maximum integer is returned
|
|    Written By: Steve Kubica
|          Date: October 15, 1995
| Modifications:
|
------------------------------------------------------------*/
int 
kdas_array_max(
   int *array, 
   int dimension)
{
   int i;
   int max = -1;
   
   for (i = 0; i < dimension; i++)
      if (array[i] > max)
	 max = array[i];
   
   return max;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_initialized 
|
|       Purpose: This routine determines if an array of size
|		 dimension has been initialized.  If the
|		 array has not been allocated or has no
|		 dimensionality, then it is assumed to 
|		 be uninitialized.  If the array is allocated,
|		 but contains the value '-1', then it is
|		 considered to be uninitialized.
|
|         Input: array     - array to check
|		 dimension - dimensionality of the array 
|
|        Output: none
|
|       Returns: TRUE if initialized, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: October 16, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_initialized(int *array, int dimension)
{
   int i;
   
   /*-- check for obvious non-initialization --*/
   if (array == NULL || dimension <= 0)
      return FALSE;
   
   for (i = 0; i < dimension; i++)
      if (array[i] == -1)
	 return FALSE;
   
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_size_initialized 
|
|       Purpose: This routine determines if an array of size
|		 dimension has been initialized.  If the
|		 array has not been allocated or has no
|		 dimensionality, then it is assumed to 
|		 be uninitialized.  If the array is allocated,
|		 but contains the values of '0' or less, then it 
|		 is considered to be uninitialized.
|
|         Input: array     - array to check
|		 dimension - dimensionality of the array 
|
|        Output: none
|
|       Returns: TRUE if initialized, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: October 16, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_size_initialized(int *array, int dimension)
{
   int i;
   
   /*-- check for obvious non-initialization --*/
   if (array == NULL || dimension <= 0)
      return FALSE;
   
   for (i = 0; i < dimension; i++)
      if (array[i] <= 0)
	 return FALSE;
   
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_bit_resample
|
|       Purpose: This routine interpolates a line of bit data.
|
|         Input: src   	  - data block containing source line
|		 ssize 	  - number of bits in source line
|		 dsize 	  - number of bits in destination line
|		 bitshift - bit shift relative to destination block
|		 offset   - overall bit offset along this line
|
|        Output: dst   - src is resampled to here
|
|       Returns: nothing
|
|    Written By: Jeremy Worley
|          Date: August 2, 1994
| Modifications:
|
------------------------------------------------------------*/
void 
kdas_bit_resample(
   kubyte *src, 
   int     ssize, 
   int     dsize,
   int     bitshift,
   double  offset,
   kubyte *dst)
{
   register kubyte mask;
   register kubyte maskout;
   register int    i;
   double          istride = (double) ssize / (double) dsize;

   for (i = 0; i < (dsize >> 3) + ((dsize % 8) != 0); i++)
      dst[i] = 0;

   for (i = 0; i < dsize; i++) 
   {
      mask    = 0x80 >> ((int) (i * istride + bitshift + offset) % 8);
      maskout = 0x80 >> (i % 8);

      if (src[ ((int)(i * istride + bitshift + offset) >> 3) ] & mask)
	 dst[(int) (i >> 3)] |= maskout;
      else
	 dst[(int) (i >> 3)] |= 0x00;
   }

   return;
}

