 /*
  * Khoros: $Id: data.c,v 1.12 1996/05/15 16:30:13 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: data.c,v 1.12 1996/05/15 16:30:13 mirage Exp $";

#endif


/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Access Routines
   >>>>		
   >>>>   Static:
   >>>>                 _kosher_presentation()
   >>>>   Public:
   >>>>                 kdms_get_data()
   >>>>                 kdms_put_data()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

static int _kosher_presentation PROTO((kdas_presentation *, char *));

/****    =============================================================
 ****    Public Data Access Routines
 ****    =============================================================
 ****/

/************************************************************
*
*  Routine Name: kdms_get_data - get data from data object
*
*       Purpose: kdms_get_data is used to obtain data that is stored
*		 in a data object.  The data that is retrieved is
*		 designated by two "corner-markers".  These are arrays
*		 which contain N integer values, where N is the
*		 dimensionality of the segment (the dimensionality of
*		 a segment can be determined with the KDMS_DIMENSION
*		 attribute).  All values in begin argument must be
*		 less than or equal to their corresponding value in
*		 the end argument.  In a two dimensional case, the
*		 begin marker is the upper left corner and the end
*		 marker is the lower right corner of a rectangle that
*		 is obtained with this function call.  The coordinate
*		 origin is the upper, left, front corner of the data
*		 set extended to N-space.  The corner markers are
*		 specified in the index order that the data set is
*		 presented in.
*
*         Input: object - the object from which the data will be obtained.
*		 segment - the segment from which the data will be obtained.
*		 begin  - the begin marker of the region of data to
*		          be retrieved.
*		 end    - the end marker of the region of data to be
*			  retrieved.
*                data      - a pointer to the region of
*                            memory that will serve as a
*                            destination for the data. If
*                            this value is NULL, then
*                            sufficient space for this
*                            operation will be allocated
*                            for this operation.  The data
*                            type kaddr is used because
*                            it indicates a generic data
*                            pointer.
*
*        Output: If "data" is not initially NULL, then the memory
*		 that it points to will be filled with the data
*		 that is being requested.
*       Returns: If "data" is not initially NULL, then the data
*                space pointed to by "data" will be returned on
*                success.  If the "data" argument is NULL, then
*                a new pointer to the requested data will be
*                returned.  Unsuccessful calls to this routine
*                are indicated by a return value of NULL.
*
*  Restrictions: This routine assumes that if the argument
*                "data" is not NULL, then it contains the
*                appropriate amount of memory with the
*                appropriate dimensionality for the
*                requested primitive.
*
*    Written By: Jeremy Worley
*          Date: October 05, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kaddr
kdms_get_data(
   kobject object,
   char   *segment,
   int    *begin,
   int    *end,
   kaddr   data)
{
   kdms_quasi_segment *segd = NULL;
   kdas_presentation  *pres = NULL;
   kaddr tdata = NULL;
   

   KDMS_INIT;

   /*-- if there is a defined segment get_data routine, then use that... --*/
   if ((segd = kdms_locate_segment_define(segment)) != NULL)
   {
      if (segd->get_data)
      {
	 kaddr return_data;
	 
	 segd->lock = TRUE;
	 return_data = segd->get_data(object, segment, begin, end, data);
	 segd->lock = FALSE;
	 return return_data;
      }
   }
   
   /*-- ...otherwise, get the presentation so we can get from that --*/   
   if ((pres = kdms_get_segment_info(object, segment)) == NULL)
   {
      kdas_set_error(KDMS_ESEG_NONEXIST);
      return NULL;
   }
   
   /*-- make sure dimension, size, index order, and such are initialized --*/
   if (!_kosher_presentation(pres, "kdms_get_data"))
      return NULL;
   
   /*-- actually go and get the data --*/
   if ((tdata = kdas_pres_get_data(pres, begin, end, data)) == NULL)
      return NULL;

   /*-- process access data callbacks --*/
   kdms_process_data_callbacks(object, pres, begin, end, KDMS_CALLBACK_ACCESS);

   /*-- return the data --*/
   return tdata;
}

/************************************************************
*
*  Routine Name: kdms_put_data - put data into object
*
*       Purpose: kdms_put_data is used to store data that 
*		 into a data object.  The data that is stored is
*		 designated by two "corner-markers".  These are arrays
*		 which contain N integer values, where N is the
*		 dimensionality of the segment (the dimensionality of
*		 a segment can be determined with the KDMS_DIMENSION
*		 attribute).  All values in begin argument must be
*		 less than or equal to their corresponding value in
*		 the end argument.  In a two dimensional case, the
*		 begin marker is the upper left corner and the end
*		 marker is the lower right corner of a rectangle that
*		 is obtained with this function call.  The coordinate
*		 origin is the upper, left, front corner of the data
*		 set extended to N-space.  The corner markers are
*		 specified in the index order that the data set is
*		 presented in.
*
*         Input: object - the object from which the data will be stored.
*		 segment - the segment from which the data will be stored.
*		 begin  - the begin marker of the region of data to
*		          be retrieved.
*		 end    - the end marker of the region of data to be
*			  retrieved.
*                data   - a pointer to the region of
*                         memory that will serve as a
*
*        Output: 
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: October 5, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_put_data(
   kobject object,
   char   *segment,
   int 	  *begin,
   int 	  *end,
   kaddr   data)
{
   kdms_quasi_segment *segd = NULL;
   kdas_presentation  *pres = NULL;
   

   KDMS_INIT;

   /*-- if there is a defined segment put_data routine, then use that... --*/
   if ((segd = kdms_locate_segment_define(segment)) != NULL)
   {
      if (segd->put_data)
      {
	 int status;
	 
	 segd->lock = TRUE;
	 status = segd->put_data(object, segment, begin, end, data);
	 segd->lock = FALSE;
	 return status;
      }
   }

   /*-- ...otherwise, get the presentation so we can put into that --*/   
   if ((pres = kdms_get_segment_info(object, segment)) == NULL)
   {
      kdas_set_error(KDMS_ESEG_NONEXIST);
      return FALSE;
   }
   
   /*-- make sure dimension, size, index order, and such are initialized --*/
   if (!_kosher_presentation(pres, "kdms_put_data"))
      return FALSE;

   /*-- process save data callbacks --*/
   kdms_process_data_callbacks(object, pres, begin, end, KDMS_CALLBACK_SAVE);

   /*-- actually go and put the data --*/
   if (!kdas_pres_put_data(pres, begin, end, data))
      return FALSE;

   /*-- process change data callbacks --*/
   kdms_process_data_callbacks(object, pres, begin, end, KDMS_CALLBACK_CHANGE);

   /*-- if we got here, then all is well --*/ 
   return TRUE;
}

/****    =============================================================
 ****    static Utilities
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kosher_presentation
|
|       Purpose: This utility routine checks to be sure that 
|		 the dimension, datatype, size, and index
|		 order have been initialized okay.
|
|         Input: pres - presentation to check
|
|        Output: none
|
|       Returns: TRUE if kosher, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: February 29, 1996
| Modifications:
|
------------------------------------------------------------*/
static int
_kosher_presentation(kdas_presentation *pres, char *routine)
{
   /*-- check for zero size, dimension and unset size --*/
   if (pres->dimension <= 0)
   {
      kdas_set_error(KDAS_ESEG_UNKNOWN_DIMENSIONALITY);
      kerror("kdataman", routine, 
	     "The dimension of the %s segment has not been initialized."
	     "Unable to access data.\n", ktoken_to_string(pres->id));
      return FALSE;
   }

   if (!kdms_legal_datatype(pres->datatype))
   {
      kdas_set_error(KDAS_ESEG_UNKNOWN_DATATYPE);
      kerror("kdataman", routine, 
	     "The data type of the %s segment has not been initialized."
	     "Unable to access data.\n", ktoken_to_string(pres->id));
      return FALSE;
   }

   if (!kdms_legal_size(pres->size, pres->dimension))
   {
      kdas_set_error(KDAS_ESEG_UNKNOWN_SIZE);
      kerror("kdataman", routine, 
	     "The size of the %s segment has not been initialized or is not "
	     "legal.  Unable to access data.\n", ktoken_to_string(pres->id));
      return FALSE;
   }
	
   if (!kdms_legal_order(pres->order, pres->dimension))
   {
      kdas_set_error(KDAS_ESEG_UNKNOWN_ORDER);
      kerror("kdataman", routine, 
	     "The index order of the %s segment has not been "
	     "initialized or is not legal.  Unable to get data.\n",
	     ktoken_to_string(pres->id));
      return FALSE;
   }

   return TRUE;
}


