 /*
  * Khoros: $Id: segment.c,v 1.14 1996/05/20 01:03:45 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: segment.c,v 1.14 1996/05/20 01:03:45 mirage Exp $";

#endif


/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Segment Utility Routines
   >>>>		
   >>>>  Private:
   >>>>                 kdms_segment_create()
   >>>>                 kdms_segment_reference()
   >>>>                 kdms_segment_sync()
   >>>>                 kdms_segment_copy_data()
   >>>>                 kdms_segment_copy_physical_attributes()
   >>>>                 kdms_segment_destruct()
   >>>>                 kdms_get_segment_info()
   >>>>                 kdms_get_segtoken_info()
   >>>>                 kdms_add_segment()
   >>>>                 kdms_destroy_segments()
   >>>>                 kdms_reference_segments()
   >>>>                 kdms_sync_segments()
   >>>>                 kdms_write_segments()
   >>>>   Public:
   >>>>                 kdms_create_segment()
   >>>>                 kdms_destroy_segment()
   >>>>                 kdms_query_segment()
   >>>>                 kdms_rename_segment()
   >>>>                 kdms_set_segment()
   >>>>                 kdms_get_segment()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


/****    =============================================================
 ****    Public Segment Routines 
 ****    =============================================================
 ****/

/************************************************************
*
*  Routine Name: kdms_create_segment - create a segment on a data object.
*
*       Purpose: This function is used to create a segment that does
*                not already exist in a data object.  If the segment
*                already exists, then this function will generate
*                an error.
*
*         Input: object  - The object to create the new segment in.
*                segment - The name of segment to create.
*
*        Output: 
*
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: August 23, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_create_segment(
   kobject object,
   char   *segment)
{
   kdms_quasi_segment *segd = NULL;
   kdas_presentation  *pres = NULL;
   
   KDMS_INIT;

   if (object == NULL || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);
   
   /*-- if there is a defined segment creation routine, then use that... --*/
   if ((segd = kdms_locate_segment_define(segment)) != NULL)
      if (segd->create)
      {
	 int status;
	 
	 segd->lock = TRUE;
	 status = segd->create(object, segment);
	 segd->lock = FALSE;
	 return status;
      }

   /*-- ...otherwise, do a normal create, first checking for existence --*/   
   if (kdms_query_segment(object, segment))
   {
      kdas_set_error(KDMS_ESEG_EXISTS);
      return FALSE;
   }

   /*-- create the segment --*/
   pres = kdms_segment_create(object, segment);

   if (pres == NULL)
      return FALSE;
   else
      return TRUE;
}

/************************************************************
*
*  Routine Name: kdms_destroy_segment - destroy a segment from a data object.
*
*       Purpose: This function is used to destroy an existing 
*                segment from a data object.  Once a segment
*                has been destroyd, any data or attributes associated
*                with that segment will be lost forever.  A new
*                segment can be created in its place.
*
*                If the segment does not exist, then an error
*                message will be issued.
*
*         Input: object  - The object containing the segment to be destroyed.
*                segment - The name of segment to be destroyed.
*
*        Output: 
*
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: August 23, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_destroy_segment(
   kobject object,
   char   *segment)
{
   kdms_quasi_segment *segd = NULL;

   KDMS_INIT;

   if (object == NULL || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   /*-- if there is a defined segment destroy routine, then use that... --*/
   if ((segd = kdms_locate_segment_define(segment)) != NULL)
      if (segd->destroy)
      {
	 int status;
	 
	 segd->lock = TRUE;
	 status = segd->destroy(object, segment);
	 segd->lock = FALSE;
	 return status;
      }

   /*-- ...otherwise, do a normal destroy --*/   
   if (!kdms_query_segment(object, segment))
   {
      kdas_set_error(KDMS_ESEG_NONEXIST);
      return FALSE;
   }

   return kdms_segment_destruct(object, segment);
}

/************************************************************
*
*  Routine Name: kdms_query_segment - determine if a data segment is available.
*
*       Purpose: This function is used to determine if a data segment
*                in an object is available.  It returns TRUE
*                if the data segment exists, FALSE otherwise.
*
*         Input: object  - The object that segment may exist in.
*                segment - The name of segment to determine the existence of.
*
*        Output: 
*
*       Returns: TRUE if the data segment exists, FALSE otherwise
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: August 23, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_query_segment(
   kobject object, 
   char   *segment)
{
   kdms_quasi_segment *segd = NULL;

   KDMS_INIT;

   if (object == NULL || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   /*------------------------------------------------------------------
   | polymorph.c in appservices takes big advantage of this.
   |   not sure if this is true anymore .... // Steve K
   \-----------------------------------------------------------------*/
   if (segment == NULL)
      return TRUE;

   /*-- if there is a defined segment query routine, then use that... --*/
   if ((segd = kdms_locate_segment_define(segment)) != NULL)
      if (segd->query)
      {
	 int status;
	 
	 segd->lock = TRUE;
	 status = segd->query(object, segment);
	 segd->lock = FALSE;
	 return status;
      }

   /*-- ...otherwise, do a normal query --*/   
   if (kdms_get_segment_info(object, segment) == NULL)
      return FALSE;

   return TRUE;
}

/************************************************************
*
*  Routine Name: kdms_rename_segment - rename a segment
*
*       Purpose: This function is used to rename a segment in 
*		 an open object.  
*
*         Input: object   - The object to rename a segment on.
*		 old_name - The current name of the segment to be renamed.
*		 new_name - The new name of the segment.
*
*        Output:
*
*       Returns: TRUE if successful, FALSE otherwise
*
*  Restrictions:
*    Written By: Jeremy Worley
*          Date: August 14, 1994
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_rename_segment(
   kobject object, 
   char   *oname, 
   char   *nname)
{
   kobject r;
   klist  *l;
   kdas_presentation *pres;
   ktoken ntoken = kstring_to_token(nname);

   if ((pres = kdms_get_segment_info(object, oname)) == NULL)
   {
      kdas_set_error(KDMS_ESEG_NONEXIST);
      return FALSE;
   }

   /*-- update our object, first delete the entry from the segment list... --*/
   object->presentations = klist_delete(object->presentations,
					(kaddr) ((long)pres->id));

   /*-- ... then, update token identifiers... --*/
   pres->id          = ntoken;
   pres->segment->id = ntoken;

   /*-- ... and finally re-add the presentation under the new name --*/
   object->presentations = klist_add(object->presentations,
				     (kaddr) ((long)ntoken), pres);

   /*-- update all reference objects by renaming them as well --*/ 
   for (l = klist_head(object->phys->references); l != NULL; l = klist_next(l))
   {
      r = (kobject) l->identifier;
      
      if ((pres = kdms_get_segment_info(r, oname)) != NULL)
      {
	 object->presentations = klist_delete(r->presentations,
					      (kaddr) ((long)pres->id));
	 pres->id = ntoken;
	 object->presentations = klist_add(r->presentations,
					   (kaddr) ((long)ntoken), pres);
      }
   }
   
   return TRUE;
}

/************************************************************
*
*  Routine Name: kdms_initialize_segment
*
*       Purpose: This routine will initialize all the data
*		 in a specified segment to the values provided.
*
*         Input: object  - the object containing segment to initialize
*		 segment - name of the segment to initialize
*		 rval    - real value to initialize with
*		 ival    - imaginary value to initialize with
*
*        Output: none
*
*       Returns: TRUE on success, FALSE otherwise
*  Restrictions:
*    Written By: Steve Kubica
*          Date: December 26, 1994
*      Verified:
*  Side Effects:
* Modifications:
*   Declaration:
*
*************************************************************/
int
kdms_initialize_segment(
   kobject object, 
   char   *segment, 
   double  rval, 
   double  ival)
{
   kaddr data;
   int  *size;
   int  *unit;
   int   beg[KDMS_MAX_DIM] = {0, 0, 0, 0, 0};
   int   end[KDMS_MAX_DIM];
   int   dimension, datatype, i;


   if (!kdms_get_attributes(object, segment, 
			    KDMS_OPTIMAL_REGION_SIZE, &unit,
			    KDMS_SIZE,                &size,
			    KDMS_DIMENSION,           &dimension, 
			    KDMS_DATA_TYPE,           &datatype,
			    NULL))
      return FALSE;

   data = kdas_memory_construct(dimension, unit, datatype, rval, ival);

   do
   {
      for (i = 0; i < dimension; i++)
	 end[i] = beg[i] + unit[i] - 1;
   
      kdms_put_data(object, segment, beg, end, data);
   }
   while (kdas_advance_position(dimension, size, unit, beg));

   kfree(data);

   return TRUE;
}


/****    =============================================================
 ****    Glue Segment Routines
 ****    =============================================================
 ****/

/************************************************************
*
*  Routine Name: kdms_set_segment - assign information to a segment directly
*
*       Purpose: This function is used to directly assign information
* 		 to a data segment.  This should only be done inside
* 		 of file format glue routines, in their input routine.
*		 The characteristics are assigned directly to the new
*		 segment.  The segment will be created if it does not
*		 yet exist.
*
*		 The data argument is optional, as the data retrieval
*		 can either be defered until it is needed, or, if the
*		 large data set function locate_segment() is defined
*		 for that format, the large data set management will
*		 be handle all reading.
*
*		 Be careful when using this routine, as its misuse can
*		 easily corrupt a data object.
*
*         Input: object    - The object to set a segment on.
*		 segment   - The name of the segment to set.
*		 dimension - The dimensionality of the segment to assign.
*		 size      - The size of the segment to assign.
*		 order     - The order of the segment to assign.
*		 datatype  - The datatype of the segment to assign.
*
*		 data      - The actual segment data, or NULL. If this
*		 	     argument is provided, the data must be
*		  	     stored in memory according to the physical
*			     characteristics provided above.  The data
*			     will not be copied, but instead it will
*			     be assigned directly, so don't free it.
*        Output:
*
*       Returns: TRUE if successful, FALSE otherwise
*
*  Restrictions:
*    Written By: Jeremy Worley and Steve Kubica
*          Date: March 22, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_set_segment(
   kobject object,
   char   *segment,
   int     dimension,
   int    *size,
   int    *order,
   int     datatype,
   kaddr   data)
{
   kdas_presentation *pres;

   if (segment == NULL)
      kdas_fail_error(KDMS_ENULL_SEGMENT);

   /*-- create the segment if it does not exist yet --*/
   if ((pres = kdms_get_segment_info(object, segment)) == NULL)
      if ((pres = kdms_segment_create(object, segment)) == NULL)
	 kdas_fail_error(KINTERNAL);

   /*----------------------------------------------------------------
   |  we want to directly assign the physical characteristics to the
   |  physical layer..   however, we want to preserve all presentation
   |  characteristics which have already been set.  on the other hand,
   |  if they *haven't* been set yet, we want to set them now.  the
   |  size is particularly tricky, as we want to preserve the order
   |  of the presentation if it has been set.
   \----------------------------------------------------------------*/

   /*-- assign physical characteristics first... --*/
   if (!kdas_segment_set_all(pres->segment, dimension, size, order, 
			     datatype, data))
      return FALSE;

   /*-- ... then initialize all uninitialized fields of the presentation --*/


   /*-- uninitialized dimension means nothing is initialized --*/
   if (pres->dimension == -1)
      kdas_pres_set_all(pres, dimension, size, order, datatype, data);

   /*-- don't preserve dimension : change it if its different --*/
   if (pres->dimension != dimension)
      kdas_pres_set_dimension(pres, dimension);

   /*-- initialize the order, only if it is not initialized --*/
   if (pres->order != NULL && !kdas_initialized(pres->order, pres->dimension))
      kdas_pres_set_order(pres, order);

   /*-- initialize the size, only if it is not initialized --*/
   if (pres->size != NULL && !kdas_size_initialized(pres->size, 
						    pres->dimension))
   {
      if (size)  /*-- make sure we have this --*/
      {
	 /*-- must initialize it to be in the presentation order --*/
	 kmemcpy(pres->size,  size,  sizeof(int) * pres->dimension);
	 kdas_array_reorder(pres->size, pres->dimension, pres->segment->order, 
			    pres->order);
      }
   }

   /*-- initialize the datatype, only if it is not initialized --*/
   if (pres->datatype == -1)
      kdas_pres_set_datatype(pres, datatype);
   
   return TRUE;
}

/************************************************************
*
*  Routine Name: kdms_get_segment - retreive information from a segment
*
*       Purpose: This function is used to directly retrieve storage
*		 information from a data segment.  This should only 
*		 be done inside of file format glue routines, in their 
*		 output routine.
*
*		 All arguments are optional, passing NULL in will
*		 cause them to be ignored.  The characteristics are
*		 assigned directly to the pointers provided, so be
*		 careful not to free the size and order arrays.
*
*		 Data retreival can be done here, the data returned
*		 will be the actual physical pointer to the memory
*		 block.  If the data set is considered to be large and
*		 the format has large data set support, the pointer
*		 returned will be NULL since the data is being
*		 buffered inside of a temporary file.
*
*		 Be careful when using this routine, as its misuse can
*		 easily corrupt a data object.
*
*         Input: object    - The object to get segment information from.
*
*		 segment   - The name of the segment to get.
*
*        Output: dimension - The dimensionality of the segment.
*
*		 size      - The size of the segment. Be careful not 
*			     to free this array, as it points to the 
*		  	     internal array used by data services.
*
*		 order     - The order of the segment. Be careful not 
*			     to free this array, as it points to the 
*		  	     internal array used by data services.
*
*		 datatype  - The datatype of the segment.
*
*		 data      - The actual segment data, if it is all in 
*			     memory. This points to data services internal 
*			     buffer, so be careful not to free it.
*
*       Returns: TRUE if successful, FALSE otherwise
*
*  Restrictions:
*    Written By: Jeremy Worley and Steve Kubica
*          Date: March 22, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_get_segment(
   kobject object,
   char   *segment,
   int    *dimension,
   int   **size,
   int   **order,
   int    *datatype,
   kaddr  *data)
{
   kdas_presentation *pres;

   if (segment == NULL)
      kdas_fail_error(KDMS_ENULL_SEGMENT);

   /*-- create the segment if it does not exist yet --*/
   if ((pres = kdms_get_segment_info(object, segment)) == NULL)
      if ((pres = kdms_segment_create(object, segment)) == NULL)
	 kdas_fail_error(KINTERNAL);

   if (dimension)
      *dimension = pres->segment->dimension;
   
   if (size)
      *size = pres->segment->size;
   
   if (order)
      *order = pres->segment->order;
   
   if (datatype)
      *datatype = pres->segment->datatype;
   
   if (data)
      *data = pres->segment->data;

   return TRUE;
}


/****    =============================================================
 ****    Segment Manifestation Methods 
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdms_segment_create
|
|       Purpose: This routine will create a new segment on a
|		 data object.  The segment presentation will be
|		 added to the presentation list of the data object.
|
|         Input: object  - object in which to create the
|		   	   new segment presentation.
|		 segment - the name of the new segment.
|
|        Output: none
|
|       Returns: segment presentation, if successful, NULL otherwise
|
|    Written By: Steve Kubica 
|          Date: February 28, 1996
| Modifications:
|
------------------------------------------------------------*/
kdas_presentation *
kdms_segment_create(
   kobject  object, 
   char    *segment)
{
   kdas_presentation *pres = NULL;
   int coupling;

   /*-------------------------------------------------------------------
   |   uncoupled if this is an input object, and we are in the midst 
   |   of instantiating the object segments via a kdms_open.
   \-------------------------------------------------------------------*/
   if ((object->phys->flags & KOBJ_READ) && object->phys->creating)
      coupling = KUNCOUPLED;
   else
      coupling = KCOUPLED; /*-- write or temp object --*/

   /*-- construct the new segment --*/
   pres = kdas_pres_construct(segment, coupling);  

   /*-- a segment likes to know where it's towel is --*/
   pres->segment->format = object->phys->format;
   pres->segment->object = object;

   /*-- finally add it to the list --*/
   kdms_add_segment(object, pres);
   
   return pres;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_segment_reference
|
|       Purpose: This routine will create a reference to a
|		 given segment. 
|
|         Input: pres - the presentation structure to be referenced
|
|        Output: none
|
|       Returns: referenced presentation, if successful, NULL otherwise
|
|    Written By: Steve Kubica 
|          Date: February 28, 1996
| Modifications:
|
------------------------------------------------------------*/
kdas_presentation *
kdms_segment_reference(kdas_presentation *pres)
{
   kdas_presentation *rpres = NULL;

   /*-- only allow reference if all phys characteristics are initialized --*/
   if (!kdms_legal_size(pres->segment->size, pres->segment->dimension) ||
       !kdms_legal_datatype(pres->segment->datatype))
   {
      kinfo(KSYSLIB, "kdms_reference_segment : the %s segment in this "
	    "object has not been fully initialized.\n  Make sure that " 
	    "the dimension, datatype, size, and index order are set "
	    "before referencing.", ktoken_to_string(pres->id));
   }

   /*-- slap a new presentation on this segment --*/
   rpres = kdas_pres_slap_on_segment(pres->segment, KUNCOUPLED);

   /*-- new presentation should match old presentation --*/
   rpres->dimension = pres->dimension;
   rpres->datatype  = pres->datatype;

   kmemcpy(rpres->size,  pres->size,  sizeof(int)*pres->dimension);
   kmemcpy(rpres->order, pres->order, sizeof(int)*pres->dimension);

   rpres->size_interpolate    = pres->size_interpolate;
   rpres->complex_convert     = pres->complex_convert;
   rpres->access_interpolate  = pres->access_interpolate;
   rpres->rpad 	      	      = pres->rpad;
   rpres->ipad 	      	      = pres->ipad;
   rpres->scaling             = pres->scaling;
   rpres->scale_factor        = pres->scale_factor;
   rpres->scale_roffset       = pres->scale_roffset;
   rpres->scale_ioffset       = pres->scale_ioffset;
   rpres->norm_min 	      = pres->norm_min;
   rpres->norm_max 	      = pres->norm_max;

   /*-- copy the attribute list --*/
   kdms_copy_attribute_list(pres->segment->object, pres->segment->object,
			    pres->id, pres->attributes, &rpres->attributes);

   /*-- references are uncoupled --*/
   rpres->coupling = KUNCOUPLED;

   return rpres;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_segment_sync
|
|       Purpose: This routine will sync a given segment 
|		 presentation in the specified direction.
|
|		 This implies either that the physical characteristics 
|		 of size, index order, and datatype are pushed down 
|		 from the presentation layer to the physical layer, 
|		 or are updated in the presentation layer from the 
|		 physical values.  Uncoupled segments will not
|		 be syncronicable.
|
|         Input: pres      - segment presentation to sync
|		 direction - direction to syncronize, either
|		  	     KPRES2PHYS or KPHYS2PRES
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley
|          Date: August 24, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_segment_sync(kdas_presentation *pres, int direction)
{
   if (pres->coupling == KUNCOUPLED)
   {
      kdas_set_error(KDMS_ESYNC_NOT);
      return FALSE;
   }

   if (direction == KPRES2PHYS)
   {
      if (pres->coupling != KUNCOUPLED)
      {
	 kdas_segment_set_size(pres->segment, pres->size);
	 kdas_segment_set_order(pres->segment, pres->order);
	 kdas_segment_set_datatype(pres->segment, pres->datatype);
      }
   }
   else if (direction == KPHYS2PRES)
   {
      if (pres->coupling != KUNCOUPLED)
      {
	 kmemcpy(pres->size, pres->segment->size, sizeof(int)*pres->dimension);
	 kmemcpy(pres->order,pres->segment->order,sizeof(int)*pres->dimension);
	 pres->datatype = pres->segment->datatype;
      }
   }
   else
   {
      kdas_set_error(KDMS_ESYNC_INVALID);
      return FALSE;
   }

   return FALSE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_segment_copy_data
|
|       Purpose: Copy the data from one segment to another.
|		 It is assumed within this routine that both
|		 presentations have the same dimensionality, 
|		 size, index order, and datatype.
|
|         Input: spres - source presentation to copy from
|		 dpres - destination presentation to copy to
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: October 6, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_segment_copy_data(
   kdas_presentation *spres, 
   kdas_presentation *dpres)
{
   int begin[KDAS_MAX_DIM] =  {0, 0, 0, 0, 0};
   int end  [KDAS_MAX_DIM] =  {0, 0, 0, 0, 0};
   int unit [KDAS_MAX_DIM] =  {1, 1, 1, 1, 1};

   kaddr block = NULL;

   int transferable;
   int i;


   /*-- check for transferablilty attribute --*/
   if (!kdms_get_attribute(spres->segment->object, 
			   ktoken_to_string(spres->id), 
			   KDMS_TRANSFERABLE, &transferable))
      return FALSE;

   /*-- if the segment is not transferable, then drop out now --*/
   if (!transferable)
      return TRUE;
   
   /*-- get the optimal region size to copy data with --*/
   kdas_optimal_region_size(spres->dimension, spres->size, 
			    spres->datatype, unit);

   /*-- don't want to generate any callbacks during this operation --*/
   kdms_disable_callbacks();

   /*-- get data from source and put into destination --*/
   do
   {
      for (i = 0; i < spres->dimension; i++)
	 end[i] = begin[i] + unit[i] - 1;

      block = kdas_pres_get_data(spres, begin, end, block);
      kdas_pres_put_data(dpres, begin, end, block);
   }
   while (kdas_advance_position(spres->dimension, spres->size, unit, begin));

   kfree(block);

   /*-- all done! --*/
   kdms_enable_callbacks();

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_segment_copy_physical_attributes
|
|       Purpose: Copy all physical attributes from one object
|		 to another.
|
|         Input: spres - source presentation to copy from
|		 dpres - destination presentation to copy to
|
|		 force_copy - TRUE if all attributes should be
|			      copied, FALSE if only attributes
|			      which have not yet been initialized 
|			      should be copied
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: March 5, 1996
| Modifications:
|
------------------------------------------------------------*/
int
kdms_segment_copy_physical_attributes(
   kdas_presentation *spres, 
   kdas_presentation *dpres,
   int                force_copy)
{
   /*-- leave alone valid values which are already set, unless forcing --*/ 

   if (force_copy || dpres->dimension <= 0)
      if (!kdas_pres_set_dimension(dpres, spres->dimension))
	 return FALSE;

   if (force_copy || !kdms_legal_datatype(dpres->datatype))
      if (!kdas_pres_set_datatype(dpres, spres->datatype))
	 return FALSE;
   
   if (force_copy || !kdms_legal_size(dpres->size, dpres->dimension))
      if (!kdas_pres_set_size(dpres, spres->size))
	 return FALSE;

   if (force_copy || !kdms_legal_order(dpres->order, dpres->dimension))
      if (!kdas_pres_set_order(dpres, spres->order))
	 return FALSE;
      
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_segment_destruct
|
|       Purpose: This routine will delete a segment from a
|		 data object.  The segment presentation will be
|		 removed from the presentation list of the data object.
|
|         Input: object  - object from which to delete the
|		   	   new segment presentation.
|		 segment - the name of the new segment.
|
|        Output: none
|
|       Returns: TRUE if successful, FALSE otherwise
|
|    Written By: Steve Kubica 
|          Date: February 29, 1996
| Modifications:
|
------------------------------------------------------------*/
int 
kdms_segment_destruct(
   kobject  object, 
   char    *segment)
{
   kdas_presentation *pres;
   ktoken  stoken = kstring_to_token(segment);
   klist  *l;
   
   /*-- find the presentation --*/
   l = klist_locate(object->presentations, (kaddr) ((long)stoken));

   pres = (kdas_presentation *) l->client_data;

   /*-- delete callbacks occur for all references --*/
   kdms_process_data_callbacks(object, pres, NULL, NULL, KDMS_CALLBACK_DELETE);

   /*-- if we are the last reference, then call all destroy callbacks --*/
   if (klist_size(pres->segment->references) == 1)
      kdms_process_data_callbacks(object, pres, NULL, NULL, 
				  KDMS_CALLBACK_DESTROY);

   /*-- delete all callbacks for this particular reference --*/
   kdms_delete_callbacks_for_object(object, pres);

   /*-- delete its reference from the object presentation list... --*/
   object->presentations = klist_delete(object->presentations,
					(kaddr) ((long)stoken));

   /*-- ...and destroy the presentation --*/ 
   kdas_pres_destruct(pres);
   
   return TRUE;
}

/****    =============================================================
 ****    Object Segment Management
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdms_get_segment_info
|
|       Purpose: This routine will return the presentation from the
|		 given object for the given segment name.
|
|         Input: object  - object from which to return the
|		   	   segment presentation.
|		 segment - name of the segment to return
|        Output: none
|
|       Returns: segment presentation, if present, NULL otherwise
|
|    Written By: Steve Kubica and Mark Young
|          Date: February 28, 1996
| Modifications:
|
------------------------------------------------------------*/
kdas_presentation *
kdms_get_segment_info(
   kobject  object, 
   char    *segment)
{
   kdas_presentation *p = NULL;
   klist             *l;

   /*-- if it's on the list, then just return it! --*/
   l = kdas_list_string_locate(klist_head(object->presentations), segment);

   if (l != NULL)
   {
      p = (kdas_presentation *) l->client_data;

      /*-----------------------------------------------------------
      |  since some segment operations are object-dependent and
      |  we are not allowed to pass the object down to kdataccess,
      |  make this object the current object for any upcoming 
      |  segment operation. 
      \-----------------------------------------------------------*/
      p->segment->object = object;
   }
   return p;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_get_segtoken_info
|
|       Purpose: This routine will return the presentation from the
|		 given object for the given segment name.
|
|         Input: object - object from which to return the
|		   	  segment presentation.
|		 stoken - tokenized name of the segment to return 
|        Output: none
|
|       Returns: segment presentation, if present, NULL otherwise
|
|    Written By: Steve Kubica and Mark Young
|          Date: February 28, 1996
| Modifications:
|
------------------------------------------------------------*/
kdas_presentation *
kdms_get_segtoken_info(
   kobject object, 
   ktoken  stoken)
{
   kdas_presentation *p = NULL;
   klist             *l;

   /*-- if it's on the list, then just return it! --*/
   l = klist_locate(klist_head(object->presentations), (kaddr) ((long)stoken));
   
   if (l != NULL)
   {
      p = (kdas_presentation *) l->client_data;

      /*-----------------------------------------------------------
      |  since some segment operations are object-dependent and
      |  we are not allowed to pass the object down to kdataccess,
      |  make this object the current object for any upcoming 
      |  segment operation. 
      \-----------------------------------------------------------*/
      p->segment->object = object;
   }

   return p;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_add_segment
|
|       Purpose: This routine will add a segment presentation 
|		 to the presentation list of an object.
|
|         Input: object - object on which to add the 
|		   	  segment presentation.
|                pres   - segment presentation to add 
|
|        Output: none
|
|       Returns: nothing
|
|    Written By: Steve Kubica, Jeremy Worley, and Mark Young
|          Date: February 28, 1996
| Modifications:
|
------------------------------------------------------------*/
int
kdms_add_segment(
   kobject            object, 
   kdas_presentation *pres)
{
   /*-- if it already exists, then delete it --*/
   if (klist_locate(object->presentations, (kaddr) ((long)pres->id)))
      object->presentations = klist_delete(object->presentations,
                                           (kaddr) ((long)pres->id));

   /*-- add the new one to the list --*/
   object->presentations = klist_add(object->presentations,
				     (kaddr) ((long)pres->id), pres);
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_destroy_segments - destroy all segments in an object
|
|       Purpose: Delete all the data segments from an object.
|
|         Input: object - the object containing segments to destroy
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Mark Young
|          Date: September 15, 1992
| Modifications:
|
------------------------------------------------------------*/
int
kdms_destroy_segments(kobject object)
{
   ktoken *stoken;
   klist  *l, *c;
   
   l = object->presentations;
   while (l != NULL)
   {
      c = l;  l = klist_next(l);

      stoken = klist_identifier(c);

      kdms_segment_destruct(object, ktoken_to_string(stoken));
   }
   klist_free(object->presentations, NULL);

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_reference_segments - creates reference segments
|
|       Purpose: Reference all segments in a source object into a
|		 destination object.
|
|         Input: sobj - the input object to reference from
|                dobj - the output object to reference to
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: April 23, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_reference_segments(kobject sobj, kobject dobj)
{
   kdas_presentation *pres;
   kdas_presentation *ref;
   klist *l;

   for (l = klist_head(sobj->presentations); l != NULL; l = klist_next(l))
   {
      pres = (kdas_presentation *) l->client_data;

      /*-- make a reference and add it to the new objects list --*/
      ref = kdms_segment_reference(pres);
      kdms_add_segment(dobj, ref);
   }

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_sync_segments
|
|       Purpose: This routine will sync all segments in a given
|		 data object. 
|
|		 This implies either that the physical characteristics 
|		 of size, index order, and datatype are pushed down 
|		 from the presentation layer to the physical layer, 
|		 or are updated in the presentation layer from the 
|		 physical values.  Uncoupled segments will not
|		 be syncronicable.
|
|         Input: object - object containing segments to sync
|
|		 direction - direction to syncronize, either
|		  	     KPRES2PHYS or KPHYS2PRES
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley
|          Date: August 24, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_sync_segments(kobject object, int direction)
{
   kdas_presentation *pres;
   klist *l;

   for (l = klist_head(object->presentations); l != NULL; l = klist_next(l))
   {
      pres = (kdas_presentation *) l->client_data;
      kdms_segment_sync(pres, direction);
   }

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_write_segments - write out all segments
|
|       Purpose: This assumes that we have large data set
|		 support, and we need to write out our segments
|		 individually.  This routine goes to each
|		 segment in the list, and writes it out.
|
|         Input: object - object with segments to write
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: February 29, 1996
| Modifications:
|
------------------------------------------------------------*/
int
kdms_write_segments(kobject object)
{
   kdas_presentation *pres;
   klist *l;

   /*-- visit each segment and write it out --*/
   for (l = klist_head(object->presentations); l != NULL; l = klist_next(l))
   {
      pres = (kdas_presentation *) l->client_data;

      pres->segment->object = object;
      kdas_segment_write(pres->segment);
   }

   return TRUE;
}
