#include <dataserv.h>

/*--*  This test checks :
 *--*      
 *--*   o  basic segment presentation functionality with large data set support
 *--*   o  presentation transposing from an input object
 *--*
 *--*  A command-line parameter varies the threshold size 
 *--*     ktest 30  : will use a threshold of 30 bytes when testing 
 *--*/

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define O_DIM    3
#define O_SIZE  {5, 3, 3}
#define O_ORDER {1, 2, 3}

/*-- the original 5x3x3 data array to consider --*/
#define DATA_ARRAY \
    {  1, 1, 1, 6, 1, \
       1, 2, 3, 6, 1, \
       1, 1, 1, 6, 1, \
                      \
       2, 2, 2, 7, 2, \
       2, 9, 1, 7, 2, \
       2, 2, 2, 7, 2, \
                      \
       3, 3, 3, 8, 3, \
       3, 4, 8, 8, 3, \
       3, 3, 3, 8, 3  }

/*-- transposed height-first array --*/
#define T_ORDER1 {2, 1, 3}
#define T_ARRAY1 \
    {  1, 1, 1,    \
       1, 2, 1,    \
       1, 3, 1,    \
       6, 6, 6,    \
       1, 1, 1,    \
                   \
       2, 2, 2,    \
       2, 9, 2,    \
       2, 1, 2,    \
       7, 7, 7,    \
       2, 2, 2,    \
                   \
       3, 3, 3,    \
       3, 4, 3,    \
       3, 8, 3,    \
       8, 8, 8,    \
       3, 3, 3 }

/*-- transposed depth-first array --*/
#define T_ORDER2 {3, 1, 2}
#define T_ARRAY2 \
    {  1, 2, 3,    \
       1, 2, 3,    \
       1, 2, 3,    \
       6, 7, 8,    \
       1, 2, 3,    \
                   \
       1, 2, 3,    \
       2, 9, 4,    \
       3, 1, 8,    \
       6, 7, 8,    \
       1, 2, 3,    \
                   \
       1, 2, 3,    \
       1, 2, 3,    \
       1, 2, 3,    \
       6, 7, 8,    \
       1, 2, 3 }

/*-- transposed everything array --*/
#define T_ORDER3 {3, 2, 1}
#define T_ARRAY3 \
    {  1, 2, 3,    \
       1, 2, 3,    \
       1, 2, 3,    \
                   \
       1, 2, 3,    \
       2, 9, 4,    \
       1, 2, 3,    \
                   \
       1, 2, 3,    \
       3, 1, 8,    \
       1, 2, 3,    \
                   \
       6, 7, 8,    \
       6, 7, 8,    \
       6, 7, 8,    \
                   \
       1, 2, 3,    \
       1, 2, 3,    \
       1, 2, 3 }


/*-- piece testing --*/
#define PIECE_LOC { -2, -1, -1 } 
#define PIECE_SIZ {  5,  4,  3 }

/*-- transposed out-of-bounds padded --*/
#define T_ARRAY3p \
    {  0, 0, 0, 0, 0,  \
       0, 0, 0, 0, 0,  \
       0, 0, 0, 0, 0,  \
       0, 0, 0, 0, 0,  \
                       \
       0, 0, 0, 0, 0,  \
       0, 0, 1, 2, 3,  \
       0, 0, 1, 2, 3,  \
       0, 0, 1, 2, 3,  \
                       \
       0, 0, 0, 0, 0,  \
       0, 0, 1, 2, 3,  \
       0, 0, 2, 9, 4,  \
       0, 0, 1, 2, 3 }

/*-- transposed out-of-bounds wrapped --*/
#define T_ARRAY3w \
    {  2, 3, 1, 2, 3,  \
       2, 3, 1, 2, 3,  \
       2, 3, 1, 2, 3,  \
       2, 3, 1, 2, 3,  \
                       \
       2, 3, 1, 2, 3,  \
       2, 3, 1, 2, 3,  \
       2, 3, 1, 2, 3,  \
       2, 3, 1, 2, 3,  \
                       \
       2, 3, 1, 2, 3,  \
       2, 3, 1, 2, 3,  \
       9, 4, 2, 9, 4,  \
       2, 3, 1, 2, 3 }
   
/*-- compare two sets of physical characteristics --*/
int 
compare_attributes(
   int 	dim1,
   int 	dim2,
   int *size1,
   int *size2,
   int *order1,
   int *order2,
   int 	datatype1,
   int 	datatype2)
{
   if (dim1 != dim2)
      FAIL("dimension mismatch\n");

   if (size1 && size2)
   {
      if (kmemcmp(size1, size2, dim1*sizeof(int)) != 0)
	 FAIL("size mismatch");
   }
   else if (!size1 || !size2)
      FAIL("size missing");

   if (order1 && order2)
   {
      if (kmemcmp(order1, order2, dim1*sizeof(int)) != 0)
	 FAIL("order mismatch");
   }
   else if (!order1 || !order2)
      FAIL("order missing");
      
   if (datatype1 != datatype2)
      FAIL("datatype mismatch");

   return TRUE;
}

/*-- initialize a position and unit in some direction --*/
void
initialize_unit(
   int  dimension,
   int  direction,
   int *size,
   int *position,
   int *unit)
{
   int i;

   for (i = 0; i < dimension; i++)
   {
      position[i] = 0;
      unit[i]     = (i == direction) ? size[i] : 1;
   }
   return;
}

/*-- compare two 3D blocks of data --*/
void
compare_data(int *data, int *cdata, int *size)
{
   int j;

#if 0
   int k;
   kprintf("\n");
   for (k = 0; k < size[1]*size[2]; k++)
   {
      for (j = 0; j < size[0]; j++)
	 kprintf(" %3d ", data[j + k*size[0]]);
      kprintf("\n");
   }
#endif

   /*-- compare the data --*/
   for (j = 0; j < size[0]*size[1]*size[2]; j++)
      if (data[j] != cdata[j])
	 FAIL("incorrect data retrieved");

   return;
}

void
main(
   int  argc,
   char *argv[])
{
   kobject i, o;

   int  odim          = O_DIM;
   int  osize[O_DIM]  = O_SIZE;
   int  oorder[O_DIM] = O_ORDER;
   int  odatatype     = KINT;

   int  array[45] = DATA_ARRAY;

   int  torder1[O_DIM] = T_ORDER1;
   int  trans1 [45]    = T_ARRAY1;

   int  torder2[O_DIM] = T_ORDER2;
   int  trans2 [45]    = T_ARRAY2;

   int  torder3[O_DIM] = T_ORDER3;
   int  trans3 [45]    = T_ARRAY3;
   int  trans3p[60]    = T_ARRAY3p;
   int  trans3w[60]    = T_ARRAY3w;

   int  p[O_DIM] = PIECE_LOC;
   int  s[O_DIM] = PIECE_SIZ;

   int  idim;
   int *isize;
   int *iorder;
   int  idatatype;

   int *tsize;

   int *data = NULL;

   int  b[O_DIM] = {0, 0, 0};
   int  e[O_DIM];
   int  j;
   int  threshold = 1000;
   int  total;


   /*-- test segment with large data set support --*/
   kdas_test_mode(KDAS_TEST_PRESENTATION, TRUE);

   /*-- handle different thresholds according to command line argument --*/
   if (argc > 1)
      threshold = atoi(argv[1]);
   kdas_set_threshold(threshold);

   kprintf("basic presentation WITH large data set support\n\n");

   for (j = 0, total = 1; j < odim; j++)
      total *= osize[j];
   total *= kdata_size(odatatype);

   kprintf("\tdata size is ");
   if (total < threshold)
      kprintf("BELOW");
   else
      kprintf("ABOVE");
   kprintf(" the threshold size of %d\n\n", threshold);

   /*-- =========================================================== --*/
   /*-- phase I : output testing                                    --*/
   /*-- =========================================================== --*/
   kprintf("\tphase  I\n");
   kprintf("\t   writing all data to output...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   for (j = 0; j < odim; j++)
      e[j] = osize[j] - 1;

   if (!kdas_object_put_data(o, b, e, array))
      FAIL("unable to put the data\n");

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");
   kprintf("successful\n");


   /*-- input testing --*/


   /*-- read the object back in --*/
   if ((i = kdas_object_open("./test.data", KOBJ_READ)) == NULL)
      FAIL("unable to open input object\n");

   if (!kdas_object_get_all(i, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");

   /*-- =========================================================== --*/
   /*-- phase II : input height-first tranposed data                --*/
   /*-- =========================================================== --*/
   kprintf("\tphase II\n");
   kprintf("\t   reading height-first transposed data...");

   /*-- set the presentation index order --*/
   if (!kdas_object_set_presentation_order(i, torder1))
      FAIL("unable to set presentation order\n");

   /*-- get the transposed size --*/
   if (!kdas_object_get_all(i, NULL, &tsize, NULL, NULL, NULL))
      FAIL("unable to retrieve input object's transposed size\n");

   for (j = 0; j < odim; j++)
      e[j] = tsize[j] - 1;

   if ((data = (int *) kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the transposed data\n");

   compare_data(data, trans1, tsize);
   kfree(data);
   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase III : input depth-first tranposed data                --*/
   /*-- =========================================================== --*/
   kprintf("\tphase III\n");
   kprintf("\t   reading depth-first transposed data...");

   /*-- set the presentation index order --*/
   if (!kdas_object_set_presentation_order(i, torder2))
      FAIL("unable to set presentation order\n");

   /*-- get the transposed size --*/
   if (!kdas_object_get_all(i, NULL, &tsize, NULL, NULL, NULL))
      FAIL("unable to retrieve input object's transposed size\n");

   for (j = 0; j < odim; j++)
      e[j] = tsize[j] - 1;

   if ((data = (int *) kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the transposed data\n");

   compare_data(data, trans2, tsize);
   kfree(data);
   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase III : input tranpose everything                       --*/
   /*-- =========================================================== --*/
   kprintf("\tphase III\n");
   kprintf("\t   reading totally transposed data...");

   /*-- set the presentation index order --*/
   if (!kdas_object_set_presentation_order(i, torder3))
      FAIL("unable to set presentation order\n");

   /*-- get the transposed size --*/
   if (!kdas_object_get_all(i, NULL, &tsize, NULL, NULL, NULL))
      FAIL("unable to retrieve input object's transposed size\n");

   for (j = 0; j < odim; j++)
      e[j] = tsize[j] - 1;

   if ((data = (int *) kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the transposed data\n");

   compare_data(data, trans3, tsize);
   kfree(data);
   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase IV : input out-of-bounds tranpose with pad            --*/
   /*-- =========================================================== --*/
   kprintf("\tphase IV\n");
   kprintf("\t   reading out-of-bounds transposed with padding data...");

   /*-- set the physical characteristics we are interested in --*/
   if (!kdas_object_set_presentation_size(i, tsize, KPAD, 0.0, 0.0))
      FAIL("unable to set presentation characteristics\n");

   for (j = 0; j < odim; j++)
      e[j] = p[j] +s[j] - 1;

   if ((data = (int *) kdas_object_get_data(i, p, e, NULL)) == NULL)
      FAIL("unable to get the transposed data\n");

   compare_data(data, trans3p, s);
   kfree(data);
   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase V : input out-of-bounds tranpose with wrapping        --*/
   /*-- =========================================================== --*/
   kprintf("\tphase V\n");
   kprintf("\t   reading out-of-bounds transposed with wrapping data...");

   /*-- set the physical characteristics we are interested in --*/
   if (!kdas_object_set_presentation_size(i, tsize, KWRAP, 0.0, 0.0))
      FAIL("unable to set presentation characteristics\n");

   for (j = 0; j < odim; j++)
      e[j] = p[j] +s[j] - 1;

   if ((data = (int *) kdas_object_get_data(i, p, e, NULL)) == NULL)
      FAIL("unable to get the transposed data\n");

   compare_data(data, trans3w, s);
   kfree(data);
   kprintf("successful\n");


   if (!kdas_object_close(i))
      FAIL("unable to close input object\n");

   kunlink("test.data");

   kexit(KEXIT_SUCCESS);
}
