#include <dataserv.h>

/*--*  This test checks :
 *--*      
 *--*   o  basic segment presentation functionality with large data set support
 *--*   o  presentation upcasting from an input object
 *--*
 *--*  A command-line parameter varies the threshold size 
 *--*     ktest 30  : will use a threshold of 30 bytes when testing 
 *--*/

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define O_DIM    3
#define O_SIZE  {5, 2, 3}
#define O_ORDER {1, 2, 3}

/*-- signed version of 5x2x3 data array --*/
#define SIGNED_ARRAY \
    {  -8, -1, -1, -6, -1, \
        1,  2, 	3,  6, 	3, \
           		   \
       	2,  2, 	2,  7, 	2, \
       -2, -9, -1, -7, -5, \
       			   \
       	5,  3, 	3,  8, 	3, \
       	3,  4, 	8,  8, 	3  }

/*-- floating point version of 5x2x3 data array --*/
#define FLOAT_ARRAY             \
    {  -8.0, -1.0, -1.0, -6.0, -1.0, \
       	1.0,  2.0,  3.0,  6.0, 	3.0, \
       				     \
       	2.0,  2.0,  2.0,  7.0, 	2.0, \
       -2.0, -9.0, -1.0, -7.0, -5.0, \
       				     \
       	5.0,  3.0,  3.0,  8.0, 	3.0, \
       	3.0,  4.0,  8.0,  8.0, 	3.0  }


#define POSITION {-1, -2, -1}
#define SIZE     {4, 4, 4} 

#define REGION    \
    {  99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
                               \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, -8.0, -1.0, -1.0, \
       99.4,  1.0,  2.0,  3.0, \
                               \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4,  2.0,  2.0,  2.0, \
       99.4, -2.0, -9.0, -1.0, \
                               \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4,  5.0,  3.0,  3.0, \
       99.4,  3.0,  4.0,  8.0 }

/*-- compare two sets of physical characteristics --*/
int 
compare_attributes(
   int 	dim1,
   int 	dim2,
   int *size1,
   int *size2,
   int *order1,
   int *order2,
   int 	datatype1,
   int 	datatype2)
{
   if (dim1 != dim2)
      FAIL("dimension mismatch\n");

   if (size1 && size2)
   {
      if (kmemcmp(size1, size2, dim1*sizeof(int)) != 0)
	 FAIL("size mismatch");
   }
   else if (!size1 || !size2)
      FAIL("size missing");

   if (order1 && order2)
   {
      if (kmemcmp(order1, order2, dim1*sizeof(int)) != 0)
	 FAIL("order mismatch");
   }
   else if (!order1 || !order2)
      FAIL("order missing");
      
   if (datatype1 != datatype2)
      FAIL("datatype mismatch");

   return TRUE;
}

/*-- compare block of data against expected block --*/
void
check_data(kaddr data, int datatype)
{
   int j;

/*-- handy macros for printing --*/
#define PRINT_INT    \
   {                                             \
      int k;                                     \
      kprintf("\n");                             \
      for (k = 0; k < 6; k++)                    \
      {                                          \
	 for (j = 0; j < 5; j++)                 \
	    kprintf(" %3d ", d[j + k*5]);        \
	 kprintf("\n");                          \
      }                                          \
   }

#define PRINT_FLOAT    \
   {                                             \
      int k;                                     \
      kprintf("\n");                             \
      for (k = 0; k < 6; k++)                    \
      {                                          \
	 for (j = 0; j < 5; j++)                 \
	    kprintf(" %g", d[j + k*5]);          \
	 kprintf("\n");                          \
      }                                          \
   }

   switch (datatype)
   {
      case KBYTE :
	 {
	    char  cdata[30] = SIGNED_ARRAY;
	    char *d = (char *) data;

	    /* PRINT_INT; */
	    for (j = 0; j < 30; j++)
	       if (d[j] != cdata[j])
		  FAIL("incorrect data retrieved");
	 } break;

      case KSHORT :
	 {
	    short  cdata[30] = SIGNED_ARRAY;
	    short *d = (short *) data;

	    /* PRINT_INT; */
	    for (j = 0; j < 30; j++)
	       if (d[j] != cdata[j])
		  FAIL("incorrect data retrieved");
	 } break;

      case KLONG :
	 {
	    long  cdata[30] = SIGNED_ARRAY;
	    long *d = (long *) data;

	    /* PRINT_INT; */
	    for (j = 0; j < 30; j++)
	       if (d[j] != cdata[j])
		  FAIL("incorrect data retrieved");
	 } break;

      case KFLOAT :
	 {
	    float  cdata[30] = FLOAT_ARRAY;
	    float *d = (float *) data;

	    /* PRINT_FLOAT; */
	    for (j = 0; j < 30; j++)
	       if (d[j] != cdata[j])
		  FAIL("incorrect data retrieved");
	 } break;

      case KDOUBLE :
	 {
	    double  cdata[30] = FLOAT_ARRAY;
	    double *d = (double *) data;

	    /* PRINT_FLOAT; */
	    for (j = 0; j < 30; j++)
	       if (d[j] != cdata[j])
		  FAIL("incorrect data retrieved");
	 } break;
   }
   
   return;
}

void
main(
   int  argc,
   char *argv[])
{
   kobject i, o;

   int  odim          = O_DIM;
   int  osize[O_DIM]  = O_SIZE;
   int  oorder[O_DIM] = O_ORDER;
   int  odatatype     = KBYTE;

   char oarray[30]    = SIGNED_ARRAY;

   int  idim;
   int *isize;
   int *iorder;
   int  idatatype;

   kaddr data = NULL;

   double *d     = NULL;
   double  r[64] = REGION;

   int  b[O_DIM] = {0, 0, 0};
   int  e[O_DIM];
   int  j;
   int  threshold = 1000;
   int  total;

   int  p[O_DIM] = POSITION;
   int  s[O_DIM] = SIZE;


   /*-- test segment with large data set support --*/
   kdas_test_mode(KDAS_TEST_PRESENTATION, TRUE);

   /*-- handle different thresholds according to command line argument --*/
   if (argc > 1)
      threshold = atoi(argv[1]);
   kdas_set_threshold(threshold);

   kprintf("basic presentation WITH large data set support\n\n");

   for (j = 0, total = 1; j < odim; j++)
      total *= osize[j];
   total *= kdata_size(odatatype);

   kprintf("\tdata size is ");
   if (total < threshold)
      kprintf("BELOW");
   else
      kprintf("ABOVE");
   kprintf(" the threshold size of %d\n\n", threshold);

   /*-- =========================================================== --*/
   /*-- phase I : output testing                                    --*/
   /*-- =========================================================== --*/
   kprintf("\tphase  I\n");
   kprintf("\t   writing all data to output...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   for (j = 0; j < odim; j++)
      e[j] = osize[j] - 1;

   if (!kdas_object_put_data(o, b, e, oarray))
      FAIL("unable to put the data\n");

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");
   kprintf("successful\n");


   /*-- input testing --*/

   /*-- read the object back in --*/
   if ((i = kdas_object_open("./test.data", KOBJ_READ)) == NULL)
      FAIL("unable to open input object\n");

   if (!kdas_object_get_all(i, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");

   /*-- =========================================================== --*/
   /*-- phase II : input byte data                                  --*/
   /*-- =========================================================== --*/
   kprintf("\tphase II\n");
   kprintf("\t   reading byte data...");

   /*-- set the presentation datatype --*/
   if (!kdas_object_set_presentation_datatype(i, KBYTE))
      FAIL("unable to set presentation datatype\n");

   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the cast data\n");

   check_data(data, KBYTE);
   kfree(data);
   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase III : input short data                                --*/
   /*-- =========================================================== --*/
   kprintf("\tphase III\n");
   kprintf("\t   reading short data...");

   /*-- set the presentation datatype --*/
   if (!kdas_object_set_presentation_datatype(i, KSHORT))
      FAIL("unable to set presentation datatype\n");

   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the cast data\n");

   check_data(data, KSHORT);
   kfree(data);
   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase IV : input long data                                  --*/
   /*-- =========================================================== --*/
   kprintf("\tphase IV\n");
   kprintf("\t   reading long data...");

   /*-- set the presentation datatype --*/
   if (!kdas_object_set_presentation_datatype(i, KLONG))
      FAIL("unable to set presentation datatype\n");

   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the cast data\n");

   check_data(data, KLONG);
   kfree(data);
   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase V : input float data                                  --*/
   /*-- =========================================================== --*/
   kprintf("\tphase V\n");
   kprintf("\t   reading float data...");

   /*-- set the presentation datatype --*/
   if (!kdas_object_set_presentation_datatype(i, KFLOAT))
      FAIL("unable to set presentation datatype\n");

   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the cast data\n");

   check_data(data, KFLOAT);
   kfree(data);
   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase VI : input double data                                --*/
   /*-- =========================================================== --*/
   kprintf("\tphase VI\n");
   kprintf("\t   reading double data...");

   /*-- set the presentation datatype --*/
   if (!kdas_object_set_presentation_datatype(i, KDOUBLE))
      FAIL("unable to set presentation datatype\n");

   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the cast data\n");

   check_data(data, KDOUBLE);
   kfree(data);
   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase VII : input out-of-bounds double data                 --*/
   /*-- =========================================================== --*/
   kprintf("\tphase VII\n");
   kprintf("\t   reading out-of-bounds double data...");

   /*-- set the presentation out-of-bounds stuff  --*/
   if (!kdas_object_set_presentation_size(i, isize, KPAD, 99.4, 0.0))
      FAIL("unable to set presentation datatype\n");


   for (j = 0; j < odim; j++)
      e[j] = s[j] + p[j] - 1;

   if ((d = (double *) kdas_object_get_data(i, p, e, NULL)) == NULL)
      FAIL("unable to get the cast data\n");

#if 0   
   {
      int k;
      kprintf("\n");
      for (k = 0; k < s[1]*s[2]; k++)
      {
	 for (j = 0; j < s[0]; j++)
	    kprintf(" %g", d[j + k*s[0]]);
	 kprintf("\n");
      }
   }
#endif

   for (j = 0; j < 64; j++)
      if (d[j] != r[j])
	 FAIL("incorrect data retrieved");

   kfree(d);
   kprintf("successful\n");


   if (!kdas_object_close(i))
      FAIL("unable to close input object\n");
   kunlink("test.data");

   kexit(KEXIT_SUCCESS);
}
