#include <dataserv.h>

/*--*  This test checks :
 *--*      
 *--*   o  basic segment presentation functionality with large data set support
 *--*   o  presentation scaling and offset from an input object
 *--*
 *--*  A command-line parameter varies the threshold size 
 *--*     ktest 30  : will use a threshold of 30 bytes when testing 
 *--*/

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define COMPARISON_TOLERANCE 0.00001

#define O_DIM    3
#define O_SIZE  {5, 2, 3}
#define O_ORDER {1, 2, 3}

/*-- floating point version of 5x2x3 data array --*/
#define FLOAT_ARRAY             \
    {  -8.0, -1.0, -1.0, -6.0, -1.0, \
       	1.0,  2.0,  3.0,  6.0, 	3.0, \
       				     \
       	2.0,  2.0,  2.0,  7.0, 	2.0, \
       -2.0, -9.0, -1.0, -7.0, -5.0, \
       				     \
       	5.0,  3.0,  3.0,  8.0, 	3.0, \
       	3.0,  4.0,  8.0,  8.0, 	3.0  }

/*-- scaled array --*/
#define SCALED_ARRAY             \
    {  -13.6,  -1.7,  -1.7, -10.2,  -1.7, \
       	 1.7,   3.4,   5.1,  10.2,   5.1, \
       				          \
       	 3.4,   3.4,   3.4,  11.9,   3.4, \
        -3.4, -15.3,  -1.7, -11.9,  -8.5, \
       				          \
       	 8.5,   5.1,   5.1,  13.6,   5.1, \
       	 5.1,   6.8,  13.6,  13.6,   5.1   }

/*-- offset array --*/
#define OFFSET_ARRAY             \
    {  83.2, 90.2, 90.2, 85.2, 90.2, \
       92.2, 93.2, 94.2, 97.2, 94.2, \
       				     \
       93.2, 93.2, 93.2, 98.2, 93.2, \
       89.2, 82.2, 90.2, 84.2, 86.2, \
       				     \
       96.2, 94.2, 94.2, 99.2, 94.2, \
       94.2, 95.2, 99.2, 99.2, 94.2  }

/*-- scaled and offset array --*/
#define SCALED_OFFSET_ARRAY      \
    {  	77.6,  89.5,  89.5,  81.0,  89.5, \
       	92.9,  94.6,  96.3, 101.4,  96.3, \
       				          \
       	94.6,  94.6,  94.6, 103.1,  94.6, \
        87.8,  75.9,  89.5,  79.3,  82.7, \
       				          \
       	99.7,  96.3,  96.3, 104.8,  96.3, \
       	96.3,  98.0, 104.8, 104.8,  96.3  }


#define POSITION {-1, -2, -1}
#define SIZE     {4, 4, 4} 

#define REGION    \
    {  99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
                               \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 77.6, 89.5, 89.5, \
       99.4, 92.9, 94.6, 96.3, \
                               \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 94.6, 94.6, 94.6, \
       99.4, 87.8, 75.9, 89.5, \
                       \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.4, 99.4, 99.4, \
       99.4, 99.7, 96.3, 96.3, \
       99.4, 96.3, 98.0, 104.8 }

/*-- compare two sets of physical characteristics --*/
int 
compare_attributes(
   int 	dim1,
   int 	dim2,
   int *size1,
   int *size2,
   int *order1,
   int *order2,
   int 	datatype1,
   int 	datatype2)
{
   if (dim1 != dim2)
      FAIL("dimension mismatch\n");

   if (size1 && size2)
   {
      if (kmemcmp(size1, size2, dim1*sizeof(int)) != 0)
	 FAIL("size mismatch");
   }
   else if (!size1 || !size2)
      FAIL("size missing");

   if (order1 && order2)
   {
      if (kmemcmp(order1, order2, dim1*sizeof(int)) != 0)
	 FAIL("order mismatch");
   }
   else if (!order1 || !order2)
      FAIL("order missing");
      
   if (datatype1 != datatype2)
      FAIL("datatype mismatch");

   return TRUE;
}

/*-- compare two 3D blocks of data --*/
void
compare_data(double *data, double *cdata, int *size)
{
   int j;

#if 0
   int k;
   kprintf("\n");
   for (k = 0; k < size[1]*size[2]; k++)
   {
      for (j = 0; j < size[0]; j++)
	 kprintf(" %g ", data[j + k*size[0]]);
      kprintf("\n");
   }
#endif

   /*-- compare the data --*/
   for (j = 0; j < size[0]*size[1]*size[2]; j++)
      if (kabs(data[j] - cdata[j]) > COMPARISON_TOLERANCE)
	 FAIL("incorrect data retrieved");

   return;
}

void
main(
   int  argc,
   char *argv[])
{
   kobject i, o;

   int  odim          = O_DIM;
   int  osize[O_DIM]  = O_SIZE;
   int  oorder[O_DIM] = O_ORDER;
   int  odatatype     = KDOUBLE;

   double array[30]  = FLOAT_ARRAY;

   double sarray[30] = SCALED_ARRAY;
   double oarray[30] = OFFSET_ARRAY;
   double barray[30] = SCALED_OFFSET_ARRAY;
   double rarray[64] = REGION;

   int  idim;
   int *isize;
   int *iorder;
   int  idatatype;

   double *data = NULL;


   int  b[O_DIM] = {0, 0, 0};
   int  e[O_DIM];
   int  j;
   int  threshold = 1000;
   int  total;

   int  p[O_DIM] = POSITION;
   int  s[O_DIM] = SIZE;


   /*-- test segment with large data set support --*/
   kdas_test_mode(KDAS_TEST_PRESENTATION, TRUE);

   /*-- handle different thresholds according to command line argument --*/
   if (argc > 1)
      threshold = atoi(argv[1]);
   kdas_set_threshold(threshold);

   kprintf("basic presentation WITH large data set support\n\n");

   for (j = 0, total = 1; j < odim; j++)
      total *= osize[j];
   total *= kdata_size(odatatype);

   kprintf("\tdata size is ");
   if (total < threshold)
      kprintf("BELOW");
   else
      kprintf("ABOVE");
   kprintf(" the threshold size of %d\n\n", threshold);

   /*-- =========================================================== --*/
   /*-- phase I : output testing                                    --*/
   /*-- =========================================================== --*/
   kprintf("\tphase  I\n");
   kprintf("\t   writing all data to output...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   for (j = 0; j < odim; j++)
      e[j] = osize[j] - 1;

   if (!kdas_object_put_data(o, b, e, array))
      FAIL("unable to put the data\n");

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");
   kprintf("successful\n");


   /*-- input testing --*/

   /*-- read the object back in --*/
   if ((i = kdas_object_open("./test.data", KOBJ_READ)) == NULL)
      FAIL("unable to open input object\n");

   if (!kdas_object_get_all(i, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");

   /*-- =========================================================== --*/
   /*-- phase II : input scaled data                                --*/
   /*-- =========================================================== --*/
   kprintf("\tphase II\n");
   kprintf("\t   reading scaled data...");

   /*-- set the presentation scaling --*/
   if (!kdas_object_set_presentation_scaling(i, 1.7, 0.0, 0.0))
      FAIL("unable to set presentation scale factor\n");

   if ((data = (double *) kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the scaled data\n");

   compare_data(data, sarray, isize);
   kfree(data);
   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase III : input offset data                                --*/
   /*-- =========================================================== --*/
   kprintf("\tphase III\n");
   kprintf("\t   reading offset data...");

   /*-- set the presentation offset --*/
   if (!kdas_object_set_presentation_scaling(i, 1.0, 91.2, 0.0))
      FAIL("unable to set presentation scale offset\n");

   if ((data = (double *) kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the offset data\n");

   compare_data(data, oarray, isize);
   kfree(data);
   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase IV : input scaled and offset data                     --*/
   /*-- =========================================================== --*/
   kprintf("\tphase III\n");
   kprintf("\t   reading scaled and offset data...");

   /*-- set the presentation scaling and offset --*/
   if (!kdas_object_set_presentation_scaling(i, 1.7, 91.2, 0.0))
      FAIL("unable to set presentation scale offset\n");

   if ((data = (double *) kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get the offset data\n");

   compare_data(data, barray, isize);
   kfree(data);
   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase VII : input out-of-bounds byte data                   --*/
   /*-- =========================================================== --*/
   kprintf("\tphase VII\n");
   kprintf("\t   reading out-of-bounds byte data...");


   /*-- set the presentation out-of-bounds stuff  --*/
   if (!kdas_object_set_presentation_size(i, isize, KPAD, 99.4, 0.0))
      FAIL("unable to set presentation datatype\n");

   for (j = 0; j < odim; j++)
      e[j] = s[j] + p[j] - 1;

   if ((data = (double *) kdas_object_get_data(i, p, e, NULL)) == NULL)
      FAIL("unable to get the cast data\n");

   compare_data(data, rarray, s);
   kfree(data);
   kprintf("successful\n");

   if (!kdas_object_close(i))
      FAIL("unable to close input object\n");
   kunlink("test.data");

   kexit(KEXIT_SUCCESS);
}
