#include <dataserv.h>

/*--*  This test checks :
 *--*      
 *--*   o  basic segment functionality with no large data set support
 *--*   o  writing to an output object 
 *--*   o  reading from an input object
 *--*   o  total access of an entire data set
 *--*   o  out of bounds region-based gets : these _should_ fail
 *--*/

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define O_DIM    2
#define O_SIZE  {6, 5}
#define O_ORDER {1, 2}

/*-- the final 6x5 data array to consider --*/
#define DATA_ARRAY \
      {  0, 3, 1, 0, 0, 0,   \
	 0, 9, 8, 1, 0, 0,   \
	 0, 0, 9, 2, 3, 3,   \
	 7, 8, 0, 0, 3, 7,   \
	 7, 7, 0, 0, 7, 3 }

#define PIECE1_LOC { 1, -2}
#define PIECE1_SIZ { 2,  4}
#define PIECE1 \
   { -4, -3,   \
     -2, -1,   \
      3,  1,   \
      9,  8 }

#define PIECE2_LOC {2, 1}
#define PIECE2_SIZ {2, 2}
#define PIECE2 \
   { 8, 1,     \
     9, 2 }
   
#define PIECE3_LOC {-3, 3}
#define PIECE3_SIZ {5, 5}
#define PIECE3 \
   { -2, -2, -2,  7,  8,   \
     -3, -3, -3,  7,  7,   \
     -4, -4, -4, -4, -4,   \
     -4, -4, -4, -4, -4,   \
     -4, -4, -4, -4, -4}

#define PIECE4_LOC {4, 2}
#define PIECE4_SIZ {9, 9}
#define PIECE4 \
   {  3,  3, -1, -1, -1, -1, -1, -1, -1,   \
      3,  7, -2, -2, -2, -2, -2, -2, -2,   \
      7,  3, -3, -3, -3, -3, -3, -3, -3,   \
     -4, -4, -4, -4, -4, -4, -4, -4, -4,   \
     -5, -5, -5, -5, -5, -5, -5, -5, -5,   \
     -6, -6, -6, -6, -6, -6, -6, -6, -6,   \
     -7, -7, -7, -7, -7, -7, -7, -7, -7,   \
     -8, -8, -8, -8, -8, -8, -8, -8, -8,   \
     -9, -9, -9, -9, -9, -9, -9, -9, -9 }


/*-- compare two sets of physical characteristics --*/
int 
compare_attributes(
   int 	dim1,
   int 	dim2,
   int *size1,
   int *size2,
   int *order1,
   int *order2,
   int 	datatype1,
   int 	datatype2)
{
   if (dim1 != dim2)
      FAIL("dimension mismatch\n");

   if (size1 && size2)
   {
      if (kmemcmp(size1, size2, dim1*sizeof(int)) != 0)
	 FAIL("size mismatch");
   }
   else if (!size1 || !size2)
      FAIL("size missing");

   if (order1 && order2)
   {
      if (kmemcmp(order1, order2, dim1*sizeof(int)) != 0)
	 FAIL("order mismatch");
   }
   else if (!order1 || !order2)
      FAIL("order missing");
      
   if (datatype1 != datatype2)
      FAIL("datatype mismatch");

   return TRUE;
}

void
main(
   int  argc,
   char *argv[])
{
   kobject i, o;

   int  odim          = O_DIM;
   int  osize[O_DIM]  = O_SIZE;
   int  oorder[O_DIM] = O_ORDER;
   int  odatatype     = KINT;

   int  piece2[4]  = PIECE2;

   int  p1[O_DIM] = PIECE1_LOC;
   int  p2[O_DIM] = PIECE2_LOC;
   int  p3[O_DIM] = PIECE3_LOC;
   int  p4[O_DIM] = PIECE4_LOC;

   int  s1[O_DIM] = PIECE1_SIZ;
   int  s2[O_DIM] = PIECE2_SIZ;
   int  s3[O_DIM] = PIECE3_SIZ;
   int  s4[O_DIM] = PIECE4_SIZ;

   int  idim;
   int *isize;
   int *iorder;
   int  idatatype;

   int  cdata[30] = DATA_ARRAY;
   int *data = NULL;
      
   int  p[O_DIM];
   int  e[O_DIM];
   int  j, k;


   /*-- test segment with no large data set support --*/
   kdas_test_mode(KDAS_TEST_SEGMENT, FALSE);


   kprintf("basic segment functionality with no large data set support\n\n");

   /*-- =========================================================== --*/
   /*-- phase I : output testing                                    --*/
   /*-- =========================================================== --*/
   kprintf("\tphase  I\n");
   kprintf("\t   writing all the data ...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   /*-- put the data all at once --*/
   for (j = 0; j < odim; j++)
   {
      p[j] = 0;
      e[j] = osize[j] - 1;
   }
   
   /*-- put all of the data --*/
   if (!kdas_object_put_data(o, p, e, cdata))
      FAIL("unable to put data\n");

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");

   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase II : input testing                                    --*/
   /*-- =========================================================== --*/
   kprintf("\tphase II\n");
   kprintf("\t   performing out-of-bounds region accesses...");

   /*-- read the object back in --*/
   if ((i = kdas_object_open("./test.data", KOBJ_READ)) == NULL)
      FAIL("unable to open input object\n");
   
   if (!kdas_object_get_all(i, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");


   /*-- first piece : out of bounds --*/
   for (j = 0; j < odim; j++)
      e[j] = p1[j] + s1[j] - 1;

   if ((data = kdas_object_get_data(i, p1, e, NULL)) != NULL)
      FAIL("out of bounds data access succeeded but *should* have failed\n");


   /*-- second piece : in bounds --*/
   for (j = 0; j < odim; j++)
      e[j] = p2[j] + s2[j] - 1;

   if ((data = kdas_object_get_data(i, p2, e, NULL)) == NULL)
      FAIL("unable to get the second piece\n");

   /*-- compare the data --*/
   for (k = 0; k < s2[1]; k++)
      for (j = 0; j < s2[1]; j++)
	 if (data[j+k*s2[0]] != piece2[j+k*s2[0]])
	    FAIL("incorrect data retrieved for second piece");
   kfree(data);


   /*-- third piece : out of bounds --*/
   for (j = 0; j < odim; j++)
      e[j] = p3[j] + s3[j] - 1;

   if ((data = kdas_object_get_data(i, p3, e, NULL)) != NULL)
      FAIL("out of bounds data access succeeded but *should* have failed\n");


   /*-- fourth piece : way out of bounds --*/
   for (j = 0; j < odim; j++)
      e[j] = p4[j] + s4[j] - 1;

   if ((data = kdas_object_get_data(i, p4, e, NULL)) != NULL)
      FAIL("out of bounds data access succeeded but *should* have failed\n");


   if (!kdas_object_close(i))
      FAIL("unable to close input object\n");

   kprintf("successful\n");

   kexit(KEXIT_SUCCESS);
}
