#include <dataserv.h>

/*--*  This test checks :
 *--*      
 *--*   o  basic segment functionality WITH large data set support
 *--*   o  writing to an output object 
 *--*   o  writing to an input/output object
 *--*   o  reading from an input object
 *--*
 *--*  A command-line parameter varies the threshold size 
 *--*      ktest 30  : will use a threshold of 30 bytes when testing 
 *--*/

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define O_DIM    2
#define O_SIZE  {6, 5}
#define O_ORDER {1, 2}

/*-- the initial 6x5 data array to consider --*/
#define INITIAL_ARRAY \
      {  0,  3,  1,  0,  0,  0,   \
         0,  9,  8,  1,  0,  0,   \
         0,  0,  9,  2,  3,  3,   \
         7,  8,  0,  0,  3,  7,   \
         7,  7,  0,  0,  7,  3 }

/*-- the final 6x5 data array to consider --*/
#define FINAL_ARRAY \
      {  0,  3,	 1,  0,	 0,  0,   \
         0,  9,	 8,  1,	 0,  0,   \
         0,  0,	32, 32,	32,  3,   \
         7,  8,	32, 32,	32,  7,   \
         7,  7,	 0,  0,	 7,  3 }

#define REGION \
      {  32, 32, 32, \
	 32, 32, 32 }

#define REGION_BEGIN {2, 2}
#define REGION_END   {4, 3}

/*-- compare two sets of physical characteristics --*/
int 
compare_attributes(
   int 	dim1,
   int 	dim2,
   int *size1,
   int *size2,
   int *order1,
   int *order2,
   int 	datatype1,
   int 	datatype2)
{
   if (dim1 != dim2)
      FAIL("dimension mismatch\n");

   if (size1 && size2)
   {
      if (kmemcmp(size1, size2, dim1*sizeof(int)) != 0)
	 FAIL("size mismatch");
   }
   else if (!size1 || !size2)
      FAIL("size missing");

   if (order1 && order2)
   {
      if (kmemcmp(order1, order2, dim1*sizeof(int)) != 0)
	 FAIL("order mismatch");
   }
   else if (!order1 || !order2)
      FAIL("order missing");
      
   if (datatype1 != datatype2)
      FAIL("datatype mismatch");

   return TRUE;
}

void
main(
   int  argc,
   char *argv[])
{
   kobject i, io, o;

   int  odim          = O_DIM;
   int  osize[O_DIM]  = O_SIZE;
   int  oorder[O_DIM] = O_ORDER;
   int  odatatype     = KINT;

   int  idim;
   int *isize;
   int *iorder;
   int  idatatype;
      
   int *data = NULL;

   int  idata[30] = INITIAL_ARRAY;
   int  fdata[30] = FINAL_ARRAY;
   int  region[6] = REGION;
      
   int  b[O_DIM];
   int  e[O_DIM];
   int  rb[O_DIM] = REGION_BEGIN;
   int  re[O_DIM] = REGION_END;
   int  j, k;
   int  threshold = 1000;
   int  total;


   /*-- test segment WITH large data set support --*/
   kdas_test_mode(KDAS_TEST_SEGMENT, TRUE);

   /*-- handle different thresholds according to command line argument --*/
   if (argc > 1)
      threshold = atoi(argv[1]);
   kdas_set_threshold(threshold);

   kprintf("basic segment functionality WITH large data set support\n\n");

   for (j = 0, total = 1; j < odim; j++)
      total *= osize[j];
   total *= kdata_size(odatatype);

   kprintf("\tdata size is ");
   if (total < threshold)
      kprintf("BELOW");
   else
      kprintf("ABOVE");
   kprintf(" the threshold size of %d\n\n", threshold);

   /*-- =========================================================== --*/
   /*-- phase I : output testing                                    --*/
   /*-- =========================================================== --*/
   printf("\tphase  I\n");
   printf("\t   writing entire data set to output ...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   for (j = 0; j < odim; j++)
   {
      b[j] = 0;
      e[j] = osize[j] - 1;
   }

   /*-- put the intial data --*/
   if (!kdas_object_put_data(o, b, e, idata))
	 FAIL("unable to put data\n");

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");

   printf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase II : input/output processing                          --*/
   /*-- =========================================================== --*/
   printf("\tphase II\n");
   printf("\t   reading input/output data set and putting region ...");

   /*-- open the object as input/output --*/
   if ((io = kdas_object_open("./test.data", KOBJ_READ|KOBJ_WRITE)) == NULL)
      FAIL("unable to open object as an input/output\n");
   
   if (!kdas_object_get_all(io, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");

   /*-- put region of data --*/
   if (!kdas_object_put_data(io, rb, re, region))
	 FAIL("unable to put data\n");

   /*-- close the object --*/
   if (!kdas_object_close(io))
      FAIL("unable to close input/output object\n");

   printf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase III : input verification                              --*/
   /*-- =========================================================== --*/
   printf("\tphase III\n");
   printf("\t   reading input data to verify ...");

   /*-- open the object as input --*/
   if ((i = kdas_object_open("./test.data", KOBJ_READ)) == NULL)
      FAIL("unable to open input object\n");
   
   if (!kdas_object_get_all(i, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");

   for (j = 0; j < idim; j++)
   {
      b[j] = 0;
      e[j] = isize[j] - 1;
   }

   /*-- get the final data --*/
   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
	 FAIL("unable to get data\n");

   for (k = 0; k < isize[1]; k++)
   {
      for (j = 0; j < isize[0]; j++)
	 kprintf(" %3d ", data[j + k*osize[0]]);
      kprintf("\n");
   }

   /*-- compare the data --*/
   for (k = 0; k < isize[1]; k++)
      for (j = 0; j < isize[0]; j++)
	 if (data[j + k*osize[0]] != fdata[j + k*osize[0]])
	    FAIL("final data incorrect!\n");

   /*-- close the object --*/
   if (!kdas_object_close(i))
      FAIL("unable to close input object\n");

   kfree(data);

   printf("successful\n");

   kexit(KEXIT_SUCCESS);
}
