#include <dataserv.h>

/*--*  This test checks :
 *--*      
 *--*   o  basic segment presentation functionality with large data set support
 *--*   o  presentation padding on reads
 *--*
 *--*   o  write an entire data set to an output object 
 *--*
 *--*   o  set larger presentation size with PADDING (pad value of 99)
 *--*   o  reading entire large-size input data
 *--*   o  reading all rows of large-size input data 
 *--*   o  reading all columns of large-size input data 
 *--*   o  reading a region of large-size input data 
 *--*
 *--*  A command-line parameter varies the threshold size 
 *--*     ktest 30  : will use a threshold of 30 bytes when testing 
 *--*/

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define O_DIM    3
#define O_SIZE  {5, 3, 2}
#define O_ORDER {1, 2, 3}

/*-- the original 3x2 data array to consider --*/
#define DATA_ARRAY \
    {  1, 1, 1, 1, 1, \
       1, 2, 3, 9, 1, \
       1, 1, 1, 1, 1, \
                      \
       2, 2, 2, 2, 2, \
       2, 9, 1, 7, 2, \
       2, 2, 2, 2, 2  }

#define BIG_ARRAY \
    {  1,  1,  1,  1,  1, 99, 99, 99, 99, \
       1,  2,  3,  9,  1, 99, 99, 99, 99, \
       1,  1,  1,  1,  1, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
  \
       2,  2,  2,  2,  2, 99, 99, 99, 99, \
       2,  9,  1,  7,  2, 99, 99, 99, 99, \
       2,  2,  2,  2,  2, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
  \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
  \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
  \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99, \
      99, 99, 99, 99, 99, 99, 99, 99, 99  }

/*-- compare two sets of physical characteristics --*/
int 
compare_attributes(
   int 	dim1,
   int 	dim2,
   int *size1,
   int *size2,
   int *order1,
   int *order2,
   int 	datatype1,
   int 	datatype2)
{
   if (dim1 != dim2)
      FAIL("dimension mismatch\n");

   if (size1 && size2)
   {
      if (kmemcmp(size1, size2, dim1*sizeof(int)) != 0)
	 FAIL("size mismatch");
   }
   else if (!size1 || !size2)
      FAIL("size missing");

   if (order1 && order2)
   {
      if (kmemcmp(order1, order2, dim1*sizeof(int)) != 0)
	 FAIL("order mismatch");
   }
   else if (!order1 || !order2)
      FAIL("order missing");
      
   if (datatype1 != datatype2)
      FAIL("datatype mismatch");

   return TRUE;
}

/*-- initialize a position and unit in some direction --*/
void
initialize_unit(
   int  dimension,
   int  direction,
   int *size,
   int *position,
   int *unit)
{
   int i;

   for (i = 0; i < dimension; i++)
   {
      position[i] = 0;
      unit[i]     = (i == direction) ? size[i] : 1;
   }
   return;
}

void
main(
   int  argc,
   char *argv[])
{
   kobject i, o;

   int  odim          = O_DIM;
   int  osize[O_DIM]  = O_SIZE;
   int  oorder[O_DIM] = O_ORDER;
   int  odatatype     = KINT;

   int  array[30]  = DATA_ARRAY;
   int  cdata[360] = BIG_ARRAY;

   int  idim;
   int *isize;
   int *iorder;
   int  idatatype;

   int  asize[O_DIM] = {9, 8, 5};
   int  interpolate  = KPAD;

   int *data = NULL;
   int *d;

   int  u[O_DIM];
   int  b[O_DIM] = {0, 0, 0};
   int  e[O_DIM];
   int  j, k, l;
   int  threshold = 1000;
   int  total;


   /*-- test segment with large data set support --*/
   kdas_test_mode(KDAS_TEST_SEGMENT, TRUE);

   /*-- handle different thresholds according to command line argument --*/
   if (argc > 1)
      threshold = atoi(argv[1]);
   kdas_set_threshold(threshold);

   kprintf("basic segment presentation WITH large data set support\n\n");

   for (j = 0, total = 1; j < odim; j++)
      total *= osize[j];
   total *= kdata_size(odatatype);

   kprintf("\tdata size is ");
   if (total < threshold)
      kprintf("BELOW");
   else
      kprintf("ABOVE");
   kprintf(" the threshold size of %d\n\n", threshold);

   /*-- =========================================================== --*/
   /*-- phase I : output testing                                    --*/
   /*-- =========================================================== --*/
   kprintf("\tphase  I\n");
   kprintf("\t   writing all data to output...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   for (j = 0; j < odim; j++)
      e[j] = osize[j] - 1;

   if (!kdas_object_put_data(o, b, e, array))
      FAIL("unable to put the data\n");

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");

   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase II : input entire data set                            --*/
   /*-- =========================================================== --*/
   kprintf("\tphase II\n");
   kprintf("\t   reading entire data set at larger size...");

   /*-- read the object back in --*/
   if ((i = kdas_object_open("./test.data", KOBJ_READ)) == NULL)
      FAIL("unable to open input object\n");

   if (!kdas_object_get_all(i, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");

   /*-- set the physical characteristics we are interested in --*/
   if (!kdas_object_set_presentation_size(i, asize, interpolate, 99.0, 0.0))
      FAIL("unable to set presentation characteristics\n");

   /*-- get the data all at once --*/
   for (j = 0; j < idim; j++)
      e[j] = asize[j] - 1;
   
   /*-- get all of the data --*/
   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get data\n");

#if 0
   kprintf("\n");
   for (l = 0; l < asize[2]; l++)
   {
      for (k = 0; k < asize[1]; k++)
      {
	 for (j = 0; j < asize[0]; j++)
	    kprintf(" %3d ", data[j + k*asize[0] + l*asize[0]*asize[1]]);
	 kprintf("\n");
      }
      kprintf("--\n");
   }
#endif

   /*-- compare the data --*/
   for (l = 0; l < asize[2]; l++)
      for (k = 0; k < asize[1]; k++)
	 for (j = 0; j < asize[0]; j++)
	 if ( data[j + k*asize[0] + l*asize[0]*asize[1]] != 
	     cdata[j + k*asize[0] + l*asize[0]*asize[1]])
	    FAIL("incorrect data retrieved");

   kfree(data);   data = NULL;

   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase III : input data set in rows                          --*/
   /*-- =========================================================== --*/

   kprintf("\tphase III\n");
   kprintf("\t   reading data set in rows at larger size...");

   /*-- retrieve the data in rows --*/
   initialize_unit(idim, 0, asize, b, u);
   do 
   {
      /*-- calculate the end point at this position ... --*/
      for (j = 0; j < odim; j++)
         e[j] = b[j] + u[j] - 1;

      /*-- ... and get the data --*/
      if ((data = kdas_object_get_data(i, b, e, data)) == NULL)
         FAIL("unable to get data\n");

#if 0
      kprintf("\n");
      for (j = 0; j < asize[0]; j++)
         kprintf(" %3d ", data[j]);
#endif

      /*-- compare the data --*/
      for (j = 0; j < asize[0]; j++)
	 if ( data[j] != cdata[j + b[1]*asize[0] + b[2]*asize[0]*asize[1]])
	    FAIL("incorrect data retrieved");

   } while (kdas_advance_position(idim, asize, u, b));

   kfree(data);   data = NULL;

   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase IV : input data set in columns                        --*/
   /*-- =========================================================== --*/

   kprintf("\tphase IV\n");
   kprintf("\t   reading data set in columns at larger size...");

   /*-- retrieve the data in columns --*/
   initialize_unit(idim, 1, asize, b, u);
   do 
   {
      /*-- calculate the end point at this position ... --*/
      for (j = 0; j < odim; j++)
         e[j] = b[j] + u[j] - 1;

      /*-- ... and get the data --*/
      if ((data = kdas_object_get_data(i, b, e, data)) == NULL)
         FAIL("unable to get data\n");

#if 0
      kprintf("\n");
      for (j = 0; j < asize[1]; j++)
         kprintf(" %3d ", data[j]);
#endif

      /*-- compare the data --*/
      for (k = 0; k < asize[1]; k++)
	 if ( data[k] != cdata[b[0] + k*asize[0] + b[2]*asize[0]*asize[1]])
	    FAIL("incorrect data retrieved");

   } while (kdas_advance_position(idim, asize, u, b));

   kfree(data);   data = NULL;

   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase V : a region of data                                  --*/
   /*-- =========================================================== --*/

   kprintf("\tphase V\n");
   kprintf("\t   reading a region of data...");

   /*-- retrieve a region of the data --*/
   b[0] = 1;   b[1] = 2;   b[2] = 0;
   u[0] = 5;   u[1] = 4;   u[2] = 5;

   /*-- calculate the end point at this position ... --*/
   for (j = 0; j < odim; j++)
      e[j] = b[j] + u[j] - 1;

   /*-- ... and get the data --*/
   if ((data = kdas_object_get_data(i, b, e, data)) == NULL)
      FAIL("unable to get data\n");

   /*-- compare the data --*/
   d = data;

#if 0
   kprintf("\n");
   for (l = 0; l < u[2]; l++)
   {
      for (k = 0; k < u[1]; k++)
      {
	 for (j = 0; j < u[0]; j++)
	    kprintf(" %3d ", data[j + k*u[0] + l*u[0]*u[1]]);
	 kprintf("\n");
      }
      kprintf("--\n");
   }
#endif
   
   for (l = b[2]; l < b[2]+u[2]; l++)
      for (k = b[1]; k < b[1]+u[1]; k++)
	 for (j = b[0]; j < b[0]+u[0]; j++)
	 {
	    if (*d != cdata[j + k*asize[0] + l*asize[0]*asize[1]])
	    {
	       kprintf("at %d %d %d\n", j, k, l);
	       FAIL("incorrect data retrieved");
	    }
	    
	    d++;
	 }
   
      
   kfree(data);   data = NULL;
   kprintf("successful\n");

   if (!kdas_object_close(i))
      FAIL("unable to close input object\n");

   kexit(KEXIT_SUCCESS);
}
