#include <dataserv.h>

/*--*  This test checks :
 *--*      
 *--*   o  basic segment presentation functionality with large data set support
 *--*   o  presentation zero-order interpolation on writes
 *--*
 *--*   o  set larger presentation size with KZERO_ORDER interpolation 
 *--*
 *--*   o  writing entire large-size input data
 *--*   o  writing all rows of large-size input data
 *--*   o  writing all columns of large-size input data
 *--*
 *--*  A command-line parameter varies the threshold size 
 *--*     ktest 30  : will use a threshold of 30 bytes when testing 
 *--*/

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define O_DIM    3
#define O_SIZE  {10, 9, 4}
#define O_ORDER {1, 2, 3}

#define DATA_ARRAY \
    {  1, 1, 1, 1, 1, \
       1, 2, 3, 9, 1, \
       1, 1, 1, 1, 1, \
                      \
       2, 2, 2, 2, 2, \
       2, 9, 1, 7, 2, \
       2, 2, 2, 2, 2  }

#define BIG_ARRAY \
    {  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 2, 2, 3, 3, 9, 9, 1, 1, \
       1, 1, 2, 2, 3, 3, 9, 9, 1, 1, \
       1, 1, 2, 2, 3, 3, 9, 9, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
  \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 2, 2, 3, 3, 9, 9, 1, 1, \
       1, 1, 2, 2, 3, 3, 9, 9, 1, 1, \
       1, 1, 2, 2, 3, 3, 9, 9, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
  \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 9, 9, 1, 1, 7, 7, 2, 2, \
       2, 2, 9, 9, 1, 1, 7, 7, 2, 2, \
       2, 2, 9, 9, 1, 1, 7, 7, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
  \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 9, 9, 1, 1, 7, 7, 2, 2, \
       2, 2, 9, 9, 1, 1, 7, 7, 2, 2, \
       2, 2, 9, 9, 1, 1, 7, 7, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2, \
       2, 2, 2, 2, 2, 2, 2, 2, 2, 2 }

/*-- compare two sets of physical characteristics --*/
int 
compare_attributes(
   int 	dim1,
   int 	dim2,
   int *size1,
   int *size2,
   int *order1,
   int *order2,
   int 	datatype1,
   int 	datatype2)
{
   if (dim1 != dim2)
      FAIL("dimension mismatch\n");

   if (size1 && size2)
   {
      if (kmemcmp(size1, size2, dim1*sizeof(int)) != 0)
	 FAIL("size mismatch");
   }
   else if (!size1 || !size2)
      FAIL("size missing");

   if (order1 && order2)
   {
      if (kmemcmp(order1, order2, dim1*sizeof(int)) != 0)
	 FAIL("order mismatch");
   }
   else if (!order1 || !order2)
      FAIL("order missing");
      
   if (datatype1 != datatype2)
      FAIL("datatype mismatch");

   return TRUE;
}

/*-- initialize a position and unit in some direction --*/
void
initialize_unit(
   int  dimension,
   int  direction,
   int *size,
   int *position,
   int *unit)
{
   int i;

   for (i = 0; i < dimension; i++)
   {
      position[i] = 0;
      unit[i]     = (i == direction) ? size[i] : 1;
   }
   return;
}

/*-- compare an entire input data set with a known array --*/
int
compare_data(
   int *cdata,
   int  odim,
   int *osize,
   int *oorder,
   int  odatatype)
{
   kobject i;

   int  idim;
   int *isize;
   int *iorder;
   int  idatatype;
   
   int *data;

   int l, k, j;

   int  b[O_DIM];
   int  e[O_DIM];

   
   /*-- read the object back in --*/
   if ((i = kdas_object_open("./test.data", KOBJ_READ)) == NULL)
      FAIL("unable to open input object\n");

   if (!kdas_object_get_all(i, &idim, &isize, &iorder, &idatatype, NULL))
      FAIL("unable to retrieve input object's characteristics\n");

   if (!compare_attributes(odim, idim, osize, isize, oorder, iorder,
			   odatatype, idatatype))
      FAIL("physical characteristics not the same\n");

   /*-- get the data all at once --*/
   for (j = 0; j < idim; j++)
   {
      b[j] = 0;
      e[j] = isize[j] - 1;
   }
   
   /*-- get all of the data --*/
   if ((data = kdas_object_get_data(i, b, e, NULL)) == NULL)
      FAIL("unable to get data\n");

#if 0
   kprintf("\n");
   for (l = 0; l < isize[2]; l++)
   {
      for (k = 0; k < isize[1]; k++)
      {
	 for (j = 0; j < isize[0]; j++)
	    kprintf(" %3d ", data[j + k*isize[0] + l*isize[0]*isize[1]]);
	 kprintf("\n");
      }
      kprintf("--\n");
   }
#endif

   /*-- compare the data --*/
   for (l = 0; l < isize[2]; l++)
      for (k = 0; k < isize[1]; k++)
	 for (j = 0; j < isize[0]; j++)
	 if ( data[j + k*isize[0] + l*isize[0]*isize[1]] != 
	     cdata[j + k*isize[0] + l*isize[0]*isize[1]])
	    FAIL("incorrect data retrieved");

   kdas_object_close(i);
   kfree(data);
   
   return TRUE;
}

void
main(
   int  argc,
   char *argv[])
{
   kobject o;

   int  odim          = O_DIM;
   int  osize[O_DIM]  = O_SIZE;
   int  oorder[O_DIM] = O_ORDER;
   int  odatatype     = KINT;

   int  array[30]  = DATA_ARRAY;
   int  cdata[360] = BIG_ARRAY;

   int  asize[O_DIM] = {5, 3, 2};
   int  interpolate  = KZERO_ORDER;

   int *data = NULL;

   int  u[O_DIM];
   int  b[O_DIM] = {0, 0, 0};
   int  e[O_DIM];
   int  j, k, l;
   int  threshold = 2000;
   int  total;


   /*-- test segment with large data set support --*/
   kdas_test_mode(KDAS_TEST_SEGMENT, TRUE);

   /*-- handle different thresholds according to command line argument --*/
   if (argc > 1)
      threshold = atoi(argv[1]);
   kdas_set_threshold(threshold);

   kprintf("basic segment presentation WITH large data set support\n\n");

   for (j = 0, total = 1; j < odim; j++)
      total *= osize[j];
   total *= kdata_size(odatatype);

   kprintf("\tdata size is ");
   if (total < threshold)
      kprintf("BELOW");
   else
      kprintf("ABOVE");
   kprintf(" the threshold size of %d\n\n", threshold);

   /*-- =========================================================== --*/
   /*-- phase I : output large data set                             --*/
   /*-- =========================================================== --*/
   kprintf("\tphase  I\n");
   kprintf("\t   writing all data to output at smaller size...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   /*-- set the physical characteristics we are interested in --*/
   if (!kdas_object_set_presentation_size(o, asize, interpolate, 99.0, 0.0))
      FAIL("unable to set presentation characteristics\n");

   for (j = 0; j < odim; j++)
      e[j] = asize[j] - 1;

   if (!kdas_object_put_data(o, b, e, array))
      FAIL("unable to put the data\n");

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");

   /*-- |||||||||||||||||||||||||||||||||||||||||||||||||||||||| --*/

   if (!compare_data(cdata, odim, osize, oorder, odatatype))
      FAIL("data set differs\n");

   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase II : output data set in rows                          --*/
   /*-- =========================================================== --*/

   kprintf("\tphase II\n");
   kprintf("\t   writing data set in rows at smaller size...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   /*-- set the physical characteristics we are interested in --*/
   if (!kdas_object_set_presentation_size(o, asize, interpolate, 99.0, 0.0))
      FAIL("unable to set presentation characteristics\n");

   /*-- store the data in rows --*/
   initialize_unit(odim, 0, asize, b, u);
   data = (int *) kmalloc((size_t) u[0] * sizeof(int));
   do 
   {
      /*-- extract the row we want to put from the big data array --*/
      for (j = 0; j < u[0]; j++)
	 data[j] = array[j + b[1]*asize[0] + b[2]*asize[0]*asize[1]];

      /*-- calculate the end point at this position ... --*/
      for (j = 0; j < odim; j++)
         e[j] = b[j] + u[j] - 1;

      /*-- ... and get the data --*/
      if (!kdas_object_put_data(o, b, e, data))
         FAIL("unable to put data\n");

   } while (kdas_advance_position(odim, asize, u, b));

   kfree(data);   data = NULL;

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");

   /*-- |||||||||||||||||||||||||||||||||||||||||||||||||||||||| --*/

   if (!compare_data(cdata, odim, osize, oorder, odatatype))
      FAIL("data set differs\n");

   kprintf("successful\n");


   /*-- =========================================================== --*/
   /*-- phase III : output data set in columns                      --*/
   /*-- =========================================================== --*/

   kprintf("\tphase III\n");
   kprintf("\t   writing data set in columns at smaller size...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   /*-- set the physical characteristics we are interested in --*/
   if (!kdas_object_set_presentation_size(o, asize, interpolate, 99.0, 0.0))
      FAIL("unable to set presentation characteristics\n");

   /*-- store the data in columns --*/
   initialize_unit(odim, 1, asize, b, u);
   data = (int *) kmalloc((size_t) u[1] * sizeof(int));
   do 
   {
      /*-- extract the row we want to put from the big data array --*/
      for (j = 0; j < u[1]; j++)
	 data[j] = array[b[0] + j*asize[0] + b[2]*asize[0]*asize[1]];

      /*-- calculate the end point at this position ... --*/
      for (j = 0; j < odim; j++)
         e[j] = b[j] + u[j] - 1;

      /*-- ... and get the data --*/
      if (!kdas_object_put_data(o, b, e, data))
         FAIL("unable to put data\n");

   } while (kdas_advance_position(odim, asize, u, b));

   kfree(data);   data = NULL;

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");

   /*-- |||||||||||||||||||||||||||||||||||||||||||||||||||||||| --*/

   if (!compare_data(cdata, odim, osize, oorder, odatatype))
      FAIL("data set differs\n");

   kprintf("successful\n");

   /*-- =========================================================== --*/
   /*-- phase IV : output data set in regions                       --*/
   /*-- =========================================================== --*/

   kprintf("\tphase IV\n");
   kprintf("\t   writing data set in regions at smaller size...");

   /*-- write out an output object --*/
   if ((o = kdas_object_open("./test.data", KOBJ_WRITE)) == NULL)
      FAIL("unable to open output object\n");

   /*-- set physical characteristics --*/
   if (!kdas_object_set_all(o, odim, osize, oorder, odatatype, NULL))
      FAIL("unable to set output object characteristics\n");

   /*-- set the physical characteristics we are interested in --*/
   if (!kdas_object_set_presentation_size(o, asize, interpolate, 99.0, 0.0))
      FAIL("unable to set presentation characteristics\n");

   /*-- store the data in regions --*/
   b[0] = 0;   b[1] = 0;   b[2] = 0;
   u[0] = 2;   u[1] = 1;   u[2] = 1;
   
   data = (int *) kmalloc((size_t) u[0]*u[1]*u[2] * sizeof(int));
   do 
   {
      /*-- extract the piece we want to put from the big data array --*/
      for (l = 0; l < u[2]; l++)
	 for (k = 0; k < u[1]; k++)
	    for (j = 0; j < u[0]; j++)
	       data[j + k*u[0] + l*u[0]*u[1]] = 
		  array[(b[0]+j) + (b[1]+k)*asize[0] 
		       +(b[2]+l)*asize[0]*asize[1]];

      /*-- calculate the end point at this position ... --*/
      for (j = 0; j < odim; j++)
         e[j] = b[j] + u[j] - 1;

      /*-- ... and get the data --*/
      if (!kdas_object_put_data(o, b, e, data))
         FAIL("unable to put data\n");

   } while (kdas_advance_position(odim, asize, u, b));

   kfree(data);   data = NULL;

   /*-- close the object --*/
   if (!kdas_object_close(o))
      FAIL("unable to close output object\n");


   /*-- |||||||||||||||||||||||||||||||||||||||||||||||||||||||| --*/

   if (!compare_data(cdata, odim, osize, oorder, odatatype))
      FAIL("data set differs\n");

   kprintf("successful\n");

   kexit(KEXIT_SUCCESS);
}
