#include <dataserv.h>
#include "io_test.h"

#define FAIL(error)               \
        {                         \
          kinfo(KFORCE,error);    \
          kexit(KEXIT_FAILURE);   \
        }

#define COORD_LIST  "coordList"
#define COORD_ARRAY "coordArray"

void define_attributes PROTO((void));

/*-- 
  -- write out a strucure to a file 
  -- . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
  --
  --*/
void define_attributes(void)
{
   kdms_define_attribute(NULL, COORD_LIST, 1, 1, KTYPE_COORD_LIST,
			 TRUE, TRUE, NULL);

   kdms_define_attribute(NULL, COORD_ARRAY, 1, 1, KTYPE_COORD_ARRAY,
			 TRUE, TRUE, NULL);
}

/*-- 
  -- write out a strucure to a file 
  -- . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
  --
  --*/
void write_structure(char *name, kaddr s, int type)
{
   int fid = kopen(name, O_WRONLY|O_CREAT|O_TRUNC, 0644);
   kwrite_generic(fid, s, (size_t) 1, type);
   kclose(fid);
}

/*-- 
  -- read in a strucure from a file 
  -- . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
  --
  --*/
void read_structure(char *name, kaddr s, int type)
{
   int fid = kopen(name, O_RDONLY, 0644);
   kread_generic(fid, s, (size_t) 1, type);
   kclose(fid);

   /*-- tidy up --*/
   kunlink(name);
}


/*-- 
  -- coord routines
  -- . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
  -- 
  --*/
coord_p create_coord(int z)
{
   coord_p c;
   
   c = (coord_p) kmalloc(sizeof(coord_t));
   c->x = (double) z;
   c->y = (double) z+1;

   return c;
}

int compare_coord(coord_p c1, coord_p c2)
{
   if (kmemcmp(c1, c2, sizeof(coord_t)) == 0)
      return TRUE;
   else
      return FALSE;
}


/*-- 
  -- coord_list routines 
  -- . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
  -- 
  --*/
coord_list_p create_coord_list(int len, int seed)
{
   coord_list_p l;

   l = (coord_list_p) kmalloc(sizeof(coord_list_t));

   /*-- create initial coord... --*/
   l->c1 = create_coord(seed++);

   kmemcpy(&l->c2, l->c1, sizeof(coord_t));
   kmemcpy(&l->c3, l->c1, sizeof(coord_t));

   /*-- ... and recursively create all subsequent coords --*/ 
   if (len == 1)
      l->next = NULL;
   else
      l->next = create_coord_list(len-1, seed);
   
   return l;
}

int compare_coord_list(coord_list_p l1, coord_list_p l2)
{
   /*-- base case --*/
   if (l1 == NULL && l2 == NULL)
      return TRUE;

   /*-- wait a sec.. they can't *both* be NULL --*/
   if (l1 == NULL || l2 == NULL)
      return FALSE;
   
   
   /*-- check our coords ... --*/
   if (!compare_coord(l1->c1, l2->c1))
      return FALSE;

   if (!compare_coord(&l1->c2, &l2->c2))
      return FALSE;

   if (!compare_coord(&l1->c3, &l2->c3))
      return FALSE;

   /*-- ... and recursively all subsequent coords --*/
   if (!compare_coord_list(l1->next, l2->next))
      return FALSE;
   
   return TRUE;
}


/*-- 
  -- coord_list routines 
  -- . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
  -- 
  --*/
coord_array_p create_coord_array(int len, int seed)
{
   coord_array_p a;
   int i;

   a = (coord_array_p) kmalloc(sizeof(coord_list_t));

   a->list.list_len = len;
   a->list.list_val = (coord_p *) kmalloc(len * sizeof(coord_p));

   for (i = 0; i < len; i++)
      a->list.list_val[i] = create_coord(seed++);
   
   return a;
}

int compare_coord_array(coord_array_p a1, coord_array_p a2)
{
   int i;
   
   if (a1 == NULL || a2 == NULL)
      return FALSE;

   if (a1->list.list_len != a2->list.list_len)
      return FALSE;

   /*-- check our coords ... --*/
   for (i = 0; i < a1->list.list_len; i++)
      if (!compare_coord(a1->list.list_val[i], a2->list.list_val[i]))
	 return FALSE;

   return TRUE;
}

void
main()
{
   kobject i, o;
   char *filename = "t.data";

   coord_list_p  olist  = NULL, ilist  = NULL;
   coord_array_p oarray = NULL, iarray = NULL;
   

   /*-- define needed datatypes and attributes --*/
   kdefine_coord_list_p();
   kdefine_coord_list_t();
   kdefine_coord_array_p();
   kdefine_coord_array_p();
   define_attributes();

   /*
    *   test a linked list of coord list structures where each contains :
    *     a coord structure, a typedef of a coord structure, 
    *     and a typedef'd pointer to a coord structure...
    */
   kprintf("\ttest 1: structures and typedefs ... ");

   /*-- output coord list --*/
   olist = create_coord_list(2, 3);
   o = kpds_open_output_object(filename);
   kpds_set_attribute(o, COORD_LIST, olist);
   kpds_close_object(o);
   
   /*-- input coord list --*/
   i = kpds_open_input_object(filename);
   kpds_get_attribute(i, COORD_LIST, &ilist);

   if (!compare_coord_list(ilist, olist))
      FAIL("bogus input coordinate list\n");

   if (!kstruct_compare(ilist, olist, KTYPE_COORD_LIST))
       FAIL("official structure compare doesn't work\n");

   /*-- tidy up --*/
   kpds_close_object(i);
   kstruct_free(olist, KTYPE_COORD_LIST, TRUE);
   kunlink(filename);
   kprintf("successful\n");


   /*
    *   test variable length array of typedef'd pointers to coord structures :
    */
   kprintf("\ttest 2: variable length array of typedef'd pointers ... ");

   /*-- output coord array --*/
   oarray = create_coord_array(2, 3);
   o = kpds_open_output_object(filename);
   kpds_set_attribute(o, COORD_ARRAY, oarray);
   kpds_close_object(o);
   
   
   /*-- input coord list --*/
   i = kpds_open_input_object(filename);
   kpds_get_attribute(i, COORD_ARRAY, &iarray);

   if (!compare_coord_array(iarray, oarray))
      FAIL("bogus input coordinate array\n");

   if (!kstruct_compare(iarray, oarray, KTYPE_COORD_ARRAY))
       FAIL("official structure compare doesn't work\n");

   /*-- tidy up --*/
   kpds_close_object(i);
   kstruct_free(oarray, KTYPE_COORD_ARRAY, TRUE);
   kunlink(filename);
   kprintf("successful\n");

   kexit(KEXIT_SUCCESS);
}
