#include "xvinclude.h"

/*
 * This program puts up an image object displaying the entire image; it
 * also creates a smaller image object to the lower right in which the
 * extracted region of interest is displayed.  This example uses a freehand
 * ROI shape, but any ROI shape can be used.
 *
 * Simply move the cursor into the image and use the mouse to trace a freehand
 * to extract the desired ROI.  The extracted ROI will be displayed
 * in the image object to the lower right.
 *
 * To write out the location of each pixel in the ROI, enter a filename in
 * the output file selection and hit <cr>.
 */

static void  get_region     PROTO((xvobject, kaddr, XEvent *));
static void  output_points  PROTO((xvobject, kaddr, kaddr));

xvobject image;
kobject  roi_object;

void main(
   int  argc,
   char *argv[])
{
        kobject  image_object;
	char     *infile = "image:mandril";
        xvobject manager, roi, position, outfile;


        /* initialize xvwidgets lib */
        if (!xvw_initialize(XVW_MENUS_XVFORMS))
        {
           kerror(NULL, "main", "unable to open display");
           kexit(KEXIT_FAILURE);
        }

        /* create the xvobject to display the image */
        image_object = roi_object = kpds_open_object(infile, KOBJ_READ);

	/* create parent */
	manager = xvw_create_manager(NULL, "parent");

	/* create output file selection to write out ascii file */
	outfile  = xvw_create_outputfile(manager, "outfile");
	xvw_set_attributes(outfile,
			   XVW_BELOW,            NULL,
			   XVW_RIGHT_OF,         NULL,
			   XVW_OUTPUTFILE_LABEL, "Ascii Output File:",
			   XVW_TACK_EDGE,        KMANAGER_TACK_HORIZ,
			   NULL);

	/* create image object to display image */
        image = xvw_create_image(manager, "image");
        xvw_set_attributes(image, 
			   XVW_IMAGE_IMAGEOBJ,  image_object,
			   XVW_IMAGE_ROI_SHAPE, KIMAGE_ROI_FREEHAND,
			   XVW_BELOW,           outfile,
			   NULL);

	/* create position object to print location of pointer in image */
	position = xvw_create_position(manager, "image");
	xvw_set_attributes(position,
                           XVW_POSITION_OBJECT, image_object,
                           XVW_BELOW,           NULL,
                           XVW_RIGHT_OF,        NULL,
                           XVW_LEFT_OF,         NULL,
                           NULL);


	/* create image object to display extracted ROI */
	roi = xvw_create_image(manager, "roi");
	xvw_set_attributes(roi,
			   XVW_BELOW,           image,
			   XVW_LEFT_OF,         NULL,
			   NULL);

        /* add an event handler to start the region of interest extraction */
        xvw_add_event(image, ButtonPressMask, get_region, (kaddr) roi);

	/* add callback to outfile selection to write out the points in ROI */
	xvw_add_callback(outfile, XVW_OUTPUTFILE_CALLBACK, output_points, roi);

	/* display and run */
        xvf_run_form();


}

/*
 *  event handler to get region of interest and display the
 *  region of interest in the smaller image visual object.
 */
static void  get_region(
    xvobject object,
    kaddr    client_data,
    XEvent   *event)
{
        xvobject roi = (xvobject) client_data;

        /*
         * simply getting the XVW_IMAGE_ROI attribute causes all
         * the work to be done automatically;  the user is prompted to
         * rubberband the area of interest, the rubberbanding is performed,
         * and a data object containing the extracted region of interest
         * is returned.
         */
        xvw_get_attribute(image, XVW_IMAGE_ROI, &roi_object);
        if (roi_object == NULL) return;

        /*
         * region of interest successfully extracted; update image object
         */
        xvw_set_attribute(roi, XVW_IMAGE_IMAGEOBJ, roi_object);
}

/*
 *  callback to write out an ascii file containing all the points inside the ROI
 */
static void output_points(
   xvobject outfile,
   kaddr    client_data,
   kaddr    call_data)
{
	kfile    *file;
	char     *filename;
        int      i, j, w, h;
        int      *value_data, *mask_data;


	/* make sure we have a filename */
	xvw_get_attribute(outfile, XVW_OUTPUTFILE_FILENAME, &filename);
	if (filename == NULL)
	{
	    kerror(NULL, "output_points", "You must enter a filename");
	    return;
	}

	/* open the file */
	file = kfopen(filename, "w"); 
	if (file == NULL)
	{
	    kerror(NULL, "output_points", "Can't open file '%s'", filename);
            return;
	}

	/* get size of ROI, set attributes prior to getting ROI data */
        kpds_get_attribute(roi_object, KPDS_VALUE_SIZE,
					 &w, &h, NULL, NULL, NULL);
        kpds_set_attribute(roi_object, KPDS_VALUE_REGION_SIZE, w, h, 1, 1, 1);
        kpds_set_attribute(roi_object, KPDS_MASK_REGION_SIZE,  w, h, 1, 1, 1);
        kpds_set_attribute(roi_object, KPDS_VALUE_POSITION,    0, 0, 0, 0, 0);
        kpds_set_attribute(roi_object, KPDS_MASK_POSITION,     0, 0, 0, 0, 0);
        kpds_set_attribute(roi_object, KPDS_VALUE_DATA_TYPE,   KINT);
        kpds_set_attribute(roi_object, KPDS_MASK_DATA_TYPE,    KINT);

        value_data = (int *) kpds_get_data(roi_object, KPDS_VALUE_REGION, NULL);
        mask_data = (int *)  kpds_get_data(roi_object, KPDS_MASK_REGION, NULL);

	for (i = 0; i < h; i++)
	{
	    for (j = 0; j < w; j++)
            {
                if (mask_data[i*w+j] != 0)
	  	    kfprintf(file, "(%d, %d) = %d\n", i, j, value_data[i*w+j]);
	    }
	}
	kfclose(file);
}
