
/*
 * Copyright (c) 1996,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 */
 

#include <design.h>
#include "poly.h"
#include "texture.h"

/**
  * the texture table is just a flat array of textures.  the idea is that many prims
  * may use the same texture, and rather than reload the texture each time a new
  * primitive is read in, we stick them here to begin with.  each time we read in
  * a new prim, we check the texture table to see if a texture of the same transport
  * name is in here.  if so, we just pass pointers around, otherwise we actually have
  * to read it in.
 **/

#define MAX_RMONSTER_TEXTURES 16
#if 0
typedef struct rmonster_texture_table
{
    rmonster_texture *rmt;	/* pointer to a plain old rmonster texture structure */
    void *hardware_texture;	/* generic pointer to a hardware texture structure (transient) */
    char tname[128];		/* the transport name for this texture. */
    
} rmonster_texture_table;
#endif

/*-----   static variables ------*/
static int num_loaded_textures =0;
static rmonster_texture_table texture_table[MAX_RMONSTER_TEXTURES];

/*------  local/static routines ------*/

static rmonster_texture *malloc_2d_texture_no_alpha PROTO((int w,int h));
static rmonster_texture *malloc_2d_texture_alpha PROTO((int w,int h));
static int load_2d_texture PROTO((kobject tobj,rmonster_texture *t));
static void register_in_texture_table PROTO((rmonster_texture *t,char *tname));
static int find_in_texture_table PROTO((char *tname,rmonster_texture_table *texture_table,int num_loaded_textures));

int
get_num_rmonster_textures()
{
    /* return the number of actively defined rmonster textures registered */
    return(num_loaded_textures);
}

rmonster_texture_table *
get_rmonster_texture_table(int n)
{
    /* return a pointer into the rmonster texture table.  valid inputs
     for "n" are 0..num_loaded_textures-1 */
    
    if ((n < 0) || (n > num_loaded_textures-1))
	return(NULL);
    else
	return(texture_table+n);
}

int
init_texturing(void)
{
    int i;
    /* called to clear out the rmonster texture table. */
    
    for (i=0;i<MAX_RMONSTER_TEXTURES;i++)
    {
	texture_table[i].rmt = NULL;
	texture_table[i].hardware_texture = (void *)NULL;
	texture_table[i].tname[0] = '\0';
    }
}


int
deinit_texturing(void)
{
    int i;
    /* called to clear out the rmonster texture table. */
    
    for (i=0;i<MAX_RMONSTER_TEXTURES;i++)
    {
	if (texture_table[i].rmt != NULL)
	{
	    free_rmonster_texture(texture_table[i].rmt);
	    kfree(texture_table[i].rmt);
	}
	texture_table[i].rmt != NULL;

	if (texture_table[i].hardware_texture != NULL)
	    kfree(texture_table[i].hardware_texture); /* transient - write the code */
	texture_table[i].hardware_texture = NULL;
	
	texture_table[i].tname[0] = '\0';
    }
}

int
load_rmonster_texture (kobject tobj, /* input */
		       rmonster_texture **tex, /* output */
		       char *tname) /* input */
{
    /**
      * load an rmonster texture into a texture table for subsequent use
      * by the renderer.  the texture data will be read in using khoros
      * data services.
     **/
    
    int status, w,h,d,has_alpha,i;
    int num_prims,*prim_list = NULL;
    rmonster_texture *t;

    status = lgisa2dtex(tobj);

    if (status != TRUE)
    {
	kerror("RMONSTER","rmonster", "the object in question is not a valid geometry texture object. Please use gcreate2dtex to make a geometry texture from an arbitrary image.");
	return(WHACKED);
    }

    /* ok, the specified transport contains a valid rmonster texture.  do a lookup
     to see if we've already loaded this texture. */

    status = find_in_texture_table(tname,texture_table,num_loaded_textures);

    if (status >= 0)		/* found in texture table */
    {
#ifdef DEBUG
	fprintf(stderr,"\tCool! found the texture in the texture table..don't have to load it again.\n");
#endif
	*tex = texture_table[status].rmt;
	return (CHILL);
    }

    status = kgeom_get_attribute(tobj,KGEOM_OBJECT,KGEOM_NUMBER_PRIMITIVES,&num_prims);
    status = kgeom_get_attribute(tobj,KGEOM_OBJECT,KGEOM_PRIMITIVE_LIST,&prim_list);

    kgeom_set_attribute(tobj,KGEOM_OBJECT,KGEOM_PRIMITIVE_POSITION,0);

    /* in anticipation of one day using one routine to load both 2 and 3d
     textures, i'm leaving the following conditional in the code.  */
    
    if (prim_list[0] == KGEOM_TEXTURE2D)
    {
	kgeom_get_attribute(tobj,KGEOM_TEXTURE2D,KGEOM_TEXTURE2D_SIZE,&w,&h);
	kgeom_get_attribute(tobj,KGEOM_OBJECT,KGEOM_HAS_ALPHA,&has_alpha);
	if (has_alpha)
	    t = malloc_2d_texture_alpha(w,h);
	else
	    t = malloc_2d_texture_no_alpha(w,h);
	load_2d_texture(tobj,t);
    }
    else if (prim_list[0] == KGEOM_TEXTURE3D)
    {
	/* not ready yet */
	kerror("RMONSTER","rmonster","thanks for your interest in 3d textures.  they are not supported yet.");
	return(WHACKED);
    }

    register_in_texture_table(t,tname);
    *tex = t;

    return(CHILL);
}

int
free_rmonster_texture (rmonster_texture *t)
{
    kfree(t->texture_union.rgb_image);
}

rmonster_texture *
malloc_2d_texture_no_alpha(int w,
			   int h)
{
    surface_colors_rgb *s1d,**s2d;
    int i,j;
    rmonster_texture *t;

    s1d = (surface_colors_rgb *)kmalloc(sizeof(surface_colors_rgb)*w*h);
    if (s1d == NULL)
    {
	kerror("RMONSTER","rmonster","unable to malloc space for a 2d texture of size %d by %d floating point RGB.",w,h);
	return(NULL);
    }

    s2d = (surface_colors_rgb **)kmalloc(sizeof(surface_colors_rgb *)*h);

    for (i=0;i<h;i++)
    {
	s2d[i] = s1d + i*w;
    }

    t = kmalloc(sizeof(rmonster_texture));
    t->dims = 2;
    t->dimensions[0] = w;
    t->dimensions[1] = h;
    t->dimensions[2] = 0;
    t->ncomponents = 3;
    t->texture_union.rgb_image = s2d;
    t->any_transparency = 0;

    return(t);
}

rmonster_texture *
malloc_2d_texture_alpha(int w,
			int h)
{
    surface_colors_rgba *s1d,**s2d;
    int i;
    rmonster_texture *t;

    s1d = (surface_colors_rgba *)kmalloc(sizeof(surface_colors_rgba)*w*h);
    if (s1d == NULL)
    {
	kerror("RMONSTER","rmonster","unable to malloc space for a 2d texture of size %d by %d floating point RGBA.",w,h);
	return(NULL);
    }

    s2d = (surface_colors_rgba **)kmalloc(sizeof(surface_colors_rgba *)*h);

    for (i=0;i<h;i++)
    {
	s2d[i] = s1d + w*i;
    }

    t = kmalloc(sizeof(rmonster_texture));
    t->dims = 2;
    t->dimensions[0] = w;
    t->dimensions[1] = h;
    t->dimensions[2] = 0;
    t->ncomponents = 4;
    t->texture_union.rgba_image = s2d;
    t->any_transparency = 0;

    return(t);
}

static int
load_2d_texture(kobject tobj,
		rmonster_texture *t)
{
    int nbytes;
    float *ptr,*p2=NULL;
    unsigned char *ival=NULL,*ival2;
    float scale;
    int size;
    int i;
    int w,h;
    int has_alpha;
#if 0
    kpds_set_attribute(tobj,KPDS_VALUE_DATA_TYPE,KFLOAT);
    kpds_set_attribute(tobj,KPDS_LOCATION_SIZE,t->dimensions[0],
		       t->dimensions[1],0,0,3);
    kpds_set_attribute(tobj,KPDS_ELEMENTS_FIRST,TRUE);

    /* scale from 0..255 to 0..1 */
    kpds_set_attribute(tobj,KPDS_VALUE_SCALING,KSCALE);
    kpds_set_attribute(tobj,KPDS_VALUE_SCALE_FACTOR,(1./255.));
    
    ptr = (float *)&(t->texture_union.rgb_image[0][0].r);
    
    p2 = kpds_get_data(tobj,KPDS_VALUE_ALL,p2);
    kmemcpy(ptr,p2,sizeof(surface_colors_rgb)*t->dimensions[0]*t->dimensions[1]);
    kfree(p2);
#endif
    kcolor_set_attribute(tobj,KCOLOR_COLORSPACE,KRGB);
    kpds_set_attribute(tobj,KPDS_MAPPING_MODE,KMAPPED); 
    
    kcolor_get_attribute(tobj,KCOLOR_HAS_ALPHA,&has_alpha);
    if (has_alpha)
	kerror("RMONSTER","rmonster","textures with alpha are not yet supported.  resulting images are bogus. ");

    size = (has_alpha) ? 4 : 3;
    kpds_set_attribute(tobj,KPDS_VALUE_SIZE,-1,-1,-1,-1,size);

    /* set the colorspace attributes so that we get RGB(a) */

    kpds_set_attribute(tobj,KPDS_VALUE_DATA_TYPE,KUBYTE);

    kpds_set_attribute(tobj,KPDS_ELEMENTS_FIRST,TRUE);
    
    ival=NULL;
    ival = kpds_get_data(tobj,KPDS_VALUE_ALL,ival);
    ival2 = ival;

    scale = 1./255.;
    w = t->dimensions[0];
    h = t->dimensions[1];
    ptr = (float *)&(t->texture_union.rgb_image[0][0].r);
    for (i=0;i<w*h*3;i++)
	*ptr++ = (*ival++) * scale;

    kfree(ival2);
}

/* stuff to support texture rendering */

static rmonster_texture *bound_texture=NULL;
int tfunc_pointsample PROTO(());
int tfunc_noop PROTO(());
int tfunc_pointsample_perspective PROTO(());

int (*tfuncs_2d[])() =
{
    tfunc_pointsample,
    tfunc_noop,
    tfunc_noop
};

int (*tfuncs_2d_perspective[])() =
{
    tfunc_pointsample_perspective,
    tfunc_noop,
    tfunc_noop
};

int (*texturefunc)() = NULL;

int
rbind_texture(rmonster_texture *new_texture,
	      int projection)
{
    bound_texture = new_texture;

    /* assign a texturing fucntion based upon the current value of
     the user-selected texturing method */
/*    texturefunc = tfunc_pointsample; */
    
    if (new_texture == NULL)
	texturefunc = tfunc_noop;
    else if (projection == KPROJECTION_PERSPECTIVE)
	texturefunc = tfuncs_2d_perspective[get_global_2dtexturing_function_index()];
    else
	texturefunc = tfuncs_2d[get_global_2dtexturing_function_index()];
}

int
runbind_texture()
{
    bound_texture = NULL;
    texturefunc = tfunc_noop;
}

int
have_bound_texture()
{
    return((bound_texture == NULL) ? 0 : 1);
}


/* texture rendering funcs */

int
tfunc_pointsample (Poly_vert *p,
		   float *in,
		   float *out)
{
    /* use input RGB pixel value as luminance attenuator.  attenuate
     point-sampled RGB value from texture by this luminance value. */
    double luminance;
    int iu,iv,i;
    
    luminance = 0.3 * in[0] + 0.6 * in[1] + 0.1 * in[2];
    point_sample_texture(p->u,p->v,out);
    for (i=0;i<3;i++)
    {
	out[i] *= luminance;
    }
}

int
tfunc_pointsample_perspective (Poly_vert *p,
			       float *in,
			       float *out)
{
    /* use input RGB pixel value as luminance attenuator.  attenuate
     point-sampled RGB value from texture by this luminance value. */
    double luminance;
    double du,dv;
    int iu,iv,i;
    
    luminance = 0.3 * in[0] + 0.6 * in[1] + 0.1 * in[2];
#if 0
    du = (p->u -0.5 ) * p->sz + 0.5;
    dv = (p->v -0.5 ) * p->sz + 0.5;
#endif

    du = p->u;
    dv = p->v;

    point_sample_texture(du,dv,out);
    
    for (i=0;i<3;i++)
    {
	out[i] *= luminance;
    }
}


int
tfunc_noop (Poly_vert *p,
	    float *in,
	    float *out)
{
    /* do nothing here */
    memcpy((char *)out,(char *)in,sizeof(float)*3);
}


int
point_sample_texture(double u,
		     double v,
		     float *out)
{
    int iu,iv;
    surface_colors_rgb *t;
    double tt;
    extern double rint();

    tt = u * (bound_texture->dimensions[0]-1);
    iu = tt;
    if (iu < 0)
    {
	fprintf(stderr," texture dimension width exceeded: %d \n",iu);
	iu = 0;
    }

    if (iu >= bound_texture->dimensions[0])
    {
	fprintf(stderr," texture dimension width exceeded: %d \n",iu);
	iu = bound_texture->dimensions[0]-1;
    }

    
    tt = v * (bound_texture->dimensions[1]-1);
    iv = tt;
    if (iv < 0)
    {
	fprintf(stderr," texture dimension height exceeded: %d \n",iv);
	iv = 0;
    }

    if (iv >= bound_texture->dimensions[1])
    {
	fprintf(stderr," texture dimension height exceeded: %d \n",iv);
	iv = bound_texture->dimensions[1]-1;
    }
    
    out[0] = bound_texture->texture_union.rgb_image[iv][iu].r;
    out[1] = bound_texture->texture_union.rgb_image[iv][iu].g;
    out[2] = bound_texture->texture_union.rgb_image[iv][iu].b;
}


/**
  * on the Imaging pane, the user can tweak a menu item that dictates
  * how 2d texture mapping is done.  the choices are basically a range
  * from quick and dirty to slow and good, along with a disable switch.
  *
  * a known inconsistency is that textures are assigned on a per-primitive
  * basis, but the texturing control knob applies to _all_ objects with
  * textures, not just the currently selected one.
  *
  * note the relationship between the order of items on the UI and the
  * order of functions below.
 **/

static int tfunc2d_index = 0;	/* what should default for this be? */

int
set_global_2dtexturing_function_index(int new_index)
{
    tfunc2d_index = new_index;
}

int
get_global_2dtexturing_function_index()
{
    return(tfunc2d_index);
}

int
get_texture_index(char *tname)
{
    /* will return a non-negative number upon success - this number is
     the index into the rmonster texture table of the texture referred
     to by "tname".  will return a -1 upon failure. */
    
    int n,status;
    rmonster_texture_table *rtable;
   
    n = get_num_rmonster_textures();
    status = find_in_texture_table(tname,texture_table,n);
    return(status);
}


static int
find_in_texture_table(char *tname,
		      rmonster_texture_table *texture_table,
		      int num_loaded_textures)
{
    /* used to determine if a texture has been registered in
     the internaal rmonster texture table.  tname is the texture
     transport name passed as input.  we do a straight linear
     (brain dead) search at this time.  14 dec 95 wes */
    
    int i;
    for (i=0;i<num_loaded_textures;i++)
    {
	if (kstrcmp(tname,texture_table[i].tname) == 0)
	    break;
    }
    if (i == num_loaded_textures) /* didn't find a match */
        i = -1;			/* -1 indicates no match was found */

    return(i);
}

static void
register_in_texture_table(rmonster_texture *t,
			  char *tname)
{
    /* assume that the variable "num_loaded_textures" points to the next
       available entry in the texture table. */

    int i;

    i = num_loaded_textures;
    
    texture_table[i].rmt = t;
    kstrcpy(texture_table[i].tname,tname);

    /* and then increment the num_loaded_textures variable */
    num_loaded_textures++;
}

int
load_constrained_size_texture(char *tname,
			      int *dims,
			      int ndims,
			      unsigned char **rval)
{
    /* loads an image from the named transport using data services
     to rescale to a given size.  returns a pointer to a 1d, elements
     first array containing the image. */
    
    unsigned char *ival=NULL;
    int size,num_prims,*prim_list;
    int has_alpha,status;
    kobject tobj;
    
    tobj = kgeom_open_object(tname,KOBJ_READ);
/*    load_rmonster_texture(textureobj,&(p->the_texture),tname); */
#if 0    
    kcolor_set_attribute(tobj,KCOLOR_COLORSPACE,KRGB);
    kpds_set_attribute(tobj,KPDS_MAPPING_MODE,KMAPPED); 
    
    kcolor_get_attribute(tobj,KCOLOR_HAS_ALPHA,&has_alpha);
#endif    
    status = kgeom_get_attribute(tobj,KGEOM_OBJECT,KGEOM_NUMBER_PRIMITIVES,&num_prims);
    status = kgeom_get_attribute(tobj,KGEOM_OBJECT,KGEOM_PRIMITIVE_LIST,&prim_list);

    kgeom_set_attribute(tobj,KGEOM_OBJECT,KGEOM_PRIMITIVE_POSITION,0);

    /* in anticipation of one day using one routine to load both 2 and 3d
     textures, i'm leaving the following conditional in the code.  */
    
    if (prim_list[0] == KGEOM_TEXTURE2D)
    {
	int w,h;
	
	kgeom_get_attribute(tobj,KGEOM_TEXTURE2D,KGEOM_TEXTURE2D_SIZE,&w,&h);
	kgeom_get_attribute(tobj,KGEOM_OBJECT,KGEOM_HAS_ALPHA,&has_alpha);
	size = (has_alpha) ? 4 : 3;

	/* reset the size of the value segment (the texture) to reflect
	 the constraints imposed by open gl*/
	
	kgeom_set_attribute(tobj,KGEOM_TEXTURE2D,KGEOM_TEXTURE2D_SIZE,
			    dims[0],dims[1]);

	/* set the colorspace attributes so that we get RGB(a) */

	kpds_set_attribute(tobj,KPDS_VALUE_DATA_TYPE,KUBYTE);
	
	kpds_set_attribute(tobj,KPDS_ELEMENTS_FIRST,TRUE);

	/* tell data services to reinterpolate the texture in the
	 following get_data() call.*/
	
	kpds_set_attribute(tobj,KPDS_VALUE_INTERPOLATE,KZERO_ORDER);

	/* and grab the data. */
	ival=NULL;
	ival = kpds_get_data(tobj,KPDS_VALUE_ALL,ival);
	*rval = ival;

	kgeom_close_object(tobj);
    }

}

void *
get_hardware_texture_ptr(int index)
{
    if ((index < 0) || (index > num_loaded_textures))
	return(NULL);
    else
	return(texture_table[index].hardware_texture);
}
