 /*
  * Khoros: $Id: llrftrain.c,v 1.12 1996/07/05 03:27:12 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: llrftrain.c,v 1.12 1996/07/05 03:27:12 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for lrftrain
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	llrftrain
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
#define ALPHA_MAX 0.995
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: llrftrain - Localized Receptive Field Classifier Trainer
* 
*       Purpose: lrftrain trains on an image for the weights used with the
*		 Localized Receptive Field classifier (see lrfclass).  The
*		 Localized Receptive Field (LRF) is based on a single layer of
*		 self-organizing, "localized receptive field" units, followed
*		 by a single layer perceptron.  The single layer of perceptron 
*		 units use the LMS or Adaline learning rule to adjust the
*		 weights.
*		 .SH "LRF network theory"
*
*		 The basic network model of the LRF consists of a two layer
*		 topology.  The first layer of "receptive field" nodes are
*		 trained using a clustering algorithm, such as K-means, or
*		 some other algorithm which can determine the receptive field
*		 centers.  Each node in the first layer computes a receptive
*		 field response function, which should approach zero as the
*		 distance from the center of the receptive field is increased.
*		 The second layer of the LRF model sums the weighted outputs of
*		 the first layer, which produces the output or response of the
*		 network.  A supervised LMS rule is used to train the weights
*		 of the second layer nodes.
*
*		 The response function of the LRF network is formulated as
*		 follows:
*
*		 !             f(x) = SUM(Ti * Ri(x))
*
*		 !where,
*
*		 !      Ri(x) = Q( ||x - xi|| / Wi )
*
*		 !      x  -  is a real valued vector in the input space,
*		 !      Ri -  is the ith receptive field response function,
*		 !      Q  -  is a radially symmetric functio with a single
*		 !            maximum at the origin, decreasing to zero at
*		 !            large radii,
*		 !      xi -  is the center of the ith receptive field,
*		 !      Wi -  is the width of the ith receptive field,
*		 !      Ti -  is the weight associated with each receptive field
*
*		 The receptive field response functions ( Ri(x) ), should be
*		 formulated such that they decrease rapidly with increasing
*		 radii.  This ensures that the response functions provide
*		 highly localized representations of the input space.  The
*		 response function used here is modeled after the gaussian, and
*		 uses the trace of the covariance matrix to set the widths of
*		 the receptive field centers.
*
*		 The weights for the output layer are found using the LMS
*		 learning rule.  The weights are adjusted at each iteration to
*		 minimize the total error, which is based on the difference
*		 between the network output and the desired result.
*
*		 Prior to using the LRF algorithm, it is necessary to run
*		 "vkmeans" on the input training image to fix the cluster
*		 centers, followed by a supervised classification of the
*		 clustered image, which assigns a desired class to each cluster
*		 center.  NOTE that the image resulting from the supervised
*		 classification MUST be appended to the "cluster center" image
*		 before running the LRF.  This is necessary since it makes the
*		 appropriate desired class assignments to the cluster centers
*		 for the training phase of the LRF.
*
*		 The number of receptive field response nodes in the first layer
*		 of the LRF is determined by the number of cluster centers in
*		 the "cluster center" image.  The number of output classes, and
*		 hence the number of output nodes in the second (ie. last)
*		 layer, is determined by the number of desired classes that was
*		 specified in the "supervised" classification phase of the
*		 clustering.  This information is contained in the last band of
*		 the cluster center image.  The number of weights in the network
*		 is determined by the number of receptive field response nodes
*		 and the number of output nodes.  That is,
*
*		 !  #Wts = (#rf_response_nodes * #output_nodes) + #output_nodes
*
*		 This routine was written with the help of and ideas from
*		 Dr. Don Hush, University of New Mexico, Dept. of EECE.
*         Input: image - the input image used for training.
*		 cc_img - the cluster center image specifying the cluster
*			  centers.
*		 var_img - the cluster variance image specifying the variances.
*		 cn_img - the cluster number image specifying which vector/pixel
*			  belongs to which cluster.
*		 converge - the convergence parameter.
*		 meu - the weight update parameter.
*		 border - the border width in pixels of the input image.
*		 max_iter - the maximum number of iterations until termination.
*		 prt_mse - the iteration interval for printing the MSE to the
*			   stats file.
*		 delta_mse - the minimum change in the MSE between iterations,
*			     for termination.
*        Output: wt_img - the resulting weight image after training.
*		 printdev - the file containing the training statistics.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: All input images except the "cluster number" image (cn_img)
*		 MUST be of data storage type FLOAT.  The "cluster number"
*		 image (cn_img) MUST be of data storage type INTEGER.  The
*		 output "weight" image (wt_img) is of data storage type FLOAT.
*    Written By: Tom Sauer and Charlie Gage
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int llrftrain (
	xvimage *image,
	xvimage *cc_img,
	xvimage *var_img,
	xvimage *cn_img,
	xvimage **wt_img,
	float converge,
	float meu,
	int border,
	int max_iter,
	int prt_mse,
	float delta_mse,
	kfile *printdev)
/* -library_def_end */

/* -library_code */
{
   xvimage  *weight_image;
   int nodes,                    /* number of nodes in the LRF layer */
       rfc_nvects,               /* number of receptive field centers */
       rfc_dim,                  /* dimension of receptive field centers */
       var_nvects,               /* variance vectors */
       var_dim,                  /* dimension of variance vectors */
       input_nvects,             /* number of input vectors */
       input_dim,                /* dimension of input vectors */
       num_classes,              /* number of desired output classes */
       iterations,               /* number of iterations */
       patrn,                    /* number of input patterns */
       prt_flag,                 /* flag indicating print MSE info */
       nr, 
       nc, 
       i, j, k, x, z, class, node, d;
   int *cn_ptr;                  /* pointer to cluster number imagedata */
   float **rf_centers,           /* pointer to receptive field centers */ 
         **cc_var,               /* pointer to cluster center variances */
         **input_data,           /* pointer to input image data */
         *class_ptr,             /* pointer to the class that corresponds */
                                 /*  to the associated cluster center */
         **R,                    /* pointer to Receptive Field values */
         **W,                    /* pointer to the Weights */
         *Y,                     /* pointer to the output node equations */
         *error,                 /* pointer to the error values */
         *desired,               /* pointer to the desired class outputs */
         *image_ptr;
   float a, q, dis, total_error;
   double alpha,                 /* MSE update parameters */
          mse, 
          prev_mse;
 
    if (prt_mse > 0)
       prt_flag = 1;
    else
       prt_flag = 0;

    nr = image->col_size;
    nc = image->row_size;

       /* initialize mse to some arbitrary large value */
    mse = 0.5;
    prev_mse = 0.0;

   /*--------------------------------------------------------------------*
    * INITIALIZE the NUMBER of NODES in the LRF, this is simply the number
    * of cluster centers previously determined 
    *--------------------------------------------------------------------*/ 
   nodes = cc_img->row_size;

   /*--------------------------------------------------------------------*
    * determine the NUMBER of DESIRED OUTPUT CLASSES
    *--------------------------------------------------------------------*/ 
   image_ptr = (float *) cc_img->imagedata;
   class_ptr = (float *) &image_ptr[cc_img->row_size * cc_img->col_size 
                                      * (cc_img->num_data_bands-1)];
   num_classes = *class_ptr; 
   for (i = 1; i < cc_img->row_size * cc_img->col_size; i++)
   {
      if (class_ptr[i] > num_classes) 
        num_classes = (int) class_ptr[i];
   }
   num_classes++;     /* add 1, since we start with zero */

   /*--------------------------------------------------------------------*
    * LOAD the RECEPTIVE FIELD CENTERS from the cluster center image,
    * but trick load_vector so that it does not see the class data
    * band attached to the end of the cluster center image
    *--------------------------------------------------------------------*/ 
   cc_img->num_data_bands -= 1;
   rf_centers = (float **) load_vector(cc_img, 0, &rfc_nvects, &rfc_dim);
   cc_img->num_data_bands += 1;

   /*--------------------------------------------------------------------*
    * LOAD the CLUSTER CENTER VARIANCES from the variance image.
    *--------------------------------------------------------------------*/ 
   cc_var = (float **) load_vector(var_img, 0, &var_nvects, &var_dim);

   /*--------------------------------------------------------------------*
    * LOAD the INPUT DATA into a vector format
    *--------------------------------------------------------------------*/ 
   input_data = (float **)load_vector(image, border, &input_nvects, &input_dim);

   /*--------------------------------------------------------------------*
    * ALLOCATE SPACE for the response functions.
    * There are N * K response functions, where N is the number of nodes
    * and K is the number of input patterns
    *--------------------------------------------------------------------*/ 
        
   R = (float **) kmalloc (sizeof (float *) * nodes);
   if (R == NULL)
   {
      kerror ("vipl", "lrftrain", "insufficient memory available\n");
      return (FALSE);
   }
   for (i = 0; i < nodes; i++) 
   {
      R[i] = (float *) kmalloc (sizeof (float) * input_nvects);
      if (R[i] == NULL) 
      {
          kerror ("vipl", "lrftrain", "insufficient memory available\n");
          return (FALSE);
       }
    }


   /*--------------------------------------------------------------------*
    * COMPUTE the RESPONSE FUNCTIONS for each node and data input
    * pattern. This is based on using the ratio of the Euclidean distance 
    * of an input data vector to a receptive field center to each 
    * diagonal element of the covariance matrix.
    *--------------------------------------------------------------------*/ 
   for (i = 0; i < input_nvects; i++) 
   {
       for (j = 0; j < nodes; j++) 
       {
                  /* compute the distance along each axis */
   
           dis = 0.0;
           for (k = 0; k < input_dim; k++) {
               a = *(input_data[i] + k) - *(rf_centers[j] + k);
               q = a * a;
               dis += q/(*(cc_var[j] +k));
           }
           *(R[j] + i) = exp(-dis);
       }
   }

   /*--------------------------------------------------------------------*
    * ALLOCATE SPACE for the number of WEIGHTS
    * number of weights = number of desired output classes * # of nodes + 1
    * the +1 is for the bias weight.
    *--------------------------------------------------------------------*/ 
   W = (float **) kmalloc (sizeof (float *) * num_classes);
   if (W == NULL)
   {
      kerror ("vipl", "lrftrain", "insufficient memory available\n");
      return (FALSE);
   }
   for (i = 0; i < num_classes; i++) 
   {
      W[i] = (float *) kmalloc (sizeof (float) * nodes + 1);
      if (W[i] == NULL) 
      {
          kerror ("vipl", "lrftrain", "insufficient memory available\n");
          return (FALSE);
       }
    }
    
   /*--------------------------------------------------------------------*
    * ALLOCATE SPACE for the OUPUT NODE EQUATIONS Y and for
    * the error for each node equation Y
    *--------------------------------------------------------------------*/ 

   Y = (float *) kmalloc (sizeof (float *) * num_classes);
   if (Y == NULL)
   {
      kerror ("vipl", "lrftrain", "insufficient memory available\n");
      return (FALSE);
   }

   error = (float *) kmalloc (sizeof (float *) * num_classes);
   if (error == NULL)
   {
      kerror ("vipl", "lrftrain", "insufficient memory available\n");
      return (FALSE);
   }

   desired = (float *) kmalloc (sizeof (float *) * num_classes);
   if (desired == NULL)
   {
      kerror ("vipl", "lrftrain", "insufficient memory available\n");
      return (FALSE);
   }
        /* END OF DATA ALLOCATION */

   (void) kfprintf (printdev, "lrftrain Statistics\n");
   (void) kfprintf (printdev, "===================\n");

   /*--------------------------------------------------------------------*
    *   INITIALIZE the WEIGHTS with a random number between -1 and 1
    *--------------------------------------------------------------------*/ 
    for (class = 0; class < num_classes; class++)
    {
      for( node = 0; node < nodes + 1; node++)
      {
         W[class][node] = (kurng() - 0.5) * 2.0;
      }
    }
        
   /*--------------------------------------------------------------------*
    * TRAIN the SINGLE LAYER PERCEPTRON for the weights using the
    *   LMS algorithm 
    *--------------------------------------------------------------------*/ 
    iterations = 0;
    do
    {
       iterations++;
       patrn = 0;
       for (x = border; x < nr - border; x++) 
       {
          for (z = border; z < nc - border; z++) 
          {
/*
                *   compute the output layer node equation 
                */
             (void) kmemset((kaddr) Y, 0, (sizeof(float) * num_classes));
             for (class = 0; class < num_classes; class++)
             {
                for (node = 1; node < nodes + 1; node++)
                {
                   Y[class] += R[node-1][patrn] * W[class][node];
                }
                Y[class] += W[class][0];
             }

                /* find the class that the current input vector belongs to
                 * and set the desired output for that class to 1 and set 
                 * all other desired outputs to -1
                 */
             cn_ptr = (int *) cn_img->imagedata;
             for (class = 0; class < num_classes; class++)
                desired[class] = -1.0;
             d = (int) class_ptr[cn_ptr[x*nc+z]];
             desired[d] = 1.0;

                 /* compute the error terms for each class
                  *
                  */
             total_error = 0.0;
             for (class = 0; class < num_classes; class++)
             {
                error[class] = desired[class] - Y[class];
                total_error += error[class] * error[class];
             }
                   
                /* update the weights for the output layer
                 *
                 */
             for (class = 0; class < num_classes; class++)
             {
                for (node = 1; node < nodes + 1; node++)
                {
                   W[class][node] = W[class][node] + meu 
                                * error[class] * R[node - 1][patrn];
                }
                W[class][0] = W[class][0] + meu * error[class];
             }
             patrn++;
          }
       }
   /*---------------------------------------------------------------------*
    * COMPUTE the MSE for the current iteration.
    *   This formula will put more weight on the previous mse 
    *   as the number of iterations increases, and less
    *   weight on the total_error for the current iteration.
    *---------------------------------------------------------------------*/
    alpha = kmin(ALPHA_MAX, (double) (iterations - 1) /(double) (iterations));
    mse = alpha*mse + ((1.0 - alpha)*(double) total_error*(double) total_error);

   /*---------------------------------------------------------------------*
    * TERMINATE program when minimum delta MSE value is reached 
    *---------------------------------------------------------------------*/
    if (fabs(mse - prev_mse) < (double)delta_mse)
    {
      kfprintf (printdev,"\nTraining was terminated at the following point:\n");
      kfprintf (printdev,"\n-----  iter = %d    mse = %.10f -----\n", iterations, mse);
      kfprintf (printdev,"\nbecause the MSE reached a plateau.\n");
      break;
    }
    prev_mse = mse;

   /*---------------------------------------------------------------------*
    * PRINT STATS at selected interval
    *---------------------------------------------------------------------*/
    if ((iterations == 1) || ((prt_flag == 1) && ((iterations % prt_mse) == 0)))
       kfprintf (printdev,"\n-----  iter = %d    mse = %.10f -----\n", iterations, mse);
   }
   while((mse > converge) && (iterations < max_iter));

   /*---------------------------------------------------------------------*
    * SAVE the WEIGHTS (W) 
    *---------------------------------------------------------------------*/
   weight_image = createimage((unsigned long) 1,       /* number of rows */
                        (unsigned long) num_classes,   /* number of columns */
                        (unsigned long) VFF_TYP_FLOAT,  /* data_storage_type */
                        (unsigned long) 1,                /* num_of_images */
                        (unsigned long) (nodes + 1),      /* num_data_bands */
                        "Weight image created by lrftrain", /* comment */
                        (unsigned long) 0,                /* map_row_size */
                        (unsigned long) 0,                /* map_col_size */
                        (unsigned long) VFF_MS_NONE,      /* map_scheme */
                        (unsigned long) VFF_MAPTYP_NONE, /* map_storage_type */
                        (unsigned long) VFF_LOC_IMPLICIT, /* location_type */
                        (unsigned long) 0);               /* location_dim */

    if (weight_image == NULL)
    {
       kerror("vipl","lrftrain", "Unable to allocate weight image!\n");
       *wt_img = NULL;
       return(FALSE);
    }
    kfree_and_NULL(weight_image->imagedata);
    weight_image->imagedata = (char *) unload_vector((char **) W, 0,
					VFF_TYP_FLOAT, num_classes, nodes + 1,
					1, num_classes);
   *wt_img = weight_image;

   /*---------------------------------------------------------------------*
    * WRITE STATS to file
    *---------------------------------------------------------------------*/
   (void) kfprintf (printdev,
                  "\n\tTotal Number of Iterations : %d\n\n",iterations);
   (void) kfprintf (printdev,
                  "\tFinal Mean Squared Error: %f\n\n",mse);
   (void) kfprintf (printdev,
                  "\tConvergence Parameter (cv): %f\n\n",converge);
   (void) kfprintf (printdev,
                  "\tWeight Update Parameter (meu): %f\n\n",meu);
   (void) kfprintf (printdev,
                  "\tMinimum DELTA MSE value (delta): %g\n\n", delta_mse);
   (void) kfprintf (printdev,
                  "\tBorder Width: %d\n\n", border);
   (void) kfprintf (printdev,
                  "\tNumber of Response Nodes: %d\n\n",nodes);
   (void) kfprintf (printdev,
                  "\tNumber of Output Classes: %d\n\n",num_classes);
   return(TRUE);
}
/* -library_code_end */
