 /*
  * Khoros: $Id: lvcltrans.c,v 1.12 1996/07/05 03:27:18 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvcltrans.c,v 1.12 1996/07/05 03:27:18 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vcltrans
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvcltrans
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
#include "lvcltrans.h"
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvcltrans - Tristimulus to tristimulus color space conversions
* 
*       Purpose: lvcltrans specifies the linear transformation matricies for
*		 conversion between ntscYIQ, cieXYZ, ntscCMY, cieUVW and
*		 ntscRGB or genericRGB colorspaces.  These transformations
*		 were taken from Pratt (Digital Image Processing, 1978).
*		 The values in the matricies are based on
*		 NTSC RGB standard phosphers, as Pratt
*		 describes. 
*
*		 A user specified transformation matrix may also
*		 be used (mflg must be set). The transformation matrix
*		 must be a 3x3 image of type FLOAT. The resulting color
*		 space model will be GENERIC. The input image will
*		 still be normalized/scaled. The input image can be
*		 any multi-spectral image.
*
*		 The input image must be of type FLOAT, and
*		 of color space model ntscYIQ, cieXYZ, ntscCMY, cieUVW,
*		 or ntscRGB or genericRGB.  The input image will be normalized
*		 by the variable norm. The default is 255. 
*
*		 If converting from ntscYIQ, cieUVW, or cieXYZ to ntscCMY
*		 the routine first converts to ntscRGB then converts to
*		 ntscCMY because
*
*		 ntscCMY = 1 - ntscRGB. 
*
*		 If converting from ntscCMY to ntscYIQ, cieXYZ, or cieUVW the
*		 routine first converts to ntscRGB then converts from ntscRGB to
*		 ntscYIQ, cieXYZ or cieUVW using the transformation matrices.
*
*		 When converting between the different color spaces,
*		 care must be taken in regards to the normalization 
*		 factor. If the image had been scaled by using 
*		 vscale, then the scale factor used in vscale
*		 should be the normalization factor used in this
*		 routine.  (Read the vcltrans(1) man page for detailed 
*		 description)
*
*		 The output values of some of the transformations will be
*		 negative, and some values will be greater than 1.
*
*		 The following transformation matricies are used:
*
*		 ! RGBtoYIQ = { { 0.299,  0.587,  0.114},
*		 !              { 0.596, -0.274, -0.322},
*		 !              { 0.211, -0.523,  0.312} };
*
*		 ! RGBtoXYZ = { { 0.607,  0.174,  0.201},
*		 !              { 0.299,  0.587,  0.114},
*		 !              { 0.000,  0.066,  1.117} };
*
*		 ! RGBtoUVW = { { 0.405,  0.116,  0.133},
*		 !              { 0.299,  0.587,  0.114},
*		 !              { 0.145,  0.827,  0.627} };
*
*		 ! YIQtoRGB = { { 1.000,  0.956,  0.621},
*		 !              { 1.000, -0.272, -0.647},
*		 !              { 1.000, -1.106,  1.703} };
*
*		 ! YIQtoXYZ = { { 0.967,  0.318,  0.594},
*		 !              { 1.000,  0.000,  0.000},
*		 !              { 1.173, -1.238,  1.870} };
*
*		 ! YIQtoUVW = { { 0.653,  0.208,  0.403},
*		 !              { 1.000,  0.000,  0.000},
*		 !              { 1.599, -0.780,  0.623} };
*
*		 ! XYZtoRGB = { { 1.910, -0.533, -0.288},
*		 !              {-0.985,  2.000, -0.028},
*		 !              { 0.058, -0.118,  0.896} };
*
*		 ! XYZtoYIQ = { { 0.000,  1.000,  0.000},
*		 !              { 1.407, -0.842, -0.451},
*		 !              { 0.932,  1.189,  0.233} };
*
*		 ! XYZtoUVW = { { 0.667,  0.000,  0.000},
*		 !              { 0.000,  1.000,  0.000},
*		 !              {-0.500,  1.500,  0.500} };
*
*		 ! UVWtoRGB = { { 2.432,  0.332, -0.576},
*		 !              {-1.519,  2.083, -0.057},
*		 !              { 1.440, -2.823,  1.803} };
*
*		 ! UVWtoYIQ = { { 0.000,  1.000,  0.000},
*		 !              { 1.403,  0.534, -0.907},
*		 !              { 1.757, -1.898,  0.470} };
*
*		 ! UVWtoXYZ = { { 1.500,  0.000,  0.000},
*		 !              { 0.000,  1.000,  0.000},
*		 !              { 1.500, -3.000,  2.000} };
*
*		 These matricies can be found in the lvcltrans.h include file.
*         Input: img      - input image  of type xvimage
*		 type     - output image colormap type.
*		 type_flg - if set the conversion is "type" otherwise defalut
*			    is RGB.
*		 norm     - normalization/scale factor.
*		 norm_flg - if set use "norm" else use 255.0.
*		 trmatrix - specifies a 3x3 user defined conversion matrix.
*		 mflg     - if set use trmatrix, else use "type" as the
*			    conversion type.
*        Output: img      - overwrites the input image  with new image data
*			    and colormap type.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: The input image must be of one of the color space models
*		 listed above.  The data type of the input must be
*		 VFF_TYP_FLOAT.
*    Written By: Donna Koechner, Tom Sauer
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvcltrans (
	xvimage *img,
	int      type,
	int      type_flg,
	float    norm,
	int      norm_flg,
	int      mflg,
	xvimage *trmatrix)
/* -library_def_end */

/* -library_code */
{

    float  *ptr1,
           *ptr2,
           *ptr3,
           *matptr,
            gen_matrix[3][3];

    register
    float   tmp1,
            tmp2,
            tmp3;

    int     i,j,
            model,
            trrow,
            trcol,
            nc,
            nr;

    char *program = "lvcltrans";

    if (!norm_flg)
        norm = 255.0; /* default normalization/scale factor */
    if (!type_flg)
        type = 1; /* default conversion to RGB */

    if (!propertype (program, img, (unsigned long) VFF_TYP_FLOAT, FALSE)) {
        kerror ("vipl", "lvcltrans", "Image must be of data type FLOAT");
        return (FALSE);
    }

      /* can only have one image*/

    if (! (proper_num_images(program,img,(unsigned long) 1,FALSE))) {
         kerror("vipl", "lvcltrans", "All inputs must contain only one image");
         return(FALSE);
     }

      /* can only have three data bands */

     if (! (proper_num_bands(program,img,(unsigned long) 3,FALSE))) {
          kerror("vipl", "lvcltrans",
		 "Input image must contain three data bands");
          return(FALSE);
      }

      /* must have map enable set to OPTIONAL */

     if (! (proper_map_enable(program,img,(unsigned long) VFF_MAP_OPTIONAL,FALSE))) {
          kerror("vipl", "lvcltrans",
		 "Input must have map enable type OPTIONAL");
          return(FALSE);
      }

    nr = img -> col_size;
    nc = img -> row_size;

    if (mflg) {  /* use the input image trmatrix as the conversion matrix */

                /* conversion matrix must be float */
        if (!propertype (program, trmatrix, (unsigned long) VFF_TYP_FLOAT, FALSE)) {
            kerror("vipl", "lvcltrans", "Conversion matrix image must be"
                   "of data type FLOAT");
            return (FALSE);
        }

                /* conversion matrix must be 3x3 */
        if (!propersize (program, trmatrix, (unsigned long) 3, (unsigned long) 3, FALSE)) {
            kerror("vipl", "lvcltrans","Conversion matrix must be of size 3x3");
            return(FALSE);
        }


      /* can only have one data band */

    if (! (proper_num_images(program,trmatrix,(unsigned long) 1,FALSE))) {
         kerror("vipl", "lvcltrans", "All inputs must contain only one image");
         return(FALSE);
     }

      /* can only have one data band */

     if (! (proper_num_bands(program,trmatrix,(unsigned long) 1,FALSE))) {
          kerror("vipl", "lvcltrans",
		 "All inputs must contain only one data band");
          return(FALSE);
      }

      /* must have map enable set to OPTIONAL */

     if (! (proper_map_enable(program,trmatrix,(unsigned long) VFF_MAP_OPTIONAL,FALSE))) {
          kerror("vipl","lvcltrans","Input must have map enable type OPTIONAL");
          return(FALSE);
      }
           /* point to input image data */
        ptr1 = (float *) &img -> imagedata[0];
        ptr2 = (float *) &img -> imagedata[nr * nc * sizeof (float)];
        ptr3 = (float *) &img -> imagedata[2 * nr * nc * sizeof (float)];

        trrow = trmatrix->col_size;
        trcol = trmatrix->row_size;

                /* need the conversion matrix to be in a 3x3 array, this is
                   for use in the #define conversion statements */

        matptr = (float *) trmatrix -> imagedata;

        for ( i = 0; i < trrow; i++) {
           for ( j = 0; j < trcol; j++) {
              gen_matrix[i][j] = *matptr;
              matptr++;
           }
        }

            /* perform the conversion */
        for (i = 0; i < nr * nc; i++) {
            tmp1 = *ptr1/norm; /* first normalize/scale */
            tmp2 = *ptr2/norm;
            tmp3 = *ptr3/norm;
            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, gen_matrix); /*actual conversion*/
            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, gen_matrix);
            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, gen_matrix);
            ++ptr1; ++ptr2; ++ptr3;
        }
                /* new colormap model is GENERIC */
        img -> color_space_model = VFF_CM_GENERIC;
    }


    else { /* the mflg was not set, so use "type" as the resulting conversion */

        model = img -> color_space_model;

                /* input image must be RGB, CMY, YIQ, XYZ, UVW */

        if (model != VFF_CM_ntscRGB &&
                model != VFF_CM_genericRGB &&
                model != VFF_CM_ntscCMY &&
                model != VFF_CM_ntscYIQ &&
                model != VFF_CM_cieXYZ &&
                model != VFF_CM_cieUVW) {
            kerror("vipl", "lvcltrans",
		   "Input image must be\n of multi-spectral typentscRGB,"
		   "genericRGB, ntscYIQ, cieXYZ, cieUVW, or ntscCMY");
            return (FALSE);
        }

        ptr1 = (float *) & img -> imagedata[0];
        ptr2 = (float *) & img -> imagedata[nr * nc * sizeof (float)];
        ptr3 = (float *) & img -> imagedata[2 * nr * nc * sizeof (float)];

         /* first switch on the input colormap model, then swtich on
            the output colormap model */
        switch (img -> color_space_model) {
            case VFF_CM_ntscRGB: 
            case VFF_CM_genericRGB: 
                switch (type) {
                    case 1:     /* convert to NTSC: RGB */
                        img -> color_space_model = VFF_CM_ntscRGB;
                        break;  /* RGB to RGB -- no conversion necessary */
                    case 2:     /* convert to generic: RGB */
                        img -> color_space_model = VFF_CM_genericRGB;
                        break;  /* RGB to RGB -- no conversion necessary */
                    case 3:     /* convert to NTSC: YIQ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, RGBtoYIQ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, RGBtoYIQ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, RGBtoYIQ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_ntscYIQ;
                        break;
                    case 4:     /* convert to NTSC: CMY */
                        for (i = 0; i < nr * nc; i++) {
                            *ptr1 = 1 - (*ptr1/norm); /* CMY = 1 - RGB */
                            *ptr2 = 1 - (*ptr2/norm);
                            *ptr3 = 1 - (*ptr3/norm);
                            ptr1++; ptr2++; ptr3++;
                        }
                        img -> color_space_model = VFF_CM_ntscCMY;
                        break;
                    case 5:     /* convert to CIE: XYZ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, RGBtoXYZ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, RGBtoXYZ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, RGBtoXYZ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_cieXYZ;
                        break;
                    case 6:     /* convert to CIE: UVW */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1(tmp1, tmp2, tmp3, RGBtoUVW); ptr1++;
                            *ptr2 = TRANS2(tmp1, tmp2, tmp3, RGBtoUVW); ptr2++;
                            *ptr3 = TRANS3(tmp1, tmp2, tmp3, RGBtoUVW); ptr3++;
                        }
                        img -> color_space_model = VFF_CM_cieUVW;
                        break;
                    default: 
                        kerror("vipl", "lvcltrans",
			       "Invalid type to convert input image to");
                        return (FALSE);
                }
                break;
            case VFF_CM_ntscYIQ: 
                switch (type) {
                    case 1:     /* convert to NTSC: RGB */
                    case 2:     /* convert to generic: RGB */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, YIQtoRGB);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, YIQtoRGB);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, YIQtoRGB);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        if (type == 1)
                           img -> color_space_model = VFF_CM_ntscRGB;
                        else
                           img -> color_space_model = VFF_CM_genericRGB;

                        break;
                    case 3:     /* convert to NTSC: YIQ */
                        img -> color_space_model = VFF_CM_ntscYIQ;
                        break;
                    case 4:     /* convert to NTSC: CMY */
                           /* use conversion matrix to RGB then transform
                               the RGB to CMY */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, YIQtoRGB);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, YIQtoRGB);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, YIQtoRGB);
                            *ptr1 = 1 - *ptr1; /* convert to CMY from RGB */
                            *ptr2 = 1 - *ptr2; /* CMY = 1 - RGB */
                            *ptr3 = 1 - *ptr3;
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_ntscCMY;
                        break;
                    case 5:     /* convert to CIE: XYZ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, YIQtoXYZ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, YIQtoXYZ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, YIQtoXYZ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_cieXYZ;
                        break;
                    case 6:     /* convert to CIE: UVW */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, YIQtoUVW);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, YIQtoUVW);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, YIQtoUVW);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_cieUVW;
                        break;
                    default: 
                        kerror("vipl", "lvcltrans",
			       "Invalid type to convert input image to");
                        return (FALSE);
                }
                break;
            case VFF_CM_ntscCMY: 
                  /* for conversion from CMY to YIQ, XYZ, UVW, first
                     must convert to RGB then perform the transformation */
                switch (type) {
                    case 1:     /* convert to NTSC: RGB */
                    case 2:     /* convert to generic: RGB */
                        for (i = 0; i < nr * nc; i++) {
                            *ptr1 = 1 - *ptr1;
                            *ptr2 = 1 - *ptr2;
                            *ptr3 = 1 - *ptr3;
                            ++ptr1;
                            ++ptr2;
                            ++ptr3;
                        }
                        if ( type == 1) 
                           img -> color_space_model = VFF_CM_ntscRGB;
                        else
                           img -> color_space_model = VFF_CM_genericRGB;

                        break;
                    case 3:     /* convert to NTSC: YIQ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = 1 - (*ptr1/norm);
                            tmp2 = 1 - (*ptr2/norm);
                            tmp3 = 1 - (*ptr3/norm);
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, RGBtoYIQ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, RGBtoYIQ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, RGBtoYIQ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_ntscYIQ;
                        break;
                    case 4:     /* convert to NTSC: CMY */
                        img -> color_space_model = VFF_CM_ntscCMY;
                        break;
                    case 5:     /* convert to CIE: XYZ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = 1 - (*ptr1/norm);
                            tmp2 = 1 - (*ptr2/norm);
                            tmp3 = 1 - (*ptr3/norm);
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, RGBtoXYZ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, RGBtoXYZ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, RGBtoXYZ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_cieXYZ;
                        break;
                    case 6:     /* convert to CIE: UVW */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = 1 - (*ptr1/norm);
                            tmp2 = 1 - (*ptr2/norm);
                            tmp3 = 1 - (*ptr3/norm);
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, RGBtoUVW);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, RGBtoUVW);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, RGBtoUVW);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_cieUVW;
                        break;
                    default: 
                        kerror("vipl", "lvcltrans",
			       "Invalid type to convert input image to");
                        return (FALSE);
                }
                break;
            case VFF_CM_cieXYZ: 
                switch (type) {
                    case 1:     /* convert to NTSC: RGB */
                    case 2:     /* convert to generic: RGB */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, XYZtoRGB);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, XYZtoRGB);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, XYZtoRGB);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        if ( type == 1)
                           img -> color_space_model = VFF_CM_ntscRGB;
                        else
                           img -> color_space_model = VFF_CM_genericRGB;
                        break;
                    case 3:     /* convert to NTSC: YIQ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, XYZtoYIQ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, XYZtoYIQ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, XYZtoYIQ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_ntscYIQ;
                        break;
                    case 4:     /* convert to NTSC: CMY */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, XYZtoRGB);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, XYZtoRGB);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, XYZtoRGB);
                            *ptr1 = 1 - *ptr1;
                            *ptr2 = 1 - *ptr2;
                            *ptr3 = 1 - *ptr3;
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_ntscCMY;
                        break;
                    case 5:     /* convert to CIE: XYZ */
                        img -> color_space_model = VFF_CM_cieXYZ;
                        break;
                    case 6:     /* convert to CIE: UVW */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, XYZtoUVW);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, XYZtoUVW);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, XYZtoUVW);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_cieUVW;
                        break;
                    default: 
                        kerror("vipl", "lvcltrans",
			       "Invalid type to convert input image to");
                        return (FALSE);
                }
                break;
            case VFF_CM_cieUVW: 
                switch (type) {
                    case 1:     /* convert to NTSC: RGB */
                    case 2:     /* convert to generic: RGB */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, UVWtoRGB);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, UVWtoRGB);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, UVWtoRGB);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        if ( type == 1)
                           img -> color_space_model = VFF_CM_ntscRGB;
                        else
                           img -> color_space_model = VFF_CM_genericRGB;
                        break;
                    case 3:     /* convert to NTSC: YIQ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, UVWtoYIQ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, UVWtoYIQ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, UVWtoYIQ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_ntscYIQ;
                        break;
                    case 4:     /* convert to NTSC: CMY */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, UVWtoRGB);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, UVWtoRGB);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, UVWtoRGB);
                            *ptr1 = 1 - *ptr1;
                            *ptr2 = 1 - *ptr2;
                            *ptr3 = 1 - *ptr3;
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_ntscCMY;
                        break;
                    case 5:     /* convert to CIE: XYZ */
                        for (i = 0; i < nr * nc; i++) {
                            tmp1 = *ptr1/norm;
                            tmp2 = *ptr2/norm;
                            tmp3 = *ptr3/norm;
                            *ptr1 = TRANS1 (tmp1, tmp2, tmp3, UVWtoXYZ);
                            *ptr2 = TRANS2 (tmp1, tmp2, tmp3, UVWtoXYZ);
                            *ptr3 = TRANS3 (tmp1, tmp2, tmp3, UVWtoXYZ);
                            ++ptr1; ++ptr2; ++ptr3;
                        }
                        img -> color_space_model = VFF_CM_cieXYZ;
                        break;
                    case 6:     /* convert to CIE: UVW */
                        img -> color_space_model = VFF_CM_cieUVW;
                        break;
                    default: 
                        kerror("vipl", "lvcltrans",
			       "Invalid type to convert input image to");
                        return (FALSE);
                }
                break;
            default: 
                kerror("vipl", "lvcltrans", "Invalid input image colormap model"
		       "must be ntscRGB, genericRGB, ntscCMY, ntscYIQ, cieXYZ, "
		       "or cieUVW");
                return (FALSE);
        }
    }

    return (TRUE);
}
/* -library_code_end */
