 /*
  * Khoros: $Id: lvfractal.c,v 1.12 1996/07/05 03:27:30 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvfractal.c,v 1.12 1996/07/05 03:27:30 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vfractal
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvfractal
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvfractal - Fractal feature extraction of the image
* 
*       Purpose: vfractal performs a fractal feature extraction of an image via
*		 calculation of the fractal dimension "D".  The fractal
*		 dimension is computed based on a calculation of the Hurst
*		 coefficient within a selected window size or "fractal region"
*		 of the image.  The Hurst coefficient is a measure of the slope
*		 of the line relating the window size vs average range of image
*		 intensity.  The Hurst coefficient is related to the fractal
*		 dimension as follows,
*
*		 !                     D = 3 - H
*		 !   where,
*		 !            D  -  is the resulting fractal dimension
*		 !            H  -  is the computed Hurst coefficient
*		 !            3  -  constant related to the Euclidian geometry
*
*		 The fractal dimension operation is computed, using overlapping
*		 windows,
*		 over the entire image.  A range of window sizes up to one
*		 third of the size of the fractal region are used to compute
*		 the fractal dimension.  The center pixel in each fractal
*		 region is updated based on the Hurst coefficient and
*		 subsequent fractal dimension calculation for that region.  
*
*		 The input image, which MUST be of data storage type FLOAT,
*		 corresponds to the -i argument.  The output image, which
*		 corresponds to the -o argument, is stored as data type FLOAT.
*		 The resulting output image will be a single band image, and
*		 will have a border consisting of pixels of value 0, since the
*		 window operation does not calculate values for border pixels.
*
*		 The size of the window for the "fractal region" is specified
*		 by the -r region size argument.  The region size argument
*		 cannot be an even number.  This means that the region
*		 dimensions must be odd (ie. 15x15, 17x17, etc.), since the
*		 center pixel is replaced with the computed value of the window
*		 data.  The default region size is 15x15.
*
*		 The initial size of the sliding window in the fractal region is
*		 specified by the -w initial sliding window size argument.  The
*		 size of the sliding window cannot be less than 2 and cannot
*		 exceed 2 minus the upper window size.  The upper window size
*		 is determined by the region size, and cannot exceed one-third
*		 of the region size.  The size of the lower window cannot exceed
*		 2 minus the upper window size, because at least two points are
*		 needed to calculate the slope of the best fit line. 
*
*		 The step size specifies the increment to use when incrementing
*		 the sliding window size from the initial size (-w argument) to
*		 the upper window size (determined from the region size).  The
*		 step size cannot exceed a value that would prevent at least two
*		 points from being used to calculate the slope of the best fit
*		 line.  The step size is specified by the -s step size argument.
*
*		 Each pixel in the image is updated with a computed value
*		 except the border pixels.  A 15x15 window will result in a
*		 border of seven pixels, a 17x17 window will result in a border
*		 of eight pixels, etc.
*
*		 The resulting feature vector will contain one band of data,
*		 which contains the computed fractal dimension "D" of the image
*		 intensity data for the particular region size selected.  Note
*		 that different results may be obtained by varying the region
*		 size and initial size of the sliding window.
*
*		 Comments - To calculate the overall fractal dimension of an
*		 image, set the region size to be the same size as the image
*		 (Note: region size must be odd), and use an initial sliding
*		 window size about one-fourth to one-third of the region size.
*		 This will result in one value for the fractal dimension of the
*		 image, which can be obtained from running vstats.
*
*		 When running this algorithm on a large image using a wide range
*		 of sliding window sizes, it may be helpful to use a generous
*		 step size to speed up the processing of the image.  For
*		 example, if a 256 by 256 image is being processed with a region
*		 size of 65 and an initial sliding window size of 3, it may be
*		 helpful to select a step size of about 6.  The following window
*		 sizes would be used in the calculation of the Hurst coefficient
*		 (slope of best fit line):  window sizes --> 3  9  15  21
*
*		 If finer resolution is needed, a smaller step size should be
*		 used at the expense of processing time.
*
*		 To view the resulting fractal dimension image in 3-D, use
*		 Xprism3 to plot the image as a surface plot.
*
*		 The input image MUST be of data storage type FLOAT. 
*
*		 The output image is of data storage type FLOAT.
*         Input: img - xvimage structure
*		 region - region window size argument
*		 lo_size - starting size of sliding window
*		 step - step size or increment of sliding window
*        Output: img - holds the resulting feature of the image.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: lvfractal will only operate on images of data storage type
*		 FLOAT. The output image is of data storage type FLOAT.
*    Written By: Charlie Gage
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 3/19/90  Charlie Gage - updated for processing speed.
****************************************************************/
/* -library_def */
int lvfractal(
	xvimage *img,
	int region,
	int lo_size,
	int step)
/* -library_def_end */

/* -library_code */
{
  int row,            /* points to image row */
      col,            /* points to image col */
      r_row,          /* points to fractal region row */
      r_col,          /* points to fractal region col */
      offset,         /* window offset in the image */
      xend,           /* step in img up to here */
      yend,           /* step in img up to here */
      r_end,          /* step in region up to here */
      nc,             /* column size of image */
      nr,             /* row size of image */
      reg_size,       /* size of fractal region */
      lo_window,      /* starting size of sliding window */
      hi_window,      /* final size of sliding window */
      win_size,       /* current size of sliding window */
      dim,            /* dimension of arrays */
      i,              /* general counters */
      j,
      k,
      p,
      index,
      left,
      right,
      top,
      bottom,
      num_pts;

float
      x;             /* points to current pixel value */

double
      max_inten,     /* holds maximum intensity value of sliding window */
      min_inten;     /* holds minimum intensity value of sliding window */

double
      tmp_slope,             /* variables used to calculate least square fit */
      tmp,
      sum_sqr,
      sum_x,
      avg,
      ave_range,     /* average range of intensity values in window */
      hurst,         /* hurst coefficient used to calculate fractal dim */
      fract_dim;     /* calculated fractal dimension of window */

float  *result;      /* pointer to resulting image */

double
       *h,           /* pointer to hurst coefficients */
       *w,           /* pointer to abcissa for least square fit calculations */
       *row_range,   /* pointer to row range values */
       *col_range;   /* pointer to col range values */

int    *count;       /* count number of windows in region for each win_size */

float  *fptr,        /* pointer to image data - type is FLOAT */
       **ptr;        /* pointer to 2-D working array */

 nr = img->col_size;              /* number of rows in image */
 nc = img->row_size;              /* number of cols in image */

 reg_size = region;               /* get dimension of fractal window */
 lo_window = lo_size;             /* get starting size of sliding window */
 hi_window = (reg_size / 3);      /* find final size of sliding window */
 dim = (hi_window - lo_window)+1;  /* find dimension of arrays used to
                                      calculate fractal dimension. */

    /* Check for proper input image type */
 if ((img->data_storage_type) != VFF_TYP_FLOAT)
  {
    kerror("vipl", "lvfractal",
	   "Error, Input image must be of data type FLOAT\n");
    return(FALSE);
  }

    /* check for odd size of fractal region */
 if((reg_size % 2) == 0)
  {
    kerror("vipl","lvfractal", "Error, region size must be odd\n");
    return(FALSE);
  }

    /* Ensure that region size does not exceed image size */
 if ((reg_size > nr) || (reg_size > nc))
  {
    kerror("vipl", "lvfractal", "Error, region size exceeds image size\n\
Image size is %d by %d\n", nr, nc);
    return(FALSE);
  }

    /* Ensure that region size does not drop below minimum of 9 x 9 */
 if (reg_size < 9)
  {
    kerror("vipl", "lvfractal",
	   "Error, region size too small to compute fractal dimension\n\
Minimum region size is 9 x 9\n");
    return(FALSE);
  }

    /* Ensure that lo_window size does not exceed fractal region size */
    /* Need at least two points to define the slope */
 if ((dim < 2) || (lo_window < 2))
  {
    kerror("vipl", "lvfractal", "Error, lo window size exceeds limits.\n\
Valid range is 2 to %d\n", (hi_window - 1));
    return(FALSE);
  }

    /* Ensure that step size is not too great */
 if ((lo_window + step) > hi_window)
  {
    kerror("vipl","lvfractal",
	   "Error, step size exceeds bounds for this image.\n\
Valid range is 1 to %d.\n", (hi_window-lo_window));
    return(FALSE);
  }

                 /* Assign input image data address to ptr */
 fptr = (float *)img->imagedata;

                 /* Allocate space for 2-D working array */
 ptr = (float **) kmalloc((unsigned int) nr * sizeof(float));
 if (ptr == NULL){
    kerror("vipl","lvfractal", "Insufficient memory available!");
    return(FALSE);
 }
 for (i = 0; i < nr; i++)
 {
  ptr[i] = (float *) kmalloc((unsigned int) nc * sizeof(float));
  if (ptr[i] == NULL)
  {
   kerror("vipl","lvfractal", "insufficient memory available!");
   return(FALSE);
  }
 }

         /* Zero out 2-D working array */
 for (i = 0; i < nr; i++)
 {
  for (j = 0; j < nc; j++)
  {
   ptr[i][j] = 0.0;
  }
 }

         /* Assign input image data to 2-D working array */
 for (i = 0; i < nr; i++)
 {
  for (j = 0; j < nc; j++)
  {
   ptr[i][j] = fptr[nc * i + j];
  }
 }

   /* Allocate space for resulting image --- of type FLOAT */
 result = (float *) kmalloc((unsigned int) nc * nr * sizeof(float));
 if (result == NULL){
    kerror("vipl","lvfractal", "Insufficient memory available!");
    return(FALSE);
 }

   /* Allocate space for data arrays */
 h = (double *) kmalloc((unsigned int) dim * sizeof(double));
 if (h == NULL){
    kerror("vipl","lvfractal", "Insufficient memory available!");
    return(FALSE);
 }
 w = (double *) kmalloc((unsigned int) dim * sizeof(double));
 if (w == NULL){
    kerror("vipl","lvfractal", "Insufficient memory available!");
    return(FALSE);
 }
 count = (int *) kmalloc((unsigned int) dim * sizeof(int));
 if (count == NULL){
    kerror("vipl","lvfractal", "Insufficient memory available!");
    return(FALSE);
 }
 row_range = (double *) kmalloc((unsigned int) dim * sizeof(double));
 if (row_range == NULL){
    kerror("vipl","lvfractal", "Insufficient memory available!");
    return(FALSE);
 }
 col_range = (double *) kmalloc((unsigned int) dim * sizeof(double));
 if (col_range == NULL){
    kerror("vipl","lvfractal", "Insufficient memory available!");
    return(FALSE);
 }


   /* Zero out space for resulting image */
 (void) kmemset((kaddr) result, 0, (size_t) (nc * nr * sizeof(float)));

 offset = reg_size / 2;   /* Find window offset from center of region */
 xend = nc - reg_size + 1;   /* step in image (x-dir) up to here */
 yend = nr - reg_size + 1;   /* step in image (y-dir) up to here */

   /* Initialize variables */
 p = 0;
 k = 0;
 (void) kmemset((kaddr) count, 0, (size_t) (dim * sizeof(int)));
 (void) kmemset((kaddr) row_range, 0, (size_t) (dim * sizeof(double)));
 (void) kmemset((kaddr) col_range, 0, (size_t) (dim * sizeof(double)));

 /*--------------------------------------------------------------------
  *   load range arrays with intensity values of first "region"
  *   in the image.
  *--------------------------------------------------------------------*/

 for (win_size = lo_window; win_size <= hi_window; (win_size += step))
  {
    r_end = reg_size - win_size + 1;
    for (r_row =0; r_row < r_end; r_row++)
    {
     for (r_col = 0; r_col < r_end; r_col++)
     {
      max_inten = (double) ptr[r_row][r_col];
      min_inten = max_inten;
          /* find range of intensities in sliding window */
      for (i = 0; i < win_size; i++)
      {
       for (j = 0; j < win_size; j++)
       {
        x = ptr[r_row + i][r_col + j];
        if (x > max_inten)
           max_inten = (double)x;
        if (x < min_inten)
           min_inten = (double)x;
       }
      }
      col_range[k] += (double)(max_inten - min_inten);
      row_range[k] += (double)(max_inten - min_inten);
      count[k]++;           /* count number of windows in region */
     }                      /*  for current window size.   */
    }
    k += step;
 
  } /* End of win_size loop */
  k = 0;

 /*---------------------------------------------------------------------
  *         perform fractal analysis of an image 
  *--------------------------------------------------------------------*/

     /*-------- loop through ROWS of image -- (y-direction) ---------*/
 for (row = 0 ; row < yend; row++)
 {
      /*-------- loop through COLS of image -- (x-direction) ---------*/
  for (col = 0 ; col < xend; col++)
  {
   /*--------------------------------------------------------------------
    *    perform fractal analysis of region by looping through 
    *    all window sizes up to one-third of fractal region.
    *     ie.  L <= 1/3(extent)
    *-------------------------------------------------------------------*/
   for (win_size = lo_window; win_size <= hi_window; (win_size += step))
   {
    r_end = reg_size - win_size + 1;      /* determine loop limits */

       /* Don't subtract first column if col = 0 */
    if (col != 0)
    {
     /*--------------------  Subtract LEFT side ------------*/
     for (left = 0; left < r_end; left++)
     {
      index = col-1;
      max_inten = (double)ptr[left + row][index];
      min_inten = max_inten;
      for (i = 0; i < win_size; i++)
      {
       for (j = 0; j < win_size; j++)
       {
        x = ptr[left + i + row][index + j];
        if (x > max_inten)
           max_inten = (double)x;
        if (x < min_inten)
           min_inten = (double)x;
       }
      }
      col_range[k] -= (double)(max_inten - min_inten);
     }
     /*--------------------  Add RIGHT side ----------------*/
     for (right = 0; right < r_end; right++)
     {
      index = col + r_end -1;
      max_inten = (double)ptr[right + row][index];
      min_inten = max_inten;
      for (i = 0; i < win_size; i++)
      {
       for (j = 0; j < win_size; j++)
       {
        x = ptr[right + i + row][index + j];
        if (x > max_inten)
           max_inten = (double)x;
        if (x < min_inten)
           min_inten = (double)x;
       }
      }
      col_range[k] += (max_inten - min_inten);
     }

    }  /*----  End of col != 0 ----*/

    /* find average range for current window size in the region */
    ave_range = (double) col_range[k] / count[k];
    if (ave_range <= 0.0)
       h[k] = 0.0;
    else
       h[k] = log10((double) ave_range);
    w[k] = log10((double) win_size);
    k += step;
  }  /*--- End of loop through all window sizes ---*/

  /*---------------------------------------------------------------
   *   find H of fractal region by performing least squares fit 
   *   of the average ranges over various window sizes.
   *---------------------------------------------------------------*/
   tmp_slope = 0.0;
   sum_x = 0.0;
   sum_sqr = 0.0;
   num_pts = 0;

   for (i = 0; i < dim; (i += step)){
       sum_x += w[i];
       num_pts++;
   }
   avg = sum_x / (double)num_pts;
   for (i = 0; i < dim; (i += step)){
       tmp = w[i] - avg;
       sum_sqr += tmp * tmp;
       tmp_slope += tmp * h[i];
   }
   hurst = tmp_slope / sum_sqr;    /* find slope of best fit line */
                                   /* slope is Hurst coefficient */
   fract_dim = 3.0 - hurst;        /* calculate fractal dimension */

           /* update center pixel of region using fractal dimension */
   result[nc * (row + offset) + offset + p] = (float)fract_dim;

                   /* reset variables and increment place counter */
   k = 0;
   p++;

 }  /*-------- End of COL loop ------------*/

if ((yend - row) > 1)
{
 /*-------------------------------------------------------------------
  *     update range arrays for next loop thru columns of image 
  *-------------------------------------------------------------------*/
 for (win_size = lo_window; win_size <= hi_window; (win_size += step))
 {
  r_end = reg_size - win_size + 1;
  /*-----------------  Subtract TOP row from region ------------*/
  for (top = 0; top < r_end; top++)
  {
   max_inten = (double)ptr[row][top];
   min_inten = max_inten;
   for (i = 0; i < win_size; i++)
   {
    for (j = 0; j < win_size; j++)
    {
     x = ptr[row + i][top + j];
     if (x > max_inten)
        max_inten = (double)x;
     if (x < min_inten)
        min_inten = (double)x;
    }
   }
   row_range[k] -= (double)(max_inten - min_inten);
  }
  /*----------------  Add BOTTOM row to region  ---------------*/
  for (bottom = 0; bottom < r_end; bottom++)
  {
   index = row + r_end;
   max_inten = (double)ptr[index][bottom];
   min_inten = max_inten;
   for (i = 0; i < win_size; i++)
   {
    for (j = 0; j < win_size; j++)
    {
     x = ptr[index + i][bottom + j];
     if (x > max_inten)
        max_inten = (double)x;
     if (x < min_inten)
        min_inten = (double)x;
    }
   }
   row_range[k] += (double)(max_inten - min_inten);
  }
                       /* update col_range array for next pass */
  col_range[k] = row_range[k];
  k += step;
 }   /*-----  End of win_size loop  -----*/

}   /*----  End of if statement  -------*/

 p = 0;       /* reset place counter */
 k = 0;
}   /*-------- End of ROW loop ------------*/

                       /* release memory from input image */
kfree_and_NULL (img->imagedata);
                       /* adjust header in the output image */
img->data_storage_type = VFF_TYP_FLOAT;
img->imagedata = (char *) result;
img->num_data_bands = 1;

return(TRUE);

}  /****  End of lvfractal  ****/
/* -library_code_end */
