 /*
  * Khoros: $Id: lvhxray.c,v 1.13 1996/07/05 03:27:43 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvhxray.c,v 1.13 1996/07/05 03:27:43 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vhxray
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvhxray
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
static void hist_stretch      PROTO((unsigned int *, unsigned int *, int, int));
static void hist_equalization PROTO((unsigned int *, unsigned int *));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvhxray - Enhance images using histogram stretch or equalization with overlapping windows.
* 
*       Purpose: This module will enhance an image by using a histogram
*		 stretch or histogram equalization method.  Vhxray will use
*		 overlapping windows over the entire image, and for each
*		 window compute the histogram and perform a stretch or
*		 equalization. The middle pixel of the window is updated
*		 according to the new histogram value.
*
*		 The window size must be odd (3x3, 5x5, 15x15, etc), since
*		 we must have a center pixel to update for each window.
*		 The default window size is 5x5.
*
*		 Since windows are being used, the border pixels will not be
*		 updated. the border pixels will be set to zero, black.  The
*		 border size can be calculated as follows:
*
*		 ! window width  / 2 = side border,
*		 ! window heigth / 2 = top & bottom border.
*
*		 The histogram stretch algorithm will stretch the histogram
*		 out as must as possible. The equalization algorithm is
*		 standard.
*         Input: image - the input image
*		 type  - 1 = do histogram equalization.
*			 2 = do histogram stretch.
*		 xsize - window width size
*		 ysize - window heigth size
*        Output: image - after the enhancement
*       Returns: TRUE (1) on success, FALSE (0) on failure *
*  Restrictions: lvhxray will only work on BYTE images.
*    Written By: Tom Sauer, Charlie Gage
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvhxray(
	xvimage *image,
	int      type,
	int      xsize,
	int      ysize)
/* -library_def_end */

/* -library_code */
{
    int row,                            /* points image row */ 
        col,                            /* points image col */ 
        g1,                             /* auxiliar variable */
        xoff,                           /* horiz. offset in the img */
        yoff,                           /* vertic. offset in the img */
        xend,                           /* step in img up to here */
        yend,                           /* step in img up to here */
        nc,                             /* column size of image */
        nr,                             /* row size of image */
        i, j, k, m;                     /* general counters */

    unsigned char med;                  /* this is the median value */

    unsigned int hist[256];             /* holds histogram of window */
    unsigned int hist2[256];            /* holds temp histogram of window */

    unsigned char *ptr,                 /* pointer to the image data char */
                  *result;              /* pointer to resulting image char */

    char *program = "lvhxray";          /* contains the library name */


   /* check for odd windows */
    if((xsize & 1) == 0 || (ysize & 1) == 0){
        (void) kerror("vipl", "lvhxray","Error window must be odd");
        return(FALSE);
    }

      /* can only have one data band */

    if (! (proper_num_images(program,image,(unsigned long) 1,FALSE))) {
         (void) kerror("vipl", "lvhxray",
		       "All inputs must contain only one image");
          return(FALSE);
     }

      /* can only have one data band */

     if (! (proper_num_bands(program,image,(unsigned long) 1,FALSE))) {
          (void) kerror("vipl", "lvhxray", 
                  "All inputs must contain only one data band");
           return(FALSE);
      }

      /* can only have data type of BYTE */

     if (! (propertype(program,image,(unsigned long) VFF_TYP_1_BYTE,FALSE))) {
          (void) kerror("vipl", "lvhxray", "Input must have data type of BYTE");
           return(FALSE);
      }

      /* must have map enable set to OPTIONAL */

     if (! (proper_map_enable(program,image,(unsigned long) VFF_MAP_OPTIONAL,FALSE))) {
          (void) kerror("vipl", "lvhxray",
                        "Input must have map enable type OPTIONAL");
           return(FALSE);
      }


    nr = image->col_size;
    nc = image->row_size;

         /* compute the border size of pixels not updated */
    xoff = xsize/2;
    yoff = ysize/2;

        /* compute where the window should stop on the right and bottom */
    xend = nc - xsize;
    yend = nr - ysize + 1;
    m = 0; 

   /* Zero histogram */
    kmemset((char *) hist, 0, 256 * sizeof(unsigned int));

   /* Allocate space for resulting image */ 
    result = (unsigned char *)kmalloc((unsigned int)nc*nr*sizeof(unsigned char));
    if(result == NULL) {
        (void) kerror("vipl", "lvhxray","Insufficient memory available");
    }

   /* Zero result image */
    kmemset((char *) result, 0, nc*nr * sizeof(unsigned char));

   /* Assign image data address to ptr */
    ptr = (unsigned char *)image->imagedata;

      /* scan each row of the image, stepping column by column */
      /* with the window */

    for (row = 0 ; row < yend; row++){

       /* setup the histogram for the current window */

        for(i = 0; i < ysize; i++){
            for(j = 0; j < xsize; j++){
                g1 = ptr[nc*(row + i) + j];
                hist[g1]++;
            }
        }

          /* compute the equalization or stretch of the window */

        switch(type) {
          case 0:
              hist_equalization(hist, hist2);
              break;
          case 1:
              hist_stretch(hist, hist2, xsize, ysize);
              break;
          default:
             kerror("vipl", "lvhxray","Invalid histogram enhancement type");
             return(FALSE);
        }

         /* get the middle pixel value and assign that image location
            with the middle pixel value mapped through the histogram */

        med = image->imagedata[nc *(row + yoff) + xoff + m];
        result[nc *(row + yoff) + xoff + m++] = (unsigned char) hist2[med];

          

           /* step across the current row with the window */

        for (col = 0 ; col < xend ; col++){

              /* adjust the histogram for the next window */

            for(k = 0; k < ysize; k++){

                   /* delete the lefmost column of the previous window*/
                g1 =  ptr[col + nc*(row + k)];
                hist[g1] = hist[g1] -1;


                   /* add the rightmost column of the current window*/
                g1 = ptr[xsize+ nc*(row + k) +col];
                hist[g1] = hist[g1] + 1;

            }
            
               /* compute the equalization or stretch of the window */

            switch(type) {
              case 0:
                  hist_equalization(hist, hist2);
                  break;
              case 1:
                  hist_stretch(hist, hist2, xsize, ysize);
                  break;
              default:
                 kerror("vipl", "lvhxray","Invalid histogram enhancement type");
                 return(FALSE);
            }

             /* get the middle pixel value and assign that image location
                with the middle pixel value mapped through the histogram */

            med = image->imagedata[nc *(row + yoff) + xoff + m];
            result[nc *(row + yoff) + xoff + m++] = (unsigned char) hist2[med];

        }

       /* Reset histogram and variables */

        kmemset((char *) hist, 0, 256 * sizeof(unsigned int));
        m = 0; 
    }

      /* assign the new image to the old image pointer */
    image->imagedata = (char *) result;

    return(TRUE);
}

/*-----------------------------------------------------------
|
|  Routine Name: hist_stretch - perform a histogram stretch
|       Purpose: Perform a histogram stretch on the histogram
|                passed in. This routine will perform the
|                maximum stretch possible.
|         Input: hist - a 256 histogram 
|        Output: hist - a 256 histogram
|    Written By: Ramiro Jordan and Marcelo Teran.
|          Date: Aug 05, 1994
------------------------------------------------------------*/
static void hist_stretch(
	unsigned int *hist,
	unsigned int *hist2,
	         int xsize,
		 int ysize)
{
    int  i, lo, hi, klo, khi, thresh;

    /* Find the histogram boundaries by locating the 1 percent levels */
    thresh = xsize * ysize * 0.01;
    lo  =  hi = -1;
    klo = khi = 0;
    for (i = 0; i < 256; i++)
      {
        klo += hist[i]; khi += hist[255-i];
        if (klo > thresh && lo == -1) lo = i;
        if (khi > thresh && hi == -1) hi = 255-i;
        if (lo != -1 && hi != -1) break;
      }

    /* Check to see if we have a reasonable contrast span. If not, re-do
       the bounds with a zero threshold. */
    if (hi == lo)
      {
        thresh = 0;
        lo  =  hi = -1;
        klo = khi = 0;
        for (i = 0; i < 256; i++)
          {
            klo += hist[i]; khi += hist[255-i];
            if (klo > thresh && lo == -1) lo = i;
            if (khi > thresh && hi == -1) hi = 255-i;
            if (lo != -1 && hi != -1) break;
          }
      } 

    /* If still have a zero span bound, quit */
    if (hi == lo) {
       kmemcpy(hist2, hist, 256*sizeof(int));
       return;
      }

    /* Re map the histogram */
    for (i = 0; i < 256; i++)
    {
        if (i < lo)
           hist2[i] = 0;
        else if (i > hi)
           hist2[i] = 255;
        else
           hist2[i] = 255*(i-lo)/(hi-lo);
    }
}

/*-----------------------------------------------------------
|
|  Routine Name: hist_equalization - perform a histogram equalization
|       Purpose: Performs a histogram equalization on a histogram.
|         Input: hist - a 256 histogram
|        Output: hist - a 256 histogram
|    Written By: Ramiro Jordan and Marcelo Teran
|          Date: Aug 05, 1994
------------------------------------------------------------*/
static void hist_equalization(
	unsigned int *hist,
	unsigned int *hist2)
{
    int  i;
    unsigned int  map[256], k;
    float f;

    /* Integrate the histogram to get the equalization map */

    k = 0;
    for (i = 0; i < 256; i++)
    {
        k += hist[i];
        map[i] = k ;
    }

    /* Compute scale factor */
    if (map[255] == 0.0)  {
       kmemcpy(hist2, hist, 256*sizeof(int));
       return;
    }

    f = 255.0/map[255];
    
    /* scale the cumulative distribution */
    for (i = 0; i < 256; i++) 
       hist2[i] = (unsigned int) ((float) map[i] * f);
}
/* -library_code_end */
