 /*
  * Khoros: $Id: lvspeckle.c,v 1.13 1996/07/05 03:28:00 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvspeckle.c,v 1.13 1996/07/05 03:28:00 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vspeckle
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvspeckle
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
static void neghull  PROTO((int, int, int, int, unsigned char *,
			    unsigned char *));
static void poshull  PROTO((int, int, int, int, unsigned char *,
			    unsigned char *));
static void symhull  PROTO((int, int, int, int, unsigned char *,
			    unsigned char *));
static void geomfilt PROTO((int, int, unsigned char *, unsigned char *));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvspeckle - Reduce speckle noise from an image by using the Crimmins algorithm.
* 
*       Purpose: lvspeckle reduces the speckle index of an image by sending
*		 the image through a geometric filter.
*		 The function takes one argument: 'n' which determines how
*		 many times the image is to be sent through the speckle
*		 reducing filter.  This filter uses the complementary hulling
*		 technique to reduce the speckle index of an image. 
*
*		 Each pixel in the image is compared with all eight of its
*		 surrounding pixels.  The pixels above and below the pixel
*		 are the N-S pair, the pixels on either side are the E-W
*		 pair, those on the diagonals are the NW-SE and NE-SW pair.
*		 The image and a pair of pixels are sent to the symmetric
*		 hull function which in turn sends the image to the positive
*		 hull function and to the negative hull function.  The
*		 positive and negative hull functions are called twice; the
*		 first time the neighboring pair of pixels is sent and the
*		 second time the complement of the pixels is sent.  The
*		 positive hull function replaces each "middle pixel" with
*		 the result of a several comparisons.  These comparison are
*		 a combination of maximums and minimums.  The final
*		 comparison is a maximum for the postive hull algorithm.
*		 The image is then sent to the negative hull function where
*		 the middle pixel is replaced with the result of more
*		 comparisons.  The final comparison for the negative hull
*		 algorithm is a minimum.  The result of all these
*		 replacements has the effect of reducing the undesired
*		 speckle noise while preserving the edges of the original
*		 image.
*         Input: image - input image, must be a byte image.
*		 n - the number of times the user wishes to send the image
*		     through the speckle removal filter.  Limited to 20.
*        Output: img2 - image with speckle removal filter applied the proper
*			number of times.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: lvspeckle accepts only BYTE images.  The number of iterations
*		 is limited to 20.
*    Written By: Caryl Peterson and R. J. Fogler
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: *		14-May-1996 - Applied bug fixes from Bob Whatmough,
*                             rjw@itd.dsto.gov.au to correct indexing problems,
*                             overflow problems, and some discrepancies
*                             between the code as stated and the original
*                             algorithm published by Crimmins in
*                             Optical Engineering, May 1986 p653.
****************************************************************/
/* -library_def */
int lvspeckle(
	xvimage *image,
	int n)
/* -library_def_end */

/* -library_code */
{
   int nc,                 /* number of columns of image */
       nr,                 /*   "    "  rows    "    "   */
       buf_nc, buf_nr, buf_size,
       i,j;

   unsigned char *imgptr, *imgbuf;  /* the imgptr will be incremented while 
                                       the imgbuf will always point to the
                                       first pixel in the image */
   unsigned char *in_buf, *g_buf, *inptr, *g_ptr;
                           /* The image is surrounded by an edge one pixel
                              wide whose value is 0.  Thus the edges of
                              the original image are preserved.  */

   char   *program = "lvspeckle";

  /* Check type,... of image */
    if (!(propertype(program, image, (unsigned long) VFF_TYP_1_BYTE, FALSE))) {
	kerror("vipl", "lvspeckle", "Image must be of type byte");
        return (FALSE);
    }
    if (!(proper_num_images (program, image, (unsigned long) 1, FALSE))) {
	kerror("vipl", "lvspeckle", "Can only work on files with one image");
        return (FALSE);
    }
    if (!(proper_num_bands (program, image, (unsigned long) 1, FALSE))) {
	kerror("vipl", "lvspeckle", "Can only work on images with 1 data band");
        return (FALSE);
    }


    nc = image->row_size; /* number of columns */
    nr = image->col_size; /* number of rows    */

    imgbuf = (unsigned char *)(image->imagedata); /* pointer to image's pixels */


/* *********************************************************** *
 * *        A L L O C A T E   I M A G E   B U F F E R S      * *
 * *********************************************************** */
    buf_nc = nc + 2;
    buf_nr = nr + 2;
    buf_size = buf_nc * buf_nr;

    /*  create in_buffer */
    in_buf = (unsigned char *) kmalloc ((size_t) buf_size);
    if (in_buf == NULL) {
      kerror("vipl", "lvspeckle", "vspeckle unable to allocate buffer");
      return(FALSE);
     }

   /* create g buffer */
    g_buf = (unsigned char *) kmalloc ((size_t) buf_size);
    if (g_buf == NULL) {
      kerror("vipl", "lvspeckle", "vspeckle unable to allocate buffer");
      return(FALSE);
     }

/* *********************************************************** *
 * *     P U T   S O U R C E   I M A G E  I N  B U F F E R   * *
 * *********************************************************** */
  imgptr = imgbuf;
  inptr = in_buf;
         /*  zero first row of buffer */
  for (i = 0; i < buf_nc; i++) {
      *inptr++ = ( unsigned char ) 0;
  }
  
  for (i = 0; i < nr ; i++) {
    *inptr++ = (unsigned char) 0;   /* zero first column */
    for (j = 0; j < nc; j++) {
       *inptr++ = *imgptr++;  /* place the image pixel's values in the buffer */
    }
    *inptr++ = (unsigned char) 0;   /* zero last column*/
     
    }
  
  for (i = 0; i < buf_nc; i++) { 
      *inptr++ = ( unsigned char ) 0;    /*  zero last row */
  }

/* *********************************************************** *
 *     P R E P A R E  G _ B U F F E R                     * *
 * *********************************************************** */
  g_ptr = g_buf;
         /*  zero first row of buffer */
  for (i = 0; i < buf_nc; i++) {
      *g_ptr++ = ( unsigned char ) 0;
  }
  for (i = 0; i < nr ; i++) {
    *g_ptr++ = (unsigned char) 0;   /* zero first column */
     g_ptr = (unsigned char *)(g_ptr + nc);
    *g_ptr++ = (unsigned char) 0;   /* zero last column*/
    }

  for (i = 0; i < buf_nc; i++) {  
      *g_ptr++ = ( unsigned char ) 0;  /* zero last row */
  }


/* *********************************************************** *
 * *    S P E C K L E   R E M O V A L   A L G O R I T H M    * *
 * *********************************************************** */


  for (j = 0; j < n; j++) {
     geomfilt(nr, nc, in_buf, g_buf);
   }

/* *********************************************************** *
 * *    P R E P A R E   R E S U L T A N T    I M A G E       * *
 * *********************************************************** */

 imgptr = imgbuf;
 inptr = in_buf;
 
 inptr = (unsigned char *) ( inptr + buf_nc);  /* skip first row*/

 for (i = 0; i < nr; i++){

   inptr++;  /* skip first column */
   
    for (j = 0; j < nc; j++) {
      *imgptr++ = *inptr++;
    }
   inptr++;   /* skip last column */
  }
  return (TRUE);
}

/* *********************************************************** *
 * *   G E O M E T R I C   F I L T E R   P R O C E D U R E   * *
 * *********************************************************** */
static void geomfilt(
	int            M,
	int            N,
	unsigned char *f,
	unsigned char *g)
{
  symhull(1, 0, M, N, f, g);   /* N-S   symmetrical hull */
  symhull(0, 1, M, N, f, g);   /* E-W   symmetrical hull */
  symhull(1, 1, M, N, f, g);   /* NE-SW symmetrical hull */
  symhull(1,-1, M, N, f, g);   /* NW-SE symmetrical hull */
}



/* *********************************************************** *
 * *    S Y M M E T R I C A L   H U L L   P R O C E D U R E  * *
 * *********************************************************** */
static void symhull(
	int            a,
	int            b,
	int            M,
	int            N,
	unsigned char *f,
	unsigned char *g)
{
          /* send the image twice to the positive hull */
  poshull( a, b, M, N, f, g);
  poshull(-a,-b, M, N, f, g);
          /* send the image twice to the negative hull */
  neghull(-a,-b, M, N, f, g);
  neghull( a, b, M, N, f, g);

}


/* *********************************************************** *
 * *       P O S I T I V E   H U L L   P R O C E D U R E     * *
 * *********************************************************** */
static void poshull(
	int            a,
	int            b,
	int            M,
	int            N,
	unsigned char *f,
	unsigned char *g)
{
  int m, n;
  int mn, ab, ba;
  unsigned int tmp, temp1, temp2; /* RJW */
  unsigned char *fmn, *gmn, *fab, *gab, *gba;
  unsigned char *fmn_p, *gmn_p, *fab_p, *gab_p, *gba_p;

  mn = (N+2) + 1;
  ba = a*(N+2) + b;
  ab = -ba;
  fmn = (unsigned char *)(f + mn);
  gmn = (unsigned char *)(g + mn);
  fab = (unsigned char *)(fmn + ab);
  gab = (unsigned char *)(gmn + ab);
  gba = (unsigned char *)(gmn + ba);

  fmn_p = fmn;
  fab_p = fab;
  gmn_p = gmn;
  for (m = 1 ; m <= M ; m++) {
    for (n = 1 ; n <= N ; n++) {
      *gmn_p = ((int) *fab_p >= (int) *fmn_p + 2) ? *fmn_p + 1 : *fmn_p;
      fmn_p++; gmn_p++; fab_p++; /* RJW */
    }
    fmn_p += 2;  /* RJW */
    gmn_p += 2;
    fab_p += 2;
  }

  fmn_p = fmn;
  gmn_p = gmn;
  gab_p = gab;
  gba_p = gba;
  for (m = 1 ; m <= M ; m++) {
    for (n = 1 ; n <= N ; n++) {
      *fmn_p = (*gba_p >= *gmn_p && *gab_p > *gmn_p) ? *gmn_p + 1 : *gmn_p;
      fmn_p++; gmn_p++; gab_p++; gba_p++; /* RJW */
    }
    fmn_p += 2;  /* RJW */
    gmn_p += 2;
    gab_p += 2;
    gba_p += 2;
  }
}

/* *********************************************************** *
 * *      N E G A T I V E   H U L L   P R O C E D U R E      * *
 * *********************************************************** */
static void neghull(
	int            a,
	int            b,
	int            M,
	int            N,
	unsigned char *f,
	unsigned char *g)
{
  int m, n;
  int mn, ab, ba;
  unsigned int tmp,  temp1, temp2;
  unsigned char *fmn, *gmn, *fab, *gab, *gba;
  unsigned char *fmn_p, *gmn_p, *fab_p, *gab_p, *gba_p;

  mn = (N+2) + 1;
  ba = a*(N+2) + b;
  ab = -ba;
  fmn = (unsigned char *)(f + mn);
  gmn = (unsigned char *)(g + mn);
  fab = (unsigned char *)(fmn + ab);
  gab = (unsigned char *)(gmn + ab);
  gba = (unsigned char *)(gmn + ba);

  fmn_p = fmn;
  gmn_p = gmn;
  fab_p = fab;
  for (m = 1 ; m <= M ; m++) {
    for (n = 1 ; n <= N ; n++) {
      *gmn_p = ((int) *fab_p <= (int) *fmn_p - 2) ? *fmn_p - 1 : *fmn_p;
      fmn_p++; gmn_p++; fab_p++; /* RJW */
    }
    fmn_p += 2;  /* RJW */
    gmn_p += 2;
    fab_p += 2;
  }

  fmn_p = fmn;
  gmn_p = gmn;
  gab_p = gab;
  gba_p = gba;
  for (m = 1 ; m <= M ; m++) {
    for (n = 1 ; n <= N ; n++) {
      *fmn_p = (*gba_p <= *gmn_p && *gab_p < *gmn_p) ? *gmn_p - 1 : *gmn_p;
      fmn_p++; gmn_p++; gab_p++; gba_p++; /* RJW */
    }
    fmn_p += 2;  /* RJW */
    gmn_p += 2;
    gab_p += 2;
    gba_p += 2;
  }
}
/* -library_code_end */
