 /*
  * Khoros: $Id: lvtilt.c,v 1.12 1996/07/05 03:28:03 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvtilt.c,v 1.12 1996/07/05 03:28:03 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vtilt
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvtilt
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvtilt - Adds a specified plane to the input image to correct for a tilt in the image luminance.
* 
*       Purpose: lvtilt adds a specified plane to the input image
*		 to correct for a tilt in the image luminance.
*		 Often an image will have a skew in the grey level
*		 values that represents a plane with a certain
*		 slope.  The tilt takes the form of a plane which
*		 has a X and Y zero crossing at the center of the
*		 input image with xslope and yslope as specified
*		 in the argument line.
*
*		 img1 is the input and output xvimage structure.
*		 This is done to save space, but you must be
*		 careful not to overwrite important data.  img2
*		 is the mask image, if mflg is set then the mask
*		 image is used to gate the operation.  A non-zero
*		 mask pixel enables the ouput pixel to contain
*		 adjusted values. A zero mask pixel just transfers
*		 the value of the corresponding pixel in the input
*		 image to the output image.  The input images must
*		 be of the same data type and size.
*
*		 xslope and yslope specify the characteristics of
*		 the plane and must be given in units per pixel.
*
*		 The output image data type is the same as the
*		 input data type.
*         Input: img1 - input/output image xvimage structure
*		 img2 - input mask image
*		 mflg - set to one if img2 to is to be used as a mask image
*		 xslope & yslope - specify the characteristics of the plane 
*				   and must be given in units per pixel.
*        Output: img1 - input/output image xvimage structure
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: lvtilt will not operate on COMPLEX or BIT data
*		 storage type images.
*    Written By: Scott Wilson, Donna Koechner
*          Date: Jul 04, 1996
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: John Rasure, Marcelo Teran, Donna Koechner
*		 Mon Jun 11 16:53:44 MDT 1990 Donna Koechner - updated to
*							       ghostwriter
*		 Fri Jun 28 16:43:19 MDT 1991 Scott Wilson - Fixed bogous
*							     pointer increment
*							     in float loop with
*							     mask
****************************************************************/
/* -library_def */
int lvtilt(
	xvimage *img1,
	xvimage *img2,
	int mflg,
	float xslope,
	float yslope)
/* -library_def_end */

/* -library_code */
{

    int i,j,k,rows,cols,rowsd2,colsd2;
    unsigned char *c1,*c2;
    short *s1,*s2;
    int *i1,*i2;
    float *f1,*f2;

    rows = img1->col_size;
    cols = img1->row_size;

    rowsd2 = rows / 2;
    colsd2 = cols / 2;

    if (rowsd2 == 0 || colsd2 == 0)
      {
        kerror("vipl","lvtilt", "Cannot tilt a one-pixel wide image!\n");
        return(FALSE);
      }

    switch (img1->data_storage_type)
      {
        case VFF_TYP_BIT :
          kerror("vipl","lvtilt", "Don't know how to TILT a BIT image!\n");
          return(FALSE);
          /* break; */
        case VFF_TYP_1_BYTE : 
          c1 = (unsigned char *)(img1->imagedata);
          if (mflg) c2 = (unsigned char *)(img2->imagedata);
          if (!mflg)
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      i = *c1 + (xslope*(k-colsd2)+yslope*(j-rowsd2));
                      if (i < 0) i = 0;
                      if (i > 255) i = 255;
                      *c1 = i;
                      c1++;
                    }
                }
            }
          else
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      if (*c2++)
                        {
                          i = *c1 + (xslope*(k-colsd2)+yslope*(j-rowsd2));
                          if (i < 0) i = 0;
                          if (i > 255) i = 255;
                          *c1 = i;
                          c1++;
                        }
                      else c1++;
                    }
                }
            }
          break;
        case VFF_TYP_2_BYTE :
          s1 = (short *)(img1->imagedata);
          if (mflg) s2 = (short *)(img2->imagedata);
          if (!mflg)
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      *s1 += xslope*(k-colsd2)+yslope*(j-rowsd2);
                      s1++;
                    }
                }
            }
          else
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      if (*s2++)
                        {
                          *s1 += xslope*(k-colsd2)+yslope*(j-rowsd2);
                          s1++;
                        }
                      else s1++;
                    }
                }
            }
          break;
        case VFF_TYP_4_BYTE :
          i1 = (int *)(img1->imagedata);
          if (mflg) i2 = (int *)(img2->imagedata);
          if (!mflg)
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      *i1 += xslope*(k-colsd2)+yslope*(j-rowsd2);
                      i1++;
                    }
                }
            }
          else
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      if (*i2++)
                        {
                          *i1 += xslope*(k-colsd2)+yslope*(j-rowsd2);
                          i1++;
                        }
                      else i1++;
                    }
                }
            }
          break;
        case VFF_TYP_FLOAT :
          f1 = (float *)(img1->imagedata);
          if (mflg) f2 = (float *)(img2->imagedata);
          if (!mflg)
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      *f1 += xslope*(k-colsd2)+yslope*(j-rowsd2);
                      f1++;
                    }
                }
            }
          else
            {
              for (j=0; j<rows; j++)
                {
                  for (k=0; k<cols; k++)
                    {
                      if (*f2++)
                        {
                          *f1 += xslope*(k-colsd2)+yslope*(j-rowsd2);
                          f1++;
                        }
                      else f1++;
                    }
                }
            }
          break;
        case VFF_TYP_COMPLEX :
          kerror("vipl","lvtilt",
		 "Don't know how to TILT a COMPLEX image!\n");
          return(FALSE);
          /* break; */
        default:
          kerror("vipl","lvtilt",
		 "Unknown data storage type, %d.\n",img1->data_storage_type);
          return(FALSE);
          /* break; */
      }
      return(TRUE);
}
/* -library_code_end */
