/*
  * Khoros: $Id: dloadvect.c,v 1.5 1996/04/05 00:56:17 fas Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: dloadvect.c,v 1.5 1996/04/05 00:56:17 fas Exp $";
#endif


/*
 * Copyright (C) 1993 - 1996, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		File Contents Title
   >>>>		
   >>>>   Public:
   >>>>                 dload_vector()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"	 	

/************************************************************
*
*  Routine Name: dload_vector - rearrange xvimage data into a vector for 1.0 DSP
*				routines
*
*       Purpose: This routine is a dsp utility which takes a multiband
*		 data file and rearanges the data so that the it is stored
*                as a list of vectors.
*                dload_vector should be called at the beginning of each
*                dsp1 routine to convert data to vector format.  It formats
*                the data according to the process direction (DSP_VECTOR,
*                DSP_BAND) passed in.
*                A **char vector data array is returned, and the values for
*                the num_vects and dimension are set.  Dimension refers
*                to the number of points in each vector.  If the data
*                is complex, dimension is the number of complex PAIRS
*                per vector.
*
*                Dload_vector checks that number of image data bands is
*                at least 1.  If data processing is in DSP_VECTOR direction,
*                Calls load_vector.  If data processing is in DSP_BAND
*                direction, assigns data to 2 dimensional vector array.
*
*                DSP_VECTOR Processing Direction:
*                Each pixel in a multiband image can be thought of
*                as a vector looking down all the bands. Thus, each
*                pixel is a vector with dimension = number of data bands.
*                This format is obtained if the parameter, process_dir,
*                passed to dload_vector is DSP_VECTOR (0).
*                Memory is allocated for the vector array by load_vector.
*
*                When process_dir = DSP_VECTOR, dload_vector will scan the
*                image row by row making a linear
*                2-dimensional array, where the first array index
*                points to a vector (which corresponds to some pixel),
*                the second array index contains the values for that
*                pixel looking down the bands on the image.
*                Memory is allocated for the vector array in dload_vector.
*
*                DSP_BAND Processing Direction:
*                If process_dir is DSP_BAND (1), the calling routine requires
*                that each data band be represented as a vector.  Therefore
*                dload_vector will return a 2-dimensional array where
*                each band is a vector of dimension rows*cols.
*
*                Data type can be of any type except BIT.
*
*         Input: image       - a pointer to the image to vectorize
*		 process_dir - a flag set to DSP_VECTOR (0) the data
*			       is to be vectorized down the bands, and
*			       set to DSP_BAND (1) if each band is to
*			       be returned as a vector.
*
*        Output: num_vects - the total number of vectors loaded into the array
*		 dimension - the dimension of each vector
*
*       Returns: an array of vectors
*
*  Restrictions: Restrictions on data or input as applicable
*    Written By: Donna Koechner
*          Date: 2/9/90
*      Verified:
*  Side Effects:
* Modifications: Converted to 2.0 by SJ
*************************************************************/
char **dload_vector(
	xvimage *image,
	int     *num_vects,
	int     *dimension,
	int      process_dir)

{
  int   border,			/* image border (set to zero for now) */
	i,j,indx;

  short *sptr, **sdata;		/* short pointer and vector array */
  int   *iptr, **idata;		/* int pointer and vector array */
  float	*fptr, **fdata;		/* float pointer and vector array */
  double *dptr, **ddata;	/* double pointer and vector array */

  char	*cptr, **data;		/* byte pointer and vector array */

  char  *prog = "dload_vector: ";


 /*
  * If image has less than one band, there is an error in the input.
  */
  
  if (image->num_data_bands < 1)
  {
    (void) kfprintf(kstderr,"dload_vector:  invalid number of data bands ");
    (void) kfprintf(kstderr,"(%d) in input image\n", image->num_data_bands);
    return(0);
  }

 /*
  * If process direction is down vectors, or the input file has
  * only one data band, call load_vector routine 
  * to rearrange data.  If the data is complex, 
  * load_vector returns dimension which is 2 times the number of bands.
  */

  if ( (process_dir == DSP_VECTOR) && (image->num_data_bands != 1) )
  {
    border = 0;
    data = load_vector(image, border, num_vects, dimension);
    if (data == NULL)
    {
      (void)kfprintf(kstderr,"load_vector failed!\n");
      return(0);
    }
  }

  else   /* processing across bands, or only one band present */
  {
   /*
    * Assign num_vects and dimension values 
    */
    *num_vects = image->num_data_bands;
    *dimension = image->col_size * image->row_size;

    switch(image->data_storage_type) 
    {
      case VFF_TYP_BIT:
           (void)kfprintf(kstderr,"%sCannot work on BIT data type\n",prog);
           return(0);

      case VFF_TYP_1_BYTE:
           /*
            * allocate the data array 
            */
            if((data = (char **) (kmalloc((unsigned int) *num_vects * 
                       sizeof(char *) ))) == NULL)
            {
              (void)kfprintf(kstderr,"dload_vector: byte malloc failed\n");
              return(0);
            }
            for(i=0; i<*num_vects; i++)
            {
              if ((data[i] = (char *) (kmalloc((unsigned int) *dimension *
                             sizeof(char) ))) == NULL)
              {
                (void)kfprintf(kstderr,"dload_vector: byte malloc failed\n");
                return(0);
              }
            }

           /*
            * Assign image data address to ptr.
            */
            cptr = (char *) image->imagedata;
        
           /*
            * Assign ptr to output data 
            */
            indx = 0;
            for (i=0; i<*num_vects; i++)
            {
              for (j=0; j<*dimension; j++)
              {
                data[i][j] = cptr[indx];
	        indx++;
              }
            }
	    break;

      case VFF_TYP_2_BYTE:
           /*
            * allocate the data array 
            */
            if((sdata = (short **) (kmalloc((unsigned int) *num_vects * 
                        sizeof(short *) ))) == NULL)
            {
              (void)kfprintf(kstderr,"dload_vector: short malloc failed\n");
              return(0);
            }
            for(i=0; i<*num_vects; i++)
            {
              if ((sdata[i] = (short *) (kmalloc((unsigned int) *dimension *
                              sizeof(short) ))) == NULL)
              {
                (void)kfprintf(kstderr,"dload_vector: short malloc failed\n");
                return(0);
              }
            }

           /*
            * Assign image data address to ptr.
            */
            sptr = (short *) image->imagedata;
        
           /*
            * Assign ptr to output data 
            */
            indx = 0;
            for (i=0; i<*num_vects; i++)
            {
              for (j=0; j<*dimension; j++)
              {
                sdata[i][j] = sptr[indx];
	        indx++;
              }
            }
            data = (char **) sdata;
	    break;

      case VFF_TYP_4_BYTE:
           /*
            * allocate the data array 
            */
            if((idata = (int **) (kmalloc((unsigned int) *num_vects * 
                        sizeof(int *) ))) == NULL)
            {
              (void)kfprintf(kstderr,"dload_vector: int malloc failed\n");
              return(0);
            }
            for(i=0; i<*num_vects; i++)
            {
              if ((idata[i] = (int *) (kmalloc((unsigned int) *dimension *
                              sizeof(int) ))) == NULL)
              {
                (void)kfprintf(kstderr,"dload_vector: int malloc failed\n");
                return(0);
              }
            }

           /*
            * Assign image data address to ptr.
            */
            iptr = (int *) image->imagedata;
        
           /*
            * Assign ptr to output data 
            */
            indx = 0;
            for (i=0; i<*num_vects; i++)
            {
              for (j=0; j<*dimension; j++)
              {
                idata[i][j] = iptr[indx];
	        indx++;
              }
            }
            data = (char **) idata;
	    break;

      case VFF_TYP_FLOAT:
           /*
            * allocate the data array 
            */
            if((fdata = (float **) (kmalloc((unsigned int) *num_vects * 
                        sizeof(float *) ))) == NULL)
            {
              (void)kfprintf(kstderr,"dload_vector: float malloc failed\n");
              return(0);
            }
            for(i=0; i<*num_vects; i++)
            {
              if ((fdata[i] = (float *) (kmalloc((unsigned int) *dimension *
                              sizeof(float) ))) == NULL)
              {
                (void)kfprintf(kstderr,"dload_vector: float malloc failed\n");
                return(0);
              }
            }

           /*
            * Assign image data address to ptr.
            */
            fptr = (float *) image->imagedata;
        
           /*
            * Assign ptr to output data 
            */
            indx = 0;
            for (i=0; i<*num_vects; i++)
            {
              for (j=0; j<*dimension; j++)
              {
                fdata[i][j] = fptr[indx];
	        indx++;
              }
            }
            data = (char **) fdata;
	    break;

      case VFF_TYP_COMPLEX:
           /*
            * allocate the data array 
            */
            if((fdata = (float **) (kmalloc((unsigned int) *num_vects * 
                        sizeof(float *) ))) == NULL)
            {
              (void)kfprintf(kstderr,"dload_vector: complex malloc failed\n");
              return(0);
            }
            for(i=0; i<*num_vects; i++)
            {
              if ((fdata[i] = (float *) (kmalloc((unsigned int) *dimension *
                              2 * sizeof(float) ))) == NULL)
              {
                (void)kfprintf(kstderr,"dload_vector: complex malloc failed\n");
                return(0);
              }
            }

           /*
            * Assign image data address to ptr.
            */
            fptr = (float *) image->imagedata;
        
           /*
            * Assign ptr to output data 
            */
            indx = 0;
            for (i=0; i<*num_vects; i++)
            {
              for (j=0; j<(*dimension*2); j++)
              {
                fdata[i][j] = fptr[indx];
	        indx++;
              }
            }
            data = (char **) fdata;
	    break;

      case VFF_TYP_DOUBLE:
           /*
            * allocate the data array 
            */
            if((ddata = (double **) (kmalloc((unsigned int) *num_vects * 
                        sizeof(double *) ))) == NULL)
            {
              (void)kfprintf(kstderr,"dload_vector: double malloc failed\n");
              return(0);
            }
            for(i=0; i<*num_vects; i++)
            {
              if ((ddata[i] = (double *) (kmalloc((unsigned int) *dimension *
                              sizeof(double) ))) == NULL)
              {
                (void)kfprintf(kstderr,"dload_vector: double malloc failed\n");
                return(0);
              }
            }

           /*
            * Assign image data address to ptr.
            */
            dptr = (double *) image->imagedata;
        
           /*
            * Assign ptr to output data 
            */
            indx = 0;
            for (i=0; i<*num_vects; i++)
            {
              for (j=0; j<*dimension; j++)
              {
                ddata[i][j] = dptr[indx];
	        indx++;
              }
            }
            data = (char **) ddata;
	    break;

      case VFF_TYP_DCOMPLEX:
           /*
            * allocate the data array 
            */
            if((ddata = (double **) (kmalloc((unsigned int) *num_vects * 
                        sizeof(double *) ))) == NULL)
            {
              (void)kfprintf(kstderr,"dload_vector: dcomplex malloc failed\n");
              return(0);
            }
            for(i=0; i<*num_vects; i++)
            {
              if ((ddata[i] = (double *) (kmalloc((unsigned int) *dimension *
                              2 * sizeof(double) ))) == NULL)
              {
                (void)kfprintf(kstderr,"dload_vector: dcomplex malloc failed\n");
                return(0);
              }
            }

           /*
            * Assign image data address to ptr.
            */
            dptr = (double *) image->imagedata;
        
           /*
            * Assign ptr to output data 
            */
            indx = 0;
            for (i=0; i<*num_vects; i++)
            {
              for (j=0; j<(*dimension*2); j++)
              {
                ddata[i][j] = dptr[indx];
	        indx++;
              }
            }
            data = (char **) ddata;
	    break;

      default:
	   (void)kfprintf(kstderr,"%sInvalid image data storage type\n",prog);
	   (void)kfprintf(kstderr,"Can only convert type BYTE, SHORT, INTEGER,");
	   (void)kfprintf(kstderr,"Float\nand Complex\n");
	   return(0);

    }  /* end of case statement */

  }
  return(data);
}
